#!/bin/bash

unset CDPATH  # CDPATH causes every "cd" command to print the name of the target
	      # directory.

# comment these two lines out if you don't want an execution log generated.
# it slows the execution down slightly.

exec 2> /tmp/BETTER_EXAMPLE.log
set -x

###############################################################
# ScriptTableViewer scripts are state machines which pass
# data back and forth to the table viewer.  In the initial
# invocation, the state is set to start.  The script prints
# intial information and transitions to some other state based
# on the needs of the script.
#
# In subsequent invocations, the state is passed back to the
# script, so that the role of the script in subsequent 
# invocations is to interpret the current state, the keys that
# the user pressed, and the data in the row and column that the
# user was in when he pressed a command key.
#
# To add commands to a display state whose name is <STATENAME>, go 
# find a function named handle<STATENAME> and pick a new key to
# let the user trigger the command.  In the handle<STATE> function,
# add a new case statement like this:
#
#    case "$cmd" in
#
#      ...
#
#      <newKey>) <new command behavior>
#                ;;
#      ...
#
#    esac
#
# Note that many new command behaviors will create new command pages.
# Command pages are associated with new STATE names.  When adding a new
# state, you must add the following things:
#
# 1.  A name for the state, for examle STATEX
# 2.  A new display function for the state, say displaySTATEX() { ... }
# 3.  A new help comment for the state, it should looke somethign like this
#          #beginHelpSTATEX
#          # some text
#          #endHelpSTATEX
# Note that your functions which display pages must always look basically like
# this:
#
#    displaySTATEX()
#    {
#        STATE=STATEX
#        CMDKEYS="<keys allowed from your new state page>"
#        TITLE="SOME TITLE FOR THIS PAGE"
#        printStateVariables
#
#        echo "table header row"
#        echo "at least 1 table body row"
#    }
#
#  Tables are too complicated to explain here.  See /home/lboggs/CXX/STV/README  
# 
#
################################################################

########################################################
# The first thing the script does is set up the empty
# state of the state variables.
########################################################

# required state variables

STATE=start         
CMDKEYS=""
TITLE=title         
HELP=""  
TAB=""


# application specific state variables

OBJECT=""  # current object being displayed

printStateVariables()
{
   #
   #  This function should be called before any other
   #  output from the script.  It should presumbably
   #  be called AFTER the new STATE values are set based on
   #  an interpretation of the current state and user's
   #  inputs.
   #
   #  Typically, functions that generate new pages will look like
   #  this:
   #
   #     STATE=NEWSTATE
   #     OBJECT="new object of some kind"
   #     TITLE="title for new page"
   #     CMDKEYS="keys active on the new page"
   #
   #     printStateVariables
   #
   #     echo "${TAB}Column1{$TAB}Column2 ..."
   #

   #-------------------------------------------------------------------------------
   #   First, echo the system variables that define the
   #   state of the displayed page
   #
   
   echo "=%==%--STATE=$STATE"
   echo "=%==%--TITLE=$TITLE"
   echo "=%==%--CMDKEYS=$CMDKEYS"
   echo "=%==%--TAB=$TAB"
   echo "=%==%--CXX_STV_REPOPULATE=1"   # 0 means don't regenerate pages when
					# returning from a child page.  1 is slower
                                        # but more accurage


   # do NOT put HELP in here
   # do NOT put the names of POPUP_FIELD_* variables here
   # do NOT put POPUP_ABORTED in here


   #-------------------------------------------------------------------------------

   #
   #  Finally, put the application specific state variables here
   #


   echo "=%==%--OBJECT=$OBJECT"
}

################################################################
# detect actual current state from command line.  command line options
# defined like this populate the new state:
#
#  --STATE_VARIABLE_NAME="value"
#
# The actual command line is long and looks basically like this:
#
#  --STATE=name1 --SVN2=value2 --SVN3=value3 ...   row col columnContents rowContents
#
# where SVN2 and SVN3 are state variable names
#
################################################################

while [ $# != 0 ]
do
   # parse options

   case "$1" in
     --*=*)

	  ;;     # process options beginning with --name=
     *)
	  break  # quit when the option does not begin with --
   esac

   name="${1/%=*/}"  # remove =value from a --name=value pair
   value="${1/#*=/}" # remove --name from the --name=value pair

   name=${name:2}    # remove leading --

   eval "$name=\"$value\""

   shift
done

################################################################
# BEGIN HELP
################################################################

if [ "$HELP" != "" ]
then

#
#  Print the prefix for all state help pages here.
#

sed -e 's/.*/help:&/1' - <<EOH

    You are looking at the help information for
    page so and so in the such and such application


EOH

#
#  Print the help information for the current state
#


cat $0 | sed -n -e "/^#beginHelp$STATE/,/^#endHelp$STATE/{ /^#begin/d; /^#end/d; s/^./help:/1; p ; }"


    exit 0
fi

################################################################
# END HELP
################################################################

if [ "$TAB" = "" ]
then
    TAB=$( echo -n -e '\t' )
fi


#beginHelpPRIMARYPAGE
#
#  ---------------------------------------------------------------
#  You are reading the help for the primary page.
#
#  Press 's' ON A HIGHLIGHTED LINE to bring up the secondary page.
#
#  Press Enter to see the parameters passed to the script
#  ---------------------------------------------------------------
#
#
#endHelpPRIMARYPAGE

displayPRIMARYPAGE()
{

   STATE=PRIMARYPAGE
   TITLE="Primary Page"
   CMDKEYS="s"                    # enable the secondary page command
   OBJECT="PRIMARY OBJECT"
   printStateVariables

   echo "${TAB}[x]Primary Object of interest"
   echo ""
   echo "[x!]   $OBJECT"
   echo ""
   echo ""
   echo "NOTE:  press F1 for help!"

}

#beginHelpSECONDARYPAGE
# 
#  ---------------------------------------------------------------
#  You are reading the help for the secondary page.
#
#  Press 'p' ON A HIGHLIGHTED LINE to bring up a popup window
#
#  Press the Enter key to see the parameters passed to the script
#  ---------------------------------------------------------------
#
#
#endHelpSECONDARYPAGE

displaySECONDARYPAGE()
{
   STATE=SECONDARYPAGE
   TITLE="Secondary Page"
   CMDKEYS='p'
   OBJECT="SECONDARY OBJECT"
   printStateVariables

   echo "${TAB}[x]Secondary Object of interest"
   echo ""
   echo "[x!]   $OBJECT"
   echo ""
   echo "Press F1 for help"
}


handlePRIMARYPAGE()
{

   cmd="$1"

   case "$cmd" in
      s)  
	  #
	  #  create a secondary page 
	  #

	  displaySECONDARYPAGE

	  ;;

      *)
	  echo "${TAB}PRIMARY PAGE Command line arguments"

	  for f in "$@"
	  do
	      echo "    $f"
	  done
	  
	  ;;
   esac


}

command_p_SECONDARYPAGE()
{
          #  the 'p' command triggers a popup in the secondary page

          # we get here both before and after the popup that the
          # x command triggers and we have to interpret what case
          # we are in
          
          if [ "$POPUP_ABORTED" = "" -a "$POPUP_FIELD_1" = "" ]
          then
              #
              # This is the first invocation of the script in this
              # configuration (requesting the popup)
              #
              # Send the description of the popup to the script table viewer
              #
              echo "=%==%--POPUP=Some Popup;Field 1;value1;Some other Field;Some other value." |
                   sed -e 's/!/!/g'   \
                       -e 's/ /!20/g' \
                       -e 's/;/!3b/g' 
              exit 0
          else
              #
              #  This is a reinvocation of the script after a popup has
              #  run, now lets decide if it was aborted or was successful.
              #
          
              if [ "$POPUP_ABORTED" != "" ]
              then
                
                  STATE=final  # state for the next invocation
                  printStateVariables
                  echo "|State = $STATE"
                  echo "YOU ABORTED!"
                  exit
          
              else
                  STATE=final  # state for the next invocation
                  printStateVariables
                  echo "|State = $STATE"
                  echo "You entered ($POPUP_FIELD_1) and ($POPUP_FIELD_2)"
		  echo ""
		  echo "Command line arguments were"
		  for f in "$@"
		  do 
		      echo "    $f"
		  done

		  echo "Script environment"
		  set | grep POPUP
		  
          
                  exit
              fi
          fi
}


handleSECONDARYPAGE()
{

   cmd="$1"

   case "$cmd" in

      p)
	  command_p_SECONDARYPAGE "$@"   # create and handle responses to the popup associated
					 # with the p command

	  ;;

      *)
	  echo "${TAB}SECONDARY PAGE Command line arguments"

	  for f in "$@"
	  do
	      echo "    $f"
	  done

	  ;;
   esac


}



#################################################################
displaystart()
{
    displayPRIMARYPAGE
}

#################################################################
#################################################################
#################################################################
# main program body

if [ "$#" = "0" ]
then
    #
    #  first time through, handle the start state
    #
    displaystart
else

    # create a command as a string then evaluate the command.

    cmd="handle$STATE"            # start with the command name which is the word
				  # "handle" followed by the state name.

    for f in "$@"                 # all all the command line parameters to the script
    do
       g="${f//\'/&quot;}"        # we are going to protect the parameters with single quotes
				  # so lets translate the existing single quotes into &quote;

       cmd="$cmd '$g'"            # append all the scripts command line arguments to the
				  # handle function name
    done

    eval "$cmd"                   # execute the handler function
fi

exit 0

