#!/bin/bash

unset CDPATH  # CDPATH causes every "cd" command to print the name of the target
	      # directory.

# comment these two lines out if you don't want an execution log generated.
# it slows the execution down slightly.

exec 2> /tmp/stv.grep.log
set -x

###############################################################
# ScriptTableViewer scripts are state machines which pass
# data back and forth to the table viewer.  In the initial
# invocation, the state is set to start.  The script prints
# intial information and transitions to some other state based
# on the needs of the script.
#
# In subsequent invocations, the state is passed back to the
# script, so that the role of the script in subsequent 
# invocations is to interpret the current state, the keys that
# the user pressed, and the data in the row and column that the
# user was in when he pressed a command key.
#
# To add commands to a display state whose name is <STATENAME>, go 
# find a function named handle<STATENAME> and pick a new key to
# let the user trigger the command.  In the handle<STATE> function,
# add a new case statement like this:
#
#    case "$cmd" in
#
#      ...
#
#      <newKey>) <new command behavior>
#                ;;
#      ...
#
#    esac
#
# Note that many new command behaviors will create new command pages.
# Command pages are associated with new STATE names.  When adding a new
# state, you must add the following things:
#
# 1.  A name for the state, for examle STATEX
# 2.  A new display function for the state, say displaySTATEX() { ... }
# 3.  A new help comment for the state, it should looke somethign like this
#          #beginHelpSTATEX
#          # some text
#          #endHelpSTATEX
# Note that your functions which display pages must always look basically like
# this:
#
#    displaySTATEX()
#    {
#        STATE=STATEX
#        CMDKEYS="<keys allowed from your new state page>"
#        TITLE="SOME TITLE FOR THIS PAGE"
#        printStateVariables
#
#        echo "table header row"
#        echo "at least 1 table body row"
#    }
#
#  Tables are too complicated to explain here.  See /home/lboggs/CXX/STV/README  
# 
#
################################################################

########################################################
# The first thing the script does is set up the empty
# state of the state variables.
########################################################

# required state variables

STATE=start         
CMDKEYS=""
TITLE=title         
HELP=""  
TAB=""


# application specific state variables

OBJECT=""  # current object being displayed

printStateVariables()
{
   #
   #  This function should be called before any other
   #  output from the script.  It should presumbably
   #  be called AFTER the new STATE values are set based on
   #  an interpretation of the current state and user's
   #  inputs.
   #
   #  Typically, functions that generate new pages will look like
   #  this:
   #
   #     STATE=NEWSTATE
   #     OBJECT="new object of some kind"
   #     TITLE="title for new page"
   #     CMDKEYS="keys active on the new page"
   #
   #     printStateVariables
   #
   #     echo "${TAB}Column1{$TAB}Column2 ..."
   #

   #-------------------------------------------------------------------------------
   #   First, echo the system variables that define the
   #   state of the displayed page
   #
   
   echo "=%==%--STATE=$STATE"
   echo "=%==%--TITLE=$TITLE"
   echo "=%==%--CMDKEYS=$CMDKEYS"
   echo "=%==%--TAB=$TAB"
   echo "=%==%--CXX_STV_REPOPULATE=1"   # 0 means don't regenerate pages when
					# returning from a child page.  1 is slower
                                        # but more accurage


   # do NOT put HELP in here
   # do NOT put the names of POPUP_FIELD_* variables here
   # do NOT put POPUP_ABORTED in here


   #-------------------------------------------------------------------------------

   #
   #  Finally, put the application specific state variables here
   #


   echo "=%==%--OBJECT=$OBJECT"
}

################################################################
# detect actual current state from command line.  command line options
# defined like this populate the new state:
#
#  --STATE_VARIABLE_NAME="value"
#
# The actual command line is long and looks basically like this:
#
#  --STATE=name1 --SVN2=value2 --SVN3=value3 ...   row col columnContents rowContents
#
# where SVN2 and SVN3 are state variable names
#
################################################################

while [ $# != 0 ]
do
   # parse options

   case "$1" in
     --*=*)

	  ;;     # process options beginning with --name=
     *)
	  break  # quit when the option does not begin with --
   esac

   name="${1/%=*/}"  # remove =value from a --name=value pair
   value="${1/#*=/}" # remove --name from the --name=value pair

   name=${name:2}    # remove leading --

   eval "$name=\"$value\""

   shift
done

################################################################
# BEGIN HELP
################################################################

if [ "$HELP" != "" ]
then

#
#  Print the prefix for all state help pages here.
#

sed -e 's/.*/help:&/1' - <<EOH


EOH

#
#  Print the help information for the current state
#


cat $0 | sed -n -e "/^#beginHelp$STATE/,/^#endHelp$STATE/{ /^#begin/d; /^#end/d; s/^./help:/1; p ; }"


    exit 0
fi

################################################################
# END HELP
################################################################

if [ "$TAB" = "" ]
then
    TAB=$( echo -n -e '\t' )
fi


#beginHelpEXPRESSIONENTRY
#
#  ---------------------------------------------------------------
#  You are reading the help information for the grep.sh STV script
#  which allows you to search a directory substructure for locations
#  in file that match a given regular expression.  Note that 
#  regular expressions can be simple strings or complex syntax
#  involving lots of weird characters.  See this page for more
#  details:  
#
#     http://www.regular-expressions.info/reference.html
#
#  Note:  To exit this page, you can press any of the following
#         key sequences:
#
#            Esc
#            ^x^k
#
#  To use this page, do the following:
#
#  1).  set the 3 fields on the form per your needs
#
#  2).  while sitting in the "Grep Expression" field, press the 
#       Enter key to invoke grep -- this could take some time.
#
#  3).  When the results page appears, you can use the ^X^W key
#       sequence to show you the contents of the file/line on which
#       the cursor is sitting.
#
#  ---------------------------------------------------------------
#
#
#endHelpEXPRESSIONENTRY

displayEXPRESSIONENTRY()
{

   STATE=EXPRESSIONENTRY
   TITLE="Enter Grep Expression"
   CMDKEYS=""                    # enable the secondary page command
   printStateVariables

   echo "${TAB}[e!]Insensitive    ${TAB}[e!]Recursive    ${TAB}[xe!]Grep Command Regular Expression"
   echo "[xe!]N${TAB}Y${TAB} "

}


#beginHelpGREPRESULTS
#
#  ---------------------------------------------------------------
#
#  You are viewing the results of a grep invocation with parameters
#  set up in a prior page.
#
#  You should see a table containing one column and many rows.  
#  These rows represent places in the directory subtree where the
#  regular expression was matched by a grep search.
#
#  You can use the ^X^W command sequence to take you open an edit
#  session on the rows of interest to you.
#
#  You can press Esc to exit.   The Q or q character will also exit
#  this page since no data entry is allowed on it.
#
#  ---------------------------------------------------------------
#
#
#endHelp

displayGREPRESULTS()
{
    TITLE="Grep Results:  $1 $2 $3"
    STATE=GREPRESULTS
    CMDKEYS=""                    # enable the secondary page command

    printStateVariables
    
    echo "${TAB}Grep Search Results"
    
    echo ""
    
    grep -n $1 $2 "$3" * 2>&1 |
    grep -v '^Binary \|/\.svn/\|.git/' 
    
    echo ""
    
    echo "[!]*EOF"

}




handleEXPRESSIONENTRY()
{

   cmd="$1"

   case "$cmd" in
      ^J)  

	  pattern="${9% }"
	  pattern="${pattern% }"
	  pattern="${pattern% }"
	  pattern="${pattern% }"
	  pattern="${pattern% }"
	  pattern="${pattern% }"
	  pattern="${pattern% }"

	  insensitive="${7% }"
	  insensitive="${insensitive% }"
	  insensitive="${insensitive% }"
	  insensitive="${insensitive% }"

	  recursive="${8% }"
	  recursive="${recursive% }"
	  recursive="${recursive% }"
	  recursive="${recursive% }"

	  if [ "$insensitive" = "N" ]
	  then
	      insensitive=""
	  else
	      insensitive="-i"  # yes!
	  fi

	  if [ "$recursive" = "N" ]
	  then
	      recursive=""
	  else
	      recursive="-r"  # yes!
	  fi


	  displayGREPRESULTS "$insensitive" "$recursive" "$pattern" 

	  ;;

      *)
	  echo "${TAB}PRIMARY PAGE Command line arguments"

	  for f in "$@"
	  do
	      echo "    $f"
	  done
	  
	  ;;
   esac


}



#################################################################
displaystart()
{
    displayEXPRESSIONENTRY
}

#################################################################
#################################################################
#################################################################
# main program body

if [ "$#" = "0" ]
then
    #
    #  first time through, handle the start state
    #
    displaystart
else

    # create a command as a string then evaluate the command.

    cmd="handle$STATE"            # start with the command name which is the word
				  # "handle" followed by the state name.

    for f in "$@"                 # all all the command line parameters to the script
    do
       g="${f//\'/&quot;}"        # we are going to protect the parameters with single quotes
				  # so lets translate the existing single quotes into &quote;

       cmd="$cmd '$g'"            # append all the scripts command line arguments to the
				  # handle function name
    done

    eval "$cmd"                   # execute the handler function
fi

exit 0

