#!/bin/bash

unset CDPATH  # CDPATH causes every "cd" command to print the name of the target
	      # directory.

# comment these two lines out if you don't want an execution log generated.
# it slows the execution down slightly.

exec 2> /tmp/log
set -x

###############################################################
# ScriptTableViewer scripts are state machines which pass
# data back and forth to the table viewer.  In the initial
# invocation, the state is set to start.  The script prints
# intial information and transitions to some other state based
# on the needs of the script.
#
# In subsequent invocations, the state is passed back to the
# script, so that the role of the script in subsequent 
# invocations is to interpret the current state, the keys that
# the user pressed, and the data in the row and column that the
# user was in when he pressed a command key.
#
# To add commands to a display state whose name is <STATENAME>, go 
# find a function named handle<STATENAME> and pick a new key to
# let the user trigger the command.  In the handle<STATE> function,
# add a new case statement like this:
#
#    case "$cmd" in
#
#      ...
#
#      <newKey>) <new command behavior>
#                ;;
#      ...
#
#    esac
#
# Note that many new command behaviors will create new command pages.
# Command pages are associated with new STATE names.  When adding a new
# state, you must add the following things:
#
# 1.  A name for the state, for examle STATEX
# 2.  A new display function for the state, say displaySTATEX() { ... }
# 3.  A new help comment for the state, it should looke somethign like this
#          #beginHelpSTATEX
#          # some text
#          #endHelpSTATEX
# Note that your functions which display pages must always look basically like
# this:
#
#    displaySTATEX()
#    {
#        STATE=STATEX
#        CMDKEYS="<keys allowed from your new state page>"
#        TITLE="SOME TITLE FOR THIS PAGE"
#        printStateVariables
#
#        echo "table header row"
#        echo "at least 1 table body row"
#    }
#
#  Tables are too complicated to explain here.  See /home/lboggs/CXX/STV/README  
# 
#
################################################################

########################################################
# The first thing the script does is set up the DEFAULT
# state of the state variables.
########################################################

# required state variables

STATE=start         
CMDKEYS=""
TITLE=title         
HELP=""             

# application specific state variables

ROOT=root      # svn root path (https://svn/something)     
TOPDIR=""      # dir from which user launched this script the first time       
VERSION=1.0    # svn commit version of something 
TAB=""         # the tab character
RELPATH=""     # relative to TOPDIR
OBJECT=""      # current object being displayed  (not always set or used)

printStateVariables()
{
   #
   #  This function should be called before any other
   #  output from the script.  It should presumbably
   #  be called AFTER the new STATE values are set based on
   #  an interpretation of the current state and user's
   #  inputs.
   #
   echo "=%==%--STATE=$STATE"
   echo "=%==%--ROOT=$ROOT"
   echo "=%==%--TITLE=$TITLE"
   echo "=%==%--TOPDIR=$TOPDIR"
   echo "=%==%--RELPATH=$RELPATH"
   echo "=%==%--CMDKEYS=$CMDKEYS"
   echo "=%==%--VERSION=$VERSION"
   echo "=%==%--TAB=$TAB"
   echo "=%==%--OBJECT=$OBJECT"
   echo "=%==%--CXX_STV_REPOPULATE=0"   # do not reread anything from svn when user exits a table

   # do NOT put HELP in here
   # do NOT put the names of POPUP_FIELD_* variables here
   # do NOT put POPUP_ABORTED in here
}

################################################################
# detect actual current state from command line.  command line options
# defined like this populate the new state:
#
#  --STATE_VARIABLE_NAME="value"
#
# The actual command line is long and looks basically like this:
#
#  --STATE=name1 --SVN2=value2 --SVN3=value3 ...   row col columnContents rowContents
#
# where SVN2 and SVN3 are state variable names
#
################################################################

while [ $# != 0 ]
do
   # parse options

   case "$1" in
     --*=*)

	  ;;     # process options beginning with --name=
     *)
	  break  # quit when the option does not begin with --
   esac

   name="${1/%=*/}"  # remove =value from a --name=value pair
   value="${1/#*=/}" # remove --name from the --name=value pair

   name=${name:2}    # remove leading --

   eval "$name=\"$value\""

   shift
done

################################################################
# BEGIN HELP
################################################################

if [ "$HELP" != "" ]
then

#
#  If the user has requested help, display state specific help.
#
#  First display a generic header

sed -e 's/.*/help:&/1' - <<EOH

SVN BROWSER:  STATE=$STATE

   The following command keys are currently active:  "${CMDKEYS}"
						     Enter,Esc,Arrows,
						     Paging keys,
						     possibly q (quit)

EOH

#
#  Then display the help text for the current state as defined by 
#  comments of the form:
#
#     #beginHelpSTATENAME
#     #  lines of help text
#     #endHelpSTATENAME
#

cat $0 | sed -n -e "/^#beginHelp$STATE/,/^#endHelp$STATE/{ /^#begin/d; /^#end/d; s/^./help:/1; p ; }"


    exit 0
fi

################################################################
# END HELP
################################################################

if [ "$TOPDIR" = "" ]
then
    TOPDIR=`pwd`
fi


if [ "$TAB" = "" ]
then
    TAB=$( echo -n -e '\t' )
fi

isRepoDir()
{

    #
    #  check the relpath to see if it is a direcotry
    #

    output=$(
	       if ! svn list -v --depth=empty "$ROOT$RELPATH@$VERSION" 2>/dev/null
	       then
		  svn list -v --depth=empty "$ROOT$RELPATH" "-r$VERSION" 2>/dev/null
	       fi
	    )
    case "$output" in
	*./)
	    return 0 # it is a directory
	    ;;
	*)
	    ;;
    esac


    return 1  # RELPATH not a directory 

}

svnInfo()
{
    svn info "$@" 2>/dev/null | 
    sed -e "s/^\([^:]*\):  *\(.*\)/SVN_\1='\2'/g" \
	-e 's/ \([^ =]\+=\)/_\1/1' \
	-e 's/ \([^ =]\+=\)/_\1/1'  \
	-e 's/ \([^ =]\+=\)/_\1/1' \
	-e 's/ \([^ =]\+=\)/_\1/1'

}

isDirREPO()
{
   # return 0 if $1 is a subdirectory of $ROOT, otherwise return 1.
   #
   #  $1 should begin with /<project>/...

   listOutput=$(
		  svn list -v --depth=empty "$ROOT$1" 2>/dev/null
		)

   case "$listOutput" in

   */)
       return 0
       ;;

   esac

   return 1

}



#
#  Make sure the ROOT variable is valid
#

if [ "$ROOT" = "root" ]
then
    if [ -d .svn ]
    then
	svninfo=$(  svnInfo . )

	if [ "$svninfo" != "" ]
	then
	   eval "$svninfo"

	   if [ "$SVN_Repository_Root" != "" ]
	   then
	       ROOT="$SVN_Repository_Root"
	   fi
	fi
	
    else

	if [ "$SVN" != "" ]
	then
	    ROOT="$SVN"
	else
	    
	    if [ -r ~/.tools_home/svnroot.txt ]
	    then
		ROOT=$( cat ~/.tools_home/svnroot.txt )
	    fi
	fi
    fi
fi




#
#  STATE variable values
#
#       start         the state that means we have no data
#                     to work with yet.  start automatically
#                     transitions to LOCALDIR or to 
#                     REPOROOT depending on the existence
#                     of file .svn in the current directory.
#
#       LOCALDIR      In this state, the directory on the
#                     disk is displayed unless it is not
#                     an svn directory -- in which case, 
#                     a transitio
#
#       REPOROOT      Display the top of the svn tree based on
#                     the $SVN variable or the contents of
#                     $HOME/.tools_home/svnroot.txt.  
#
#       LOCALFILE     Display a file in the current directory
#
#       LOCALDIFF     Diff the local file with its checked in
#                     copy.
#
#       REPOFILE      Display a file version extracted from the
#                     svn repository
#
#       REPODIFF      Diff two versions of files stored in the
#                     svn repository
#       
#       REPODIR       Displaying the contents of a directory as
#                     extracted from the repository rather than
#                     what is necessarily in the current directory.
#
#       REPOHIST      Viewing a repository history log file --
#                     created by svn log -v.
#
#       LOCALSTATUS   Display the status (svn st output) for a local
#                     file or directory.
#
#       LOCALANN      Display the annotated version of a checked in
#                     file (or the immediate ancestor of a local file)
#                 
#       REPOANN       Display the annotated version of a file in the 
#                     repository.
#
#       REPOBRACH     Display the list of ancestors of the the current
#                     repository directory
#
#       LOCALBRACH    Display the list of ancestors of the branch 
#                     displayed in the current local directory.
#
#       LOCALCOMP     Display a list of the files which have changed 
#                     between the current directory and some other
#                     branch as selected in state LOCALBRACH.
#   
#       LOCALDELTA    Display the changes to a specific file after
#                     having selected a file instate LOCALCOMP
#                     


#################################################################


#beginHelpLOCALANN     DO NOT DELETE THE HELP SECTION!
#
#  You are viewing the svn annotation of a file in your local
#  directory -- but the changes you might have made to it are
#  NOT being shown.  Each line of the files is identified by the
#  svn change version and the author.
#
#  The following commands are applicable:
#
#    q and ESC to exit this display
#
#    Enter display's the svn diff that created the specific line
#          the cursor is sitting on.
# 
#            
#endHelpLOCALANN

displayLOCALANN()
(
   TITLE="$TAB Annotation:  $OBJECT"
   CMDKEYS=""
   STATE=LOCALANN
   printStateVariables

   echo "$TAB[x]Contents"
 
   if ! svn annotate "$OBJECT" 2>&1
   then
      echo ""
      echo "<data not available>"
   fi | expand | sed -e 'y/(/-/; y/)/-/; s/.*/[x!]&/1'


   echo "*EOF"
)
	       
displayLOCALDIFF()
{
  STATE=LOCALDIFF
  TITLE="LOCALDIFF:$OBJECT"
  printStateVariables


  echo "${TAB}Line by line diferences"

  {
      svn diff --depth=immediates -x -b -x -w -x --ignore-eol-style "$OBJECT" 2>/dev/null 
    
      if [ $# != 0 ]
      then
	 echo "<data not found>"
      else
	 echo "*EOF"
      fi
  } | expand



}

#beginHelpLOCALHIST     DO NOT DELETE THE HELP SECTION!
#  You are viewing an svn log of the changes made to
#  a file or directory.  
#
#  If you got to this page using the capital-L or 
#  capital-H command, then you are seeing the complete
#  history of the object shown in the page title, but
#  if you used a lower-case command letter then you are
#  seeing branch specific changes only.
#  
#
#  The following commands are applicable:
#
#    q and ESC to exit this dispaly
# 
#            
#endHelpLOCALHIST

displayLOCALHIST()
{
    #
    #  display this branch's changes to the specified file or directory.
    #  The variable OBJECT contains the full pathname of the file or directory
    #  to be displayed.  Objects which are not versioned in svn will produce an
    #  error
    #

    STATE=LOCALHIST
    CMDKEYS=""
    TITLE="LOCALHIST:$OBJECT"
    printStateVariables
    
    
    echo "${TAB}[x]History${TAB}Version"
    
    #
    #  Log the history of specified object within the current branch
    #

    svn log --stop-on-copy . |   # log entries  backward from current until the copy
    grep '^r[0-9]' |             # ignore output that is not the a version number id
    tail -1 |                    # ignore all but the last version
    while read rev rest
    do 
	# viable rev now contains the branch point for the current
	# directory (actually the COPY point)
    
	if ! svn log -v -rHEAD:$rev "$OBJECT"
	then
	    echo "ERROR:"
	    echo "       No history available for $OBJECT"
	    echo ""
	    echo "       Try using the 's' command to show 'svn st' output"
	    echo "       for this file."
	fi |
	sed -e "
		/^r[0-9]\+/,/^$/{

				 /^r[0-9]/{
					     #
					     #  when a new log entry is found, save the version number
					     #  in the sedhold buffer
					     #

					     h                        # copy work buffer to hold buffer
					     s/^\(r[0-9]\+\).*/\1/1   # remove all but version id
					     x                        # swap hold with work

					  }

				 /^  *[AM] \//{
						  #
						  #  when an add or modify record is found, turn it into
						  #  an editable field containing the relative
						  #  path and a second field containing the version.
						  #  

						  s/[()]/@/g
						  s/.*/[x!]&${TAB}/1
						  G                    # append version number
						  s/\n//g              # remove injected newlines
					      }
				}
	       "
    
    done


}

displayGLOBALHIST()
{
    # same as displayLOCALHIST except that you get the whole log
    # associated with the current file -- slow to generate


    STATE=LOCALHIST  # not global hist state, LOCALHIST!
    CMDKEYS=""
    TITLE="GLOBALHIST:$OBJECT"
    printStateVariables
    
    
    echo "${TAB}[x]History${TAB}Revision"
    
    #
    #  Log the complete history of specified object 
    #

    
    if ! svn log -v -rHEAD:1 "$OBJECT"
    then
	echo "ERROR:"
	echo "       No history available for $OBJECT"
	echo ""
    fi |
    sed -e "/^r[0-9]\+/,/^\$/{
				 /^r[0-9]/{
					     #
					     #  when a new log entry is found, save the version number
					     #  in the sedhold buffer
					     #

					     h                        # copy work buffer to hold buffer
					     s/^\(r[0-9]\+\).*/\1/1   # remove all but version id
					     x                        # swap hold with work

					  }
			     /^  *[AM] \//{
					      s/[()]/@/g
					      s/.*/[x!]&${TAB}/1
						  G                    # append version number
						  s/\n//g              # remove injected newlines

					  }
			    }
	   "
    
    


}

displayGLOBALHIST_question()
{
   TITLE="ARE YOU SURE?"
   STATE=GLOBALHISTQ # Q IS question only state that is invoked as a
		     # warning about how long it takes to display the 
		     # global history

   printStateVariables

   echo "|[x]Why you are being asked this:"

   echo ""
   echo "  Displaying the complete history can be very time consuming?"
   echo "  Perhaps 2 or more minutes!  Do you really want to?  If so"
   echo "  Select YES below."
   echo ""
   echo "[x!]     YES"
   echo ""
   echo "  If not, presses Q or Esc to exit"
}

handleGLOBALHISTQ()
{
   displayGLOBALHIST
    
}





#beginHelpLOCALFILE     DO NOT DELETE THE HELP SECTION!
#  You are viewing a file in your local directory.  
#  The following commands are applicable:
#
#    Enter - Lets you get a diff between the current file
#            content and the original file from which you
#            checked out.
#
#    d,D   - Diff, just like the Enter key provides.
#
#    a,A   - Annotate, show annotated version of the checked in
#            version of this file.
#    
#    h     - Change log (local to this branch)
#
#    H     - full change log
#              +  VERY SLOW TO GENERATE (2 minutes)
#    
#                                                                 .
#endHelpLOCALFILE

displayLOCALFILE()
{
   STATE=LOCALFILE
   TITLE="LOCALFILE:$OBJECT"
   CMDKEYS="dDah"
   printStateVariables

   echo "$TAB[x!]File Contents"
   
   expand "$OBJECT" 2>&1  | sed -e 's/.*/[x]&/1'

   echo ""
   echo "*EOF"
}

handleLOCALFILE()  # handle commands when a local file is displayed
{
    case "$1" in
        "^J") displayLOCALDIFF "$@" 
            ;;
        d) displayLOCALDIFF "$@" 
            ;;
        D) displayLOCALDIFF "$@" 
            ;;

	a) 
	   displayLOCALANN
	   ;;

	A) 
	   displayLOCALANN
           ;;

	h) displayLOCALHIST
	   ;;

	H) displayGLOBALHIST
	   ;;


        *)
           echo "$TAB[x!]Invalid LOCALFILE command '$1'"
           for f in "$@" 
           do
              echo "$f"
           done
           ;;
    esac
}


#beginHelpLOCALDIR     DO NOT DELETE THE HELP SECTION!
#  You are viewing a local directory in your user account.  The following commands 
#  are applicable:
#
#    ?     - display the checkout status of all files in the current subtree.
#              + SLOW (30+ seconds) TO GENERATE
#    a     - display the annotation for the file under the cursor.  Normal files only.
#    A     - add a file to svn
#    B     - display branching for the current directory (last 4 ancestors)
#              + SLOW (10+ seconds) TO GENERATE
#    C     - compare this branch to another branch (checked in files only)
#              + SLOW (10-30 seconds) TO GENERATE
#    d     - diff the file under the cursor with its immediate predecessor.
#    D     - diff the current directory against its immediate predecessor.
#    h     - display current branch history of the file under the cursor
#    H     - display whole history of the file under the cursor
#              + VERY SLOW (2 minutes or more) to GENERATE
#    I     - display svn info for the current directory
#    l     - display the branch specific changes to the current directory
#    L     - display whole history of changes to the current directory
#              + VERY SLOW (2 minutes or more) to GENERATE
#    m     - display a log of the true modifications to the file under the cursor
#              + SLOW (10+seconds)
#    R     - display the repository root directory
#    r     - display the repository information for the current local directory
#    s     - gives you the edit status of the file under the cursor
#    S     - gives you the edit status of all files in the current directory (only)
#    U     - prompts for a user name and displays changes by that user in this directory
#
#    Enter - gives you a display of the file or directory on 
#            the current line.
#                                                                                          .
#endHelpLOCALDIR

displayLOCALDIR()
{
   TITLE="LOCALDIR:$OBJECT"
   STATE=LOCALDIR
   CMDKEYS="${CMDKEYS}sSdDhHlLRrBCamI?UA"
   printStateVariables

   echo "$TAB[x]Type${TAB}User${TAB}Group${TAB}size${TAB}Date${TAB}Name"

   ls -l "$OBJECT" |  
   sed \
       -e 1d \
       -e "s/\([^ ]\+\) *\([^ ]\+\) *\([^ ]\+\) *\([^ ]\+\) *\([^ ]\+\) *\([^ ]* *[^ ]* [^ ]*\) *\(.*\)/\1${TAB}\3${TAB}\4${TAB}\5${TAB}\6${TAB}\7/1" \
       -e 's/.*/[x!]&/1'

   echo "*EOF"

}

fileLOCALDIR()
{
     # handle a user Enter key on a file or directory

     OBJECT="$TOPDIR$RELPATH/$1"

     if [ -d "$OBJECT" ]
     then

       STATE=LOCALDIR  # should already be, but ..
       RELPATH="$RELPATH/$1"
       displayLOCALDIR

     else

       displayLOCALFILE  # uses OBJECT

     fi
}


#beginHelpLOCALSTATUS     DO NOT DELETE THE HELP SECTION!
#  You are viewing the svn status command output on files in a
#  local directory (only).  The commands available are:
#
#    Enter, d -  View a diff of files which have been added, deleted,
#                or modified.
#
#
#  NOTE:  non-modified and non-configured files will not accept any
#         commands.
#
#endHelpLOCALSTATUS     DO NOT DELETE THE HELP SECTION!

STATUSTITLES="${TAB}[x]Changes${TAB}[x]Version${TAB}[x]LastChange${TAB}User${TAB}File"

STATUSFILTER="
	       /^?/s/\(.....\)  *\(.*\)/\1$TAB $TAB $TAB $TAB \2/1

	       /^?/!s/^\(.....\)  *\([^ ]*\)  *\([^ ]*\)  *\([^ ]*\)  *\(.*\)/\1$TAB\2$TAB\3$TAB\4$TAB\5/1
               /^[MAD]/{
                          # when files are added deleted or changed
                          # let the user perform diff
                          s/.*/[x!]&/1
                       }
	     "

statusLOCALFILE()
{
   #
   #  display svn status on local directory files
   #

   OBJECT="$TOPDIR$RELPATH/$1"
   TITLE="EDIT STATUS:$OBJECT"
   STATE=LOCALSTATUS
   CMDKEYS="d"
   printStateVariables

   echo "$STATUSTITLES"

   (
       cd "$TOPDIR$RELPATH"
       if ! svn st --depth=immediates -v "$1"
       then
           echo "Error:  status of $OBJECT not found"
       fi

   ) |
     sed -e "$STATUSFILTER"

   echo ""

}

statusLOCALDIR()
{

   TITLE="LOCALSTATUS:$OBJECT"
   STATE=LOCALSTATUS
   CMDKEYS="d"
   printStateVariables

   echo "$STATUSTITLES"

   (
       cd "$TOPDIR$RELPATH"
       if ! svn st -v --depth=immediates 2>&1
       then
           echo "Error:  status of $OBJECT not found"
       fi

   ) |
     sed -e "$STATUSFILTER"
	    


   echo ""

}

handleLOCALSTATUS()
{
   cmd="$1"

   case "$cmd" in

       d)  
	   (
	       TITLE="diff $1"
	       STATUS=LOCALDIFF
	       OBJECT="$TOPDIR$RELPATH$1"
               CMDKEYS=""
	       printStateVariables
	       
               echo "${TAB}Changed lines"
               cd "$TOPDIR$RELPATH"
	       shift 10
	       if ! svn diff -x -b -x -w -x --ignore-eol-style "$1"  2>/dev/null 
	       then
		  echo "<error -- diff not available"
	       fi | expand
               echo "*EOF"
	   )
           ;;

       ^J) (
               TITLE="diff $1"
               STATUS=LOCALDIFF
               OBJECT="$TOPDIR$RELPATH$1"
               CMDKEYS=""
               printStateVariables
	       
               echo "${TAB}Changed lines"
               cd "$TOPDIR$RELPATH"
	       shift 10
	       if ! svn diff -x -b -x -w -x --ignore-eol-style "$1"  2>/dev/null
	       then
		  echo "<error -- diff not available"
	       fi | expand
               echo "*EOF"
	   )
           ;;

       *)
	   echo "${TAB}Invalid LOCALSTATUS command '$1'"
           for f in "$@" 
           do
              echo "$f"
           done
	   ;;

   esac
}

#beginHelpREPODIR     DO NOT DELETE THE HELP SECTION!
#     
#------------------------------------------------------------------------
#  You are viewing a directory documented in the svn repository. It
#  may or may not match files found in your local directories.
#
#    a     - display annotation of the file under the cursor
#
#    B     - display branching of current directory (last 4 branches)
#              + SLOW (10+ seconds) TO GENERATE
#   
#    h     - display the changes to file under the cursor in this branch
#
#    H     - display the full history up to this point for the file
#            under the cursor.  VERY SLOW (2+ minutes) TO GENERATE
#
#    l     - display changes to the current directory in this branch
#
#    m     - list revisions which modify the file under the cursor.
#              +  SLOW (10+ seconds) TO GENERATE
#
#    R     - display repository root directory
#
#    P     - display project history.
#              +  SUPER SLOW TO CREATE THE FIRST TIME!  10 Minutes or 
#                 more After the initial creation, it still may take 30
#                 seconds to load and is slow to page due to its huge 
#                 size.
#
#    Enter - gives you a display of the file or directory on 
#            the current line.
#------------------------------------------------------------------------
#endHelpREPODIR

displayREPODIR()
{
   TITLE="REPODIR:$ROOT$RELPATH@$VERSION"
   CMDKEYS="lhHadDBPm"
   STATE=REPODIR
   printStateVariables

   echo "${TAB}[x] Version ${TAB}[x] User ${TAB}[x] Size ${TAB}[x] Date ${TAB}[x] PROJECT"

   f='  *\([^ ]\+\)'

   {
       svn list -v "$ROOT$RELPATH@$VERSION" 2>/dev/null

       if [ $? != 0 ]
       then
           svn list -v "$ROOT$RELPATH" -r"$VERSION" 2>/dev/null  #svn is goofy about @VERSION specs some times

           if [ $? != 0 ]
           then
               echo "<data not found>"  # make the display work correctly 
           fi
       fi

   } |
   sed -e "
              /\.\/$/d  # get rid of the directory entry
              #
              # convert tabs to spaces
              #

              y/\t/ / 

              #
              #  Split the line into executable fields
              #

              /.*\/$/ {
                          # directory lines end in /
                          # directories do not have a file size field so they have a
                          # different format than non-directories
                          #
                          # note that leading and trailing spaces appear on al lfields
                          # except the project

                          s/^$f$f  *\($f$f$f\)  *\(.*\)/ \1 ${TAB} \2 ${TAB} -${TAB}${TAB}\3 ${TAB}\7/1
                      }

              /.*\/$/!{
                          # non-directory lines do NOT end in /

                          #       vers       id        size
                          s/^ *\([^ ]*\) *\([^ ]*\) *\([^ ]*\) *\($f$f$f\) *\(.*\)/ \1 $TAB \2 $TAB \3 $TAB\4$TAB\8/1

                      }


              #
              #  On all projects, except the one named "./", enable command
              #  handling.
              #
	      / \.\/\$/!{ s/.*/[x!]&/; }  

          "

    echo "*EOF"
}


repoLOCALDIR()
{
   # show the repository information for the current local directory.

   eval $( svnInfo "$TOPDIR$RELPATH" )

   url="$SVN_URL"

   RELPATH="${url#$SVN_Repository_Root}/"
   VERSION="$SVN_Revision"

   displayREPODIR


}

#beginHelpLOCALBRACH     DO NOT DELETE THE HELP SECTION!
#
#  You are viewing a listing of the ancestor branches
#  starting with the repository branch directory currently
#  viewed.  You may select an ancestor for viewing.
#  
#
#    Enter - View the selected ancestor branch directory. 
#                                                            .
#endHelpLOCALBRACH

brachiateLOCALDIR()
{
   #
   #  display branching of the current directory
   #

   executable=`which svnBrachiation`

   if [ "$executable" = "" ]
   then
       echo "|Error"
       echo ""
       echo "Cannot find script 'svnBrachiation' in your path"
       echo ""
   else

       STATE=LOCALBRACH
       CMDKEYS=""
       printStateVariables

       echo "${TAB}[x]Branching history"
       "$executable" "$TOPDIR$RELPATH"  | sed -e '1!{s/.*/[x!]&/1;}'
   fi



}

compareLOCALDIR()
{
   # display a list of branches to compare the current directory to.
   # checked in files only.  When the user picks a branch, produce
   # a diff summary

   eval `svnInfo`

   relpath="${SVN_URL#$ROOT}"
   project="${relpath/\//}"
   project="${project/\/*/}"

   executable=`which svnMainBranches`
   
   if [ "$executable" = "" ]
   then
       echo "|Error"
       echo ""
       echo "Can not find script 'svnMainBranches' -- put it in your path"
       echo "The script prints the urls of alternative branches to the"
       echo "the current directory."
       echo ""
       echo "relpath=$relpath"
       echo "executable=$executable"
       echo "project=$project"
   else

       inTopDir=true

       if [ "which svnTopDir 2>/dev/null" = "" ]
       then
	  
           echo "|Error"
           echo ""
	   echo "Can not find helper script 'svnTopDir'"
           echo ""

	   exit 0

       fi


       if [ "$RELPATH" != "" ]
       then
           inTopDir=false
       else
	   if [ "$PWD" != `svnTopDir` ]
	   then
	       inTopDir=false
	   fi
       fi



       if [ "$inTopDir" = "false" ]
       then
	   echo "|Error"
	   echo ""
	   echo "This command can only be executed from the top of a "
	   echo "checked out directory tree.  In this case, from"
	   echo "   `svnTopDir`"
	   echo ""
       else
           STATE=LOCALCOMP
           TITLE="List of other branches"
           CMDKEYS=""
           printStateVariables
	   echo "${TAB}[x]Other Branches"
	   $executable "$project" | 
	   sed -e "
		    s%$SVN_Repository_Root%%1
		    s/.*/[x!]&/1
		  "
           echo ""
       fi

   fi


}

handleLOCALCOMP()
{
   # handle the selection of an alternative branch to compare the current branch to

   cmd="$1"

   eval `svnInfo`

   relpath="$6"

   otherBranchPath="$SVN_Repository_Root$relpath"

   localRelpath="${SVN_URL#$SVN_Repository_Root}"

   case "$cmd" in
       ^J)
	   TITLE="DELTA:$relpath"
	   STATE="LOCALDELTA"
	   CMDKEYS=""
	   printStateVariables
	   echo "${TAB}[x]Diff summary${TAB}Relative path"

	   echo "Comparing"
	   echo "   source = Checked out version"
	   echo "               ($SVN_URL)"
	   echo "   to     = Another Branch"
	   echo "               ($otherBranchPath)"
	   echo ""
	   echo ""

	   #note:  svn diff's output is relative to the left side of the
	   #       diff operation.  That is, when files are added, deleted
	   #       or changed, the url that is used is relative to the left
	   #       operand, if the object was added in the right hand side.


	   svn diff --summarize "$SVN_URL" "$otherBranchPath" | 
	       sed -e "
			s%$SVN_Repository_Root$localRelpath%%1
			s%.*%[x!]&${TAB}$relpath%1
		      "
	   echo ""
	   ;;
       *)
	   echo "${TAB}Invalid LOCALCOMP command '$1'"
           for f in "$@" 
           do
              echo "$f"
           done
	   ;;
   esac

}

handleLOCALDELTA()
{

   #
   #  handle commands when the user is seeing a list of files that
   #  are different between the current local directory and some
   #  other branch.
   #

   cmd="$1"

   case "$cmd" in

      ^J)
	  subpath=$( echo "$7" | cut -c8-999 | sed -e 's/^  *//1' )

	  eval $( svnInfo )

	  url="$SVN_URL/$subpath"
	  
	  RELPATH="$8"

	  echo "${TAB}Delta"

	  svn diff -x -b -x -w -x --ignore-eol-style "$url" "$SVN_Repository_Root/$RELPATH$subpath" 2>&1 | expand
	  echo "[!]--End of Diff"
	  

	  ;;
       *)
	   echo "${TAB}Invalid LOCALDELTA command '$1'"
           for f in "$@" 
           do
              echo "$f"
           done
	   ;;
   esac
}



#beginHelpLOCALDIRMOD     DO NOT DELETE THE HELP SECTION!
#
#  You are viewing a page containing a list of modifications to the file
#  under the cursor.  This a "svn log" containing only actual changes
#  to the data in the file as opposed to just any old modification
#  which might include property changes, branch name changes, etc.
#
#
#    Enter - View the differences in the change under the cursor
#                                                                         .
#endHelpLOCALDIRMOD
modificationsLOCALDIR()
{

    script=`which svnListChanges`

    if [ "$script" = "" ]
    then

	echo "|Error"
	echo ""
	echo "Cannot find script, svnListChanges"

	return
    fi


    TITLE="$OBJECT"
    CMDKEYS=""
    STATE=LOCALDIRMOD
    printStateVariables
    echo "${TAB}[x]File Content Changes"
    "$script" "$OBJECT" |
    sed -e '
	      /^-.*/{s/.*//1}  # replace ---- with a blank line

	      /^r[0-9]* |/{
			    s/.*/[x!]&/1   # make version number lines clickable
			  }


	   '

    echo ""
}




handleLOCALDIRMOD()
{
    cmd="$1"
    version="${6%% *}"

    case "$cmd" in

       ^J)
	   TITLE="Diff:$OBJECT@$version"
	   printStateVariables
	   echo "${TAB}Changed lines"
	   {
	       if ! svn diff -x -b -x -w -x --ignore-eol-style "-c$version" "$OBJECT@$version"  2>/dev/null
	       then
		  if ! svn diff -x -b -x -w -x --ignore-eol-style "-c$version" "$OBJECT"  2>&1
		  then
		       echo "<data not found>"
		  else
		       echo "*EOF"
		  fi
	       else
		  echo "*EOF"
	       fi       
	   } | expand


	   ;;



       *)
	   echo "${TAB}Invalid LOCALDIRMOD command '$1'"
           for f in "$@" 
           do
              echo "$f"
           done
	   ;;
    esac
}


#beginHelpSUBTREESTATUSLOCALDIR     DO NOT DELETE THE HELP SECTION!
#
# You are viewing a page that describes the checkout status of all
# files within a subtree hierarchy.
#
#   q, Q, ESC -- Exit this display
#   
#           i -- checkin the file under the cursor
#                                                                         .
#endHelpSUBTREESTATUSLOCALDIR

subtreeStatusLOCALDIR()
{
    TITLE="SUBTREE STATUS:$TOPDIR$RELPATH"

    STATE=SUBTREESTATUSLOCALDIR

    CMDKEYS="ix"

    printStateVariables

    echo "${TAB}[x]STATUS $TOPDIR$RELPATH"

    tmpfile=/tmp/subtreestatus$$.log


    if ! svn st >$tmpfile 2>&1
    then
       echo "Error:  could not get the status of checked out files"
       echo ""
       cat $tmpfile
    else

       echo "[_]MODIFIED FILES"
       echo ""

       sed -n -e '/^[ ?]/!{s/.*/[!x]&/1; p}' $tmpfile

       echo ""
       echo "[_]UNKNOWN FILES"
       echo ""

       sed -n -e '/^?/{s/.*/[!x]&/1; p }' $tmpfile

       echo ""
       echo "[_]MERGED FILES"
       echo ""

       sed -n -e '/^ M/{s/.*/[!x]&/1; p }' $tmpfile


    fi

    rm -f $tmpfile
}

checkinLOCALDIR()
{
  file="$1"

  #  the 'i' command triggers a popup that asks you if you want
  #  to checkin the file under the cursor

  # we get here both before and after the popup that the
  # U command triggers and we have to interpret what case
  # we are in
  
  if [ "$POPUP_ABORTED" = "" -a "$POPUP_FIELD_1" = "" ]
  then
      #
      # This is the first invocation of the script in this
      # configuration (requesting the popup)
      #
      # Send the description of the popup to the script table viewer
      #
      echo "=%==%--POPUP=Checkin the file;SVN Comment;                         " |
	   sed -e 's/!/!/g'   \
	       -e 's/ /!20/g' \
	       -e 's/;/!3b/g' 
      exit 0
  else
      #
      #  This is a reinvocation of the script after a popup has
      #  run, now lets decide if it was aborted or was successful.
      #
  
      if [ "$POPUP_ABORTED" != "" ]
      then
	
	  STATE=final  # state for the next invocation
	  printStateVariables
	  echo "|State = $STATE"
	  echo "YOU ABORTED!"
	  exit
  
      else

	  comment="$POPUP_FIELD_1"

	  TITLE="CHECKIN: $OBJECT"

	  printStateVariables

	  echo "${TAB}Checkin log..."

	  svn ci -m "$comment" "$file" 2>&1

	  echo ""
	  
      fi
  fi
}

handleSUBTREESTATUSLOCALDIR()
{
  cmd="$1"

  file=`echo "$6" | cut -c8-999 | sed -e 's/^  *//1'  `


  case "$cmd" in

      "^J")
	   case "$6" in
	       [?MA\ ]*)

		   if [ -d "$file" ]
		   then
		       echo "${TAB}VIEWING $file"

		       ls -l "$file" |
			  sed -e 's/.*/[]&/1'
     
                       echo ""
		       echo "[!]*EOF"
		   else
		       fileLOCALDIR "$file"
		   fi


		   ;;

	      *)
		  echo "${TAB}Invalid command"
                  echo ""
                  echo "${TAB}  No command supported on this line"
		  echo ""
		  echo "[!]*EOF"


	   esac


	   ;;


       i)  checkinLOCALDIR "$file"
	   ;;


       *)
	   echo "${TAB}Invalid SUBTREESTATUSLOCALDIR command '$1'"
           for f in "$@" 
           do
              echo "$f"
           done
	   ;;
   esac



}

#beginHelpUSERLOGLOCALDIR     DO NOT DELETE THE HELP SECTION!
#
# You are viewing changes made by a specific user that affect
# the current directory.
#
#   q, Q, ESC -- Exit this display
#
#   Enter -- view the specific files that were modified.
#                                                                         .
#endHelpUSERLOGLOCALDIR

userLogLOCALDIR()
{
  #  the 'U' command triggers a popup that asks you the userid of the
  #  person whose changes you want to see

  # we get here both before and after the popup that the
  # U command triggers and we have to interpret what case
  # we are in
  
  if [ "$POPUP_ABORTED" = "" -a "$POPUP_FIELD_1" = "" ]
  then
      #
      # This is the first invocation of the script in this
      # configuration (requesting the popup)
      #
      # Send the description of the popup to the script table viewer
      #
      echo "=%==%--POPUP=Enter an svn user id;User ID;" |
	   sed -e 's/!/!/g'   \
	       -e 's/ /!20/g' \
	       -e 's/;/!3b/g' 
      exit 0
  else
      #
      #  This is a reinvocation of the script after a popup has
      #  run, now lets decide if it was aborted or was successful.
      #
  
      if [ "$POPUP_ABORTED" != "" ]
      then
	
	  STATE=final  # state for the next invocation
	  printStateVariables
	  echo "|State = $STATE"
	  echo "YOU ABORTED!"
	  exit
  
      else

	  user="$POPUP_FIELD_1"

	  TITLE="USERLOG: $OBJECT"
	  STATE=USERLOGLOCALDIR

	  printStateVariables

	  echo "${TAB}[x]Changes by $user"

	  svn log "$OBJECT" -l2000 |
	  fold -s |
	  sed -n -e "/^r[0-9]\+ *|  *$user  *|/,/^------------------------------------------/{
			/^r[0-9]/{
				     s/.*/[x!]&/1
		       }
		       p
		     }                                   
		    "

	  echo ""
	  echo "*EOF"


	  
      fi
  fi
}

handleUSERLOGLOCALDIR()
{
  cmd="$1"
  VERSION="${6%% *}"

   case "$cmd" in

      ^J)
	   STATE=LOCALHIST
	   TITLE="LOCALHIST:$OBJECT@$VERSION"
	   CMDKEYS="x"
	   printStateVariables
	   echo "${TAB}[x]Changed files in this version"
	   { 
	       if ! svn log -v -l1 "$OBJECT@$VERSION" 
	       then
		   if ! svn log -v -l1 -r$VERSION "$OBJECT"
		   then
		       echo "ERROR:  change log info for version $version not available"
		   fi
	       fi
	   } |
	   sed -e 's/^   [MA] .*/[x!]&/1'

	   ;;

       *)
	   printStateVariables
	   echo "${TAB}Invalid USERLOG command '$1'"
	   echo "OBJECT=$OBJECT"
	   echo "version=$version"
           for f in "$@" 
           do
              echo "$f"
           done
	   ;;
   esac
    
}




handleLOCALDIR()  # handle commands when a local directory is displayed
{
  cmd="$1"
  shift 11    # eat rows and colums and line contents 
  file="$1"



   case "$cmd" in

       "?")  subtreeStatusLOCALDIR
	     ;;

       "^J") fileLOCALDIR "$file" "$2"
	     ;;

       "s")  statusLOCALFILE "$file"
	     ;;

       "S")  statusLOCALDIR
	     ;;

       "d")  
	     OBJECT="$TOPDIR$RELPATH/$file"

	     displayLOCALDIFF
	     ;;

       "D")  
	     OBJECT="$TOPDIR$RELPATH"

	     displayLOCALDIFF
	     ;;

       "h")  
	     OBJECT="$TOPDIR$RELPATH/$file"

	     displayLOCALHIST
	     ;;

       "H")  
	     OBJECT="$TOPDIR$RELPATH/$file"

	     displayGLOBALHIST
	     ;;

       "I")  
	     echo "${TAB}SVNINFO:  $TOPDIR$RELPATH"
	     echo ""
	     svn info
	     ;;

       "l")  
	     OBJECT="$TOPDIR$RELPATH"

	     displayLOCALHIST
	     ;;

       "L")  
	     OBJECT="$TOPDIR$RELPATH"
	     displayGLOBALHIST_question
	     ;;

       R)    displayREPOROOT
             ;;

       r)    repoLOCALDIR
	     ;;

       B)    brachiateLOCALDIR
	     ;;

       C)    compareLOCALDIR
	     ;;

       m)    
	     OBJECT="$TOPDIR$RELPATH/$file"
	     modificationsLOCALDIR
	     ;;


       "a")

	   if [ ! -d "$TOPDIR$RELPATH/$file" ]
	   then
	       OBJECT="$TOPDIR$RELPATH/$file"

	       displayLOCALANN

	   else
	       echo "|Error"
	       echo ""
	       echo "Annotation not available for directories"
	       echo ""
	   fi
	   ;;

       "A")
	       OBJECT="$TOPDIR$RELPATH/$file"
	       if svn add "$OBJECT" >/tmp/svn.add.log 2>&1
	       then
		  echo "|Done"
		  echo ""
		  echo "svn add log:"
		  echo ""
		  cat /tmp/svn.add.log
		  rm /tmp/svn.add.log
	       else
		   echo "|Error"
		   echo ""
		   cat /tmp/svn.add.log
		   rm /tmp/svn.add.log
		   echo ""
	       fi
	   ;;

       U)  userLogLOCALDIR
           ;;



       *)
	   echo "${TAB}Invalid LOCALDIR command '$1'"
           for f in "$@" 
           do
              echo "$f"
           done
	   ;;
   esac

}


################################################################
#beginHelpREPOROOT     DO NOT DELETE THE HELP SECTION!
#
#  You are viewing the root directory of the svn repository defined
#  in either your SVN variable or in the 
#
#      ~/.tools_home/svnroot.txt file.
#  
#  You are in this mode of svn display because you executed the
#  svn browser from a directory which does not contain a .svn 
#  subdirectory.
#
#    v     - prompts you for a version number and shows you the
#            repository at the time of that version
#
#    Enter - gives you a display of the file or directory on 
#            the current line.
#                                                                     .
#endHelpREPOROOT

displayREPOROOT()
{
   TITLE="SVN REPOSITORY: $ROOT"
   STATE=REPOROOT
   CMDKEYS="v"
   printStateVariables

   echo "${TAB}[x!] Version ${TAB}[x!] User ${TAB}[x!] Date ${TAB}[x!] PROJECT"

   f='  *\([^ ]\+\)'

   { 
	svn list -v "$ROOT" 2>/dev/null

	if [ $? != 0 ]
	then
	    echo "<data not found>" 
	fi

   } |
   sed -e "
	      /\.\/$/d  # get rid of the directory entry
	      #
	      #  Split the line into executable fields with leading and trailing
	      #  spaces in the displayed table except for the project name which
	      #  has no leading space -- see handleREPOROOT()
	      #

	      y/\t/ / 
	      s/^$f$f  *\($f$f$f\)  *\(.*\)/ \1 ${TAB} \2 ${TAB}\3 ${TAB}\7/1

	      #
	      #  On all projects, except the one named "./", enable command
	      #  handling.
	      #
	      / \.\/\$/!{ s/.*/[x]&/; }  

	  "
}




fileREPOROOT()
{
   # handle commands on lines in the REPOROOT state display

   project="$1"
   cmd="$2"

   # assumes VERSION is set correctly for the current row

   # right now, just assume all commands open the project

   RELPATH="/$project"

   displayREPODIR

}


versionREPOROOT()
{

  project="$1"

  #  the 'v' command triggers a popup that asks you the version of the
  #  project under the cursor

  # we get here both before and after the popup that the
  # x command triggers and we have to interpret what case
  # we are in
  
  if [ "$POPUP_ABORTED" = "" -a "$POPUP_FIELD_1" = "" ]
  then
      #
      # This is the first invocation of the script in this
      # configuration (requesting the popup)
      #
      # Send the description of the popup to the script table viewer
      #
      echo "=%==%--POPUP=Select version for project $project;version number;$VERSION" |
	   sed -e 's/!/!/g'   \
	       -e 's/ /!20/g' \
	       -e 's/;/!3b/g' 
      exit 0
  else
      #
      #  This is a reinvocation of the script after a popup has
      #  run, now lets decide if it was aborted or was successful.
      #
  
      if [ "$POPUP_ABORTED" != "" ]
      then
	
	  STATE=final  # state for the next invocation
	  printStateVariables
	  echo "|State = $STATE"
	  echo "YOU ABORTED!"
	  exit
  
      else

	  VERSION="$POPUP_FIELD_1"

	  fileREPOROOT $project "^J"
	  
      fi
  fi
}



handleREPOROOT() # handle commands when the svn repository directory is displayed
{
   # expecting command line arguments like this:
   #
   # common components
   #
   #   1=user command key pressed (^J=enter)
   #   2=current row
   #   3=current col
   #   4=rows
   #   5=cols
   #   6=current cell contents
   #
   # page specific parts (see displayREPOROOT)
   #
   #   7 ... 99 = all cells on this row
   #      7=version
   #      8=user
   #      9=date
   #      10=project


   cmd="$1"                # no leading/trailing spaces

   VERSION="$7"
   VERSION="${VERSION# }"  # the version field has both leading and trailing space
   VERSION="${VERSION% }"  

   shift 9
   project="$@"            # no leading or tailing space


   case "$cmd" in
       "^J") fileREPOROOT "$project" "$cmd" 
	     ;;

       v)    versionREPOROOT "$project" "$cmd"
	     ;;

       *)
	   echo "$TAB[x!]Invalid REPOROOT state command '$cmd'"
           for f in "$@" 
           do
              echo "$f"
           done
	   ;;
   esac

}

#beginHelpREPOFILE     DO NOT DELETE THE HELP SECTION!
#
#  You are viewing a listing of a file extracted 
#  your SVN repository.
#  
#
#    Enter -     Enter, d, and D give you a diff between
#    d,D         between this version of the file and
#                its predecessor.
#   
#    a     -     Gives an annotation of this version of
#                the file.
#                
#    h     -     Gives a listing of the changes to the
#                file in its branch.
#
#    H     -     Gives a list of all changes to the file
#                up to the point you are viewing.
#                    +  VERY SLOW TO GENERATE (2+ minutes)
#
#    R     -     Display the repository as it was at the time
#                of this commit.
#                                                              .
#endHelpREPOFILE

displayREPOFILE()
{
    TITLE="REPOFILE:$RELPATH@$VERSION"
    STATE=REPOFILE
    CMDKEYS="ahHdDR"
    printStateVariables

    echo "${TAB}[x]Contents"


    { 
	{
	    if ! svn cat "$ROOT$RELPATH@$VERSION"  2>/dev/null
            then

	       if ! svn cat -r$VERSION "$ROOT$RELPATH"  2>/dev/null # sometimes svn gives error when you use the @VERSION in the url
               then
                   echo "<data not found>"  # blank line to make the display work properly
               fi
            fi

	} | expand | sed -e 's/.*/[x]&/1'




    } 
}

#beginHelpREPOHIST     DO NOT DELETE THE HELP SECTION!
#  You are viewing an svn log of the changes made to
#  a file or directory at some earlier time in the
#  svn repository history.  
#
#  Only files in the current branch (ie svn copy) are
#  shown.
#
#  Note that since directories and files can be copied and 
#  renamed, different versions of a given URL can have 
#  profoundly different histories, and different files with 
#  different names can have almost identical histories.
#
#  The following commands are applicable:
#
#    q and ESC to exit this dispaly
# 
#            
#endHelpREPOHIST

displayREPOHIST()
{
    # display the branch specific log of the currently display RELPATH

    STATE=REPOHIST
    CMDKEYS=""
    TITLE="REPOHIST:$RELPATH@$VERSION"
    printStateVariables
    
    
    echo "${TAB}[x]History${TAB}Revision"
    
    #
    #  Log the history of specified object within the current branch
    #
    {
	# print the single change log entry that indicates the copy of the
	# specified relpath and version

	if ! svn log --stop-on-copy "$ROOT$RELPATH@$VERSION" -r1:$VERSION -l1
	then
	   if ! svn log --stop-on-copy "$ROOT$RELPATH" -r1:$VERSION -l1 2>&1
	   then
	      echo "<error:  log not found for $RELPATH@$VERSION"
	   fi
	fi
    } | 
    grep '^r[0-9]' | 
    tail -1 | 
    while read rev rest
    do 
	# we go through this loop either 0 or 1 times -- and the rev
	# is the copy point for the current branch (ie svn copy)

        # viable rev now contains the branch point for the current
        # directory (actually the COPY point)
    
	if ! svn log -v "-r$VERSION:$rev" "$ROOT$RELPATH@$VERSION" 
        then
	    if ! svn log -v "-r$VERSION:$rev" "$ROOT$RELPATH" 2>&1
	    then
		echo "ERROR:"
		echo "       No history available for $RELPATH"
		echo ""
	    fi
        fi |
	sed -e "/^r[0-9]\+/,/^\$/{
				     /^r[0-9]\+/ {
						   #
						   #  save the change
						   #  revision
						   #
						   s/^\(r[0-9]\+\).*/\1/1
						   h
						 }

				     /^  *[AM] \//{
						      s/[()]/@/g # make there are no parens
								 # to avoid script errors

						      s/.*/[x!]&$TAB/1  # make this line executable

						      G # append the change rev from the hold
							# buffer

						      s/\n//1
						  }
                                }
	       "
    
    done



}


#beginHelpREPOLOG     DO NOT DELETE THE HELP SECTION!
#  You are viewing an svn log of the changes made to
#  a file or directory.  
#
#  All changes made to the file that affect the
#  selected revision are shown.
#
#  The following commands are applicable:
#
#    q and ESC to exit this dispaly
# 
#            
#endHelpREPOLOG

displayREPOLOG()
{
    # display the whole history of a file or directory

    STATE=REPOHIST      # not REPOLOG!
    CMDKEYS=""
    TITLE="REPOHIST:$RELPATH@$VERSION"
    printStateVariables
    
    
    echo "${TAB}[x]History${TAB}Version"
    
    #
    #  Log the history of specified object within the current branch
    #

    
    if ! svn log -v -r$VERSION:1 "$ROOT$RELPATH@$VERSION"
    then
	if ! svn log -v -r$VERSION:1 "$ROOT$RELPATH" 2>&1
	then
	    echo "ERROR:"
	    echo "       No history available for $RELPATH"
	    echo ""
	fi
    fi |
    sed -e "/^r[0-9]\+/,/^\$/{
			       /^r[0-9]* /{
					     #
					     #  on version number lines, save the number
					     #  to the hold buffer
					     #
					     s/ .*//g  # delete all but the version number
					     h
					  }

			       /^  *[AM] \//{
						s/[()]/@/g
						s/.*/[x!]&$TAB/1
				       
						G # append hold to pattern
						s/\n//1  # remove the newline separator

						
					    }
                            }
	   "
    
    



}


fileREPODIR()
{

    # display a URL as either a file or a directory
    # depending on whether or not the name ends in /


    file="$1"
    cmd="$2"

    RELPATH="$RELPATH$file"  # relpath has a trailing /

    case "$RELPATH" in
       */)
	    displayREPODIR
	    ;;
       *)
	   displayREPOFILE
	   ;;
    esac

}

#beginHelpREPOBRACH     DO NOT DELETE THE HELP SECTION!
#
#  You are viewing a listing of the ancestor branches
#  starting with the repository branch directory currently
#  viewed.  You may select an ancestor for viewing.
#  
#
#    Enter - View the selected ancestor branch directory.
#                                                          .
#endHelpREPOBRACH

brachiateREPODIR()
{
    # display last 4 ancestors branches of the current branch

    executable=`which svnBrachiation`

    if [ "$executable" = "" ]
    then
	 echo "|Error"
	 echo ""
	 echo "Cannot find script 'svnBrachiation' in your path"
	 echo ""
    else
	STATE=REPOBRACH
	TITLE="BRANCHIATION:$ROOT$RELPATH@$VERSION"
	CMDKEYS=""
	printStateVariables
	echo "${TAB}[x]Brachiation"
	svnBrachiation "$ROOT$RELPATH" | sed -e '1!{s/.*/[x!]&/1;}'
    fi


}

handleREPOBRACH()
{
    cmd="$1"
    selection="$6"

    v="${selection#*@}"
    p="${selection%@*}"
    r="${p#$ROOT}"


    case "$cmd" in

       ^J) RELPATH="$r"
	   VERSION=$v
	   displayREPODIR
	   ;;

       *)
	   echo "$TAB[x!]Invalid REPOBRACH state command '$cmd'"
           for f in "$@" 
           do
              echo "$f"
           done
	   echo "ROOT=$ROOT"
	   echo "RELPATH=$RELPATH"
	   echo "VERSION=$VERSION"
	   echo "r=$r"
	   echo "v=$v"
	   ;;

    esac
}

handleLOCALBRACH()
{
    cmd="$1"
    selection="$6"

    v="${selection#*@}"
    p="${selection%@*}"
    r="${p#$ROOT}"


    case "$cmd" in

       ^J) RELPATH="$r"
	   VERSION=$v
	   displayREPODIR
	   ;;

       *)
	   echo "$TAB[x!]Invalid LOCALBRACH state command '$cmd'"
           for f in "$@" 
           do
              echo "$f"
           done
	   echo "ROOT=$ROOT"
	   echo "RELPATH=$RELPATH"
	   echo "VERSION=$VERSION"
	   echo "r=$r"
	   echo "v=$v"
	   ;;

    esac


}

annotateREPODIR()
{
    file="$1"
    cmd="$2"

    RELPATH="$RELPATH$file"  # relpath has a trailing /

    case "$RELPATH" in
       */)
	    echo "|Error"
	    echo ""
	    echo "Cannot annotate directories"
	    echo ""
	    ;;
       *)
	   TITLE="$RELPATH@$VERSION"
           STATE=REPOANN
	   printStateVariables
	   echo "${TAB}[x]Annotation"

	   {
	       svn annotate "$ROOT$RELPATH@$VERSION" 2>/dev/null

	       if [ $? != 0 ]
	       then
		  if ! svn annotate -r$VERSION "$ROOT$RELPATH" 2>&1
		  then
		     echo "<data not available>"
		  fi


	       fi


	   } | expand |
	   sed -e 's/.*/[x!]&/1'

	   ;;
    esac

}

diffREPODIR()
{
    file="$1"
    cmd="$2"

    RELPATH="$RELPATH$file"  # relpath has a trailing /
    TITLE="$RELPATH@$VERSION"
    printStateVariables
    echo "${TAB}Diff of just version $VERSION"
    
    {
	svn diff -x --ignore-eol-style -x -b -x -w -c$VERSION "$ROOT$RELPATH@$VERSION" 2>/dev/null
    
	if [ $? != 0 ]
	then

           if ! svn diff -x --ignore-eol-style -x -b -x -w  -c$VERSION "$ROOT$RELPATH" 2>&1
	   then
	      echo "<data not available>"
	   fi
    
	fi
    
    
    } | expand

    echo "*EOF"


}

# help will come from the displayREPOHIST function

displayProjectHistory()
{
   project="${RELPATH#/}"
   project="${project%%/*}"

   log="$HOME/.tools_home/STV-SVN-PROJECT-$project.log"

   if [ ! -r "$log" ]
   then
       if ! svn log -v "$SVN/$project" >"$log" 2>&1
       then
	   rm -f "$log"
	   echo "|Error:  could not create change log"
	   echo ""
	   echo "svn log command:  svn log -v $SVN/$project >$log"
	   echo ""
	   return 1
       fi
   fi


    STATE=REPOHIST
    CMDKEYS=""
    TITLE="project log: $log"
    printStateVariables
    
    echo "${TAB}[x]History${TAB}Revision"

    cat "$log" |
	sed -e "/^r[0-9]\+/,/^\$/{
				     /^r[0-9]\+/ {
						   #
						   #  save the change
						   #  revision
						   #
						   s/^\(r[0-9]\+\).*/\1/1
						   h
						 }

				     /^  *[AM] \//{
						      s/[()]/@/g # make there are no parens
								 # to avoid script errors

						      s/.*/[x!]&$TAB/1  # make this line executable

						      G # append the change rev from the hold
							# buffer

						      s/\n//1
						  }
                                }
	       "
}

#beginHelpREPODIRMOD     DO NOT DELETE THE HELP SECTION!
#
#  You are viewing a page containing a list of modifications to the file
#  under the cursor.  This is an "svn log" containing only actual changes
#  to the data in the file as opposed to just any old modification
#  which might include property changes, branch name changes, etc.
#
#
#    Enter - View the changes in the version under the cursor
#                                                                           .
#endHelpREPODIRMOD
modificationsREPODIR()
{

    script=`which svnListChanges`

    if [ "$script" = "" ]
    then

	echo "|Error"
	echo ""
	echo "Cannot find script, svnListChanges"

	return
    fi
    
    # display the branch specific log of the currently display RELPATH

    STATE=REPODIRMOD
    CMDKEYS=""
    TITLE="Mods:$RELPATH@$VERSION"
    printStateVariables

    echo "${TAB}[x]File Content Changes"

    "$script" "$ROOT$RELPATH@$VERSION" |
    sed -e '
	     s/^r[0-9]* |.*/[x!]&/1
	   '

    echo "*EOF"
    


}

handleREPODIRMOD()
{
    echo "${TAB}Changed Lines"

    version="${6%% *}"

    if ! svn diff -c$version "$ROOT$RELPATH@$version" 2>/dev/null
    then
	svn diff -c$version "$ROOT$RELPATH" 2>&1
    fi

    echo "*EOF"


}


handleREPODIR()
{
   # expecting command line arguments like this:
   #
   # common components
   #
   #   1=user command key pressed (^J=enter)
   #   2=current row
   #   3=current col
   #   4=rows
   #   5=cols
   #   6=current cell contents
   #
   # page specific parts (see displayREPOROOT)
   #
   #   7 ... 99 = all cells on this row
   #      7=version
   #      8=user
   #      9=size
   #      10=month
   #      11=day
   #      12=year
   #      13=file

   savedVersion="$VERSION"  # some commands don't use the version on the current line
			    # of the directory display

   cmd="$1"
   VERSION="$7"
   VERSION="${VERSION# }"  # remove leading and trailing blank injected for display formatting.
   VERSION="${VERSION% }"  #  see displayREPODIR

   shift 10
   file="$@"


   case "$cmd" in
       "^J") fileREPODIR "$file" "$cmd"
	     ;;

       a)    annotateREPODIR "$file" $cmd
	     ;;

       d)    diffREPODIR "$file" "$cmd"
	     ;;

       h)    RELPATH="$RELPATH$file"
	     displayREPOHIST  # display branch log of file under cursor
	     ;;

       H)    RELPATH="$RELPATH$file"
	     displayREPOLOG   # display full log of file under cursor
             ;;

       l)    VERSION="$savedVersion"
	     displayREPOHIST  # display current branch log of this directory
	     ;;

       m)    VERSION="$savedVersion"
	     RELPATH="$RELPATH$file"
	     modificationsREPODIR
	     ;;

       R)    displayREPOROOT
             ;;

       B)    brachiateREPODIR
	     ;;

       P)    displayProjectHistory
	     ;;

       *)
	   echo "$TAB[x!]Invalid REPODIR state command '$cmd'"
           for f in "$@" 
           do
              echo "$f"
           done
	   ;;
   esac
}


annotateREPOFILE()
{
    #
    # this assumes that the user was looking at a file from the repository and pressed the 'a' key.
    #

    TITLE="$RELPATH"
    STATE=REPOANN
    CMDKEYS="ahHdD"
    printStateVariables

    echo "${TAB}[x]Annotation"


    { 
	{
	    if ! svn annotate "$ROOT$RELPATH@$VERSION"  2>/dev/null
            then

	       if ! svn annotate -r$VERSION "$ROOT$RELPATH"  2>/dev/null # sometimes svn gives error when you use the @VERSION in the url
               then
                   echo "<data not found>"  # blank line to make the display work properly
               fi
            fi

	} | expand | sed -e 's/.*/[x!]&/1'




    } 
}

diffREPOFILE()
{
    TITLE="DIFF:$RELPATH@$VERSION"
    STATE=REPOFILE
    CMDKEYS="ahHdD"
    printStateVariables

    echo "${TAB}Diff of just version $VERSION"


    { 
	{
	    if ! svn diff -x -b -x -w -x --ignore-eol-style -c$VERSION "$ROOT$RELPATH@$VERSION"  2>/dev/null
            then

	       if ! svn diff -x -b -x -w -x --ignore-eol-style -c$VERSION "$ROOT$RELPATH"  2>/dev/null # sometimes svn gives error when you use the @VERSION in the url
               then
                   echo "<data not found>"  # blank line to make the display work properly
               fi
            fi

	} | expand | sed -e 's/.*/[x]&/1'


	echo "*EOF"

    } 
}



handleREPOFILE()
{
   cmd="$1"
    
   case "$cmd" in
       h)  displayREPOHIST
	   ;;
       H)  displayREPOLOG
	   ;;

       a)  annotateREPOFILE
	   ;;

       d)  diffREPOFILE
	   ;;

       R)  RELPATH="/"
	   displayREPODIR
	   ;;

       *)
	   echo "$TAB[x!]Invalid REPOFILE state command '$cmd'"
           for f in "$@" 
           do
              echo "$f"
           done
	   echo ""
	   echo "RELPATH=$RELPATH"
	   ;;
   esac
    
}

handleLOCALHIST()
{
   cmd="$1"
   if [ $# -ge 8 ]
   then
	VERSION="$8"
   fi

   RELPATH=`
	      echo "$7" |
	      sed -e '
		       s/^  *[AMRD]  *//1  # remove leading blanks and change type

		       s/  *@from.*//1  # remove trailing copy-from source
		     '
	   `

   isdir=0

   if isRepoDir
   then
	isdir=1
	RELPATH="$RELPATH/"
   else
	isdir=0
   fi


   case "$cmd" in
       ^J) if [ "$isdir" != "0" ]
	   then
                displayREPODIR
	   else
		displayREPOFILE
	   fi
	   ;;
       *)
	   echo "$TAB[x!]Invalid LOCALHIST state command '$cmd'"
           for f in "$@" 
           do
              echo "$f"
           done
	   echo ""
	   echo "RELPATH=$RELPATH"
	   echo "OBJECT=$OBJECT"
	   echo "ROOT=$ROOT"
	   echo "TOPDIR=$TOPDIR"
	   ;;
   esac
   
}

handleREPOHIST()
{
   cmd="$1"
   VERSION="$8" 
   RELPATH=`
	      echo "$7" |
	      sed -e '
		       s/^  *[AMRD]  *//1  # remove leading blanks and change type

		       s/  *@from.*//1  # remove trailing copy-from source
		     '
	   `

   isdir=0

   if isDirREPO "$RELPATH@$VERSION"
   then
	isdir=1
	RELPATH="$RELPATH/"
   else
	isdir=0
   fi


   case "$cmd" in
       ^J) if [ "$isdir" != "0" ]
	   then
		displayREPODIR
	   else
		displayREPOFILE
	   fi
	   ;;
       *)
	   echo "$TAB[x!]Invalid LOCALHIST state command '$cmd'"
           for f in "$@" 
           do
              echo "$f"
           done
	   echo ""
	   echo "RELPATH=$RELPATH"
	   echo "OBJECT=$OBJECT"
	   echo "ROOT=$ROOT"
	   echo "TOPDIR=$TOPDIR"
	   ;;
   esac
}

handleLOCALANN()
{
   #  handle annotation page commands from local directory
   #  annotations -- the variable, OBJECT, contains the name
   #  of the file

   cmd="$1"
   version=$(
		echo "$7" | sed -e 's/^  *//1;  s/ .*//g' 
	    )



   case "$cmd" in

      ^J)
	   TITLE="diff:$OBJECT@"
	   printStateVariables

	   echo "${TAB}Changes in version $version"

	   if ! svn diff -x --ignore-eol-style -x -w -x -b -c$version "$OBJECT" 2>&1 
	   then
	       echo ""
	       echo "Earlier version may not exist -- was the file created in the"
	       echo "selected version?"
	       echo ""
	   fi | expand

	   echo "*EOF"

	   ;;


       *)
	   echo "$TAB[x!]Invalid LOCALANN state (localdir annotation) command '$cmd'"
           for f in "$@" 
           do
              echo "$f"
           done
	   echo ""
	   echo "RELPATH=$RELPATH"
	   echo "OBJECT=$OBJECT"
	   echo "ROOT=$ROOT"
	   echo "TOPDIR=$TOPDIR"
	   ;;
   esac


}


handleREPOANN()
{
   #  handle annotation page commands from a repository file

   cmd="$1"
   version=$(
		echo "$7" | sed -e 's/^  *//1;  s/ .*//g' 
	    )


   case "$cmd" in

      ^J)
	   TITLE="diff:$RELPATH"
	   printStateVariables

	   echo "${TAB}Changes in version $version"

	   if ! svn diff -x --ignore-eol-style -x -w -x -b -c$version "$ROOT$RELPATH" 2>&1
	   then
	       if ! svn diff -x --ignore-eol-style -x -w -x -b -c$version "$ROOT$RELPATH@$version" 2>&1 
	       then
		   echo ""
		   echo "Earlier version may not exist -- was the file created in the"
		   echo "selected version?"
		   echo ""
	       fi 
	   fi | 
	   expand
	   echo "*EOF"

	   ;;


       *)
	   echo "$TAB[x!]Invalid REPOANN state (localdir annotation) command '$cmd'"
           for f in "$@" 
           do
              echo "$f"
           done
	   echo ""
	   echo "RELPATH=$RELPATH"
	   echo "OBJECT=$OBJECT"
	   echo "ROOT=$ROOT"
	   echo "TOPDIR=$TOPDIR"
	   ;;
   esac


}


#################################################################
displaystart()
{
    if [ -d .svn ]
    then
       OBJECT="$TOPDIR$RELPATH"
       displayLOCALDIR
    else
       if [ "$ROOT" = "root" ]
       then
	  echo "|Error:  ROOT variable is not set"
	  echo "Press Escape to exit"
       else
	  displayREPOROOT
       fi
    fi
}

#################################################################
#################################################################
#################################################################
# main program body

if [ "$#" = "0" ]
then
    #
    #  first time through, handle the start state
    #
    displaystart
else

    # create a command as a string then evaluate the command.

    cmd="handle$STATE"            # start with the command name which is the word
				  # "handle" followed by the state name.

    for f in "$@"                 # all all the command line parameters to the script
    do
       g="${f//\'/&quot;}"        # we are going to protect the parameters with single quotes
				  # so lets translate the existing single quotes into &quote;

       cmd="$cmd '$g'"            # append all the scripts command line arguments to the
				  # handle function name
    done

    eval "$cmd"                   # execute the handler function
fi

exit 0

