#include <cxxtls/point.h>
#include <cxxtls/ascii_plot.h>
#include <iostream>
#include <fstream>
#include <string>
#include <stdlib.h>
#include <list>
#include <limits>
#include <math.h>
///@file
///  Defines the program, aplot, which reads a list of data points from files
///  and produces and ascii plot of those points.

///@page aplot_program
///
///<h1>APLOT -- a program that plots 2D dimensional points in an ASCII format</h1>
///
///   A plot reads one or more input files and for each produces an ASCII plot
///   of the 2 dimensional data points stored in each input file.


#ifdef _MSC_VER
#define isnan _isnan
#endif

using namespace std;
using namespace cxxtls;

static void printHelp();
static void processFilename(string const &filename);

typedef cxxtls::geometry::Point<2> Point2D;

int majorErrorOccurred=0;

int main(int argc, char **argv)
{
    ++argv;  // skip the program name

    while(*argv)
    {
       string arg(*argv++);

       if(arg.empty() || arg[0] != '-')
       {
          --argv;  // put the argument back in the stream

          break;
       }
       if(arg == "-h" || arg == "--help")
       {
           printHelp();
           exit(1);
       }
       else
       {
           cerr << "error:  invalid option " << arg << endl;
           exit(1);
       }

    }

    while(*argv)
    {
        string arg(*argv++);

        processFilename(arg);

    }



    exit(majorErrorOccurred);

}

static void plotFile(istream &infile);

static void processFilename(string const &filename)
{
   if(filename == "-")
   {
      plotFile(cin);
      return;
   }

   ifstream infile(filename.c_str());

   if(!infile.good())
   {
      cerr << "Error:  unable to open file '" << filename << "'" << endl;

      majorErrorOccurred=true;
      return;
   }

   cout << "Plotting file " << filename << endl;

   plotFile(infile);

}


static void plotFile(istream &infile)
{
    // read the data points from the file
    
    list<Point2D> points;


    Point2D ll, ur;

    size_t lineNumber=0;

    for(;;)
    {

        string line;

        if(!std::getline(infile, line))
           break;

        ++lineNumber;

        if(line.empty())
          continue;

        if(line[0] == '#')
          continue;

        double x = numeric_limits<double>::quiet_NaN();
        double y = numeric_limits<double>::quiet_NaN();

        sscanf(line.c_str(), "%lf %lf", &x, &y);

        if( isnan(x) || isnan(y))
        {
            cerr << "Error at line " << lineNumber << ", invalid floating point numbers" << endl;
        }


        Point2D thisPoint(x,y);

        if(points.empty())
        {
           ll = ur = thisPoint;
        }
        else
        {
             ll = ll.lesser(thisPoint);
            ur = ur.greater(thisPoint);
        }

        points.push_back( thisPoint );
    }

    AsciiPlotter  plotter(ll.get(0), ll.get(1), ur.get(0), ur.get(1), points.size());

    CXXTLS_FOREACH(Point2D const &cur, points)
    {
       plotter.plot(cur.get(0), cur.get(1), '*');
    }

    cout << plotter << endl;



}
static void printHelp()
{
    cout << "aplot:  print an ascii plot of points in 2 dimensions" << endl;
}


