#include <cxxtls/point.h>
#include <cxxtls/GaussianLpSolver.h>
#include <cxxtls/lexical_cast.h>
#include <cxxtls/ascii_plot.h>
#include <cxxtls/strtool.h>
#include <cxxtls/foreach.h>
#include <cxxtls/stlstream.h>

#include <iostream>
#include <stdlib.h>
#include <list>

using namespace std;
using namespace cxxtls;

static void printHelp(); //  print's help and then exits(-1);

int  predictorOrder= 0;
int  futurePoints  = 0;
bool Verbose       = false;    // really verbose
bool verbose       = false;    // a little verbose
bool graph         = false;
size_t maxPoints   = -1;
double scale       = 1;   // y scale when plotting
int    xScale      = 1;   // x scale when plotting

static bool createPredictorEquation(list<double> const &points, list<double> &outputCoefficients, size_t predictorOrder, size_t maxPoints=size_t(-1));
    //  @return true if it is possible to create the predictor equations in the outputParameters vector
    //   from the given data points and required predictorOrder.  On success, outputParameters.size() should
    //   be predictorOrder.
    //
    //  @param points  -- a list of input data points for use in constructing the equation's outputCoefficents.  Not all input points
    //                    may be used.  See maxPoints.
    //
    //  @param outputCoefficients -- a list of output coefficients produced by this function from the input @ref points list.
    //
    //  @param predictorOrder -- the number of outputCoefficients to be produced.
    //  @param maxPoints -- the maximum number of input points to be used.  The default is infinite.
    //

static double computeErrors(list<double> const &points, list<double> &equationCoefficients, list<double> &errorOutput);
    //  Given an equation, which is derived from the points, reuse the points as way of determining how accurate
    //  the equation is.  This only works if the number of points is greater than the predictorOrder -- otherwise, you
    //  won't have any points to compare against.  The more the better for error detection purposes.
    //
    //  The approach is simple, given the equation, start with the first predictorOrder count of points, assume the error is zero
    //  within that group, and use the points in that range to predict the next point.  Since we know the correct answer
    //  we can then compute the error at that point.  We repeat for all other points that we know.  This allows us to 
    //  create the errorOutput list.  As we do this, we can compute the standard deviation of the errors and return it.
    //
    //  With the errorOutput in hand, one can then use this data as a new series and create a predictor equation from it
    //  and use that equation as a guess as to how bad the equationCoefficients perform.
    // 

static void predict(list<double> const &points, list<double> &equation, int extraPoints, list<double> &output);
   // given a list of input points and a modeling equation, compute extraPoints new points.  Insert the
   // original list of points into the output and then follow them by the predicted points.

static void plotTimeSeries(list<double> const &data, size_t inputCount, list<double> const &errorPrediction);



bool enoughPoints(int predictorOrder, size_t pointCount, size_t &minNeeded)
    // return false if there are not enough points to perform a prediction
{
    minNeeded = predictorOrder * 2;

    return size_t(pointCount) >= (minNeeded);
}


int main(int argc, char **argv)
{
    ++argv;

    while(*argv)
    {
       string arg(*argv++);

       if(arg == "-h")
       {
          printHelp(); // and die
       }
       else
       if(arg == "-V")
       {
           Verbose = true;
           verbose = true;
       }
       else
       if(arg == "-v")
       {
           verbose = true;
       }
       else
       if(arg == "-g")
       {
           graph = true;
       }
       else
       if(arg == "-p" && *argv)
       {
          predictorOrder = lexical_cast<int>(*argv++);
       }
       else
       if(arg == "-f" && *argv)
       {
          futurePoints = lexical_cast<int>(*argv++);
       }
       else
       if(arg == "-m" && *argv)
       {
          maxPoints = lexical_cast<int>(*argv++);
       }
       else
       if(arg == "-y" && *argv)
       {
          scale = lexical_cast<double>(*argv++);
       }
       else
       if(arg == "-x" && *argv)
       {
          xScale = lexical_cast<int>(*argv++);
       }

    }

    if(predictorOrder == 0 || futurePoints == 0)
       printHelp(); // and die


    list<double> points;

    string line;

    int lineNo = 0;

    while(std::getline(cin, line))
    {

       ++lineNo;

       StrTool::stringlist_t strings;

       StrTool::parse_words(line, &strings, 100000, StrTool::Is_Space());

       CXXTLS_FOREACH(string const &cur, strings)
       {
           double d;

           int count = sscanf(cur.c_str(), "%lf", &d);

           if(count)
              points.push_back(scale * d);
          else
          {
             cout << "Error:  at line " << lineNo << ":  expected floating point number, but found:  " << cur << endl;
             exit(1);
          }
       }
    }

    size_t minNeeded=0;

    if(!enoughPoints(predictorOrder, points.size(), minNeeded) )
    {
       cout << "Error:  predictorOrder of " << predictorOrder << " requires " << minNeeded << " points of data\n"
            << "        but only " << points.size() << " series points have been provided." 
            << endl;

       exit(1);
    }

    if(maxPoints < minNeeded )
    {
       cout << "Error:  predictorOrder of " << predictorOrder << " requires " << minNeeded << " points of data\n"
            << "        but only maxPoints= " << maxPoints << " have been provided " 
            << endl;

       exit(1);
    }


    if(verbose)
    {
        cout << "Input Points" << endl;

        CXXTLS_FOREACH(double const &cur, points)
        {
           cout << cur << ' ';
        }
        
        cout << endl;
    }

    list<double> equation;
    list<double> errors;


    if(Verbose)
        cout << "Creating primary predictor equation" << endl;

    bool found = false;

    double minStdDev=1e139;

    size_t workingPoints = points.size();
    size_t foundPoints   = 0;

    if(maxPoints < workingPoints)
       workingPoints = maxPoints;

    while(workingPoints > minNeeded)
    {
        if(createPredictorEquation(points, equation, predictorOrder, workingPoints))
        {
            list<double>  tmpErrors;

            double errorStdDev = computeErrors(points, equation, tmpErrors); 

            if(errorStdDev < minStdDev)
            {
               minStdDev = errorStdDev;
               foundPoints = workingPoints;
               found=1;
               errors.swap(tmpErrors);
            }
        }
    
        --workingPoints;
    
    }

    if(!found)
    {
           cout << "Unable to create predictor equation from the last " << 2 * predictorOrder << " points" << endl;
           exit(1);
    }

    maxPoints = foundPoints;

    if(verbose)
    {
        cout << "Min error stddev maxPoints = " << maxPoints << endl;
        
        cout << "Predicted error points\n\t" ;
        
        CXXTLS_FOREACH(double const &cur, errors)
        {
           cout << cur << ' ';
        }
        
        cout << endl;
    }

    list<double> errorEquation;

    if(Verbose)
        cout << "Creating error predictor equation" << endl;


    if(errorEquation.empty())
    {
        workingPoints = maxPoints;

        int errorPredictionOrder = predictorOrder;

        while(errorPredictionOrder > 0)
        {
            list<double> tmpErrorEquation;
        
            if(createPredictorEquation(errors, tmpErrorEquation, errorPredictionOrder, workingPoints))
            {
                errorEquation.swap(tmpErrorEquation);
                break;
            }
        
            --errorPredictionOrder;
        
        }
    }


    if(errorEquation.empty())
    {
        workingPoints = maxPoints;
        
        while(workingPoints > minNeeded)
        {
            list<double> tmpErrorEquation;
        
            if(createPredictorEquation(errors, tmpErrorEquation, predictorOrder, workingPoints))
            {
                errorEquation.swap(tmpErrorEquation);
                break;
            }
        
            --workingPoints;
        
        }
    }


    if(errorEquation.empty())
    {
       cout << "Unable to create predictor equation for the errors from the last " << 2 * predictorOrder << " points" << endl;
    }

    list<double> output;

    predict(points, equation, futurePoints, output);

    cout << "Predicted output points\n\t" ;

    CXXTLS_FOREACH(double const &cur, output)
    {
       cout << cur << ' ';
    }

    cout << endl;

    list<double> errorPrediction;

    predict(errors, errorEquation, futurePoints, errorPrediction);

    if(verbose)
    {
        cout << "Standard Deviation of the Error:  " << minStdDev << endl;
        
        cout << "Predicted output Errors\n\t" ;
        
        CXXTLS_FOREACH(double const &cur, errorPrediction)
        {
           cout << cur << ' ';
        }
        
        cout << endl;
    }


    if(graph)
    {
        cout << endl;
        cout << "Graph of Points";
        plotTimeSeries(output, points.size(), errorPrediction);
    }



}


static void printHelp()
{
     cout << 
     "\n"
     "   predict:  Given a sequence of data points, a predictive equation size,  and a count of\n"
     "             future points to forecast, produce future output points.  Assume that the input\n"
     "             data points are a time series. \n"
     "\n"
     "\n"
     "   usage:    predict  -p predictiveEquationSize  -f futurePoints <inputPoints  [otherOptions] \n"
     "\n"
     "   Description:\n"
     "\n"
     "      Given the number of points to include in th predictive equation and a sequence of known \n"
     "      data points which is 2 * predictiveEquationSize or higher, predict futurePoints new data\n"
     "      points as inferred by the inputPoints.\n"
     "\n"
     "      Input points are observed values of some real world function\n"
     "\n"
     "   Other Options: \n"
     "     -h   Get Help\n"
     "     -v   Low verbosity level\n"
     "     -V   High verbosity\n"
     "     -g   Print a graph of the data points, see -x and -y\n"
     "     -m p Limit the predictor equation to p points (1 ... p)\n"
     "     -x v When graphing, scale x values by v\n"
     "     -y v when graphing, scale y values by v\n"
     << endl;
     ;



     exit(1);
}

static bool createPredictorEquation(list<double> const &points, list<double> &outputParameters, size_t predictorOrder, size_t maxPoints)
    /// @return true if it is possible to contruct a predictor equation whose parameters will be writen to
    /// outputParameters.
{
    outputParameters.resize(0);

    // To create a predictor equation, of order 2, you need 4 data points, at least.
    //
    // Lets assume that you have a sequence of input points like this:
    // 
    //  points[0]
    //  points[1]
    //  points[2]
    //  points[3]
    //
    //  Which means that we have 5 points from which to infer the predictor equation of order 2.
    //
    //  Here is the set of simultaneous equations that we will solve for in order to produce the order 2
    //  predictor equation:
    //
    //      points[3] = points[2] * X +  points[1] * Y;
    //      points[2] = points[1] * X +  points[0] * Y;
    //
    //  Note that is is sufficient for a solution in most cases.
    //
    //  Then, with solved values for X and Y, we can preduct any future point, say point 4, given points 
    //  point[3] and point[2].  Like this:
    //
    //     points[4] = points[3] * X + points[2] * Y;
    //  
    //  Of course, the question remains, "How do we compute X and Y?"
    //
    //  And since the above equation is only a prediction, how bad is the error?
    //
    //  Luckily, a famous mathmetician, Calrf Friedrich Gauss, figured out to solve linear equations, around
    //  1808, and we can thus use his method by contructing a class of type GaussianLpSolver solver(2);  We
    //  can then invoke the solve() method and extract out the values of X and Y.
    //

    GaussianLpSolver solver(predictorOrder);

    size_t minNeeded=0;

    if(!enoughPoints(predictorOrder, points.size(), minNeeded))
       throw int(predictorOrder);

    if(!enoughPoints(predictorOrder, maxPoints, minNeeded))
       throw int(predictorOrder);


    // to compute the predictor equation, we are going to work like this:
    //
    // There will be predictorOrder equations in the solver.  Each each equation has a 
    // rigth hand side which defines the block of points used in an equation.  Each equation's
    // right hand side value is a one location less than the previous equation's right hand side.



    list<double>::const_iterator blockStartIterator = points.end();

    if(maxPoints < points.size())
    {
        // if the max points is being limited, only create an equation based on the limited subset.


        blockStartIterator = points.begin();

        size_t count = maxPoints;

        while(count--)
        {
            blockStartIterator++;
        }
    }


    --blockStartIterator;

    for(size_t row = 0 ; row < predictorOrder; ++row )
    {
        list<double>::const_iterator parameterIterator = blockStartIterator;

        solver.setRHS(row, *parameterIterator--);

        for(size_t param = 0;  param < predictorOrder; ++param)
        {
           solver.setParameter(param, row, *parameterIterator--);
        }


        --blockStartIterator; // setup for the next row
    }

    // at this point, we have created predictorOrder equations, with parameters
    // for predictorOrder unknowns, plus 1 RHS for each equation.


    if(!solver.solve())
    {
        if(Verbose)
        {
           cout << "Unable to solve predictor equation set:";
           cout << solver << endl;
        }

        return false;  // could not solve the equations.
    }
    else
    if(Verbose)
    {
       cout << "Predictor equation matrix" << endl;
       cout << solver << endl;
    }

    for(size_t row = 0; row < predictorOrder; ++row)
    {
        outputParameters.push_back(solver.result(row));
    }

    return true;
}

static void predict(list<double> const &points, list<double> &equation, int extraPoints, list<double> &output)
{

    output.resize(0);

    if(points.size() < equation.size())
       throw int(points.size());

    if(equation.empty())
       return;

    output = points;  // copy input to output

    while(extraPoints--)
    {
       // repeatedly add 1 extra point to the data in the output vector
       // which consists of the data from the input vector plus extra points created, 1 at  time.
       //
       // Note that when each new point is added into the output, any previously predicted points are included
       // in the equations, so that the error in the prediction gets worse over time...

       list<double>::iterator last = output.end(); --last;

       list<double>::const_iterator eqf = equation.begin(), eql = equation.end();

       double newPoint = 0;

       while(eqf != eql)
       {
           newPoint += *eqf++ * *last--;
       }

       output.push_back(newPoint);

    }




}

typedef cxxtls::geometry::Point<2> Point2D;


static void plotTimeSeries(list<double> const &data, size_t inputCount, list<double> const &errorPrediction)
{

    bool printErrorEnvelope = (errorPrediction.size() == data.size());
    
    list<Point2D> points;
    list<Point2D> errorUpperBound;
    list<Point2D> errorLowerBound;

    Point2D ll, ur;

    int count=0;

    list<double>::const_iterator errorPtr = errorPrediction.begin();

    for(list<double>::const_iterator f = data.begin(), l = data.end();
        f != l;
        ++f, ++count
       )
    {
        int x = xScale * count;

        Point2D thisPoint(x,*f);

        if(points.empty())
        {
           ll = ur = thisPoint;
        }
        else
        {
             ll = ll.lesser(thisPoint);
             ur = ur.greater(thisPoint);
        }

        points.push_back( thisPoint );

        if(printErrorEnvelope)
        {
           Point2D upperbound = thisPoint;
           Point2D lowerbound = thisPoint;

           upperbound.set(1,1 + ceil(*f + *errorPtr));
           lowerbound.set(1,-1 + floor(*f - *errorPtr));

           ll = ll.lesser(lowerbound);
           ur = ur.greater(upperbound);
           
           ++errorPtr;
        }


    }


    AsciiPlotter  plotter(ll.get(0), ll.get(1), ur.get(0), ur.get(1), points.size());


    errorPtr = errorPrediction.begin();

    CXXTLS_FOREACH(Point2D const &cur, points)
    {

       if(printErrorEnvelope)
       {
           double top = ( cur.get(1) + *errorPtr);
           double bot = ( cur.get(1) - *errorPtr);


           plotter.plot(cur.get(0), top, '^');
           plotter.plot(cur.get(0), bot, 'v');


           ++errorPtr;
       }

       char c = '@';

       if(inputCount)
       {
          --inputCount;
          c = '*';  // indicating an input point
       }

       plotter.plot(cur.get(0), cur.get(1), c);

    }

    cout << plotter << endl;



}

static double computeErrors(list<double> const &points, list<double> &equationCoefficients, list<double> &errorOutput)
    //  Given an equation, which is derived from the points, reuse the points as way of determining how accurate
    //  the equation is.  This only works if the number of points is greater than the predictorOrder -- otherwise, you
    //  won't have any points to compare against.  The more the better for error detection purposes.
    //
    //  The approach is simple, given the equation, start with the first predictorOrder count of points, assume the error is zero
    //  within that group, and use the points in that range to predict the next point.  Since we know the correct answer, it is the
    //  value at points[predictorOrder], we can then compute the error at that point.  We repeat for all other points that we know.  
    //  This allows us to create the errorOutput list.  As we do this, we can compute the standard deviation of the errors and return it.
    //
    //  With the errorOutput in hand, one can then use this data as a new series and create a predictor equation from it
    //  and use that equation as a guess as to how bad the equationCoefficients perform.
    // 
    //  @note  errorOutput values are always >= 0.
{
    errorOutput.resize(0);

    if(points.size() < equationCoefficients.size())
       throw int(points.size());

    list<double>::const_iterator startPoint = points.begin(), endPoint = points.end();

    size_t pointNumber = 0;

    for(size_t i = 0; i < equationCoefficients.size(); ++i)
    {

       ++startPoint;
       errorOutput.push_back(0);
    }

    double sumOfSquares=0;

    while(startPoint != endPoint)
    {
       ++pointNumber;

       list<double>::const_iterator last = startPoint;

       // evaluate the predictor equation on the known points just before start

       --last;   

       list<double>::const_iterator eqf = equationCoefficients.begin(), eql = equationCoefficients.end();

       double newPoint = 0;  // ie the point that should be at *startPoint

       while(eqf != eql)
       {
           newPoint += *eqf++ * *last--;
       }

       double diff = *startPoint - newPoint;

       sumOfSquares += (diff * diff);

       double errorValue = sqrt(sumOfSquares);

       errorOutput.push_back(errorValue);

       ++startPoint;
     }

   // return sqrt( sumOfSquares / points.size() );
   return sqrt( sumOfSquares );
}

