#ifdef __linux__
//
// Copyright 2013, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.


//@File
//
///   This file ties together a PseudoTerminal, threads, and a ControlledProcess
///   to turn gdb into a slave that drives a child process and captures pure gdb output
///   separately from child program output.
///
///   This is not a good encapsulation, it is just a mostly working example.
///
///   It lets you enter commands from the command line and sends them to a controlled
///   gdb process which is writing child process output to the pseudo terminal.
///   This is important because it is allows pure separation of debugged program output
///   from gdb's output which can be difficult to do otherwise.
///
///   To demonstrate the program, run make at the cxx directory level, then run the
///   program like this:
///@code
///
///   bin/gdb
///
///@endcode
///
///   When the gdb prompt is presented, "|< (gdb) ", enter "file bin/cpptagdb.exe".
///   This will load program bin/cpptagdb.exe but not start it running.  To run it, enter
///   this command:
///@code
///
///   run bin/*.cxx
///
///@endcode
///
///   Gdb will then run the child process and it's output will be prefixed with "|{ ".  
///
///   When the output completes, type exit and hit return a couple of times.
///   


#include <cxxtls/pseudoterminal.h>
#include <cxxtls/controlledprocess.h>
#include <cxxtls/thread.h>
#include <stdlib.h>
#include <iostream>
#include <memory>
#include <time.h>
#include <unistd.h>

using namespace cxxtls;
using namespace std;

void shortNap()
{
     struct timespec delay, rv;

     delay.tv_sec = 0;
     delay.tv_nsec=  100 * 1000 * 1000;   // 100,000,000 nanoseconds or 1/10 of a second

     nanosleep(&delay, &rv);
}


static void *readAndPrintFILE(void *p)
{
    // p is assumed to be a normal FILE*

    FILE *f = reinterpret_cast<FILE*>(p);

    CXXTLS_DEBUG("starting[readAndPrintFILE]");

    char buffer[20000];

    while(fgets(buffer, sizeof(buffer), f))
    {
        CXXTLS_COUT_LOCK(var);

        cout << "|< " << buffer; // buffer alread has a \n
    }



    CXXTLS_DEBUG("exiting[readAndPrintFILE]");

    return 0;
}

static void *readAndPrintFD(void *p)
{
    // p is assumed to be an int * which points to a file descriptor

    int *fdPtr = reinterpret_cast<int*>(p);

    int fd = *fdPtr;

    CXXTLS_DEBUG("starting[readAndPrintFD]");

    bool nl =true;

    for(;;)
    {
      char c;

      int rc = read(fd, &c, 1);

      if(rc < 0)
          break;  // end of file

      if(rc == 0)
      {
         shortNap();
         continue;
      }

      CXXTLS_COUT_LOCK(var);

      if(nl)
      {
          nl = false;
         cout << "|{ ";
      }

      cout << char(c);

      if(c == '\n' || c == '\r')
      {
         nl = true;
      }
    }

    CXXTLS_DEBUG("exiting[readAndPrintFD]");

    return 0;
}


int main(int argc, char **argv)
{
#ifndef __linux__
  cout << "OOh!  No no no no no!  No can do big guy!  No way, Jose Cuervo" << endl;
  exit(1);
#endif

  string line;

  PseudoTerminal pt;

  if(pt.open(false) < PseudoTerminal::good)
  {
    fprintf(stderr, "couldn't create pseudo terminal\n");
    exit(1);
  }

  int masterFD = pt.masterFD();

  ThreadId slaveReader = createThread(readAndPrintFD, &masterFD);

  string command = "exec gdb --interpreter=mi";

  command += " -tty ";
  command += pt.slaveTerminalName();

  
  std::auto_ptr<ControlledProcess> 
     cpPtr(ControlledProcess::start(command, 
                                    ControlledProcess::LINE_BUFFERED, 
                                    ControlledProcess::CHARACTER_BUFFERED));
  if(cpPtr->status() != 0)
  {
    fprintf(stderr, "couldn't start process\n");
    exit(1);
  }

  ThreadId gdbReader = createThread(readAndPrintFILE, cpPtr->process_stdout());

  while(std::getline(cin, line))
  {
      fputs(line.c_str(), cpPtr->process_stdin());
      fputs("\n",         cpPtr->process_stdin());

  }

  joinThread(slaveReader);
  joinThread(gdbReader);

}

#else
    int main()
    {
       // on windows do nothing
    }
#endif
