//
// Copyright 2004, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.

///@file
/// This file implementa a program, tree, that prints directory
/// listings to stdout based on command line options.  Its output
/// is intended to draw a picture of the directory structure,
/// not just list the names.  Line drawing characters are used to
/// describe the selected directory structures.

#include <cxxtls/file.h>
#include <iostream>
#include <iomanip>
#include <cxxtls/options.h>
#include <set>

extern "C"
{
#include <fnmatch.h>
};


using namespace std;
using namespace cxxtls;

typedef FileName::Tree                   Tree;
typedef std::auto_ptr<Tree>        TreePtr;
typedef Tree::children_t::const_iterator const_iterator;
typedef Tree::children_t::iterator       iterator;

bool include_attributes=false;
bool include_time=false;
bool include_size=false;

set<string> ignored_names;
set<string> name_filters;

void print_children(Tree const* p, string prefix)
{
  const_iterator f=p->children().begin(), l = p->children().end();

  while(f != l)
  {
    Tree const *child = *f++;


    {
      // compare the current file name against all ignore patterns

      set<string>::const_iterator a, b;

      bool match=false;

      for(a = ignored_names.begin(), b = ignored_names.end(); a != b; ++a)
      {
        if(0 == fnmatch(const_cast<char*>(a->c_str()), 
                        const_cast<char*>(child->name().c_str()),
                        1) )
        {
          match=true;
          break;
        }
      }

      if(match)
        continue; 

    }

    if(!child->status().mode.is_dir())
    {
      // compare the current file name against all required patterns

      set<string>::const_iterator a, b;

      bool match=false;

      for(a = name_filters.begin(), b = name_filters.end(); a != b; ++a)
      {
        if(0 != fnmatch(const_cast<char*>(a->c_str()), 
                        const_cast<char*>(child->name().c_str()),
                        1) )
        {
          match=true;
          break;
        }
      }

      if(match)
        continue; 

    }



    bool is_last = (f==l);

    char c = is_last ? '`' : '|';

    if(include_attributes)
      cout << child->status().mode << " ";

    if(include_time)
      cout << child->status().time << " ";

    if(include_size)
      cout << setw(8) << child->status().size << " ";

    cout << prefix << c << "-- " << child->name() << endl;

    string new_prefix = prefix;

    if(is_last)  
      new_prefix += "    ";
    else
      new_prefix += "|   ";


    print_children(child, new_prefix);

  }


}


int main(int argc, char **argv, char **environ)
{

  ProgramOptions options("-d,-h,-a,-t,-s,-i;-p;-b;-A", argc, argv, environ);

  if(options.option("-h"))
  {
    cout << "tree.exe:  Output a directory tree either starting with the\n"
         << "           current directory or with a named directory.\n"
         << "\n"
         << "Usage:\n"
         << "\n"
         << "  tree.exe [options] [directory_name]\n"
         << "\n"
         << "Options:\n"
         << "  -h:  produce this help message\n"
         << "  -A:  all files\n"
         << "  -d:  show only directories in the output\n"
         << "  -a:  include file attributes\n"
         << "  -t:  include file modification time\n"
         << "  -s:  include file size\n"
         << "  -i pattern: ignore files/dirs with this filename pattern\n"
         << "  -p pattern: only show files named 'n'\n"
         << "  -b:  print debug information\n"
         << endl;
    return 0;
  }

  if(options.option("-a"))
    include_attributes=true;

  string dirp=".";

  if(options.argv.size() > 1)
    dirp = options.argv[1];
  
  bool only_dirs = false;

  if(options.option("-d"))
    only_dirs = true;

  if(options.option("-t"))
    include_time = true;

  if(options.option("-a"))
    include_attributes = true;

  if(options.option("-s"))
    include_size = true;

  if(options.option("-i"))
  {
    ProgramOptions::value_iterator f, l;

    f = options.option("-i").begin();
    l = options.option("-i").end();

    while(f !=l)
    {
      ignored_names.insert(*f);
      ++f;
    }


  }

  if(options.option("-p"))
  {
    ProgramOptions::value_iterator f, l;

    f = options.option("-p").begin();
    l = options.option("-p").end();

    while(f !=l)
    {
      name_filters.insert(*f);
      ++f;
    }


  }


  FileName dir(dirp);

  bool debug=false;

  if(options.option("-b"))
     debug=true;

  bool allFiles=false;

  if(options.option("-A"))
     allFiles=true;

  TreePtr tree(dir.tree(only_dirs, allFiles, debug));

  cout << tree->name() << endl;

  print_children(tree.get(), "");


}
