//
// Copyright 2002-2013, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.

//@file
///
///   This file defines the implementation of a viewer which is a text _editor_.
///   That is, you don't just get to look at the file, you get to change it
///   as well.  For consistencey sake the cursor movement and paging keys
///   obey the standard convention in CursorWindow::func_*.  The func map
///   also properly defines the function key mapping.
/// 
///   There are however additional function key mappings which would not
///   apply to a pure viewer.
/// 
///   These additional mappings allow for inserting and deleting characters
///   as well as replacing strings, deleting lines, etc.  Whenever possible,
///   single key bindings are used, but there are simply more functions than
///   can reasonably be bound to one key stroke -- and thus multi-key bindings
///   are required.  These bindings are typically attached to both a
///   function key and (F1 is help, for example, F2, is next window, etc.).
/// 
///   The Editor is implemented in two major layers:  The outer layer is almost
///   empty -- it consists of the Viewer member functions -- operator(), and
///   handle_events.  The inner layer is where all the work gets done.  The
///   inner layer is represented as the class, TextEditor::Impl.  Such an
///   object is the guts of a text editor.  It basically is a collection of
///   helper functions used by the outer layer's mechanism.
/// 
///   The TextEditor::handle_events() function maps keys into edit function
///   pointers -- which are members of TextEditor::Impl.  It then applies these
///   functions to the Impl object -- which means performing edit actions and
///   updating the screen.  Each editor command is mapped in this way to a
///   single Impl class member.
/// 
///   In order to support undoing of edit commands, each 'action' that one
///   would change the edited file gets an "undo stack entry" which is
///   executable by the editor's "undo" command.
/// 
///   Most edit commands operate on single lines or blocks of lines.  This is
///   by design.  Unlike emacs, this editor is a line oriented editor --
///   despite its page viewing nature.
/// 
///   The edit "buffer" is a skiprope of std::string.  When the file is first
///   read, the strings in the file are stored in a tab-packed state.  That is,
///   tabs are not expanded.  When you change a line, it is unpacked.  You
///   don't get back to the tab-packed state until you save the file.
/// 
///   The lines in the file are represented by an EditLine object.  That is,
///   the edit session is a skiprope<EditLine>.  A skiprope is like a vector
///   but which offers only O(ln(N)) access.  No single operation is more than
///   O(ln(N)) -- unlike a real vector where random inserts/deletes are O(N).
/// 
///   The edit buffer is of type 'rep_type' in the text below.  Since the
///   rep_type is conceptually like a vector, rather than like a 'list', you
///   must basically deal with line numbers rather than iterators.  If you are
///   just reading data, iterators are fine.  If you ever start inserting or
///   deleting, you need to be doing that in terms of line numbers.
/// 
///   There is also an iterator, TextIterator, that knows how to span lines.
///   operator*() on the iterator gives you the character under the iterator.
///   However, the iterator lets you address the 'end' of each line.  If you
///   do operator* on the end of the line, you get '\n'.  If you iterate past
///   last valid character on the line, you get an iterator to the '\n'.  If
///   you continue iterating, you end up pointing into the next line as you
///   might reasonably expect.  Thus, the iterator lets you read the file
///   one character at a time and have it act like you were reading a real
///   unix file.  I say "unix" file because DOS files have the '\r\n' sequence.
///   Unix has only the '\n' sequence, and the fools at Apple, only use the
///   \r sequence.  What !#$!@#$!@#$ drugs were all these morons on?  One
///   character line termination if FINE!
/// 
///   Note that the TextIterator is very expensive to use.  You should only
///   use it when you need a function that processes characters across multiple
///   lines.
/// 
///   The class, UndoOperation, is mainly a holder for data -- it is not a
///   stand alone utility.
/// 
///   Associated with every edit command is an entry in the 'func' map.  This
///   entry is created during static init time by objects of type
///   AddMapping.  Every edit command has an AddMapping object.  The constructor
///   of which is designed to update the global variable, TextEditor::Impl::func.
/// 
/// 
//

#include <cxxtls/options.h>
#include <cxxtls/varobj.h>
#include <cxxtls/fmtd.h>
#include <texteditor.hxx>
#include <cxxtls/viewermanager.h>
#include <cxxtls/skiprope.h>
#include <fstream>
#include <portable_strstream.h>
#include <cxxtls/strtool.h>
#include <map>
#include <memory.h>
#include <cxxtls/strtool.h>
#include <algorithm>
#include <portable_io.h>
#include <cxxtls/etagsdb.h>
#include <errno.h>
#include <string.h>
#include <ctype.h>
#include <cxxtls/simple_regex.h>
#include <cxxtls/cursesinterface.h>
#include <cxxtls/foreach.h>
#include <cxxtls/cpptagdb.h>
#include <cxxtls/muSED.h>
#include <cxxtls/utf.h>



#include <fstream>


using namespace cxxtls;

typedef CursorWindow::row_col     row_col;
typedef CursorWindow::viewport    viewport;
typedef viewport::repaint_handler repaint_handler;
typedef CursorWindow::input_event input_event;

struct LineAnnotation
    ///  A wrapper around the annotation state of a line of text.  The annotation state
    ///  is controlled primrarily by the listen() function and typically refers to annotations
    ///  being controlled by other viewers, rather than the editor itself.
    ///
    ///  Use operator bool() to test for any annotations at all.
    ///  Use displayValue() to get the character for use in displaying the annotaitons
    ///
    ///  Use hasBreakpoint() and setBreakpoint() to test and set the breakpoint annotation
    ///  flag.
    ///  Use hasExternalMark() and setExternalMark() to test and set the breakpoint annotation
    ///  flag.

{
    char value_;

    enum annotationMaskBits
    {
        breakpoint    = 1,
        externalMark  = 2,
    };

    LineAnnotation()
    :  value_(0)
    {
    }

    operator bool() const { return value_; }  ///< return true if any annotations exist

    char displayValue() const
       
    {
        if(value_ == 0)
           return ' ';     // no annotation

        if(  (value_ & breakpoint) && (value_ & externalMark) )
        {
           return '+';  
        }

        if( (value_ & breakpoint) )
        {
           return '*';
        }

        if( (value_ & externalMark) )
           return '>';
          
        return ' ';
    }

    void setBreakpoint(bool v)
    {
       if(v)
           value_ |= breakpoint;
       else
           value_ &= ~breakpoint;
    }

    void setExternalMark(bool v)
    {
       if(v)
           value_ |= externalMark;
       else
           value_ &= ~externalMark;
    }

    bool hasBreakpoint() const 
    {
        return value_ & breakpoint;
    }

    bool hasExternalMark() const
    {
        return value_ & externalMark;
    }

   
};



struct EditLine
  ///  A datastructure holding a line of text in an edit session.
  ///  Lines of text are tab expanded before display and binary
  ///  characters are converted into ~'s for display purposes.
  ///
  ///  If you don't change a line of text, the tab expansion never
  ///  gets written back to the file, even if you save the edit
  ///  session, but if you do save, any changed lines will be
  ///  tab packed before writing.
  ///
  ///  This data structure keeps two pointers -- one pointer if the
  ///  data is never modified and the other if it is.  It could be done
  ///  with a flag and a pointer, but it would take up the same space on
  ///  most machines.
{

  //  Only one of the folllowing 2 pointers will be non-null

  std::string *original_data_;   ///  original unmodified data
  std::string *modified_data_;   ///  data after characters are changed

  bool crRemoved_; ///< A carriage return was removed from this line during
                   ///  the file reading process (linux/unix only).

  LineAnnotation annotation_; 
                    ///<  A bit mask that indicates the special nature of
                    ///   the line.  @see LineAnnotation for details.

  EditLine(std::string* original_data)
  :  original_data_(original_data),
     modified_data_(0),
     crRemoved_(0)
  {
    ///  A constructor -- you must supply the original data.
  }

  ~EditLine()
    /// A destructor -- what'll they thank uv next, pawww?
  {
    // does nothing, you are responsible for deleting the data yourself.
    // that saves an extraneous copy construction everywhere
  }

  std::string* data()
    ///  Get the raw, un-tab-expanded copy of the edit record
  {
    return original_data_ ? original_data_ : modified_data_;
  }

  operator std::string()
    ///  Get the tab expanded, version (that has the non-ascii chars turned to ~)
  {
    if(original_data_)
      return StrTool::expand_tabs(*original_data_);
    else
      return *modified_data_;

  }

  void replace(std::string const &s)
    ///  Replace the edit line's data with that specified in the string
  {
    if(original_data_)
      delete original_data_;

    original_data_ = 0;

    if(modified_data_)
      delete modified_data_;

    modified_data_ = new std::string(s);
  }

  std::string &maybe_tabbify()
    ///  Return a tab compressed version of this line -- but only if the
    ///  line was modified.  If the original_data is unmodified, just
    ///  return it.
  {
    static std::string rv;

    if(original_data_)
      return *original_data_;

    rv = StrTool::pack_tabs(*modified_data_);

    return rv;
  }



};

typedef skiprope<EditLine>        rep_type;   // edit session representation


struct TextIterator
  /// An iterator to the characters in a skiprope of EditLines.
  ///
  /// This iterator is expensive!
  ///
  /// The character after the last valid date in the line accessed
  /// by this iterator -- even though it doesn't exist.  That character
  /// is returned as '\n'.  Thus, the iterator can be used to count lines
  /// even though it spans lines when you use operator++ or minus.
  ///
  /// Do not change the edit session while iterating over it!
  ///
{
   int           line_;    // line number we are talking about
   std::string   text_;    // text of a line in the edit session
   size_t        length_;  // length of text_
   size_t        col_;     // column within text_ (can be same as length!)
   rep_type&     rep_;     // the container of lines

   TextIterator(TextIterator const& r)
   : line_(r.line_),
     text_(r.text_),
     length_(r.length_),
     col_(r.col_),
     rep_( const_cast<rep_type&>(r.rep_))
   {
   }

   void operator=(TextIterator const &r)
   {
     line_   = r.line_;
     text_   = r.text_;
     length_ = r.length_;
     col_    = r.col_;
     
     if(&rep_ != &r.rep_)
     {
        // fatal error -- the two must be compatible!
     }

   }


   TextIterator(rep_type &rep, int line, size_t column)
   : line_(line),
     col_(column),
     rep_(rep)
   {
     if(line < rep.size())
       text_ = rep[line];
     else
       text_ = "";

     length_ = text_.size();
   }

   void correct_overruns() // called after incrementing the col_
   {
     if(col_ > length_)  // NOT ">="  !
     {
        col_ = 0;
        
        ++line_;
        
        if(line_ >= rep_.size())
        {
          line_ = rep_.size();
          text_ = "";
        }
        else
        {
          text_ =  rep_[line_];
        }
        
        length_ = text_.size();
        
     }
   }

   void correct_underruns() // called after decrementing the col_
   {
     if(col_ >= length_)    // yes >= (and length_ is an unsigned!
     {
        
        --line_;
        
        if(line_ < 0)
        {
          line_ = 0;
          text_ = "";
        }
        else
        {
          text_ =  rep_[line_];
        }
        
        length_ = text_.size();
        col_ = length_;     // yes, past the end of line so the operator* will
                            // return '\n'
                        
        
     }
   }


   char operator*() ///         fetch the character under the iterator
   {
     return col_ >= length_ ? '\n' : text_[col_];
   }


   TextIterator &operator++ ()
   {
     ++col_;

     correct_overruns();

     return *this;
   }


   TextIterator operator++ (int )
   {
     TextIterator rv= *this;
     ++col_;
     correct_overruns();
     return rv;
   }

   TextIterator &operator-- ()
   {
     --col_;

     correct_underruns();

     return *this;
   }


   TextIterator operator-- (int )
   {
     TextIterator rv= *this;
     --col_;
     correct_underruns();
     return rv;
   }

   bool operator==( TextIterator const &r )
   {
     return col_ == r.col_ && line_ == r.line_;
   }

   bool operator!=( TextIterator const &r )
   {
     return col_ != r.col_ || line_ != r.line_;
   }

};


struct UndoOperation
  ///  A data structure that holds information needed to reverse
  ///  a change to the file.
{
   enum ActionToTake
   {
      replace_line,
      insert_line,
      erase_line,
      join_lines,       // created when you split
      insert_block,
      delete_block,
      replace_block,
      break_lines,      // created when you join
      nested_undo_ops,  // replace all strings' uses this to treat the undoing
                        // all changes in a single operation
                        
                        
                        
      last_code         // MUST BE LAST!
   };

   int code_;          // action to take
   int count_;         // of lines in a deleted block

   std::string text_;  // to replace_with

   int lineno_;        // line number upon which to act

   std::list<std::string> block_;  // block of lines to insert at lineno_


   typedef std::list<UndoOperation*> undo_stack;

   undo_stack *nested_undo_stack_;  // replace_first()

   int nested_first_mark_; // only valid if using the nested undo stack
   int nested_last_mark_;

private: 

   void* operator new(size_t t)
     /// prevent accidental abuse of the undo logic -- you are supposed to
     /// use Impl::push_undo_stack()
   {
     return new char[t];
   }

public:

   void operator delete(void *p)
   {
     char *cp = (char*)(p);

     delete[] cp;
   }

   static UndoOperation* make_new()
     // don't call this -- it is meant only
     // to be used by Impl::push_undo_stack()
   {
     return new UndoOperation;  
   }

};


struct TextEditor::Impl
  /// This class actually implements the text editor.  The Viewer which
  /// holds this datastructure is just a mechanism for routing events
  /// to this class.
  ///
  /// Editor commands are implemented as member functions of this class
  /// and the mapping between editor commands and these functions is stored
  /// in a static member, func.
  ///
  /// The records in the edit session are stored in a skiprope. This
  /// container pretends to be a vector, and has all its limitations, but
  /// gives O(ln(N)) access to the array elements -- including inserts
  /// and deletes.
  ///
{
  FileName fullname_;       ///  name of file being edited
  int      left_;           ///  left column in viewport where lines are drawn
  bool     active_;         ///  is this window active?
  int      first_row_;      ///  where editor data appears
  int      page_top_;       ///  offset from beginning of file till top of page
  size_t   page_left_;      ///  offset from beginning of line till left of page

  int      cur_;            ///  current line in the file
  int      first_mark_;     ///  first line of the marked block of the file
  int      last_mark_;      ///  1 + last marked line in the block

  size_t   col_;            ///  current column within the line (may be past end
                            ///  of valid data on the line
                        
  int      bottom_line_;    ///  line no of last visible line ( may not be valid line no )

  size_t   htab_;           ///  horizontal tab size
  bool     use_tabs_;       ///  compress using tabs when saving the file

  size_t   left_margin_;    ///  column to put cursor in on line split operations


  rep_type lines_;  ///  container of line text

  std::string word_chars_; ///< string containing character which can comprise
                           ///  words.
                        
  std::string search_text_; ///< current string to search for
  std::string replacement_; ///< replaces search_text_ in replace_first and _next ops.
  bool        is_regex_;    ///  is the search_text_ a regular expression?
  SimpleRegex regexp_;      ///         the regexp to use if needed

  bool key_kludge_;  ///  this shouldn't exist, it is just a development mechanism to
                     ///  allow 2 key command processing.
                
  bool word_kludge_; ///< Kludges on top of kludges -- oh my ;-(

  bool slider_kludge_; ///< Kludges on kludges on kludges...  This one handles left and right arrows
                       ///  that slide marked blocks left and right by one character.

  bool is_writeable_; ///<  can we write the data in this file?

  bool is_new_;       ///  Was this file newly created for this edit session
                
  UndoOperation::undo_stack  undo_stack_;  ///  push front and pop front only!

  bool is_dirty_; // are there unsaved edit changes

  bool nonstandard_encoding_;  // utf8 or something

  UndoOperation* push_undo_stack()
  {
    UndoOperation*rv = UndoOperation::make_new();

    undo_stack_.push_front(rv);

    is_dirty_ = true;

    return rv;
  }


  ViewerManager* manager_;


  bool fetchAllLines(std::list<std::string> &outout); ///< extract all lines of the edit session

  int      size() { return lines_.size(); } ///<  Number of lines in the edit session

    /// Get a reference to the text of the current line.  Don't cache these
    /// values because edit activities can create or delete them and your cache
    /// will be out of sync.

  void     display_line(viewport* vp, int line);

  void     read_file(); ///< Read the file into this edit session.  Use the name
                        ///  found in 'fullname_'.

  void  paint_page(viewport* vp, Viewer *v) { (*v)(vp, 0); }
    /// Well, what do you expect it to do?  It repaints the editor page,
    /// title and all

  void  restore_cursor(viewport* vp)
    ///  Put the cursor where it should be.
  {
    vp->set_curpos(first_row_ + (cur_ - page_top_),
                   col_ - page_left_ + left_
                  );
  }

  void push_undo_line_change(int line, std::string const &old_text)
    ///  Push a change to a line on the undo stack -- unless the current
    ///  line already has a change -- multiple changes per line are not
    ///  saved for space reasons.
  {
    if(undo_stack_.begin() == undo_stack_.end())
    {
      // stack is empty, create the first element

      UndoOperation* op = push_undo_stack();

      op->code_   = UndoOperation::replace_line;
      op->text_   = StrTool::expand_tabs(old_text);
      op->lineno_ = line;


    }
    else
    {
      // the stack isn't empty, see if the most recent item is the
      // same as this item

      UndoOperation* op = undo_stack_.front();

      if(op->code_    != UndoOperation::replace_line ||
         op->lineno_  != line
        )
      {
        // this is a new one, so push it on the stack

        UndoOperation* op = push_undo_stack();
        
        op->code_   = UndoOperation::replace_line;
        op->text_   = StrTool::expand_tabs(old_text);
        op->lineno_ = line;
        

      }

    }

  }


  void replace_line(int line_no, std::string const &new_text, bool save_undo=true)
    ///  This is the central location for changes to the file.  The
    ///  file is changed, one whole line at a time.  Every character you
    ///  insert, individually, results in a call to this function.  This
    ///  function knows how to implement 'undo'.
  {
     if(save_undo)
       push_undo_line_change(line_no, *line_text(line_no).data());

     lines_[line_no].replace(new_text);
  }

  rep_type::iterator line_address(int line_no)
    ///  Get an iterator to a specified line number
  {
    return lines_.find(line_no);
  }

  void erase_line(rep_type::iterator location,
                  int                corresponding_line,
                  bool               save_undo_info = true
                 )
    ///  Erase a line given an iterator to it (see line_address above)
  {
    if(save_undo_info)
    {
      EditLine& line  = *location;

      UndoOperation *op = push_undo_stack();

      op->text_   = *line.data();
      op->lineno_ = corresponding_line;
      op->code_   = UndoOperation::insert_line;

    }

    lines_.erase(location);
  }


  EditLine& line_text(int line_no)
    ///  Get the current line's representation in the edit session.
  {

    static EditLine empty( new std::string("") );

    if( line_no >= lines_.size())
    {
       return empty;
    }

    return lines_[line_no];
  }


  void  make_col_visible(viewport*, Viewer*);
    /// Make sure col_ is visible -- ie handle horizontal scrolling to
    /// ensure that it is.

  void  make_line_visible(viewport*, Viewer*, bool center=true);
    /// Make sure cur_ is visible -- ie handle vertical scrolling to
    /// ensure that it is.  Also, this function makes sure that cur_ is
    /// within a valid extent of the file if possible.  The third
    /// parameter, center, indicates that an attempt should be made to
    /// make the current line centered in the viewport if it is not
    /// already visible.  When scrolling one line at a time, you'd
    /// NOT want to do this, but in most other cases you would.

  size_t vwidth(row_col &vp_size) { return vp_size.col_ - left_; }
    /// given a specified viewport size, return how many character columns
    /// are available for drawing edit text.

  size_t vheight(row_col &vp_size) { return vp_size.row_ - first_row_; }
    /// given a specified viewport size, return how many character rows
    /// are available for drawing edit text.

  bool  is_marked(int line) { return line >= first_mark_ && line < last_mark_; }
    /// returns true if the specified line is in the marked block

  typedef bool (Impl::*func_handler)(viewport*, Viewer*, input_event const *);
    /// A data type representing a member of the TextEditor::Impl class

  static std::map<int, std::pair<func_handler, bool> > func;
    /// The table of valid editor functions.  This a mapping between an
    /// integer and a pointer to a member of the Impl class.  The 'int' that
    /// gets mapped into a pointer to member is the result of taking the
    /// input event and using it as an index in the CursorWindow::func map.
    /// The function id (an int) returned is used as an index into the
    /// Impl class func map to get a pointer to the function to implement the
    /// desired function.  The map actually contains a pair<ptr, bool> where
    /// the pointer is the func_handler and the bool is a flag indicating that
    /// this function is modifies the edit session or only moves around within it.
    /// The integers in the CursorWindow;:func map are defined in enum
    /// CursorWindow::edit_function_names.

  bool find_next_helper(viewport*, Viewer*, input_event const *);
    /// This function is the actual search routine.  It assumes that the
    /// search_text_ string is set and that the cursor is position at the
    /// start of the search.  This function does not wrap around at the
    /// bottom of the file.
    ///
    /// The caller must save the cur_ and col_ values if needed.  This
    /// function returns true if the string is found and leaves col_ and cur_
    /// pointing to the beginning of the found string.  If the string is not
    /// found, col_and cur_ point to the end of the file or there abouts.

  bool find_prev_helper(viewport*, Viewer*, input_event const *);
    /// Same as find_next_helper but scans up through the file, not down.

  bool replace_next_helper(viewport*, Viewer*, input_event const *);
    /// Same as find_next_helper except that it actually replaces
    /// the text found in the search match with replacement_ and leaves
    /// the cursor in the first column after the replacement.  If this function
    /// returns false, it did not change anything.  The caller is responsible
    /// for restoring the cursor position, painting pages, etc.

  void save         (viewport*, TextEditor*, input_event const *);
    /// Save the edit session to the file from whence it came


  bool is_word_char(char c)
    ///  Check to see if a specified character is a 'word character'
    ///  by consulting the 'word_chars_' table.
  {
    if(std::find( word_chars_.begin(), word_chars_.end(), c) == word_chars_.end())
      return false;

    return true;
  }

  //  The following functions return iterators into the edit session that will
  //  span lines -- don't modify the session by creating, deleting, or changing
  //  lines as this will invalidate the iterators.

  TextIterator top_pointer();  ///  get iterator        to first character in the edit session
  TextIterator cur_pointer();  ///  get iterator        to char under the cursor
  TextIterator bot_pointer();  ///  get iterator to end of edit session

  //
  // The following class, StringCompare, is a character comparison functor.  It is
  // meant to be used as a parameter to the std::search algoritm
  //

  bool mixed_case_;    // controls CharCompare
  bool match_words_;   // find and replace only whole words

  struct CharCompare
  {
    TextEditor::Impl* impl_;

    bool operator() (char a, char b) const
    {
        if(impl_->mixed_case_)
        {
          if( isupper(a) )
            a = tolower(a);
        
          if( isupper(b) )
            b = tolower(b);
        
          return a == b;
        }
        
        return a == b;
    }

    CharCompare(TextEditor::Impl* t)
    : impl_(t)
    {
    }

  };


  //
  // Declarations for the various editor member functions whose job is to
  // directly implement an edit command
  //
  bool page_down      (viewport*, Viewer*, input_event const *);
  bool page_up        (viewport*, Viewer*, input_event const *);
  bool char_left      (viewport*, Viewer*, input_event const *);
  bool char_right     (viewport*, Viewer*, input_event const *);
  bool char_matching  (viewport*, Viewer*, input_event const *);
  bool tab_right      (viewport*, Viewer*, input_event const *);
  bool tab_left       (viewport*, Viewer*, input_event const *);
  bool char_delprv    (viewport*, Viewer*, input_event const *);
  bool char_delete    (viewport*, Viewer*, input_event const *);
  bool char_insert    (viewport*, Viewer*, input_event const *);
  bool line_home      (viewport*, Viewer*, input_event const *);
  bool line_end       (viewport*, Viewer*, input_event const *);
  bool line_split     (viewport*, Viewer*, input_event const *);
  bool line_up        (viewport*, Viewer*, input_event const *);
  bool line_down      (viewport*, Viewer*, input_event const *);
  bool line_del       (viewport*, Viewer*, input_event const *);
  bool line_mark      (viewport*, Viewer*, input_event const *);
  bool line_clreol    (viewport*, Viewer*, input_event const *);
  bool line_goto      (viewport*, Viewer*, input_event const *);
  bool line_join      (viewport*, Viewer*, input_event const *);
  bool line_insert    (viewport*, Viewer*, input_event const *);
  bool file_top       (viewport*, Viewer*, input_event const *);
  bool file_bottom    (viewport*, Viewer*, input_event const *);
  bool file_insert    (viewport*, Viewer*, input_event const *);
  bool word_next      (viewport*, Viewer*, input_event const *);
  bool word_prev      (viewport*, Viewer*, input_event const *);
  bool word_delete    (viewport*, Viewer*, input_event const *);
  bool snatch_edit    (viewport*, Viewer*, input_event const *);
  bool show_symbols   (viewport*, Viewer*, input_event const *);
  bool word_sfind     (viewport*, Viewer*, input_event const *);
  bool find_first     (viewport*, Viewer*, input_event const *);
  bool find_next      (viewport*, Viewer*, input_event const *);
  bool find_prev      (viewport*, Viewer*, input_event const *);
  bool block_delete   (viewport*, Viewer*, input_event const *);
  bool block_copy     (viewport*, Viewer*, input_event const *);
  bool block_paste    (viewport*, Viewer*, input_event const *);
  bool block_insert   (viewport*, Viewer*, input_event const *);
  bool block_adjust   (viewport*, Viewer*, input_event const *);
  bool block_save     (viewport*, Viewer*, input_event const *);
  bool block_transform(viewport*, Viewer*, input_event const *); // use mused to change the block
  bool replace_first  (viewport*, Viewer*, input_event const *);
  bool replace_next   (viewport*, Viewer*, input_event const *);
  bool undo           (viewport*, Viewer*, input_event const *);
  bool etags_first    (viewport*, Viewer*, input_event const *);
  bool cpptags_first  (viewport*, Viewer*, input_event const *);
  bool tags_next      (viewport*, Viewer*, input_event const *);
  bool etags_all      (viewport*, Viewer*, input_event const *); // list all matching functions
  bool cpptags_all    (viewport*, Viewer*, input_event const *);
  bool to_edit        (viewport*, Viewer*, input_event const *);
  bool reread         (viewport*, Viewer*, input_event const *); // cannot be undone
  bool switch_source  (viewport*, Viewer*, input_event const *);

  void adjustHelper(int adjustment, viewport *, Viewer*);  ///<  Slide block left or right by adjustment
                                                           ///   characters.

};

typedef TextEditor::Impl  Impl;  // to simplfy typing


std::map<int, std::pair<Impl::func_handler, bool> > Impl::func;
    // the mapping between CursorWindow::func_*
    // values and TextEditor command handlers.

const bool Modifying=   true;  // use as parm to AddMapping constructor
const bool NonModifying=false;

struct AddMapping
  ///  An AddMapping object exists only for the purpose of statically
  ///  initializing the Impl::func map with valid entries.
  ///  You declare a member function of the Impl class then you declare an
  ///  object of type AddMapping to cause the function to appear in the
  ///  func table.  You only need to do this once, and it is the intent of
  ///  addMapping's design that you do this at file scope -- not that it
  ///  matters.
  ///
  ///  The Impl::func map is a mapping between function key identifiers,
  ///  defined in enum CursorWindow::edit_function_names.  Other editor
  ///  functions exist but do not get initialized in this map.  They are
  ///  called directly upon interpretation in the editor's input function itself.
{
  AddMapping(int cmd, Impl::func_handler f, bool modifying)
  {
    Impl::func[cmd] = std::pair<Impl::func_handler, bool>(f, modifying);
  }
};

bool TextEditor::fetchAllLines(std::list<std::string> &output)
{
   if(impl_.get())
   {
      return impl_->fetchAllLines(output);
   }

   return false;
}

void 
TextEditor::
listen(std::string const &type, std::string const &msg)
{
    static std::string breakpointSet("breakpoint");
    static std::string stopPoint    ("stopped");
    static std::string deleteIt("delete");

    if(type == breakpointSet)
    {
       VarObj::ParseResult  parseResult = VarObj::parse(msg);

       if(parseResult.rv_)
       {
           VarObj *filePtr = parseResult.rv_->findNestedChild("file");
           VarObj *linePtr = parseResult.rv_->findNestedChild("line");
           VarObj *actPtr  = parseResult.rv_->findNestedChild("action");

           if(filePtr && linePtr)
           {
               std::string file = filePtr->value();
               int         line = -1;

               sscanf(linePtr->value().c_str(), "%d", &line);

               if(line > 0 && file == impl_->fullname_)
               {
                  EditLine &editLine = impl_->line_text(line-1);

                  if(actPtr && deleteIt == actPtr->value())
                      editLine.annotation_.setBreakpoint(false); 
                  else
                      editLine.annotation_.setBreakpoint(true);

                  if(manager()->is_visible(this))
                  {
                     CursorWindow::viewport *vp = manager()->find_viewport(this);

                     (*this)(vp,1);  // display as inactive
                  }

               }

           }

           delete parseResult.rv_;
       }
    }
    else
    if(type == stopPoint)
    {
       VarObj::ParseResult  parseResult = VarObj::parse(msg);

       if(parseResult.rv_)
       {
           VarObj *filePtr = parseResult.rv_->findNestedChild("file");
           VarObj *linePtr = parseResult.rv_->findNestedChild("line");
           VarObj *actPtr  = parseResult.rv_->findNestedChild("action");

           std::string file = filePtr->value();
           int         line = -1;
           
           sscanf(linePtr->value().c_str(), "%d", &line);
           
           if(line > 0 && file == impl_->fullname_)
           {
              EditLine &editLine = impl_->line_text(line-1);
           
              if(actPtr && deleteIt == actPtr->value())
                  editLine.annotation_.setExternalMark(false);
              else
                  editLine.annotation_.setExternalMark(true);
            
             if(manager()->is_visible(this))
              {
                 CursorWindow::viewport *vp = manager()->find_viewport(this);
           
                 (*this)(vp,1);  // display as inactive
              }
           
           }

           delete parseResult.rv_;
       }
    }
}


void
TextEditor::
operator() ( CursorWindow::viewport *     vp,
             int                          cmd
           )
  /// repaint the editor window in its current state
{
  // WARNING:  do not save vp -- it gets deleted regularly!
  
  // get page size and origin

  row_col size    = vp->size();
  row_col origin  = vp->origin();

  int     avail_rows  = impl_->vheight(size);
  // determine if the page is active or not

  if(cmd == repaint_handler::activate)
    impl_->active_ = true;
  else
  if(cmd == repaint_handler::deactivate)
    impl_->active_ = false;

  // handle the page left margin

  if(origin.col_)
  {
    impl_->left_ = 1;

        
    for(int i=0; i < size.row_; ++i)
    {
      vp->set_curpos(i, 0);

      if(i == size.row_ -1 || i == 0)
       vp->set_text_attribute(CursorWindow::bold_ul);
      else
       vp->set_text_attribute(ViewerManager::inactive_mark_att);

      using namespace CursesInterface;

      vp->write(line_chars[VL_MIDDLE]);
    }

  }
  else
    impl_->left_ =0;

  // handle the title line

  vp->set_curpos(0, impl_->left_);
  vp->set_text_attribute(impl_->active_? ViewerManager::active_title_att
                                       : ViewerManager::inactive_title_att
                                       );
  if( impl_->is_new_ )
    vp->write("Creating ");
  else
  if( impl_->is_writeable_ )
    vp->write("Editing ");
  else
    vp->write("Reading ");


  if(size.col_ > 70 )
  {
      vp->write(impl_->fullname_.shorten(size.col_ - 8 - 26) );
  }
  else
      vp->write(impl_->fullname_.shorten(size.col_ - 8) );



  vp->fill_to_eol();

  // paint the body of the text

  if(impl_->cur_ - impl_->page_top_ >= avail_rows)
  {
     // current cursor row is not visible -- make ti be visible.

     impl_->page_top_ = impl_->cur_ - avail_rows/2;

     if(impl_->page_top_ < 0)
     {
        impl_->page_top_ = 0;
     }

  }


  int row = impl_->first_row_;

  int line = impl_->page_top_;

  impl_->bottom_line_ = line + avail_rows - 1 ;  // for bottom line highlighting

  int row_count = 0;

  while(line < impl_->size() && row_count < avail_rows )
  {
     vp->set_curpos(row, impl_->left_);
     impl_->display_line(vp, line);
     ++row;
     ++line;
     ++row_count;
  }


  // paint the lines after body of the text

  bool first=true;

  while(row < size.row_)
  {
    vp->set_curpos(row, impl_->left_);

    if(first)
    {
      vp->set_text_attribute(ViewerManager::active_title_att);
      vp->write("*eof*");
      vp->set_text_attribute(ViewerManager::normal_att);
      first = false;
    }
    else
    {
      if(row == size.row_ -1)
        vp->set_text_attribute(ViewerManager::bottom_att);
      else
        vp->set_text_attribute(ViewerManager::normal_att);
    }
    vp->fill_to_eol();

    ++row;
  }

  // place the cursor

  impl_->restore_cursor(vp);

}
                        
bool
TextEditor::
handle_event( input_event const * e,
              viewport          * vp
            )
{

  if(!impl_->is_writeable_)
  {
     // in view only mode, treat 'q' as an exit command

     if(e->type_ == CursorWindow::input_event::DataKey)
     {
        if(e->value_ == 'q' || e->value_ == 'Q')
        {
           return true;
        }
     }
  }


  if(e->type_ == CursorWindow::input_event::ForceExitKey)
  {
    return true;
  }

  bool rv = false;


  if(impl_->slider_kludge_)
  {
     // the user has pressed either ^X -> or ^X <-.  We have already moved the
     // marked block at least once and are now handling secondary move operations.

     bool markActive = false;

     if(impl_->first_mark_ < impl_->last_mark_)
        markActive = true;

     if(!markActive)
     {
        impl_->slider_kludge_ = false; // no point in a slider kludge without a mark
     }
     else
     switch(e->value_)
     {
         default:  
           // we don't know what the current key means, so turn off the slider kludge and
           // go back to normal key handling.   Interpret the current key as normal
           impl_->slider_kludge_ = false;
           break;

        case CursorWindow::key_right:
        {
            impl_->adjustHelper(1,vp, this);
            goto done; // exit and paint the current cursor position
        }

        case CursorWindow::key_left:
        {
            impl_->adjustHelper(-1,vp, this);
            goto done; // exit and paint the current cursor position
        }

     }

  }

  
  if(impl_->key_kludge_)
  {
    //
    // Most recent key was '^X', so handle the second key of a multi-key sequence
    // Note:  this is stupid.  I build a great mechanism for function key sequence
    // processing and I am not using it....
    //

    impl_->key_kludge_ = false;      // reset the ^X flag

    if(e->value_ == ('='))       // ^X = switches to the next source alternative
    {
      impl_->switch_source(vp,this,e);
    }
    else
    if(e->value_ == ('L'-'@'))       // ^X ^L rereads edit session
    {
      impl_->reread(vp,this,e);
    }
    else
    if(e->value_ == ('S' - '@'))     // ^X ^S saves the session
    {
      if(impl_->word_kludge_)
      {
        impl_->word_sfind(vp,this,e);
        impl_->word_kludge_ = false;
      }
      else
        impl_->save(vp,this,e);
    }
    else
    if(e->value_ == ('K' - '@') ||     // ^X ^K quits the session
       e->value_ == 'k'         ||     // ^X k quits the session
       0
      )
    {
      if(impl_->is_dirty_)
      {
        vp->beep();
        
        CursorWindow::Dialog d("Unsaved edit changes exist");
        d += new CursorWindow::Dialog::String("ok",
                                              "Quit Anyway (Yes or No)",
                                              "No",
                                              4
                                             );
        
        if(d.popup( manager_->window() ))
        {
          return false;
        }
        
        std::string ok_value = d.element_value("ok");
        
        if(ok_value.size() > 2)
        {
          ok_value.erase(2, ok_value.size());
        }
        
        if(ok_value != "Ye" && ok_value != "ye" )
        {
          vp->beep();
          return false;
        }
        
      }
      return true;
    }
    else
    if(e->value_ == ('G' - '@'))
      rv = impl_->line_goto(vp,this,e);
    else
    if(e->value_ == ('E' - '@'))
      rv= impl_->snatch_edit(vp,this,e);
    else
    if(e->value_ == ('W' - '@'))
    {
      impl_->key_kludge_ = true;   // if handling ^X, ignore ^W for the moment
      impl_->word_kludge_ = true;   // if handling ^X, ignore ^W for the moment
    }
    else
    if(e->value_ == '!')
    {
      impl_->etags_first(vp,this,e);
    }
    else
    if(e->value_ == '.')
    {
      impl_->cpptags_first(vp,this,e);
    }
    else
    if(e->value_ == '?')
    {
      impl_->etags_all(vp,this,e);
    }
    else
    if(e->value_ == '/')
    {
      impl_->cpptags_all(vp,this,e);
    }
    else
    if(e->value_ == ',')
    {
      impl_->tags_next(vp,this,e);
    }
    else
    if(e->value_ == ('A' - '@'))
    {
      impl_->key_kludge_ = true;   // if handling ^X, ignore ^A for the moment
      impl_->word_kludge_ = true;   // if handling ^X, ignore ^A for the moment
    }
    else
    if(e->value_ == ('I' - '@'))
    {
      impl_->block_insert(vp,this,e);  // ^X ^A ^I maps to 'insert text into block'
    }
    else
    if(e->value_ == ('B' - '@'))
    {
      impl_->block_adjust(vp,this,e);  // ^X ^A ^B maps to 'adjust block left and right'
    }
    else
    if(e->value_ == ('U' - '@'))
    {
      impl_->undo(vp,this,e); // ^X ^U is also undo.
    }
    else
    if(e->value_ == '0')
    {
      impl_->show_symbols(vp,this,e); // run cpptagdb.exe on this file and show the results
    }
    else
    if(e->value_ == ('R' - '@') )
    {
        impl_->file_insert(vp,this,e);
    }
    else
    if(e->value_ == ('O' - '@'))
    {
       impl_->block_save(vp,this,e);
    }
    else
    if(e->value_ == ('T' - '@'))
    {
       impl_->block_transform(vp,this,e);
    }
    else
    if(e->value_ == CursorWindow::key_right)
    {
       if(impl_->first_mark_ < impl_->last_mark_)
       {
           impl_->adjustHelper(1, vp, this);

           impl_->slider_kludge_ = true;
       }
    }
    else
    if(e->value_ == CursorWindow::key_left)
    {
       if(impl_->first_mark_ < impl_->last_mark_)
       {
           impl_->adjustHelper(-1, vp, this);
           impl_->slider_kludge_ = true;
       }
    }
    else
    if(e->value_ == ' ')
    {
        static std::string type("fileLocation");

        static std::string space(" ");
        static std::string file("{file=\"");
        static std::string line("\",line=");
        static std::string deleteIt(",action=\"delete\"");

        std::string qdelete;  // empty by default

        EditLine &editLine = impl_->line_text(impl_->cur_);

        if(editLine.annotation_.hasBreakpoint())
        {
           qdelete = deleteIt;
        }


        std::string msg = file                                   + 
                          impl_->fullname_                       + 
                          line                                   +
                          std::string(fmtd<>(impl_->cur_ + 1))   + 
                          qdelete                                +
                          "}";  // VarObj encoding format

        manager()->broadcast(type, msg);
    }
    else
    {
       vp->beep();
    }

  }
  else
  {
    impl_->word_kludge_ = false;

    std::pair<Impl::func_handler, bool> map_entry = Impl::func[ CursorWindow::func[e->value_] ];

    Impl::func_handler f         =  map_entry.first;
    bool               modifying =  map_entry.second;


    if(f)
    {
       if(!impl_->is_writeable_ && modifying)
         vp->beep();
       else
         rv = (impl_.get()->*f)(vp, this, e);
    }
    else
    if(e->value_ == 0x1b)
    {
      vp->beep();  // this should bring up a command menu
    }
    else
    if(e->value_ == ('X' - '@'))
    {
      impl_->key_kludge_ = true;
    }
    else
      manager_->window()->beep();
  }

done:

  if(rv == false && (this == manager()->active_viewer()))
  {
     char buffer[256];



     size_t lineOffset = impl_->cur_;
     size_t charOffset = impl_->col_;


     snprintf(buffer, 256, "L=%-8d C=%-3d", int(lineOffset+1), int(charOffset+1));

     CursorWindow::row_col windowSize = vp->size();

     if(windowSize.col_ > 70)
     {
         
         vp->set_curpos(0, windowSize.col_ - 26);
         
         vp->set_text_attribute(impl_->active_? ViewerManager::active_title_att
                                              : ViewerManager::inactive_title_att
                               );
         vp->fill_to_eol();
         
         
         
         vp->set_curpos(0, windowSize.col_ - strlen(buffer));
         vp->write(buffer);
         
         impl_->restore_cursor(vp);
     }

  }


  return rv;
}

        
void
TextEditor::
help()
{
  typedef std::list<std::string> help_text;

  std::auto_ptr< help_text > viewer_help( new help_text );

  extern char const *editversion;

  std::string htext("Help for the text editor ");

  htext += editversion;

  viewer_help->push_back(htext);

  viewer_help->push_back("");

  viewer_help->push_back("(Use ^S to search)");

  viewer_help->push_back("");


  viewer_help->push_back("  In addition to the standard key bindings");
  viewer_help->push_back("  shown later, the text editor interprets");
  viewer_help->push_back("  the following parochial key bindings:");
  viewer_help->push_back("    ^X -> -- enter block slider mode and slide the marked block right");
  viewer_help->push_back("    ^X <- -- enter block slider mode and slide the marked block left");
  viewer_help->push_back("");
  viewer_help->push_back("    NOTE:  In block slider mode, repeated right and left arrow key presses");
  viewer_help->push_back("           continually slide the currently marked block left and right.  Any");
  viewer_help->push_back("           other key ends the mode and the key is executed as normal.");
  viewer_help->push_back("");
  viewer_help->push_back("    ^X ^G -- goto specified line");
  viewer_help->push_back("    ^X ^S -- saves the edit session");
  viewer_help->push_back("    ^X ^K -- quits the edit session");
  viewer_help->push_back("    ^X  k -- quits the edit session");
  viewer_help->push_back("    ^X ^l -- rereads the file being edited, discarding changes");
  viewer_help->push_back("    ^X ^W ^E -- edit the file whose name appears");
  viewer_help->push_back("             under the cursor");
  viewer_help->push_back("    ^X ^W ^S -- find the next instance of the word under the cursor");
  viewer_help->push_back("    ^X  . -- Symbol name search (TAGPP.tagpp database)");
  viewer_help->push_back("    ^X  / -- All matching symbols searched (TAGPP.tagpp database)");
  viewer_help->push_back("    ^X  ! -- etags search FIRST matching symbol");
  viewer_help->push_back("    ^X  ? -- etags search ALL matching symbols");
  viewer_help->push_back("    ^X  , -- Search for next matching symbol");
#ifdef USE_muSED
  viewer_help->push_back("    ^X ^T -- Transform a block using SED commands:");
  viewer_help->push_back("             Note:  if any *.mused files are being edited,");
  viewer_help->push_back("             one of them will be selected as the script,");
  viewer_help->push_back("             otherwise, the user will be prompted for a 3 line script.");
#endif
  viewer_help->push_back("    ^X ^U -- undo (same as ^_ and ^?)");
  viewer_help->push_back("    ^X 0  -- show the symbols in the current file (only)");
  viewer_help->push_back("    ^X ^R -- read a file into the edit session");
  viewer_help->push_back("    ^X ^O -- write a selected block to a file.");
  viewer_help->push_back("");
  viewer_help->push_back("    ^X ^A ^I  -- insert column of text into");
  viewer_help->push_back("                 the currently marked block");
  viewer_help->push_back("    ^X ^A ^B  -- add or delete columns of");
  viewer_help->push_back("                 blanks to the current block");
  viewer_help->push_back("    F12  --  jump back and forth between source file variants:  .c, .h, etc");
  viewer_help->push_back("");
  viewer_help->push_back("  Special environment variables:");
  viewer_help->push_back("");
#ifdef _MSC_VER
  viewer_help->push_back("   ETAGSPATH --  A ';' separated list of directories");
#else
  viewer_help->push_back("   ETAGSPATH --  A ':' separated list of directories");
#endif
  viewer_help->push_back("                 in which to look for the TAGS file");
  viewer_help->push_back("                 which is used when looking for symbol definitions");
  viewer_help->push_back("                 and for source file alternates (.c/.h) -- see F12");
#ifdef _MSC_VER
  viewer_help->push_back("                 For example:  .:..:/home/user");
#else
  viewer_help->push_back("                 For example:  .;..;/home/user");
#endif
  viewer_help->push_back("                 ");
  viewer_help->push_back("                 The script, makeCppTags, can be used to create");
  viewer_help->push_back("                 the tags database.  Run it after a successful make.");
#ifdef _MSC_VER
  viewer_help->push_back("   SOURCEPATH--  A ';' separated list of directories and '*'s");
#else
  viewer_help->push_back("   SOURCEPATH--  A ':' separated list of directories and '*'s");
#endif
  viewer_help->push_back("                 in which to look for the source file alternates file");
  viewer_help->push_back("                 which is used when looking for symbol definitions");
  viewer_help->push_back("                 and for source file alternates (.c/.h) -- see F12");
#ifdef _MSC_VER
  viewer_help->push_back("                 For example:  ..;/project/main/include;/project/main/src;/home/*");
#else
  viewer_help->push_back("                 For example:  ..:/project/main/include:/project/main/src:/home/*");
#endif
  viewer_help->push_back("                 Searched when ^X ^E or ^X ^W ^E are pressed.");
#ifdef _MSC_VER
  viewer_help->push_back("   SOURCEALTS--  A ';' separated list of file name extensions");
#else
  viewer_help->push_back("   SOURCEALTS--  A ':' separated list of file name extensions");
#endif
  viewer_help->push_back("                 which are used by the F12 key (switch source) to decide");
  viewer_help->push_back("                 how to find the alternative source file given the current");
#ifdef _MSC_VER
  viewer_help->push_back("                 source file's name.  EG: .;.c;.h;.cc;.cxx;.cpp.  See F12");
#else
  viewer_help->push_back("                 source file's name.  EG: .:.c:.h:.cc:.cxx:.cpp.  See F12");
#endif



  impl_->manager_->help_helper(*viewer_help);

}


TextEditor::
TextEditor(ViewerManager*vm, FileName fullname)
: Viewer(vm),
  impl_(new Impl)
{
  impl_->nonstandard_encoding_ = false;
  impl_->fullname_    = fullname;
  impl_->left_        = 0;
  impl_->active_      = false;
  impl_->first_row_   = 1;
  impl_->page_top_    = 0;
  impl_->page_left_   = 0;
  impl_->cur_         = 0;
  impl_->col_         = 0;
  impl_->htab_        = 8;      // changed when file read
  impl_->use_tabs_    = true;   // changed when file read
  impl_->bottom_line_ = 10000;  // an obviously invalid value
  impl_->first_mark_  = -1;
  impl_->last_mark_   = -1;     // pair of marked lines
  impl_->left_margin_ = 0;
  impl_->manager_     = vm;
  impl_->key_kludge_  = false;   // by by default, we are not in a multi-key sequence
  impl_->word_kludge_ = false;   // by default we are not in a ^X ^W ^S, or ^X ^B ^A mode...
  impl_->slider_kludge_ = false; // by default, we are not in block slider mode
  impl_->mixed_case_  = true;   // mixed case find string.
  impl_->match_words_ = false;  // only match whole words
  impl_->is_writeable_= true;   // true unless otherwise set to false by read_file
  impl_->is_new_      = false;  // set to true by read_file if file does not exist
  impl_->is_regex_    = false;  // not regex searches by default

  impl_->read_file();

  impl_->is_dirty_    = false;  // no unsaved edit changes exist yet

  {
    // define word characters

    std::string &word_chars = impl_->word_chars_;

    char c;

    word_chars += '_';

    for(c='a'; c <= 'z'; ++c) word_chars += c;
    for(c='A'; c <= 'Z'; ++c) word_chars += c;
    for(c='0'; c <= '9'; ++c) word_chars += c;
    word_chars += ':';

  }

}

TextEditor::
~TextEditor()
{
  if(isTemporary())  // isTemporary is inherited from CursorWindow::repaint_handler
  {
      impl_->fullname_.remove();
  }


  rep_type::iterator first = impl_->lines_.begin(),
                     last  = impl_->lines_.end();
                
  while(first != last)
  {
    EditLine cur = *first++;

    delete cur.data();
  }
}

Viewer*
TextEditor::
app(ViewerManager*vm, std::string &fullname)
{
  if(fullname.size() == 0)
  {
    CursorWindow::FileSelector d("Edit File Name", "*", "");

    fullname = d.popup(vm->window(), 0);

    if(fullname == "")
      return 0;


    FileName tmp(fullname);
    tmp.convert_to_absolute_path();

    fullname = tmp;

  }
  return new TextEditor(vm,fullname);

}

std::string
TextEditor::
app_name="TextEditor";

std::string
TextEditor::
description() const
{
  return impl_->fullname_;
}

bool
TextEditor::
is_dirty() const
{
  return impl_->is_dirty_;
}

void
Impl::
display_line(viewport* vp, int line)
{
  int oppositeAttribute=0;
  int normalAttribute=0;

  if(is_marked(line))
  {
    if(active_)
    {
      normalAttribute   =  ViewerManager::highlighted_att;
      oppositeAttribute =  ViewerManager::inactive_mark_att;
    }
    else
    {
      normalAttribute    = ViewerManager::inactive_mark_att;
      oppositeAttribute  = ViewerManager::highlighted_att;
    }
  }
  else
  if(line == bottom_line_)
  {
    normalAttribute   =  ViewerManager::bottom_att;
    oppositeAttribute =  ViewerManager::normal_att;
  }
  else
  {
    normalAttribute   = ViewerManager::normal_att;
    oppositeAttribute = ViewerManager::bottom_att;
  }

  vp->set_text_attribute(normalAttribute);

  if(line < size() )
  {
    EditLine &editLine = line_text(line);

    std::string str = editLine;

    CursorWindow::row_col start_curpos;

    if(editLine.annotation_ && page_left_ == 0)
    {
        start_curpos = vp->curpos();
    }


    if(str.size() >= (size_t)(page_left_))
    {
      std::string::iterator first = str.begin() + page_left_,
                            last  = str.end();

      vp->write(first,last);
    }

    if(editLine.annotation_ && page_left_ == 0)
    {
        CursorWindow::row_col end_curpos = vp->curpos();

        vp->set_curpos(start_curpos);

        vp->set_text_attribute(oppositeAttribute);

        vp->write(editLine.annotation_.displayValue());

        vp->set_text_attribute(normalAttribute);

        vp->set_curpos(end_curpos);

    }



      vp->fill_to_eol();

  }
  else
      vp->fill_to_eol();
}


////////////////// page down handler ///////////////////////////////////

AddMapping page_down_func(CursorWindow::func_next, &Impl::page_down, NonModifying);

bool
Impl::
page_down(viewport* vp, Viewer*v, input_event const *e)
{
  size_t page_offset = cur_ - page_top_;

  page_top_ += vp->size().row_ - first_row_;

  if(page_top_ >= size())
    page_top_ = size() -1;

  if(page_top_ < 0)
    page_top_ = 0;

  cur_ = page_top_ + page_offset;

  if(cur_ >= size())
    cur_ = size() -1;

  if(cur_ < 0)
    cur_ = 0;

  paint_page(vp,v);

  return false;
}

////////////////// page down handler ///////////////////////////////////
////////////////// page up handler ///////////////////////////////////

AddMapping page_up_func(CursorWindow::func_prior, &Impl::page_up, NonModifying);

bool
Impl::
page_up(viewport* vp, Viewer*v, input_event const *e)
{
  size_t page_offset = cur_ - page_top_;

  if(page_top_ == 0)
  {
    cur_ = 0;
  }
  else
  {
    page_top_ -= vp->size().row_ - first_row_;

    if(page_top_ < 0)
      page_top_ = 0;

    cur_ = page_top_ + page_offset;

    if(cur_ < 0)
      cur_ = size() -1;

    if(cur_ < 0)
      cur_ = 0;
  }

  paint_page(vp,v);

  return false;
}

////////////////// page up handler ///////////////////////////////////

////////////////// tab right handler ///////////////////////////////////

AddMapping tab_right_func(CursorWindow::func_tab, &Impl::tab_right, NonModifying);

bool
Impl::
tab_right(viewport* vp, Viewer*v, input_event const *e)
{
  if(cur_ >= size() || cur_ < 0)
    return false;

  col_ += htab_;

  make_col_visible(vp,v);

  return false;

}

////////////////// tab right handler ///////////////////////////////////

////////////////// char right handler ///////////////////////////////////

AddMapping char_right_func(CursorWindow::func_right, &Impl::char_right, NonModifying);

bool
Impl::
char_right(viewport* vp, Viewer*v, input_event const *e)
{
  if(cur_ >= size() || cur_ < 0)
    return false;

  ++col_;

  make_col_visible(vp,v);

  return false;

}

void
Impl::
make_col_visible(viewport* vp, Viewer* v)
{
  row_col size = vp->size();

  size_t  avail_cols = vwidth(size);

  if(col_ < page_left_)
  {
    // need to scroll left

     page_left_ = col_;
     page_left_ /= htab_;
     page_left_ *= htab_;

     paint_page(vp,v);
  }
  else
  if(col_ - page_left_ >= avail_cols)
  {
    // need to scroll right

    size_t target_col =  col_ / htab_;
           target_col += 1;
           target_col *= htab_;
        
    page_left_ = target_col - avail_cols;

    paint_page(vp,v);

  }
  else
  {
    restore_cursor(vp);
  }

}

////////////////// char right handler ///////////////////////////////////

////////////////// char left handler ///////////////////////////////////

AddMapping char_left_func(CursorWindow::func_left, &Impl::char_left, NonModifying);

bool
Impl::
char_left(viewport* vp, Viewer*v, input_event const *e)
{
  if(cur_ >= size() || cur_ < 0)
    return false;

  --col_;

  if( (int)(col_) < 0)
    col_ = 0;

  make_col_visible(vp,v);

  return false;

}

////////////////// char left handler ///////////////////////////////////

////////////////// line down handler ///////////////////////////////////

AddMapping line_down_func(CursorWindow::func_down, &Impl::line_down, NonModifying);

bool
Impl::
line_down(viewport* vp, Viewer*v, input_event const *e)
{
  ++cur_;

  make_line_visible(vp,v,false);

  return false;

}

void
Impl::
make_line_visible(viewport* vp, Viewer* v, bool center)
{
  if(cur_ >= size())
    cur_ = size() - 1;

  if(cur_ < 0)
    cur_ = 0;

  row_col size = vp->size();

  int avail_rows = vheight(size);

  int row_offset = cur_ - page_top_;

  if(row_offset < 0)
  {
    if(center)
      page_top_ = cur_ - avail_rows/2;
    else
      page_top_ = cur_;

    if(page_top_ < 0)
      page_top_ = 0;

    paint_page(vp,v);
  }
  else
  if(row_offset >= avail_rows)
  {
    if(center)
      page_top_ = cur_ - (avail_rows/2 - 1);
    else
      page_top_ = cur_ - (avail_rows - 1);

    if(page_top_ < 0)
      page_top_ = 0;

    paint_page(vp,v);

  }
  else
  {
    restore_cursor(vp);
  }


}


////////////////// line down handler ///////////////////////////////////

////////////////// line up handler ///////////////////////////////////

AddMapping line_up_func(CursorWindow::func_up, &Impl::line_up, NonModifying);

bool
Impl::
line_up(viewport* vp, Viewer*v, input_event const *e)
{
  --cur_;

  make_line_visible(vp,v,false);

  return false;

}

////////////////// line up handler ///////////////////////////////////

////////////////// line home handler ///////////////////////////////////

AddMapping line_home_func(CursorWindow::func_home, &Impl::line_home, NonModifying);

bool
Impl::
line_home(viewport* vp, Viewer*v, input_event const *e)
{
  col_ = 0;

  make_col_visible(vp,v);

  return false;

}

////////////////// line home handler ///////////////////////////////////

////////////////// line end handler ///////////////////////////////////

AddMapping line_end_func(CursorWindow::func_end, &Impl::line_end, NonModifying);

bool
Impl::
line_end(viewport* vp, Viewer*v, input_event const *e)
{

  col_ = 0;

  if(size() > cur_)
    col_ = std::string(line_text(cur_)).size();

  make_col_visible(vp,v);

  return false;

}

////////////////// line end handler ///////////////////////////////////

////////////////// file top handler ///////////////////////////////////

AddMapping file_top_func(CursorWindow::func_top, &Impl::file_top, NonModifying);

bool
Impl::
file_top(viewport* vp, Viewer*v, input_event const *e)
{

  cur_ = 0;

  make_line_visible(vp,v);

  return false;

}

////////////////// file top handler ///////////////////////////////////

////////////////// file bottom handler ///////////////////////////////////

AddMapping file_bottom_func(CursorWindow::func_bottom, &Impl::file_bottom, NonModifying);

bool
Impl::
file_bottom(viewport* vp, Viewer*v, input_event const *e)
{

  cur_ = size()-1;

  if(cur_ < 0)
    cur_ = 0;

  make_line_visible(vp,v);

  return false;

}

////////////////// file bottom handler ///////////////////////////////////


////////////////// line mark handler ///////////////////////////////////

AddMapping line_mark_func(CursorWindow::func_mark, &Impl::line_mark, NonModifying);

bool
Impl::
line_mark(viewport* vp, Viewer*v, input_event const *e)
{
  if(cur_ >= first_mark_ && cur_ < last_mark_ )
  {
    // clear extant mark
    first_mark_ = last_mark_ = -1;
  }
  else
  if(first_mark_ == last_mark_ )
  {
    // no previous mark:  mark this one line

    first_mark_ = cur_;
    last_mark_ = first_mark_ + 1;
  }
  else
  if(cur_ < first_mark_)
    first_mark_ = cur_;
  else
  if(cur_ >= last_mark_)
    last_mark_ = cur_ + 1;

  paint_page(vp,v);

  return false;
}

////////////////// line mark handler ///////////////////////////////////

////////////////// Impl::read_file ////////////////////////////////////

struct Accumulate_File_Lines
: public FileName::accumulate_lines
  //
  ///  This struct implements the bridge between the FileName::read_lines
  ///  method and the lines_ data member of the Impl class.  The lines_
  ///  data is a collection of pointers to std::string, not std::strings directly
  ///  so this class performs the needed copies and dereferences to make the
  ///  interface look right to FileName::read_lines.
  //
{

  rep_type &lines_;

  Accumulate_File_Lines(rep_type &lines)
  : lines_(lines)
  {
  }

  void push_back(std::string const &new_line)
  {
    lines_.push_back(new std::string(new_line));
  }

  std::string& back()
  {
    return *lines_.back().data();
  }

  void setBackHasCR(bool flag)
  {
     lines_.back().crRemoved_ = true; 
  }


};

void
Impl::
read_file()
{
  //
  //  Read the file into the lines_ data structure
  //

#ifdef _MSC_VER
    bool removeTrailingCROnUnix = false;
#else
    bool removeTrailingCROnUnix = true;
#endif

  Accumulate_File_Lines accum(lines_);


  { // handle utf encoding 

     // ask the user if conflating UTF extended characters plain ascii is ok

     utf::encoding type = utf::fileEncoding(fullname_.c_str());

     if(type != utf::utf7_enc && type != utf::utfunk_enc)
     {
        // we have some form of utf encoding.

        is_writeable_ = false;


        CursorWindow::Selection d("UTF Encoded File:  " + fullname_);
        
        std::string encodingName = std::string("    type = ") +  std::string(utf::encodingName(type));

        d += "Encoding type is not supported";
        d += encodingName;
        d += "Press Enter to down-convert to plain old ANSI characters";
        d += "Or, press ESC to abandon the edit attempt";
        
        if(d.popup(manager_->window()))
        {
            return;
        }

        nonstandard_encoding_ = true;
        
        FileContents contents;

        fullname_.slurp(&contents);

        if(!contents.ok())
        {
            CursorWindow::Selection d("Error reading file " + fullname_);
            d += "";
            d += "Press Enter to continue";

            d.popup(manager_->window());

            return;
        }

        char const *scan = contents.begin();
        char const *end   = contents.end();

        utf::streamEncoding(scan,end); // skip the encoding stuff

        utf::LineDecoder ld;

        while(scan < end)
        {
            std::wstring curline = ld.decode(type, scan, end);    

            std::wstring::const_iterator f = curline.begin(), l = curline.end();

            std::string tmp;

            while(f != l)
            {
                char c = utf::conflate(*f++);

                if(c  == '\r' && f == l)
                   break;

                tmp += c;
            }

            accum.push_back(tmp);

        }


        return; // populate lines with the conflated form of the file, then return 


     }

     
  } // handle utf encoding



  if(! fullname_.exists() )
  {
    manager_->window()->beep();
    is_new_ = true;
  }
  else
  {

    
    std::string errors = fullname_.read_lines(accum,0,removeTrailingCROnUnix);
    
    if(errors.size() != 0)
    {
      CursorWindow::Message d("Error reading file " + fullname_);
    
      d += errors;
      d += "";
      d += "Press Enter to continue";
    
      d.popup(manager_->window());
    
      return;
    
    }

    is_writeable_ = fullname_.is_writeable();

  }

  std::string extension = fullname_.extension();
  std::string basename  = fullname_.basename();

  use_tabs_=true;


  if(extension == ".c"   ||
     extension == ".cc"  ||
     extension == ".h"   ||
     extension == ".cpp" ||
     extension == ".cxx" ||
     extension == ".C"   
    )
  {
    htab_     = 2;
    use_tabs_ = false;
  }
  else
  if(basename == "makefile" ||
     basename == "Makefile"
    )
  {
    htab_ = 8;
  }
  else
    htab_ = 4;


  // check for binary data

  if(lines_.size())
  {
      {
          std::string::iterator   scan;

          utf::encoding type = utf::streamEncoding(scan = lines_[0].data()->begin(), lines_[0].data()->end());

          if(type != utf::utf7_enc)
          {
              
              manager_->window()->beep();
     
             CursorWindow::Message d("Warning:  Non-ANSII character Encoding");
     
             d += "";
             d += "";
             d += std::string("Warning: this file contains data of type:  ") + std::string(utf::encodingName(type));
             d += "";
             d += "This editor is not meant to be used with this kind of data";
             d += "Saving changes is disabled to prevent data destruction.";
             d += "";
             d += "You can use ^Z to re-enable writing to the file, but do";
             d += "so at your own risk.";
             d += "";
     
             
             d.popup(manager_->window());
             is_writeable_ = false;

             return;
          }
      }
      
      
      int count=0;
      bool hasBinary = false;


      for( ;   (!hasBinary)
            && (count < 10)
            && (count < lines_.size()); 
            ++count
         )
      {

          std::string const *p = lines_[count].data();

          if(!p)
            continue;
          

          std::string::const_iterator l=p->begin(),
                                      g=p->end();

          while(   !hasBinary 
                && (l != g)
               )
          {
            unsigned char c = *l++;

            if( c > '~')
            {
               hasBinary=true;
               break;
            }

            if(c < ' ')
            {
               if(    (c != '\r')
                  &&  (c != '\n')
                  &&  (c != 0x09)  // tab
                 )
               {
                   hasBinary = true;
                   break;
               }
            }

          }

      }  // processing characters in first 10 lines of the fil

      if(hasBinary)
      {
         manager_->window()->beep();

        CursorWindow::Message d("Warning:  Binary data");

        d += "";
        d += "";
        d += "Warning: this file contains binary data.";
        d += "";
        d += "This editor is not meant to be used with binary data.";
        d += "Saving changes is disabled to prevent data destruction.";
        d += "";
        d += "You can use ^Z to re-enable writing to the file, but do";
        d += "so at your own risk.";
        d += "";

        
        d.popup(manager_->window());
        is_writeable_ = false;

         
      }

  }  // check for binary



}
////////////////// Impl::read_file ////////////////////////////////////

////////////////// word next ////////////////////////////////////
AddMapping word_next_func(CursorWindow::func_nextwd, &Impl::word_next, NonModifying);

bool
Impl::
word_next(viewport* vp, Viewer*v, input_event const *e)
{
  std::string line = line_text(cur_);

  std::string::iterator first = line.begin(),
                        last  = line.end();

  if(col_ >= line.size())
  {

    if(cur_ < size())
    {
      ++cur_;
      col_ = 0;
      make_line_visible(vp,v);
    }
  }
                        
  if(cur_ < size() && col_ < line.size())
  {
    // not off the end of the line

    std::string::iterator start = first + col_;

    std::string::iterator next_word_start = StrTool::find_next_word(start,
                                                                    last,
                                                                    word_chars_.begin(),
                                                                    word_chars_.end()
                                                                   );
   if(next_word_start != last)
   {
      col_  += next_word_start - start;

   }
   else
   {
      col_ = line.size();
   }

   make_col_visible(vp,v);
  }

  return false;
}
////////////////// word next ////////////////////////////////////


////////////////// word prev ////////////////////////////////////
AddMapping word_prev_func(CursorWindow::func_prevwd, &Impl::word_prev, NonModifying);

bool
Impl::
word_prev(viewport* vp, Viewer*v, input_event const *e)
{
  std::string line = line_text(cur_);

  std::string::iterator first = line.begin(),
                        last  = line.end();

  if(col_)
  {
    std::string::iterator start = first + col_;

    std::string::iterator word_begin = StrTool::find_prev_word(start,
                                                               first,
                                                               last,
                                                               word_chars_.begin(),
                                                               word_chars_.end()
                                                              );
    col_ = word_begin - first;

    make_col_visible(vp,v);


  }
  else
  {
    if(cur_)
    {
      // move back to the end of the previous line

      --cur_;

      line = line_text(cur_);

      col_ = line.size();

      make_line_visible(vp,v);

    }
  }

  return false;
}
////////////////// word prev ////////////////////////////////////


////////////////// word delete ////////////////////////////////////
AddMapping word_del_func(CursorWindow::func_dw, &Impl::word_delete, Modifying);

bool
Impl::
word_delete(viewport* vp, Viewer*v, input_event const *e)
{
  std::string line = line_text(cur_);

  if(col_ >= line.size())
  {

    if(cur_ < size())
    {
      ++cur_;
      col_ = 0;
      line = line_text(cur_);
      make_line_visible(vp,v);
    }
  }

  std::string::iterator first = line.begin(),
                        last  = line.end();


                        
  if(cur_ < size() && col_ < line.size())
  {
    // not off the end of the line

    std::string::iterator start = first + col_;

    std::string::iterator next_word_start = StrTool::find_next_word(start,
                                                                    last,
                                                                    word_chars_.begin(),
                                                                    word_chars_.end()
                                                                   );
    size_t bytes_till_next_word;
                                                                
    if(next_word_start != last)
    {
       bytes_till_next_word = next_word_start - start;

    }
    else
    {
       bytes_till_next_word = line.size() - col_;;
    }


    last = first + ( col_ + bytes_till_next_word );

    line.erase(start,last);

    replace_line(cur_, line);

    row_col curpos = vp->curpos();

    make_col_visible(vp,v);

    vp->set_curpos(curpos.row_, left_);

    display_line(vp, cur_);

  }

  restore_cursor(vp);

  return false;
}
////////////////// word delete ////////////////////////////////////


////////////////// char insert ////////////////////////////////////
AddMapping char_insert_func(CursorWindow::func_data, &Impl::char_insert, Modifying);

bool
Impl::
char_insert(viewport* vp, Viewer*v, input_event const *e)
{
  if(cur_ >= size())
  {
    if(cur_ == 0)
    {
      lines_.push_back(new std::string(""));
      paint_page(vp,v);
      restore_cursor(vp);
    }
    else
      return false;  // something is wrong with the data structures
  }


  std::string line = line_text(cur_);

  std::string::iterator first = line.begin();

  if(col_ <= line.size())
  {
    line.insert(first+col_, e->value_);
  }
  else
  {
     line.append(col_ - line.size(), ' ');
     line += (char)(e->value_);
  }

  ++col_;

  replace_line(cur_, line);

  row_col curpos = vp->curpos();

  make_col_visible(vp,v);

  vp->set_curpos(curpos.row_, left_);

  display_line(vp, cur_);

  restore_cursor(vp);

  return false;
}
////////////////// char insert ////////////////////////////////////


////////////////// char delete ////////////////////////////////////
AddMapping char_delete_func(CursorWindow::func_dc, &Impl::char_delete, Modifying);

bool
Impl::
char_delete(viewport* vp, Viewer*v, input_event const *e)
{
  if(cur_ >= size())
    return false;


  std::string line = line_text(cur_);

  if(col_ >= line.size())
  {
    line_join(vp,v,e);

    line = line_text(cur_);

    col_ = line.size();

    make_col_visible(vp,v);

    return false;

  }


  std::string::iterator first = line.begin();

  if(col_ < line.size())
  {
    line.erase(first+col_);
  }

  replace_line(cur_,line);

  row_col curpos = vp->curpos();

  make_col_visible(vp,v);

  vp->set_curpos(curpos.row_, left_);

  display_line(vp, cur_);

  restore_cursor(vp);

  return false;
}
////////////////// char delete ////////////////////////////////////

////////////////// line delete ////////////////////////////////////

AddMapping line_del_func(CursorWindow::func_dl, &Impl::line_del, Modifying);

bool
Impl::
line_del(viewport* vp, Viewer*v, input_event const *e)
{
  if(cur_ >= size())
    return false;

  // erase the extant paste buffer

  Viewer::paste_buffer_type &paste_buffer = *v->paste_buffer();

  paste_buffer.erase(paste_buffer.begin(), paste_buffer.end());

  paste_buffer.insert(paste_buffer.begin(), *line_text(cur_).data() );

  v->commitPasteBuffer();

  erase_line( line_address(cur_), cur_ );

  if(cur_ >= size())
    cur_ = size() -1;

  if(cur_ < 0)
    cur_ = 0;

  if(first_mark_ != last_mark_)
  {
    //
    // handle adjustments to the marked region
    //
    if(cur_ >= first_mark_ && cur_ < last_mark_)
    {
      // delete a line in the marked region

      if(cur_ == first_mark_)
        ++first_mark_;
      else
        --last_mark_;
    }
    else
    if(cur_ < first_mark_)
    {
      --first_mark_;
      --last_mark_;
    }
  }

  paint_page(vp,v);

  return false;
}
////////////////// line delete ////////////////////////////////////


////////////////// line split ////////////////////////////////////

AddMapping line_split_func(CursorWindow::func_enter, &Impl::line_split, Modifying);

bool
Impl::
line_split(viewport* vp, Viewer*v, input_event const *e)
{
  if(cur_ >= size())
  {
    // no lines defined yet

    lines_.push_back( new std::string("") );
    col_ = 0;
    paint_page(vp,v);
    return false;
  }

  std::string line = line_text(cur_);

  size_t first_non_blank_col=0;

  while(first_non_blank_col < line.size())
  {
    if(line[first_non_blank_col] != ' ')
      break;

    ++first_non_blank_col;
  }

  if(first_non_blank_col >= line.size())
    first_non_blank_col = 0;
  else
    left_margin_ = first_non_blank_col;

  if(col_ >= line.size())
  {
    // already past the end of line, just insert a new one

    rep_type::iterator cur_line_ptr = lines_.find(cur_);

    ++cur_line_ptr;

    lines_.insert(cur_line_ptr, new std::string(""));

    ++cur_;
    col_ = left_margin_;
    make_col_visible(vp,v);
    make_line_visible(vp,v);

    paint_page(vp,v);

    UndoOperation* op = push_undo_stack();

    op->code_    = UndoOperation::erase_line;
    op->lineno_  = cur_;


    return false;

  }

  // ok, add a new line which is the tail of the current line
  // and rip the tail off the current line


  rep_type::iterator cur_line_ptr = lines_.find(cur_);

  ++cur_line_ptr;

  std::string new_line = line;

  new_line.erase(new_line.begin(), new_line.begin() + col_);

  StrTool::remove_leading(new_line,' ');

  {
    UndoOperation *op = push_undo_stack();


    op->code_ = UndoOperation::join_lines;
    op->lineno_ = cur_;
    op->text_ = line;

  }


  std::string line_to_insert(left_margin_, ' ');
              line_to_insert += std::string(new_line);

  lines_.insert(cur_line_ptr, new std::string(line_to_insert) );

  line.erase(line.begin() + col_, line.end());

  replace_line(cur_,line,false);

  ++cur_;
  col_ = left_margin_;

  make_line_visible(vp,v);
  paint_page(vp,v);


  return false;
}
////////////////// line split ////////////////////////////////////

////////////////// line clreol ////////////////////////////////////

AddMapping line_clreol_func(CursorWindow::func_clreol, &Impl::line_clreol, Modifying);

bool
Impl::
line_clreol(viewport* vp, Viewer*v, input_event const *e)
{
  if(cur_ >= size())
    return false;

  std::string line = line_text(cur_);

  std::string::iterator first = line.begin();

  if(col_ < line.size())
  {
    line.erase(first+col_, line.end());
  }

  replace_line(cur_,line);

  row_col curpos = vp->curpos();

  make_col_visible(vp,v);

  vp->set_curpos(curpos.row_, left_);

  display_line(vp, cur_);

  restore_cursor(vp);

  return false;
}
////////////////// line clreol ////////////////////////////////////


////////////////// char delprv ////////////////////////////////////
AddMapping char_delprv_func(CursorWindow::func_dc_prev, &Impl::char_delprv, Modifying);

bool
Impl::
char_delprv(viewport* vp, Viewer*v, input_event const *e)
{
  if(cur_ >= size())
    return false;

  std::string line = line_text(cur_);

  if(col_ == 0)
  {
    // join this line to the previous line

    if(cur_)
    {
      // if not at top of the file

      std::string prevline = line_text(cur_-1);

      size_t prevlength = prevline.size();

      prevline += line;

      replace_line(cur_-1,prevline);

      erase_line(line_address(cur_), cur_);

      --cur_;

      col_ = prevlength;

      make_col_visible(vp,v);
      make_line_visible(vp,v);
      paint_page(vp,v);


    }

    return false;
  }
  else
  if(col_ > line.size())
  {
    // just move the cursor back one character

    --col_;
    make_col_visible(vp,v);
    restore_cursor(vp);
    return false;
  }

  // ok, delete the previous character and leave the cursor
  // in tht position

  --col_;
  make_col_visible(vp,v);

  return char_delete(vp,v,e);

}
////////////////// char delprv ////////////////////////////////////

////////////////// find first ////////////////////////////////////
AddMapping find_func(CursorWindow::func_find, &Impl::find_first, NonModifying);

bool
Impl::
find_first(viewport* vp, Viewer*v, input_event const *e)
{
  CursorWindow::Dialog dialog("Find string");

  char const *mix = "Y";

  if(mixed_case_ == false)
    mix = "N";

  dialog += new CursorWindow::Dialog::String("case",
                                             "Mixed case",
                                             mix,
                                             2
                                            );


  dialog += new CursorWindow::Dialog::String("mode",
                                             "(W)ord, (R)egexp, or (S)tring",
                                             (is_regex_ ? "R" : (match_words_ ? "W" : "S")),
                                             2
                                            );

  dialog += new CursorWindow::Dialog::String("text",
                                             "Search for",
                                             search_text_,
                                             40
                                            );

  dialog += new CursorWindow::Dialog::Ok("ok","");

  dialog.set_first_input_field("text");


  if( dialog.popup(vp->window()) )
    return false;

  std::string mixed_case = dialog.element_value("case");

  if(   mixed_case.size() 
     && (mixed_case[0] == 'n' || mixed_case[0] == 'N')
    )
  {
    mixed_case_ = false;
  }
  else
  if(   mixed_case.size() 
     && (mixed_case[0] == 'y' || mixed_case[0] == 'Y')
    )
  {
    mixed_case_ = true;
  }


  std::string match_words = dialog.element_value("mode");

  match_words_ = false;
  is_regex_    = false;

  if(   match_words.size() 
     && (match_words[0] == 'r' || match_words[0] == 'R')
    )
  {
    is_regex_ =  true;
  }
  else
  if(   match_words.size() 
     && (match_words[0] == 'w' || mixed_case[0] == 'W')
    )
  {
    match_words_ = true;
  }

  std::string search_text = dialog.element_value("text");

  if(search_text.size() == 0)
    return false; // ignore searches for nothing

  search_text_ = search_text;

  int    saved_line   = cur_;
  size_t saved_col    = col_;

  --col_;  // first time through, let it try the current column
           // too.  find next always skips the current character

  if(is_regex_ && search_text_[0] == '^')
    --cur_;  // find next will increment it

  if(!find_next_helper(vp,v,e))
  {
    cur_ = saved_line;
    col_ = saved_col;
    vp->beep();
  }
  else
  {
    make_col_visible(vp,v);
    make_line_visible(vp,v);
  }

  return false;
}
////////////////// find first ////////////////////////////////////

////////////////// find prev ////////////////////////////////////
AddMapping find_func_prev(CursorWindow::func_findprv, &Impl::find_prev, NonModifying);

bool
Impl::
find_prev(viewport* vp, Viewer*v, input_event const *e)
{
  int    saved_line   = cur_;
  size_t saved_col    = col_;

  if(!find_prev_helper(vp,v,e))
  {
    cur_ = saved_line;
    col_ = saved_col;
    vp->beep();
  }
  else
  {
    make_col_visible(vp,v);
    make_line_visible(vp,v);
  }

  return false;
}

bool
Impl::
find_prev_helper(viewport* vp, Viewer*v, input_event const *e)
  /// return true if the string is found
{
  // the caller must save/restore cur_ and col_

  --col_;


  if(is_regex_)
  {
    SimpleRegex *p = &regexp_;

    p->SimpleRegex::~SimpleRegex();  // destruct current regex

    new(&regexp_) SimpleRegex(search_text_,  (mixed_case_ ? "i" : "") );

    //    regexp_ = search_text_;

    if(search_text_[0] == '^')
    {
      ++cur_;
      col_ = 0;
    }
  }
  
  


  bool force_col_to_end = false;

  while(cur_ >= 0)
  { // search all lines till beginning of file

    // get the text of the current line, tab expanded

    std::string curline = line_text(cur_);

    // make sure the line isn't empty and that the cursor
    // variable is properly positioned

    if(force_col_to_end)
    {
      col_ = curline.size()-1;

      if(col_ < 0)
      {
        --cur_;
        continue;
      }
    }

    // since backwards searches are hard, lets use a forward search
    // and make a list of all matches, as column numbers in this line,
    // then search around in that list for column numbers

    std::vector<int> matching_columns;

    for(int scan = 0; size_t(scan) < curline.size(); )
    {

      std::string::iterator where;


      if(is_regex_)
      {
        int startpos = scan;
        
        if(search_text_[0] == '^')
          startpos = 0;

        int count = regexp_(curline, startpos, curline.size());
        
        if(count > 0)
        {
          where = (curline.begin() + startpos + regexp_.matches()[0].offset);
        }
        else
          where = curline.end();
        
      }
      else
      {
        where = std::search(curline.begin() + scan,
                            curline.end(),
                            search_text_.begin(),
                            search_text_.end(),
                            CharCompare(this)
                           );
      }

                                        
      if(where != curline.end())
      {
        // found a match
        
        scan = where - curline.begin();
        
        matching_columns.push_back( scan );
        
        ++scan;
      }
      else
      {
        break;
      }
    }

    // matching_columns now has a list of all matches on this line


    if( int(col_) >= 0 )
    {
      // not past beginning of current line -- so try a search here

      // scan across the list of matches, until the current match
      // is farther to the right than the current column

      int match;

      for(match = 0; size_t(match) < matching_columns.size(); ++match)
      {
        if(matching_columns[match] > int(col_) )
          break;
      }

      --match;

      if(match >= 0)
      {
          std::string::iterator where = curline.begin() + matching_columns[match];
        
         if(match_words_)
         {
           // if matching only whole words, check to see if this is in fact a word
           // and if not, don't stop searching
        
           bool begin_starts_word = (where == curline.begin());
        
           if(!begin_starts_word)
             begin_starts_word = !is_word_char(where[-1]);
        
           std::string::iterator word_end = where + search_text_.size();
        
           bool end_ends_word = (word_end == curline.end());
        
           if(!end_ends_word)
             end_ends_word = !is_word_char(*word_end);
        
           if(begin_starts_word && end_ends_word)
           {
             col_ = where - curline.begin();
             return true;
           }
        
           col_ = where - curline.begin();
           --col_;
           force_col_to_end = false;
        
           continue;
        
        
         }
         else
         {
           // any old match will do, so quit
           col_ = where - curline.begin();
           return true;
         }
      }
    }

    --cur_;

    force_col_to_end=true;

  }


  return false;

}
////////////////// find prev ////////////////////////////////////

////////////////// find next ////////////////////////////////////
AddMapping find_func_next(CursorWindow::func_findnxt, &Impl::find_next, NonModifying);

bool
Impl::
find_next(viewport* vp, Viewer*v, input_event const *e)
{
  int    saved_line   = cur_;
  size_t saved_col    = col_;

  if(!find_next_helper(vp,v,e))
  {
    cur_ = saved_line;
    col_ = saved_col;
    vp->beep();
  }
  else
  {
    make_col_visible(vp,v);
    make_line_visible(vp,v);
  }

  return false;
}



bool
Impl::
find_next_helper(viewport* vp, Viewer*v, input_event const *e)
  /// return true if the string is found
{
  // the caller must save/restore cur_ and col_

  ++col_;  // note that replace_next() decrements this prior to the call to
           // replace_next_helper() which calls this function.

  if(is_regex_)
  {
    SimpleRegex *p = &regexp_;

    p->SimpleRegex::~SimpleRegex();  // destruct current regex

    new(&regexp_) SimpleRegex(search_text_,  (mixed_case_ ? "i" : "") );

    //    regexp_ = search_text_;

    if(search_text_[0] == '^')
    {
      ++cur_;
      col_ = 0;
    }
  }


  while(cur_ < lines_.size())
  { // search all lines till eof

    std::string curline = line_text(cur_);

    if(col_ < curline.size())
    {

      int startpos = col_;

      // not past end of current line -- so try a search here

      std::string::iterator where;


      if(is_regex_)
      {
        
        if(search_text_[0] == '^')
          startpos = 0;

        int count = regexp_(curline, startpos, curline.size());
        
        if(count > 0)
        {
          where = (curline.begin() + regexp_.matches()[0].offset);
        }
        else
          where = curline.end();
        
      }
      else
      {
        where = std::search(curline.begin() + col_,
                                                  curline.end(),
                                                  search_text_.begin(),
                                                  search_text_.end(),
                                                  CharCompare(this)
                                                 );
      }
                                        
      if(where != curline.end())
      {
        
         if(match_words_ && !is_regex_)
         {
           // if matching only whole words, check to see if this is in fact a word
           // and if not, don't stop searching
        
           bool begin_starts_word = (where == curline.begin());
        
           if(!begin_starts_word)
             begin_starts_word = !is_word_char(where[-1]);
        
           std::string::iterator word_end = where + search_text_.size();
        
           bool end_ends_word = (word_end == curline.end());
        
           if(!end_ends_word)
             end_ends_word = !is_word_char(*word_end);
        
           if(begin_starts_word && end_ends_word)
           {
             col_ = where - curline.begin();
             return true;
           }
        
           col_ = where - curline.begin();
           ++col_;
        
           continue;
        
        
         }
         else
         {
           // any old match will do, so quit
           col_ = where - curline.begin();



           return true;
         }
      }
    }

    ++cur_;
    col_ = 0;

  }


  return false;

}

////////////////// find next ////////////////////////////////////




////////////////// replace first ////////////////////////////////////

AddMapping replace_first_func(CursorWindow::func_repl, &Impl::replace_first, Modifying);

bool
Impl::
replace_first(viewport* vp, Viewer*v, input_event const *e)
  // Prompt the user for a find/replacement pair and replace the first
  // match.
{
  CursorWindow::Dialog dialog("Replace string");


  char const *mix = "Y";

  if(mixed_case_ == false)
    mix = "N";

  dialog += new CursorWindow::Dialog::String("case",
                                             "Mixed case",
                                             mix,
                                             2
                                            );

  dialog += new CursorWindow::Dialog::String("mode",
                                             "(W)ord, (R)egexp, or (S)tring",
                                             (is_regex_ ? "R" : (match_words_ ? "W" : "S")),
                                             2
                                            );

  dialog += new CursorWindow::Dialog::String("all", "Replace (a)ll or (n)ext",
                                             "n",
                                             2
                                            );

  dialog += new CursorWindow::Dialog::String("find",
                                             "Search for",
                                             search_text_,
                                             40
                                            );
                                        
  dialog += new CursorWindow::Dialog::String("replacement",
                                             "Replace with",
                                             replacement_,
                                             40
                                            );
                                        

  dialog += new CursorWindow::Dialog::Ok("ok","");

  dialog.set_first_input_field("find");

  if( dialog.popup(vp->window()) )
    return false;

  std::string mixed_case = dialog.element_value("case");

  if(   mixed_case.size() 
     && (mixed_case[0] == 'n' || mixed_case[0] == 'N')
    )
  {
    mixed_case_ = false;
  }
  else
  if(   mixed_case.size() 
     && (mixed_case[0] == 'y' || mixed_case[0] == 'Y')
    )
  {
    mixed_case_ = true;
  }

  std::string match_words = dialog.element_value("mode");

  match_words_ = false;
  is_regex_    = false;

  if(   match_words.size() 
     && (match_words[0] == 'r' || match_words[0] == 'R')
    )
  {
    is_regex_ =  true;
  }
  else
  if(   match_words.size() 
     && (match_words[0] == 'w' || mixed_case[0] == 'W')
    )
  {
    match_words_ = true;
  }


  std::string search_text = dialog.element_value("find");
  std::string replacement = dialog.element_value("replacement");

  if(search_text.size() == 0)
    return false;

  search_text_ = search_text;
  replacement_ = replacement;

  int    saved_line = cur_;
  size_t saved_col  = col_;


  if(is_regex_ && search_text_[0] == '^')
    --cur_;  // replace_next_helper, through find next helper, will increment it

  if( tolower(dialog.element_value("all")[0]) == 'a')
  {
    // replace all matches

    UndoOperation::undo_stack* nested_changes = new UndoOperation::undo_stack;

    undo_stack_.swap(*nested_changes);  // temporarily swap the real undo
                                        // stack withe the new one just
                                        // created so we can then place the
                                        // changes caused by replace_next_helper
                                        // to be accumulated into the new
                                        // stack -- which will then be placed on
                                        // the real undo stack

    bool found_one=false;

    for(;;)
    {
        --col_;  // will be incremented immediately by replace_next_helper()
                 // through it's call to find_next_helper()

        if(!replace_next_helper(vp,v,e))
           break;

      found_one = true;
    }

    undo_stack_.swap(*nested_changes);  // put the old undo stack back


    UndoOperation* op = push_undo_stack();

    op->code_ = UndoOperation::nested_undo_ops;

    op->nested_undo_stack_ = nested_changes;

    op->nested_first_mark_ = first_mark_;
    op->nested_last_mark_  = last_mark_;

    if(!found_one)
    {
      vp->beep();
      cur_ = saved_line;
      col_ = saved_col;
    }
    else
    {
      // cursor left at end of replacement
    
      make_col_visible(vp,v);
      make_line_visible(vp,v);
      paint_page(vp,v);

    }

  }
  else
  {
    // replace only the next match

  --col_;  // replace_next_helper will increment col_
           // immediately through it's call to find_next_helper

    if(replace_next_helper(vp,v,e))
    {
      // cursor left at end of replacement
    
      make_col_visible(vp,v);
      make_line_visible(vp,v);
      paint_page(vp,v);
    
    }
    else
    {
      cur_ = saved_line;
      col_ = saved_col;
      vp->beep();
    }
  }


  return false;
}

bool
Impl::
replace_next_helper(viewport* vp, Viewer*v, input_event const *e)
  ///  replace at most one string -- leaves cursor at end of replacement
  ///  or some random place -- remember to fix it back the way it was
  ///  if the replace fails.  In any case make sure the cursor is visible
{

  // note:  replace_next() decrements col_ before calling this function
  //        and find_next_helper increments it before doing anything else

  if(!find_next_helper(vp,v,e))
  {
    return false;
  }

  if(is_regex_)
  {
    std::string curline = line_text(cur_);

    std::string replacement = regexp_.substitute(curline, replacement_);

    replace_line(cur_,replacement);

    col_ += regexp_.matches()[0].length;

    col_ += replacement.size() - curline.size();

  }
  else
  {

    // found it, cursor is at the beginning of the word

    std::string curline = line_text(cur_);

    std::string::iterator target = curline.begin() + col_;

    curline.erase(target, target + search_text_.size());

    curline.insert(col_, replacement_);

    replace_line(cur_,curline);

    col_ += replacement_.size();

  }

  return true;

}


////////////////// replace first ////////////////////////////////////

////////////////// replace next ////////////////////////////////////
AddMapping replace_next_func(CursorWindow::func_replnext, &Impl::replace_next, Modifying);

bool
Impl::
replace_next(viewport* vp, Viewer*v, input_event const *e)
{

  int    saved_line   = cur_;
  size_t saved_col    = col_;


  --col_;  // it will be incremented immediately find_next_helper which is called
           // the first thing in find_next_helper

  if(!replace_next_helper(vp,v,e))
  {
    cur_ = saved_line;
    col_ = saved_col;
    vp->beep();
  }
  else
  {

    make_col_visible(vp,v);
    make_line_visible(vp,v);

    vp->set_curpos(first_row_ + cur_ - page_top_, left_);
    display_line(vp, cur_);

    restore_cursor(vp);

  }

  return false;
}

////////////////// replace next ////////////////////////////////////

////////////////// delete block ////////////////////////////////////
AddMapping delete_block_func(CursorWindow::func_db, &Impl::block_delete, Modifying);

bool
Impl::
block_delete(viewport* vp, Viewer*v, input_event const *e)
  // cut the marked block to the paste buffer
{
  if(first_mark_ == last_mark_)
  {
    vp->beep();
    return false;  // no marked block
  }


  // erase the extant paste buffer, fill it with the deleted
  // items

  Viewer::paste_buffer_type &paste_buffer = *v->paste_buffer();

  paste_buffer.erase(paste_buffer.begin(), paste_buffer.end());

  // Create an undo operation to hold the deleted lines

  UndoOperation *op = push_undo_stack();

  op->code_   = UndoOperation::insert_block;
  op->lineno_ = first_mark_;

  // delete the lines and copy them to the paste
  // buffer

  while(last_mark_ != first_mark_)
  {
    // delete marked lines starting at tail to simplify work

    --last_mark_;

    rep_type::iterator scan = line_address(last_mark_);

    if(scan != lines_.end())
    {
      EditLine &el =*scan;

      paste_buffer.insert(paste_buffer.begin(), el);
      op->block_.push_back(el);

      erase_line(scan, last_mark_, false);  // erase but don't save undo info
                                            // since we are doing it already
    }

  }

  cur_ = first_mark_;
  col_ = 0;
  page_left_ = 0;
  page_top_  = first_mark_;

  first_mark_ = last_mark_ = -1;

  paint_page(vp, v);

  v->commitPasteBuffer();


  return false;
}
////////////////// delete block ////////////////////////////////////

////////////////// paste block ////////////////////////////////////
AddMapping paste_block_func(CursorWindow::func_paste, &Impl::block_paste, Modifying);

bool
Impl::
block_paste(viewport* vp, Viewer*v, input_event const *e)
  // cut the marked block to the paste buffer
{
  Viewer::paste_buffer_type &paste_buffer = *v->paste_buffer();

  Viewer::paste_buffer_type::iterator first = paste_buffer.begin(),
                                      last  = paste_buffer.end();


  UndoOperation* op = push_undo_stack();

  op->code_ = UndoOperation::delete_block;
  op->count_= 0;
  op->lineno_ = cur_;
                                
  while(first != last)
  {
    std::string const *insertedLine = &( *first++ );

    std::string buffer;

    if(use_tabs_)
    {
       // if we are going to save this file with tabs, then we should 
       // tab compress lines from the paste buffer by replacing groups of 8 
       // spaces with tabs.

       buffer = StrTool::pack_tabs(insertedLine->c_str());

       insertedLine = &buffer;

    }
    else
    {
       // if we are not saving with tabs, expand any tabs that might be in the line

       buffer = StrTool::expand_tabs(insertedLine->c_str());

       insertedLine = &buffer;

    }


    lines_.insert( lines_.find(cur_), new std::string(*insertedLine) );

    ++cur_;
    ++op->count_;
  }

  make_line_visible(vp, v);
  paint_page(vp,v);

  return false;
}
////////////////// paste block ////////////////////////////////////



////////////////// file insert ////////////////////////////////////
AddMapping file_insert_func(CursorWindow::func_insfile, &Impl::file_insert, Modifying);

bool
Impl::
file_insert(viewport* vp, Viewer*v, input_event const *e)
  // read a file into the current buffer above the current line
{

  FileName f(fullname_);

  CursorWindow::FileSelector d("Insert File",  f.dirname() + "*", "");

  std::string rv = d.popup(manager_->window(), v);

  if(rv.empty())
     return false;


  std::ifstream s(rv.c_str());

  if(!s.good())
  {
      CursorWindow::Message  d("Error");

      d += std::string("Can not open file ") + rv;

      d.popup(manager_->window());

      return false;
  }

  UndoOperation* op = push_undo_stack();

  op->code_ = UndoOperation::delete_block;
  op->count_= 0;
  op->lineno_ = cur_;
      

  std::string line;

  while(std::getline(s, line))
  {
    if(!use_tabs_)
       line = StrTool::expand_tabs(line);

    lines_.insert( lines_.find(cur_), new std::string(line) );

    ++cur_;
    ++op->count_;
  }

  make_line_visible(vp, v);
  paint_page(vp,v);

  return false;
}
////////////////// file insert ////////////////////////////////////

////////////////// copy block ////////////////////////////////////
AddMapping copy_block_func(CursorWindow::func_cb, &Impl::block_copy, NonModifying);

bool
Impl::
block_copy(viewport* vp, Viewer*v, input_event const *e)
  // copy the marked block to the paste buffer
{
  if(first_mark_ == last_mark_)
    return false;  // no marked block

  Viewer::paste_buffer_type &paste_buffer = *v->paste_buffer();

  paste_buffer.erase(paste_buffer.begin(), paste_buffer.end());

  // copy the lines and copy them to the paste
  // buffer

  while(last_mark_ != first_mark_)
  {
    // copy marked lines starting at tail to simplify work

    --last_mark_;

    rep_type::iterator scan = lines_.find(last_mark_);

    if(scan != lines_.end())
    {
      EditLine &el =*scan;

      paste_buffer.insert(paste_buffer.begin(), *el.data());

    }

  }

  v->commitPasteBuffer();  // write the paste buffer to the os

  first_mark_ = last_mark_ = -1;

  paint_page(vp, v);


  return false;
}
////////////////// copy block ////////////////////////////////////

////////////////// save block ////////////////////////////////////
AddMapping save_block_func(CursorWindow::func_cb, &Impl::block_copy, NonModifying);

bool
Impl::
block_save(viewport* vp, Viewer*v, input_event const *e)
  // save the block to a file
{
  if(first_mark_ == last_mark_)
  {
    vp->beep();

    return false;  // no marked block
  }

  FileName f(fullname_);

  CursorWindow::FileSelector d("Save Block to File",  f.dirname() + "*", "");

  std::string target = d.popup(manager_->window(), v);

  if(target.empty())
     return false;

  std::ofstream  targetFile(target.c_str());

  if(!targetFile.good())
  {
     manager_->window()->beep();
     return false;
  }

  for(int i = first_mark_; i < last_mark_; ++i)
  {
    rep_type::iterator scan = lines_.find(i);

    if(scan != lines_.end())
    {
      EditLine &el =*scan;

      std::string data = StrTool::expand_tabs(*el.data());

      targetFile << data << '\n';

    }

  }


  paint_page(vp, v);


  return false;
}
////////////////// save block ////////////////////////////////////

////////////////// transform block ////////////////////////////////////
AddMapping transform_block_func(10000, &Impl::block_transform, Modifying);

bool
Impl::
block_transform(viewport* vp, Viewer*v, input_event const *e)
  // cut the marked block to the paste buffer
{
#ifdef USE_muSED
  
  if(first_mark_ == last_mark_)
  {
    vp->beep();
    return false;  // no marked block
  }


  static std::list<std::string>  script;  // maintained across multiple invocations

  if(script.size() == 0)
  {
      script.push_back("");
      script.push_back("p");
      script.push_back("");   // always guarantee an empty line at the end
  }
  else
  {
     // make sure that there's always a blank line to add one more command if needed.


     if(!script.back().empty())
     {
        script.push_back("");
     }
  }

  bool promptUser=true;
  
  {
      typedef std::pair<Viewer*, FileName> ViewerInfo;


      std::list<ViewerInfo> viewers;

      manager_->findMatchingViewers(FileName("*.mused"), viewers);

      if(!viewers.empty())
      {

         script.clear();


         Viewer* activeViewer = viewers.front().first;

         if(viewers.size() > 1)
         {

             CursorWindow::Selection d("Choose a mused script:");
         
             CXXTLS_FOREACH(ViewerInfo &cur, viewers)
             {
                d += cur.second;
             }


             if(d.popup(manager_->window()))
               return false;
         
             CXXTLS_FOREACH(ViewerInfo &cur, viewers)
             {
                 if(cur.second == d.selection_)
                 {
                    activeViewer = cur.first;
                    break;
                 }
             }
             
         }


         if(!activeViewer->fetchAllLines(script))
         {
            
             CursorWindow::Message d("Error:  cannot access script");
             
             d += "Press Enter to continue";
             
             d.popup(manager_->window());
             
             return false;
         }
         else
         {
            promptUser = false;
         }
      }

  }


  
  if(promptUser)
  {
      // create a multi-line string dialog that lets the user
      // enter a script.
    
      CursorWindow::Dialog d("Enter a short MUSED script");
    
      int count=0;
    
      CXXTLS_FOREACH(std::string const &cur, script)
      {
          char name  [40];
          char label [40];
    
          ++count;
    
          snprintf(name,  sizeof(name),  "cmd%d", count);
          snprintf(label, sizeof(label), "%d", count);
    
    
          d += new CursorWindow::Dialog::String(name,
                                                label,
                                                cur,
                                                50
                                                );
      }
      
      // let the user edit the dialog text
    
      if(d.popup( manager_->window() ))
      {
        return false;
      }
    
      // fetch the text out of the dialog
    
      count=0;
      CXXTLS_FOREACH(std::string &cur, script)
      {
          char name  [40];
          ++count;
    
          snprintf(name,  sizeof(name),  "cmd%d", count);
    
          cur = d.element_value(name);
    
      }
  }

  int saved_first_mark = first_mark_;
  int saved_last_mark  = last_mark_;

  UndoOperation::undo_stack* nested_changes = new UndoOperation::undo_stack;
  
  undo_stack_.swap(*nested_changes);  // temporarily swap the real undo
                                      // stack withe the new one just
                                      // created so we can then place the
                                      // changes caused by replace_next_helper
                                      // to be accumulated into the new
                                      // stack -- which will then be placed on
                                      // the real undo stack

  std::list<std::string>  inputBlock;



  // Create an undo operation to hold the deleted lines
  {
      UndoOperation *op = push_undo_stack();
    
      op->code_   = UndoOperation::insert_block;
      op->lineno_ = first_mark_;
    
      // delete the lines and copy them to the paste
      // buffer
    
      while(last_mark_ != first_mark_)
      {
        // delete marked lines starting at tail to simplify work
    
        --last_mark_;
    
        rep_type::iterator scan = line_address(last_mark_);
    
        if(scan != lines_.end())
        {
          EditLine &el =*scan;

          std::string tmp = el;

          inputBlock.push_front(tmp);   // records are processed in reverse order
                                        // because of the way undo works, 
                                        // but we need them in forward order
                                        // for the muSED::apply operation.
    
          op->block_.push_back(tmp);
    
          erase_line(scan, last_mark_, false);  // erase but don't save undo info
                                                // since we are doing it already
        }
    
      }
  }


  std::list<std::string> output;

  std::string error = muSED::apply(script, inputBlock, output);

  {
    UndoOperation* op = push_undo_stack();
    
    op->code_ = UndoOperation::delete_block;
    op->count_= 0;
    op->lineno_ = first_mark_;

    cur_ = first_mark_;
    
    std::list<std::string>::iterator first = output.begin(),
                                     last  = output.end();

    while(first != last)
    {


      std::string insertedLine = *first++;
    
      std::string buffer;
    
      if(use_tabs_)
      {
         // if we are going to save this file with tabs, then we should 
         // tab compress lines from the paste buffer by replacing groups of 8 
         // spaces with tabs.
    
         buffer = StrTool::pack_tabs(insertedLine.c_str());
    
         insertedLine = buffer;
    
      }
      else
      {
         // if we are not saving with tabs, expand any tabs that might be in the line
    
         buffer = StrTool::expand_tabs(insertedLine.c_str());
    
         insertedLine = buffer;
    
      }
    
      lines_.insert( lines_.find(cur_), new std::string(insertedLine) );
    
      ++cur_;
      ++op->count_;
    
    }
  }

  last_mark_ = cur_;

  cur_ = first_mark_;

  {
    undo_stack_.swap(*nested_changes);  // put the old undo stack back


    UndoOperation* op = push_undo_stack();

    op->code_ = UndoOperation::nested_undo_ops;

    op->nested_undo_stack_ = nested_changes;

    op->nested_first_mark_ = saved_first_mark;
    op->nested_last_mark_  = saved_last_mark;
  }


  if(!error.empty())
  {
      CursorWindow::Message d("Script Error");

      d += error;

      d.popup(manager_->window());

  }

  paint_page(vp, v);
#endif

  return false;
}
////////////////// tranform block ////////////////////////////////////


////////////////// save //////////////////////////////////////////

void
Impl::
save(viewport* vp, TextEditor*v, input_event const *e)
{
  if(nonstandard_encoding_)
  {
    CursorWindow::Selection d("Warning:  Non ANSI File Encoding");

    d += "";
    d += "Saving this file will change its endoing to ANSI";
    d += "";
    d += "Corruption of the file data is highly likely";
    d += "";
    d += "Press ESC to abort the save";
    d += "Or press Enter to continue";


    if(d.popup(manager_->window()))
    {
        return;
    }

    nonstandard_encoding_ = false; // its ansi now...
  }
  
  
  if(!is_writeable_)
  {
    std::string write_anyway("Change to write mode?");

    CursorWindow::Selection d("File not opened for write");

    d += "Abandon attempt to save?";
    d += write_anyway;

    if(d.popup(manager_->window()))
      return;

    if(d.selection_ == write_anyway)
    {
      is_writeable_ = true;
    }
    else
    {
      return;
    }
  }


  FILE *f = fopen(fullname_.c_str(), "wb");

  row_col window_size = vp->window()->size();

  is_dirty_ = false;

  if(!f)
  {
    vp->beep();

    CursorWindow::Message m( std::string("Error opening ") +
                             fullname_.shorten(window_size.col_ - 27) +
                             " for write"
                           );
                        
    char buffer[40];

    sprintf(buffer, "errno is %s", strerror(errno));

    m += buffer;
    m += "Press enter to continue";

    m.popup(vp->window());

    is_writeable_ = false;

    return;

  }

  bool force_unix_eol=false;

  #ifdef _MSC_VER
    if( 
        getenv("HOME") && 
        (getenv("SFUDIR") || getenv("CYGWIN") )
      )
      force_unix_eol = true;
  #endif

  rep_type::iterator first = lines_.begin(),
                     last  = lines_.end();

  while(first != last)
  {
    std::string data;

    bool addCR = (*first).crRemoved_;

    
    if(use_tabs_)
      data = (*first++).maybe_tabbify();
    else
      data = *first++; // call the string function not the data() function because the string conversion
                       // expands tabs.  If we are not using tabs to save, then remove them from all lines

    fwrite(data.data(), data.size(), 1, f);
    
    if(force_unix_eol)
    {
      char eol='\n';
      fwrite(&eol, 1, 1, f);
    }
    else
    {
      if(addCR)
      {
         // on unix and linux, this flag might be set but on
         // dos it won't ever be.

         fwrite("\r\n", 2, 1, f); // add full dos eol sequence

       
      }
      else
           fwrite(&EOL_Sequence, sizeof(EOL_Sequence), 1, f);
    }

  }


  fclose(f);
}
/////////////////// save //////////////////////////////////////////

////////////////// line goto ////////////////////////////////////
AddMapping line_goto_func(CursorWindow::func_goto, &Impl::line_goto, NonModifying);

bool
Impl::
line_goto(viewport* vp, Viewer*v, input_event const *e)
  // prompt the user and goto a specified line
{

  StreamableString ss;

  ss << "Goto a line between 1 and " << lines_.size();


  CursorWindow::Dialog dialog(ss);

  StreamableString lineno;

  lineno << (1+cur_);

  dialog += new CursorWindow::Dialog::String("line",
                                             "Goto line",
                                             lineno,
                                             12
                                            );
                                        

  dialog += new CursorWindow::Dialog::Ok("ok","");

  if( dialog.popup(vp->window()) )
    return false;

  int target_line=1;

  sscanf(dialog.element_value("line").c_str(), "%d", &target_line);

  --target_line;


  if(target_line < 0 || target_line >= lines_.size())
    return false;

  cur_ = target_line;

  make_line_visible(vp,v);

  restore_cursor(vp);

  return false;
}

////////////////// line goto //////////////////////////////////

////////////////// undo the most recent change  ///////////////
AddMapping line_undo(CursorWindow::func_undo, &Impl::undo, Modifying);

bool
Impl::
undo(viewport* vp, Viewer*v, input_event const *e)
  // pop the undo stack and perform the reverse operation of that
  // action
{
  bool painting=true;  // update screen after changes

  if(!e)
    painting = false; // nested undo's don't paint, only when finished


  if(undo_stack_.begin() != undo_stack_.end())
  {
    // there is something to undo

    std::auto_ptr<UndoOperation> op(undo_stack_.front());

    undo_stack_.pop_front();

    is_dirty_ = true;

    switch(op->code_)
    {
      case  UndoOperation::replace_line:
          replace_line(op->lineno_, op->text_, false);
          cur_ = op->lineno_;
          if(painting)
          {
            make_line_visible(vp,v);
            paint_page(vp,v);
          }
        break;

      case  UndoOperation::insert_line:
          lines_.insert( lines_.find(op->lineno_), new std::string(op->text_));
          cur_ = op->lineno_;
          if(painting)
          {
            make_line_visible(vp,v);
            paint_page(vp,v);
          }
        break;
      case  UndoOperation::erase_line:
          lines_.erase( lines_.find(op->lineno_) );
          cur_ = op->lineno_;
          if(painting)
          {
            make_line_visible(vp,v);
            paint_page(vp,v);
          }
        break;
        
      case  UndoOperation::break_lines:
          lines_.erase( lines_.find(op->lineno_) );
          cur_ = op->lineno_;
          {
            std::list<std::string>::iterator first = op->block_.begin(),
                                             last  = op->block_.end();
                                        
            while(last != first)
            {
              --last;
              std::string &r = *last;
        
              lines_.insert(lines_.find(cur_), new std::string(r));
            }
          }
        
          if(painting)
          {
            make_line_visible(vp,v);
            paint_page(vp,v);
          }
        break;
        

      case  UndoOperation::join_lines:

          if(op->lineno_ < lines_.size()-1)
          {
            std::string current_line_text = line_text(op->lineno_);
        
            col_ = current_line_text.size();
        
            cur_ = op->lineno_;
        
            lines_.erase( lines_.find(cur_+1) );
        
            replace_line(cur_, op->text_, false);
        
            if(painting)
            {
              make_line_visible(vp,v);
              paint_page(vp,v);
            }
          }
        break;
        
      case  UndoOperation::insert_block:
        {
          std::list<std::string>::iterator first = op->block_.begin(),
                                           last  = op->block_.end();
                                        
          while(first != last)
          {
            lines_.insert(lines_.find(op->lineno_), new std::string(*first));
            ++first;
          }

          if(painting)
            paint_page(vp,v);
        
        }
        break;
        
      case  UndoOperation::replace_block:
        {
          std::list<std::string>::iterator first = op->block_.begin(),
                                           last  = op->block_.end();
                                        
          while(first != last)
          {
            replace_line(op->lineno_++, std::string(*first), false);
        
            ++first;
          }

          if(painting)
            paint_page(vp,v);
        
        }
        break;
        
        
      case  UndoOperation::delete_block:
        {
          // this algorithm is necessary because lines_ is a
          // vector and whenever you insert or delete into a vector
          // _all_ iterators are invalid
        
        
          while(op->count_)
          {
            rep_type::iterator first = lines_.find(op->lineno_);
        
            lines_.erase(first);
        
             --op->count_;
          }
                
          cur_ = op->lineno_;
        
          if(painting)
          {
            make_line_visible(vp,v);
            paint_page(vp,v);
          }
        
        }
        break;
        
      case UndoOperation::nested_undo_ops:
        {
           UndoOperation::undo_stack* nested_changes = op->nested_undo_stack_;
        
           undo_stack_.swap(*nested_changes);
        
           bool something_to_do = !undo_stack_.empty();
        
           while(!undo_stack_.empty())
           {
             undo(vp,v,0);  // the 0 for e is flag saying don't paint
           }
        
           if(something_to_do)
             is_dirty_ = true;
        
           undo_stack_.swap(*nested_changes);
        
           delete nested_changes;

           first_mark_ = op->nested_first_mark_;
           last_mark_  = op->nested_last_mark_;
        
           if(painting)
           {
             make_line_visible(vp,v);
             paint_page(vp,v);
           }
        
        }
        break;
    }

  }
  else
  {
    vp->beep();
    is_dirty_ = false;
  }

  return false;
}

////////////////////////// show tags /////////////////////////////////////////////
bool
Impl::
show_symbols(viewport* vp, Viewer*v, input_event const *e)
{


  std::string name = fullname_ + ".cpptagdb";

  Viewer* oldie = manager_->find_viewer(name);

  if(oldie)
  {
    manager_->activate(oldie);
  }
  else
  {

    manager_->window()->close();

    Viewer *newbie = manager_->open(name);  // name is the "current" file name plus .cpptagdb.  The
                                            // symbol viewer knows to use cpptagdb.exe to get the
                                            // correct output instead of actually reading files whose
                                            // name ends in .cpptagdb

    manager_->window()->open();

    if(!newbie)
      manager_->window()->beep();
    else
    {
      manager_->hsplit(newbie, name);
    }

  }


  return false;
}



////////////////////////// word edit /////////////////////////////////////////////


struct Is_Path_Separator
{
  bool operator() (char c) const
  {
    return c == FileName::PATH_separator ;
  }

};

bool
Impl::
snatch_edit(viewport* vp, Viewer*v, input_event const *e)
  //
  // Snatch the text under the cursor and assume it is a filename and bring up
  // an edit session on that file.  Also, if the text under the cursor looks
  // like a recognized compiler error message or grep output, make sure the
  // current line is positioned correctly in the edited file.
  //
{

  std::string s = line_text(cur_);

  size_t rLength = s.size();

  if( col_ >= rLength || s[col_] == ' ')
  { // not sitting on a word

    manager_->window()->beep();

    return false;

  } // not sitting on a word

  std::string tmp( s.begin() + col_, s.end());

  FileName name;
  int      number=0;

  if(!StrTool::snatch_file_info(tmp, name, number))
  {
    manager_->window()->beep();

    return false;

  }

  name = manager_->find_file_in_standard_locations(name, fullname_.dirname());

  if(name.empty())
  {
    manager_->window()->beep();

    return false;

  }

  Viewer* oldie = manager_->find_viewer(name);

  if(oldie)
  {
    oldie->set_row_col_hint(number-1,0);
    manager_->activate(oldie);
  }
  else
  {
    if(!name.exists())
    {
       CursorWindow::Message m(name + " does not exist");

       m += "Sorry, can't find it";
       m += "";
       m += "Press Enter to continue";

       m.popup(manager_->window());

       return false;
    }

    Viewer *newbie = manager_->edit(name);

    if(!newbie)
      manager_->window()->beep();
    else
    {
      newbie->set_row_col_hint(number-1, 0);  // 0 based line numbers!
      manager_->vsplit(newbie, name);
    }

  }


  return false;
}

void
TextEditor::
set_row_col_hint(size_t line_setting, size_t col, bool force_repaint)
{

   if(int(line_setting) >= impl_->size())
      return;   // attempt to seek to non-existent line

   // first, check to see if the line setting is within the currently
   // visible display region -- if any.

  if(line_setting >= size_t(impl_->page_top_) )
  {
      CursorWindow::viewport* vp = impl_->manager_->find_viewport(this);

      if(vp)
      {
          int row_offset       = line_setting - impl_->page_top_;
          int available_rows   = vp->size().row_ - impl_->first_row_;

          if(row_offset < available_rows)
          {
              
              // the line we are looking for is actually on the
              // currently displayed page.
              
              // leave the top as it is and just set the line offset
              // within the page
              
              impl_->cur_ = line_setting;
              
              impl_->col_ = col;
              impl_->page_left_ = col;
              
              return;
          }
      }
  }


      
  int line = line_setting;
  
  if(line >= 0 && line < impl_->size())
  {
    impl_->cur_ = line;
    if(line > 2)
        impl_->page_top_ = line-2;
    else
        impl_->page_top_ = line;

    impl_->col_ = col;
    impl_->page_left_ = col;
  }

}

void
TextEditor::
set_read_only()
{
  impl_->is_writeable_ = false;
}




bool
Impl::
word_sfind(viewport* vp, Viewer*v, input_event const *e)
  ///  Snatch the word under the cursor out of the edit buffer and
  ///  put it in the find string buffer.  Then perform a find next on it.
{
  std::string curline = line_text(cur_);

  if(col_ >= curline.size() || curline[col_] == ' ')
  {
    vp->beep();
    return false;
  }

  std::string::const_iterator first = curline.begin() + col_,
                              last  = curline.end();
                        
  search_text_ = "";

  while(first != last && is_word_char(*first) )
  {
    search_text_ += *first++;
  }

  if(search_text_.size() == 0)
  {
    vp->beep();
    return false;
  }

  size_t col_save = col_;
  int    cur_save = cur_;

  if(find_next_helper(vp, v, e))
  {
    make_line_visible(vp,v);
    make_col_visible(vp,v);
  }
  else
  {
    vp->beep();
    col_ = col_save;
    cur_ = cur_save;
  }

  return false;
}

bool
Impl::
etags_first(viewport* vp, Viewer*v, input_event const *e)
  ///  Snatch the word under the cursor out of the edit buffer and
  ///  look it up in the etags database
{
  EtagsDB* etags=(manager_->default_etags_db());  // pts to static object


  if(!etags || !etags->ok())
  {
    // error message will have been handled by ViewerManager code
    return false;
  }

  std::string symbol;
  std::string file;
  int         line;
  std::string matching_text;

  std::string curline = line_text(cur_);

  if( !(col_ >= curline.size() || curline[col_] == ' ') )
  {
    std::string::const_iterator first = curline.begin() + col_,
                                last  = curline.end();
                        
    while(first != last && is_word_char(*first) )
    {
      symbol += *first++;
    }
  }


  CursorWindow::Dialog d("Search TAGS (etags file) For A Symbol");

  d += new CursorWindow::Dialog::String("symbol",
                                        "Symbol Name",
                                        symbol,
                                        40
                                        );
                                
  if(d.popup(manager_->window()))
    return 0;

  symbol = d.element_value("symbol");

  if(symbol.size() == 0)
  {
    vp->beep();
    return false;
  }

  bool rv = etags->find_first(symbol, &file, &line, &matching_text);

  if(rv)
  {
    Viewer* oldie = manager_->find_viewer(file);

    if(oldie)
    {
      oldie->set_row_col_hint(line-1,0);
      manager_->activate(oldie);
    }
    else
    {

      Viewer* newbie = manager_->edit(file);

      if(newbie)
      {
        newbie->set_row_col_hint(line-1,0);

        manager_->vsplit(newbie, file);

      }
      else
        vp->beep();
    }

  }

  return false;
}

bool
Impl::
cpptags_first(viewport* vp, Viewer*v, input_event const *e)
  ///  Snatch the word under the cursor out of the edit buffer and
  ///  look it up in the cpptags database
{
  CppTagDB::Info* cpptags=(manager_->default_cpptags_db());  // pts to static object


  if(!cpptags)
  {
    // error message will have been handled by ViewerManager code
    return false;
  }

  std::string symbol;
  std::string file;
  int         line;
  std::string matching_text;

  std::string curline = line_text(cur_);

  if( !(col_ >= curline.size() || curline[col_] == ' ') )
  {
    std::string::const_iterator first = curline.begin() + col_,
                                last  = curline.end();
                        
    while(first != last && is_word_char(*first) )
    {
      symbol += *first++;
    }
  }


  CursorWindow::Dialog d("Search TAGPP.tagpp (cpptagdb.exe file) For A Symbol");

  d += new CursorWindow::Dialog::String("symbol",
                                        "Symbol Name",
                                        symbol,
                                        40
                                        );
                                
  if(d.popup(manager_->window()))
    return 0;

  symbol = d.element_value("symbol");

  if(symbol.size() == 0)
  {
    vp->beep();
    return false;
  }

  std::vector<cxxtls::CppTagDB::SymbolInfo const *> matches;

  cpptags->findMatchingMembers(symbol, matches);

  if(!matches.empty())
  {
    std::vector<ViewerManager::FileLocation>  locations;

    CXXTLS_FOREACH(CppTagDB::SymbolInfo const *cur, matches)
    {
       locations.push_back( ViewerManager::FileLocation(cur->file_, cur->line_) );
    }

    manager_->setFileLocations(locations);

    ViewerManager::FileLocation tmp;

    manager_->popFileLocation(tmp);

    file = tmp.filename_;
    line = tmp.lineNumber_;


    Viewer* oldie = manager_->find_viewer(file);

    if(oldie)
    {
      oldie->set_row_col_hint(line-1,0);
      manager_->activate(oldie);
    }
    else
    {

      Viewer* newbie = manager_->edit(file);

      if(newbie)
      {
        newbie->set_row_col_hint(line-1,0);

        manager_->vsplit(newbie, file);

      }
      else
        vp->beep();
    }

  }
  else
     vp->beep();

  return false;
}


bool
Impl::
etags_all(viewport* vp, Viewer*v, input_event const *e)
  ///  Snatch the word under the cursor out of the edit buffer and
  ///  look it up in the etags database.  List all matches in grep format.
{
  EtagsDB* etags=(manager_->default_etags_db()); // pts to static object

  if(!etags || !etags->ok())
  {
    // error message will have been handled by ViewerManager code
    return false;
  }

  std::string symbol;
  std::string file;
  int         line;
  std::string matching_text;

  std::string curline = line_text(cur_);

  if( !(col_ >= curline.size() || curline[col_] == ' ') )
  {
    std::string::const_iterator first = curline.begin() + col_,
                                last  = curline.end();
                        
    while(first != last && is_word_char(*first) )
    {
      symbol += *first++;
    }
  }


  CursorWindow::Dialog d("Search TAGS (etags file) For All Matching Symbols");

  d += new CursorWindow::Dialog::String("symbol",
                                        "Symbol Name",
                                        symbol,
                                        40
                                        );
                                
  if(d.popup(manager_->window()))
    return 0;

  symbol = d.element_value("symbol");

  if(symbol.size() == 0)
  {
    vp->beep();
    return false;
  }

  bool rv = etags->find_first(symbol, &file, &line, &matching_text);

  if(rv)
  {
      CursorWindow::Selection d("Symbols matching " + symbol);
    
      d += "";

      while(rv)
      {
         std::string tmp = file;

         char buffer[40];

         snprintf(buffer, 40, ":%d", line);

         tmp += buffer;
         tmp += ' ';
         tmp += matching_text;

         d += tmp;

         rv = etags->find_next(&file, &line, &matching_text);

      }

      d += "";

      d += "Press Enter to continue";

      if(!d.popup(manager_->window()))
      {
          // user didn't abort

          std::string const &selection = d.selection_;

          size_t colon = selection.find_first_of(':');

          if(colon >= selection.size())
          {
            vp->beep();  // should never get here
          }
          else
          {
            // there is a filename:lineNumber pair

            file = selection.substr(0, colon);

            std::string tmp = selection.substr(colon+1, selection.size() - colon -1);

            line = 0;

            sscanf(tmp.c_str(), "%d", &line);

            Viewer* oldie = manager_->find_viewer(file);
        
            if(oldie)
            {
              oldie->set_row_col_hint(line-1,0);
              manager_->activate(oldie);
            }
            else
            {
        
              Viewer* newbie = manager_->edit(file);
        
              if(newbie)
              {
                newbie->set_row_col_hint(line-1,0);
        
                manager_->vsplit(newbie, file);
        
              }
              else
                vp->beep();
            }

          }
          

      }

  }
  else
    vp->beep();

  return false;
}

bool
Impl::
cpptags_all(viewport* vp, Viewer*v, input_event const *e)
  ///  Snatch the word under the cursor out of the edit buffer and
  ///  look it up in the TAGPP.tagpp database.  List all matches in grep format.
{

  CppTagDB::Info* cpptags=(manager_->default_cpptags_db());  // pts to static object

  if(!cpptags)
  {
    // error message will have been handled by ViewerManager code
    return false;
  }

  std::string symbol;
  std::string file;
  int         line;
  std::string matching_text;


  std::string curline = line_text(cur_);

  if( !(col_ >= curline.size() || curline[col_] == ' ') )
  {
    std::string::const_iterator first = curline.begin() + col_,
                                last  = curline.end();
                        
    while(first != last && is_word_char(*first) )
    {
      symbol += *first++;
    }
  }


  CursorWindow::Dialog d("Search TAGPP.tagpp (cpptagdb.exe file) For All Matching Symbols");

  d += new CursorWindow::Dialog::String("symbol",
                                        "Symbol Name",
                                        symbol,
                                        40
                                        );
                                
  if(d.popup(manager_->window()))
    return 0;

  symbol = d.element_value("symbol");

  if(symbol.size() == 0)
  {
    vp->beep();
    return false;
  }

  std::vector<cxxtls::CppTagDB::SymbolInfo const *> matches;

 cpptags->findMatchingMembers(symbol, matches, false);  // insensitive comparison

  if(!matches.empty())
  {
      CursorWindow::Selection d("Symbols matching " + symbol);
    
      d += "";

      std::vector<cxxtls::CppTagDB::SymbolInfo const *>::const_iterator first, last;

      for(first = matches.begin(), last = matches.end(); first != last; ++first)
      {
         std::string tmp = (*first)->file_;

         char buffer[40];

         snprintf(buffer, 40, ":%d", (*first)->line_);

         tmp += buffer;

         d += tmp;

      }

      d += "";

      d += "Press Enter to continue";

      if(!d.popup(manager_->window()))
      {
          // user didn't abort

          std::string const &selection = d.selection_;

          size_t colon = selection.find_first_of(':');

          if(colon >= selection.size())
          {
            vp->beep();  // should never get here
          }
          else
          {
            // there is a filename:lineNumber pair

            file = selection.substr(0, colon);

            std::string tmp = selection.substr(colon+1, selection.size() - colon -1);

            line = 0;

            sscanf(tmp.c_str(), "%d", &line);

            Viewer* oldie = manager_->find_viewer(file);
        
            if(oldie)
            {
              oldie->set_row_col_hint(line-1,0);
              manager_->activate(oldie);
            }
            else
            {
        
              Viewer* newbie = manager_->edit(file);
        
              if(newbie)
              {
                newbie->set_row_col_hint(line-1,0);
        
                manager_->vsplit(newbie, file);
        
              }
              else
                vp->beep();
            }

          }
          

      }

  }
  else
    vp->beep();

  return false;
}



bool
Impl::
tags_next(viewport* vp, Viewer*v, input_event const *e)
  ///  Repeat the most recent etags_first search.
  ///
  ///  OR repeat the most recent cpptags_first search -- whichever
  ///  happened most recently.
{

  std::string file;
  int         line;
  std::string matching_text;

  ViewerManager::FileLocation loc;


  if(manager_->popFileLocation(loc))
  {
     // there is a saved file location


     file = loc.filename_;
     line = loc.lineNumber_;

  }
  else
  {
      EtagsDB *etags=(manager_->default_etags_db());  // pts to static object

      if(etags && etags->ok())
      {
        if(etags->find_next(&file, &line, &matching_text))
        {
           // found file and line
        }
        else
        {
           // no file location and unfortunately the etags
           // database does not contain the requested data.
      
      
           vp->beep();
           return false;
        }
      }
      else
      {
        // neither etags nor saved file locations could be accessed
      
        vp->beep();
      
      
        return false;
      }
  }

  Viewer* oldie = manager_->find_viewer(file);
  
  if(oldie)
  {
    oldie->set_row_col_hint(line-1,0);
    manager_->activate(oldie);
  }
  else
  {
  
    Viewer* newbie = manager_->edit(file);
  
    if(newbie)
    {
      newbie->set_row_col_hint(line-1,0);
  
      manager_->add(newbie, file);
  
    }
    else
      vp->beep();
  }


  return false;
}

////////////////// find matching character  ///////////////
AddMapping matching_char(CursorWindow::func_matching, &Impl::char_matching, NonModifying);



TextIterator
Impl::
top_pointer()
{
  return TextIterator(lines_, 0, 0);
}


TextIterator
Impl::
bot_pointer()
{
  return TextIterator(lines_, lines_.size(), 0);
}

TextIterator
Impl::
cur_pointer()
{
  std::string curline = line_text(cur_);

  if(cur_ >= lines_.size())
    return TextIterator(lines_, lines_.size(), 0);
  else
  if(col_ >= curline.size())
    return TextIterator(lines_, cur_, curline.size());

  return TextIterator(lines_, cur_, col_);
}

bool
Impl::
char_matching(viewport* vp, Viewer*v, input_event const *e)
{
  // find the character matching the one under the cursor

  TextIterator top = top_pointer(),  // get iterator to top of the edit session
               cur = cur_pointer(),  // get iterator to current file position
               bot = bot_pointer();  // get iterator to the bottom of the file

  if(cur == bot)
  {
    //
    // at end of file...
    //
    vp->beep();
    return false;
  }


  char match = *cur;
  bool down;
  char orig  = match;

  switch(match)
  {
    case '{':  match = '}'; down=true; break;
    case '(':  match = ')'; down=true; break;
    case '<':  match = '>'; down=true; break;
    case '[':  match = ']'; down=true; break;

    case '}':  match = '{'; down=false; break;
    case ')':  match = '('; down=false; break;
    case '>':  match = '<'; down=false; break;
    case ']':  match = '['; down=false; break;

    default:

      vp->beep();
      return false;
  }

  if(down)
  {
    // parse C++ tokens forward through the file until the match is found

    int depth = 0;

    while(cur != bot)
    {
      char c = *cur;

      if(c == orig)
        ++depth;
      else
      if(c == match)
      {
        --depth;
        if(depth == 0)
          break;
      }
      else
      if(c == '\'')
      {
        ++cur;
        
        while(cur != bot && *cur != '\'')
        {
          // skip quoted strings
        
          if(*cur == '\\') // skip any escape sequence
            ++cur;
        
          if(cur != bot)  // skip the current char
            ++cur;
        }
        
      }
      else
      if(c == '"')
      {
        ++cur;
        
        while(cur != bot && *cur != '"')
        {
          // skip quoted strings
        
          if(*cur == '\\') // skip any escape sequence
            ++cur;
        
          if(cur != bot)  // skip the current char
            ++cur;
        }
        
      }
      else
      if(c == '/')
      {
        // handle c++ style comments

        ++cur;

        if( (cur != bot) )
        {
            if((*cur == '/') )
            {
              while(  (cur != bot) && (*cur  != '\n') )
                ++cur;
            }
            else
            {
              --cur;  // one slash should be ignored, but two is a comment.
            }
        }

      }
      
      if(cur != bot)
        ++cur;
    }

  }
  else
  {
    // parse C++ tokens backwards through the file until the match is found

    int depth = 0;

    while(cur != top)
    {
      char c = *cur;

      if(c == orig)
        ++depth;
      if(c == match)
      {
        --depth;
        if(depth == 0)
          break;
      }
      else
      if(c == '\'' )
      { 

        // skip the trailing single quoted string
        // scan backwards from the last valid char in the string to the first
        // ' in the string -- taking into account this kind of annoyance:  '\''
        //                                                                   ^
        
        --cur;

        while(cur != top)
        {
           if(*cur != '\'')
             --cur;
           else
           {
             // ok, we have a ' -- but is this:  'asdf\'adsf'
             //                                        ^

             --cur; 

             if(*cur != '\\' )
             {
               ++cur;  // cur will get decremented later in a the outer loop so make sure
                       // the cursor is not already decrement once by mistake
               break;  
             }
             else
               --cur;
           }
        }

      } 
      else
      if(c == '"')
      { 

        // skip the trailing single quoted string
        // scan backwards from the last valid char in the string to the first
        // " in the string -- taking into account this kind of annoyance:  "\""
        //                                                                   ^
        
        --cur;

        while(cur != top)
        {
           char c = *cur;  // for debugging purposes

           (void)(c); // shut the stupid warning the F* off.

           if(*cur != '"')
             --cur;
           else
           {
             // ok, we have a ' -- but is this:  "asdf\"adsf"
             //                                        ^

             --cur; 

             if(*cur != '\\' )
             {
               ++cur;  // cur will get decremented later in a the outer loop so make sure
                       // the cursor is not already decrement once by mistake
               break;  
             }
             else
               --cur;
           }
        }

      } 

      if( (cur != top) && (*cur == '\n') )
      {
        // step back to the previous line and remove any c++ style comments on it

        --cur;

        TextIterator here(cur);

        char c = *cur;

        c= c;  // silence unused variable warnings;

        //        while( (cur != top) && (*cur != '\n') && (*cur != '"') && (*cur != '\'') )

        while( (cur != top) && (*cur != '\n') )
        {
          c = *cur;
          --cur;
        }

        // cur now sits at the beginning of the previous line -- scan forward until // is found
        // or until cur == here -- if // is found, stop and make that be 'here'

        for(;;)
        {
          if(cur == here)
            break;

          if(*cur == '/' && (cur != here) )
          {
            ++cur;

            if(cur != here && (*cur == '/'))
            {
               here = cur;
               break; 

               // leave here pointing to the comment -- it won't cause any trouble
            }

          }
          else
            ++cur;
        }

        cur = here;

        ++cur; // so the following decrement will work...
      }


      if(cur != top)
        --cur;
    }
  }
    
  if(cur == bot || cur == top)
  {
    //
    // found no match
    //
    vp->beep();
    return false;
  }


  cur_ = cur.line_;
  col_ = cur.col_;

  make_line_visible(vp,v);
  make_col_visible(vp,v);

  return false;
}

////////////////// block_insert /////////////////////////////////////////////////

bool
Impl::
block_insert(viewport* vp, Viewer*v, input_event const *e)
{
  // if a block is marked, prompt the user for a string to
  // insert into the block at the current column.  Each marked line
  // gets the inserted text.


  if(first_mark_ == last_mark_)
  {
    vp->beep();
    return false;
  }

  CursorWindow::Dialog d("Enter a string to insert into the block");

  d += new CursorWindow::Dialog::String("name",
                                        "Text to insert",
                                        "",
                                        40
                                       );
                                
  d += new CursorWindow::Dialog::Ok("ok","");

  if(d.popup(vp->window()))
    return false;

  std::string text = d.element_value("name");

  if(text.size() == 0)
    return false;

  int i;

  UndoOperation* op = push_undo_stack();

  op->code_   =  UndoOperation::replace_block;
  op->lineno_ =  first_mark_;

  for(i=first_mark_; i < last_mark_; ++i)
  {
    std::string curline = line_text(i);

    op->block_.push_back(lines_[i]);

    if(curline.size() < col_)
    {
       size_t pad = col_ - curline.size();

       curline += std::string(pad, ' ');
    }

    curline.insert(col_, text);

    replace_line(i, curline, false);  // don't include undo info, we've done it alreay

  }

  paint_page(vp,v);

  return false;
}

////////////////// block_adjust /////////////////////////////////////////////////


void
Impl::
adjustHelper(int adjustment, viewport* vp, Viewer*v)
    ///  A helper function called by block_adjust() when the user wants to
    ///  specify a specific width to slide a block left or right.
{
  UndoOperation* op = push_undo_stack();

  op->code_   =  UndoOperation::replace_block;
  op->lineno_ =  first_mark_;

  std::string spaces_to_insert( adjustment > 0 ? adjustment : 0, ' ');

  int i;

  for(i=first_mark_; i < last_mark_; ++i)
  {
    std::string curline = line_text(i);

    op->block_.push_back(lines_[i]);

    if(curline.size() < col_)
    {
       size_t pad = col_ - curline.size();

       curline += std::string(pad, ' ');
    }

    if(adjustment > 0)
      curline.insert(col_, spaces_to_insert);
    else
    {
      size_t line_length = curline.size();

      size_t chars_to_delete = -adjustment;

      if(col_ + chars_to_delete > line_length)
        chars_to_delete = line_length - col_;
        
      curline.erase(col_, chars_to_delete);

    }

    replace_line(i, curline, false);  // don't include undo info, we've done it alreay

  }

  paint_page(vp,v);
}


bool
Impl::
block_adjust(viewport* vp, Viewer*v, input_event const *e)
{
  // Prompt the user for a number of columns to insert into the
  // marked block.  That is, insert a column of spaces at the current
  // cursor column into all lines in the marked block.

  if(first_mark_ == last_mark_)
  {
    vp->beep();
    return false;
  }

  CursorWindow::Dialog d("Insert/delete colums");

  d += new CursorWindow::Dialog::String("name",
                                        "Columns to insert/delete",
                                        "0",
                                        10
                                       );
                                
  d += new CursorWindow::Dialog::Ok("ok","");
                                
                                
  if(d.popup(vp->window()))
    return false;

  std::string text = d.element_value("name");

  if(text.size() == 0)
    return false;

  int adjustment =0;

  sscanf(text.c_str(), "%d", &adjustment);

  if(adjustment == 0)
    return false;

  adjustHelper(adjustment, vp, v);


  return false;
}
////////////////// block_adjust /////////////////////////////////////////////////



////////////////// line_join /////////////////////////////////////////////////
AddMapping join_lines_func(CursorWindow::func_join, &Impl::line_join, Modifying);

bool
Impl::
line_join(viewport* vp, Viewer*v, input_event const *e)
{
  // join this line with the next line.

  if(size() == 0 || cur_ == size() - 1)
    return false;  // nothing to do, you can't join the end of the file to this line

  std::string next_line = line_text(cur_ + 1);

  std::string cur_line = line_text(cur_);

  if(next_line.size())
  {
    // next line is not blank -- remove leading blanks from that
    // line  -- there's no point in joining blanks.

    size_t col_tmp = cur_line.size();

    std::string::const_iterator first = next_line.begin(),
                                last  = next_line.end();
                                
    while(first != last && *first == ' ')
      ++first;

    {
      size_t s = first - next_line.begin();

      if(s)
        next_line.erase(0, first - next_line.begin());
    }

    // at this point, the local copy of the next line of text has
    // had leading blanks removed.  Now remove trailing blanks from cur_line

    first = cur_line.begin();
    last  = cur_line.end();

    while(first != last)
    {
      --last;

      if(*last != ' ')
      {
        ++last;
        break;
      }
    }

    // at this point, last points to the end of the string
    // or to the first blank character after the last non-blank
    // in cur_line

    {
       size_t new_length = (last - cur_line.begin());
       
       cur_line.erase( new_length, cur_line.size() - new_length);

       col_tmp = new_length;
    }

    cur_line += next_line;

    // cur_line is now the final product

    col_ = col_tmp;

  }

  UndoOperation* op = push_undo_stack();

  op->code_    = UndoOperation::break_lines;
  op->lineno_ = cur_;

  op->block_.push_back(std::string(lines_[cur_]));
  op->block_.push_back(std::string(lines_[cur_+1]));

  erase_line( lines_.find(cur_+1), cur_+1, false);
  erase_line( lines_.find(cur_), cur_, false);
  lines_.insert(lines_.find(cur_), new std::string(cur_line));

  make_col_visible(vp,v);
  make_line_visible(vp,v);
  paint_page(vp,v);

  return false;
}
////////////////// line_join /////////////////////////////////////////////////

////////////////// line_insert /////////////////////////////////////////////////
AddMapping insert_lines_func(CursorWindow::func_insline, &Impl::line_insert, Modifying);

bool
Impl::
line_insert(viewport* vp, Viewer*v, input_event const *e)
{
  if(size() == 0)
  {
    lines_.insert(lines_.begin(), new std::string(""));
  }
  else
  {
    lines_.insert(lines_.find(cur_), new std::string(""));
  }

  push_undo_stack();
  undo_stack_.front()->code_   = UndoOperation::erase_line;
  undo_stack_.front()->lineno_ = cur_;


  make_line_visible(vp,v);
  paint_page(vp,v);

  return false;
}
////////////////// line_insert /////////////////////////////////////////////////

////////////////// to_edit /////////////////////////////////////////////////

AddMapping convert2edit(CursorWindow::func_to_edit, &Impl::to_edit, NonModifying);

bool
Impl::
to_edit(viewport* vp, Viewer*v, input_event const *e)
  // this function makes the current read only session writeable -- ie converts a 
  // view into an edit
{
  if(!is_writeable_)
  {
    is_writeable_ = true;
    paint_page(vp,v);
  }
  else
    vp->beep();

  return false;
}

////////////////// to_edit /////////////////////////////////////////////////

////////////////// reread_edit /////////////////////////////////////////////////

AddMapping reread_edit(CursorWindow::func_reread, &Impl::reread, NonModifying);

bool
Impl::
reread(viewport* vp, Viewer*v, input_event const *e)
  // check to see if the file being edited has changed on disk.
{
  // cannot be undone!

  if(!fullname_.exists())
  {
    // if there's nothing to read, warn the user then quit.

    CursorWindow::Message d("Error reading file " + fullname_);
  
    d += "";
    d += "Press Enter to continue";
  
    d.popup(manager_->window());
  
    return false;
  }
  else
  {
     if(is_dirty_)
     {
       vp->beep();

       CursorWindow::Dialog d("Unsaved edit changes exist");

       d += new CursorWindow::Dialog::String("ok",
                                             "Discard them? (Yes or No)",
                                             "No",
                                             4
                                            );
       
       if(d.popup( manager_->window() ))
       {
         return false;
       }
       
       std::string ok_value = d.element_value("ok");
       
       if(ok_value.size() > 2)
       {
         ok_value.erase(2, ok_value.size());
       }
       
       if(ok_value != "Ye" && ok_value != "ye" )
       {
         vp->beep();
         return false;
       }
     }

     // we are going to reread the file -- so eliminate all the lines currently 
     // found -- and the block top and bottom -- 
     // and finally go to the top

     file_top(vp, v, e);

     first_mark_ = 0;
     last_mark_  = size();

     block_delete(vp, v, e);

     cur_ = 0;

     // size_t s = size();  // should be 0

     // now empty the undo stack

     while(!undo_stack_.empty())
     {
       delete undo_stack_.front();
       undo_stack_.pop_front();
     }

     read_file();

     make_line_visible(vp, v);

     paint_page(vp,v);

     return false;


  }


  return false;
}

////////////////// reread_edit /////////////////////////////////////////////////

////////////////// switch_source /////////////////////////////////////////////////

struct is_colon
{
  bool operator() (char c) const { return c == ':'; }
};

struct is_pathsep
{
  bool operator() (char c) const { return c == '/' || c == '\\'; }
};


AddMapping switch_source_file(CursorWindow::func_switch, 
                              &Impl::switch_source, 
                              NonModifying);

bool
Impl::
switch_source(viewport* vp, Viewer*v, input_event const *e)
  // switch to an alternate file based on the current filename extension
  // and the directories listed in the SOURCEPATH and CDPATH environment
  // variable.  The filename extensions of interest are specified in
  // environment variable SOURCEALTS like this:
  //
  //   .c:.h:.:.inl:
  //
  // A single ., with no other lettering means that 'no extension' is 
  // meant -- such as the new c++ standard for system headers.
{
  StrTool::stringlist_t dirs;

  //
  // Collected needed environment variables and handle errors
  //

  std::string srcalts= ProgramOptions::global_options_->getenv("SOURCEALTS").value();
  std::string srcpath= ProgramOptions::global_options_->getenv("SOURCEPATH").value();
  std::string cdpath = ProgramOptions::global_options_->getenv("CDPATH").value();


  if(srcalts.size() == 0)
  {
    vp->beep();

    CursorWindow::Message d("SOURCEALTS environment variable is not set");

    d += "";
    d += "Set the SOURCEALTS environment variable to a colon";
    d += "separated list of file name extensions which count";
    d += "as source files";
    d += "";
    d += "For example:";
    d += "  SOURCEALTS=:.c:.h:.:.inl";
    d += "This example makes the following file name";
    d += "extensions valid source alternatives:";
    d += "  .c, .h, none, .inl";
    d += "";
    d += "Press Enter to continue";
    
    d.popup(manager_->window());
    
    return false;
  }


  if(srcpath.size() == 0 && cdpath.size() == 0)
  {
    vp->beep();

    CursorWindow::Message d("No SOURCEPATH or CDPATH set");
    
    d += "Please set one or both of the following";
    d += "environment variables:";
    d += "  SOURCEPATH";
    d += "  CDPATH";
    d += "Both variables are ':' separated lists";
    d += "directory names.";
    d += "";
    d += "Press Enter to continue";
    
    
    d.popup(manager_->window());
    
    return false;
  }

  //
  // Environent variables are good -- use them
  //

  StrTool::stringlist_t src_types;  // "srcalts split up (.c, .h, etc)"
  StrTool::stringlist_t src_dirs;   // "srcpath split up" + cdpath split up
  StrTool::stringlist_t src_path;   // "srcpath split up"
  StrTool::stringlist_t cd_dirs;    // "srcpath split up"

  src_dirs.push_back(".");

  StrTool::parse_words(srcalts, &src_types, 20000, is_colon());
  StrTool::parse_words(srcpath, &src_path,  20000, is_colon());
  StrTool::parse_words(cdpath,  &cd_dirs,  20000, is_colon());

  StrTool::stringlist_t::iterator sbegin, send;

  //
  // expand any directory name patters in the last basename's of src_path
  // elements
  //

  for(sbegin = src_path.begin(), send = src_path.end();
      sbegin != send;
      ++sbegin
     )
  {
    FileName cur(*sbegin);

    if(!cur.is_absolute_path())
    {
      cur = fullname_.dirname() + cur;
    }

    if(cur.basename() == "*")
    {
      FileName::sorted_names_t tmp;

      FileName::find_matching(cur, &tmp, 0x03);  
        // get directories only and full paths

      FileName::sorted_names_t::iterator a=tmp.begin(), b=tmp.end();
      while(a != b)
      {
        src_dirs.push_back(*a++);
      }

    }
    else
    {
      src_dirs.push_back(cur);
    }
  }

  //
  // include the cdpath in the src_dirs
  //
  for(sbegin = cd_dirs.begin(), send = cd_dirs.end();
      sbegin != send;
      ++sbegin
     )
  {
    src_dirs.push_back(*sbegin);
  }


  //
  // handle the 'none' file name extension
  //

  sbegin =  src_types.begin();
  send   =  src_types.end();

  while(sbegin != send)
  {
    if(*sbegin == ".")
      *sbegin = "";

    ++sbegin;
  }


  //
  //  Verify that this file is described by SOURCEALTS
  //

  std::string ext = fullname_.extension();
  std::string base= fullname_.basename(true);


  sbegin =  src_types.begin();
  send   =  src_types.end();

  while(sbegin != send)
  {
    if(ext == *sbegin)
      break;

    ++sbegin;
  }

  if(sbegin == send)
  {
    vp->beep();

    std::string tmp("  "); tmp += srcalts;

    CursorWindow::Message d("This file is not listed in SOURCEALTS");

    d += "In order to use the 'switch_source' editor function,";
    d += "you must be editing a file whose file name extension";
    d += "is listed in the environment variable, SOURCEALTS.";
    d += "";
    d += "That variable contains the following: ";
    d += tmp;
    d += "";
    d += "But you are editing a file whose extension is '" + ext + "'";
    d += "";
    d += "Press Enter to continue";

    d.popup(manager_->window());
    
    return false;
  }

  //
  // Find files matching the requirements
  //

  {
     typedef std::set<std::string> matches_t;

     matches_t matches;

     for(sbegin  =src_dirs.begin(),
           send  =src_dirs.end();
         sbegin != send;
         ++sbegin
        )
     {
        StrTool::stringlist_t::iterator f = src_types.begin(),
                                        l = src_types.end();
        while(f != l)
        {
          std::string cur_ext = *f;

          FileName test(*sbegin + "/" + base + cur_ext);

          if(test.exists() && !test.is_dir())
          {
            test.convert_to_absolute_path();
            matches.insert(test);
          }

          ++f;

        }
     }

     //
     // If less than 2 matches found, let user see the list of
     // source dirs
     //

     if(matches.size() < 2)
     {
       CursorWindow::Message d("FYI:  less than two matches found in the following dirs:");

       for(sbegin = src_dirs.begin(), send = src_dirs.end();
           sbegin != send;
           ++sbegin
          )
       {
         d += std::string("  ") + *sbegin;
       }

       d += "Press enter to continue";
     
       d.popup(manager_->window());
     }



     //
     //  Prompt the user with the list and let him select one
     //


     {
      CursorWindow::Selection d("Select an alternate source file below:");

      matches_t::iterator sbegin, send;

      for(sbegin = matches.begin(), send = matches.end();
          sbegin != send;
          ++sbegin
         )
      {
        d += *sbegin;
      }
  
      if(d.popup(manager_->window()))
        {
          return false;
        }

      //
      // Load the selected file into an editor
      // 

      {
        FileName file(d.selection_);

        Viewer* oldie = manager_->find_viewer(file);
      
        if(oldie)
        {
          manager_->activate(oldie);
        }
        else
        {
      
          Viewer* newbie = manager_->edit(file);
      
          if(newbie)
          {
            manager_->add(newbie, file);
      
          }
          else
            vp->beep();
        }
      
      }
      
     }
  }


  return false;
}
////////////////// switch_source /////////////////////////////////////////////////

bool
Impl::
fetchAllLines(std::list< std::string > &output)
{
    rep_type::iterator first = lines_.begin(), last = lines_.end();

    while(first != last)
    {
        output.push_back(*first++);
    }

    return true;
}

