#ifndef text_editor_included_h
#define text_editor_included_h
//
// Copyright 2002, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.

#include <cxxtls/viewer.h>
#include <cxxtls/file.h>


class TextEditor
: public cxxtls::Viewer
    ///  A viewer that implements an ascii text editor which operates on a single file.
    ///
    ///  This viewer lets the upser change and save ascii files.  It is meant for programmers
    ///  not general purpose text editing.  For example, it understands that source code files
    ///  should not be saved with tab but that other files, like makefiles should.
    ///
    ///  It understands the following broadcast messages:
    ///
    ///  @li fileLocation --  sent to indicate that the user has pressed ^X-space to indicate
    ///                       that a breakpoint should be set.
    ///
    ///  @li breakpoint   --  received to indicate the marking or the clearing of a breakpoint mark.
    ///
    ///  @li stopPoint    --  received to indicate the marking or the clearing of a stop mark sent
    ///                       from the debugger.
    ///
    ///  See the listen() method for interpretation of broadcast messages
{
public:

  struct Impl;     ///<  Implemented texteditor.cxx.  This class actually implements
		   ///   the texteditor.  The TextEditor class is just a wrapper that
		   ///   presents itself to the ViewerManager, but the Impl class is
		   ///   where all the real editor functionality lies.

private:

  std::auto_ptr<Impl> impl_;

public:

  virtual void operator() ( cxxtls::CursorWindow::viewport *  vp,
			    int                               cmd
			  );
			
  virtual bool handle_event( cxxtls::CursorWindow::input_event const * e,
			     cxxtls::CursorWindow::viewport          * vp
			   );
  virtual void help();

  TextEditor(cxxtls::ViewerManager*vm, cxxtls::FileName fullname);

  ~TextEditor();

  static cxxtls::Viewer* app(cxxtls::ViewerManager*vm, std::string &fullname);
   ///<
   ///  The function used by the viewer manager to prompt the user
   ///  for a new text viewer.  See @@ViewerManager::application@@.

  static std::string app_name; ///<  The name of this application

  std::string const& application_name() const { return app_name; }
    ///<
    ///  Return the name of the application to use when describing this
    ///  viewer in the @@ViewerManager::appChooser@@ or in the
    ///  @@ViewerManager::ViewerSelector@@.

  void set_row_col_hint(size_t line, size_t col, bool force_repaint);
      ///<  handle the row /col hint that comes from the main program's
      ///   command line to indicate which line of the file to show.

  std::string description() const;
    ///<
    ///	 This function returns the name	of the file being edited.  It is
    ///  the description of this edit session.

  bool is_dirty() const ;
      ///<  Some real, honest to goodness change has been made to the data in
      ///   the edit session.

  void set_read_only() ;  
    ///<
    ///  Mark this edit session as read only.  Can be undone by the user


  virtual bool fetchAllLines(std::list<std::string> &output);
      ///<  See cxxtls::Viewer::fetchAllLines

  virtual void listen(std::string const &type, std::string const &msg);
     ///<  Listen for broadcast messages.  The msg is assumed to be encoded
     ///   in VarObj style.

};

#endif
