#ifndef cxxtls_threadGate_header_included_p
#define cxxtls_threadGate_header_included_p
//
// Copyright 2002-2013, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.
//
// Warning:  not all files in this directory structure may be covered by the
// same copyright.  Some of them may be part of the GNU source distribution
// and you must obey the GPL copyright for any such files.

///@file threadGate
///
///  Defines class ThreadGate
///
///
#include <cxxtls/thread.h>

class  ThreadGate
    ///  A ThreadGate is an object that can be used by one thread to wait until multiple events have occurred as signaled
    ///  by other threads.  Basically, it is a thread safe event counter.
    ///
    ///  The idea is that a main thread will use a ThreadGate to synchronize the actions of other threads.  The main thread
    ///  will call the following members:
    ///
    ///  @li @ref ThreadGate(size_t count)
    ///  @li @ref reset(size_t count)
    ///  @li @ref wait()
    ///
    ///  The constructor and reset() calls start the event processing sequence and indicate that the number of events that
    ///  the wait() function is expecting to receive from other threads as they perform some sort of service and then call the
    ///  signal() method to indicate that they are finished. 
    ///
    ///  The wait() method suspends the caller until the sequence is complete.  The wait function() may return immediately
    ///  or may delay the caller indefinitely -- it all depends on the arrival of events.  Events arrive when other threads
    ///  make calls to the signalEvent() method.
    ///
    ///  Note the ThreadGate object can be reused but before any secondary sequence of operations, call reset(count).
    ///
    ///  Note that more than one thread can call the wait() method.  This allows for group synchronization.  Basically,
    ///  you can use a pair of ThreadGates to regulate the starting of a bunch of threads.  The creator of the group
    ///  can wait until all worker threads have started and signaled that they are awake.  Once started, the worker threads
    ///  can wait on a gate that the creator thread signals.  That way, you can line up your horses, have them waiting at the
    ///  starting gate, then fire the starter pistol, and they will all race away at the same time.  Er, uh ... maybe I've had too much
    ///  coffee.
    ///
{
    cxxtls::Mutex             guardian_;
    cxxtls::ConditionVariable sequencer_;

    size_t            eventsReceived_;
    size_t            eventsExpected_;

public:

    struct UnexpectedSignalException {};

    ThreadGate(size_t count)
    :  eventsReceived_(0)
    ,  eventsExpected_(count)
        /// Construct an event gate and set the count of expected events.
    {
        CXXTLS_DEBUG("construct ThreadGate(" << count << ", " << std::hex << (long )(this) << std::hex << ")");
    }

    void signal()
       /// Worker threads call this function to signal the completion of a task.  This function
       /// increments the eventsRecieved_ flag;  In the event that too many events arrive, the signal function will
       /// throw an UnexpectedEventException.  When the count of received events equals the number of expected events
       /// 
    {
        cxxtls::ScopedMutexLock lock(guardian_);

        ++eventsReceived_;

        CXXTLS_DEBUG("signal ThreadGate(received=" << eventsReceived_ << ", expected=" << eventsExpected_ << ", " << std::hex << (long )(this) << std::hex << ")");


        if(eventsReceived_ > eventsExpected_)
        {
           sequencer_.notifyAll();               // release all the waiters but thrown an exception.
                                                 // A different thread may get the wait signal than
                                                 // gets the exception.
           throw UnexpectedSignalException();
        }
        else
        if(eventsReceived_ == eventsExpected_)
            sequencer_.notifyAll();   
    }

    void wait()
    {
        cxxtls::ScopedMutexLock lock(guardian_);

        if(eventsExpected_ == eventsReceived_)
        {
           CXXTLS_DEBUG("wait already complete on call to wait ThreadGate(received=" << eventsReceived_ << ", expected=" << eventsExpected_ << ", " << std::hex << (long )(this) << std::hex << ")");

           eventsExpected_ = 0;
           eventsReceived_ = 0;
           return;
        }

        CXXTLS_DEBUG("waiting on ThreadGate(received=" << eventsReceived_ << ", expected=" << eventsExpected_ << ", " << std::hex << (long )(this) << std::hex << ")");

        sequencer_.wait(lock);   // wait till the receivedCount_ == expected count;

        eventsExpected_ = 0;
        eventsReceived_ = 0;

        CXXTLS_DEBUG("wait complete ThreadGate(received=" << eventsReceived_ << ", expected=" << eventsExpected_ << ", " << std::hex << (long )(this) << std::hex << ")");


    }

    void reset(size_t count)
    {

       cxxtls::ScopedMutexLock lock(guardian_);

       eventsExpected_ = count;
       eventsReceived_ = 0;

       CXXTLS_DEBUG("resetting ThreadGate(received=" << eventsReceived_ << ", expected=" << eventsExpected_ << ", " << std::hex << (long )(this) << std::hex << ")");

    }

    bool sequenceComplete()
    {
       cxxtls::ScopedMutexLock lock(guardian_);

       return eventsExpected_ == eventsReceived_;
    }




};


#endif
