#ifndef cxxtls_thread_monitored_set_header_included_p
#define cxxtls_thread_monitored_set_header_included_p

#include <cxxtls/thread.h>
#include <set>
#include <list>

namespace cxxtls
{
template<class T>
struct IntReasonType
   /// a trivial implementation of the ExitReasonType used by the following template.
   /// This kind of exit reason type exists to simplify many uses of ThreadMonitoredSet 
   /// though not all of them.  See uses of this class in the ThreadMonitoredSet::wait()
{
    int value_;

    IntReasonType(int value=0)
    :  value_(value)
    {
    }

    operator int() const { return value_; }

    static IntReasonType<T> unset;
    static IntReasonType<T> shuttingDown;

};

template<class T> IntReasonType<T> IntReasonType<T>::unset(-1);
template<class T> IntReasonType<T> IntReasonType<T>::shuttingDown(-2);

template<class T, class ExitReasonType=IntReasonType<T> >
class ThreadMonitoredSet
    ///  A ThreadMonitoredSet is a set of items whose existence can be monitored by multiple threads.
    ///  The goal here is to provide a way to insert an item in the set and to be informed when it
    ///  is removed.  
    ///
    ///  @warning  Do not use std::string as T!  Do not use any reference counted object
    ///  as T.  See more warnigns below.
    ///
    ///  @warning  ExitReasonType must be defined carefully for use by this class.  Here are the rules:
    ///
    ///    @li  ExitReasonType must not containin any pointers to objects outside itself.  It must tbe
    ///         mostly Plain Old Data (POD).
    ///    @li  But, it must must have 2 static data members of type ExitReasonType, which have the
    ///         following names and meanings:  1)  unsetValue, 2)  shuttingDown.
    ///
    ///  That is, the ExitReasonType class must be designed for use with this class which must communicate
    ///  3 kinds of information to the caller of the wait() method:
    ///
    ///     @li  the object, T, is still in use.
    ///     @li  the thread monitored set is shutting down, so you should act as if T is gone from the list
    ///     @li  the object, T, is finished and has returned a value.
    ///
    ///  This means you need 3 basic values:  still in use, shutting down, all other successful completions.
    ///  See the notes below, in this comment, about using std::string in the T type and apply the same 
    ///  thinking for your ExitReasonType class.
    ///
    ///  To use this class, you do the following:
    ///
    ///  @li use the add(t) function to insert something into the set.
    ///  @li use the wait(t) function to wait until that object has been removed.
    ///  @li use the remove(t,reasontype) function to remove it
    ///  @li you can also use the check(t) function to see if it is still there without waiting.
    ///  @li you use the shutdown() function to wake all the waiters up and tell them that all the
    ///      data is gone.  This will delete all the items in the set and mark the set as unusable
    ///      for future operations.
    ///  @li you can use getList() to fetch a copy of all the contents in a thread safe manor, but this
    ///  will not overcome the problems associated with std::string.  See getList() and the warnings below
    ///  for more details.
    ///
    ///  @warning Multithreading warning:  all items in this set should be thread safe.  For integers, this 
    ///  is easy.  But for std::string and any object that has pointers to data that is shared across
    ///  threads, this can be problematic.
    ///
    ///  To use a std::string as T or <b>as a member of T</b>, you must make sure that the string that gets
    ///  stuck in the T (or becomes the T) is unique across threads.  Some implementations of std::string allow multiple
    ///  std::string objects to share the same "text body" -- using reference counters to delete it when
    ///  the last customer is destroyed.  But this causes multithreading bugs:  crashes or memory leaks. 
    ///  
    ///  It is generally best not to pass std::string objects to from one thread to another.  It can seem
    ///  very much like it is working, then years later, after you have left the project, some poor guy will
    ///  be stuck dealing with a crash or an out of memory situation that no one on the project can figure
    ///  out because the crash is occurring to a super-rare race condition that magically becomes ubiquitous
    ///  for no apparent reason.
    ///
    ///  To pass a string to a thread, use operator new to create a heap packet and pass that.
    ///
    ///  An example of using a std::string in a T object:
    ///
    ///@begincode
    ///
    ///    struct T
    ///    {
    ///         std::string member_;
    ///
    ///         T(std::string const &value)
    ///         {
    ///             member_.assign(value.begin(), value.end());  // makes member_'s contents unique
    ///         }
    ///
    ///         T(T const &rhs)
    ///         {
    ///             member_.assign(rhs.member_.begin(), rhs.member_.end());  // makes member_'s contents unique
    ///         }
    ///
    ///         T& operator=(T const &rhs)
    ///         {
    ///             member_.assign(rhs.member_.begin(), rhs.member_.end());  // makes member_'s contents unique
    ///         }
    ///    }
    ///@endCode
    ///
    ///  @note the above use of the std::string::assign() method does not make T thread safe, but it does
    ///  make each copy of the member_ string unique.  So combining that behavior of T with ThreadMontoredSet's
    ///  mutex locking, you should be good.
    ///
    ///  Note, if your ExitReasonType contains strings or objects with pointers outside of themselves, they need to follow
    ///  rules similar to struct T in the above example.
    ///
{
       std::set<T> items_;
       Mutex       itemsGuard_;
       bool        shutdown_;

       ConditionVariable waiters_;

       // in order to orchestrate the passing of "exit codes" from one thread to another, the following items
       // are included in this container:

       struct WaitInfo
       {
          int              count_;
          ExitReasonType   reasonCode_;
       };

       std::map<T,WaitInfo> waitCounts_;      ///< for every T in the set, a count of waiting threads is kept, as is the return code


public:

       struct DuplicateEntryException{};
       struct MissingItemException{};


       ThreadMonitoredSet()
       :  shutdown_(false)
       {
       }

       ~ThreadMonitoredSet()
       {
           ScopedMutexLock lock(itemsGuard_);

           items_.clear();

           shutdown_ = true;

       }

       void add(T const &t)
          /// add a new, previouly non existing object to the set.
       {
          ScopedMutexLock lock(itemsGuard_);

          if(items_.find(t) != items_.end())
              throw DuplicateEntryException();

          if(shutdown_)
             return;  // just ignore the attempt to insert.  If you attempt to wait
                      // for the t object, wait will return immediately telling you it is
                      // gone.

          items_.insert(t);

          // don't need to notifyall becuase we are the only thread that
          // cares about the item we just inserted.

          // we will of course be calling wait(t) shortly.  But when we do
          // that, it is entirely possible that the t we just inserted will have
          // been removed.  But so what, the wait function will return immediately
          // because t will have been removed in this case.  On the other hand,
          // if t has not been removed by the time we call wait, then wait will
          // hang until someone else removes it -- by design.

       }

       void remove(T const &t, ExitReasonType const & reasonCode)
          ///  remove the specified item from the set and notify all waiters that
          ///  something has happened.  Each one will wake up and check to see if the
          ///  removed items is the thing they are waiting on.  If is not, they will go
          ///  back to waiting, otherwise they will stop waiting and continue about their
          ///  business. 
          ///
          ///  @param t           The item in the set being kept track of.
          ///  @param reasonCode  An "exit code", or some other state in place at the time
          ///                     of the remove call.  Could mean anything you want.
          /// 
       {
          ScopedMutexLock lock(itemsGuard_);

          std::set<T>::iterator where = items_.find(t);

          if(where == items_.end())
              throw MissingItemException();

          items_.erase(where);

          std::map<T,WaitInfo>::iterator waiter = waitCounts_.find(t);

          if(waiter != waitCounts_.end())
          {
             // at least one thread waits for this one

             waiter->second.reasonCode_ = reasonCode;
          }


          waiters_.notifyAll();

       }

       void check(T const &t) 
       {
          ScopedMutexLock lock(itemsGuard_);

          if(items_.find(t) != items_.end())
             return true;

           return false;
       }

       bool isShutdown() 
       {
          ScopedMutexLock lock(itemsGuard_);

          return shutdown_;
          
       }

       ExitReasonType wait(T const &t)
           /// wait forever until the item in the set whose value is t gets removed.  If it is already removed
           /// return immediately.
           ///
           /// Return the value stored in the return code map, waitCounts_.  The remove() function sets the value
           /// that gets returned.
           ///
           /// A return code of -1 means that the item being waited for was not found in the waitCounts_ map.  ? got me dude.
           /// A return code of -2 means that the system is shutting down.  
           ///
           /// @warning  The ExitReasonType object must not contain pointers to objects owned by the thread being
           /// monitored because when that thread dies, the returned pointer will be trash.
           ///
           /// @note The default constructor for the ExitReasonType objects should indicate the state of being "unset",
           ///       and it is up to the using application to ensure this.
       {
           bool firstTimeThrough=true;

           for(;;)
           {

               ScopedMutexLock lock(itemsGuard_);

               if(shutdown_)
               {
                  waitCounts_.clear();
                  return ExitReasonType::shuttingDown;
               }

               if(items_.find(t) == items_.end())
               {
                  std::map<T,WaitInfo>::iterator where = waitCounts_.find(t);

                  if(where == waitCounts_.end())
                     return ExitReasonType::unset;

                  ExitReasonType rc = where->second.reasonCode_;

                  --where->second.count_;

                  if(where->second.count_ == 0)
                  {
                     waitCounts_.erase(  where );
                  }

                  return rc;   
               }
               else
               if(firstTimeThrough)
               {
                   firstTimeThrough = false;

                   std::map<T,WaitInfo>::iterator where = waitCounts_.find(t);

                   if(where == waitCounts_.end())
                   {
                      WaitInfo tmp;

                      tmp.count_      = 1;
                      tmp.reasonCode_ = ExitReasonType::unset;

                      waitCounts_[t] = tmp;
                   }
                   else
                   {
                      ++where->second.count_;
                   }

               }

               waiters_.wait(lock);

               // if we get here, someone has awakened us -- either because
               // some item, possibly ours but possibly not, has been removed
               // from the set.  Or the shutdown() function has been called
               // in which case, all have been removed.

           }

       }


       void shutdown()
       {
           ScopedMutexLock lock(itemsGuard_);

           shutdown_ = true;

           items_.clear();

           waiters_.notifyAll();

       }



       void getList(std::list<T>  &output) 
          ///  Copy all the items in the set to a list and so so in a thread
          ///  guarded operation.  Note that this does not solve the problem
          ///  of std::string objects being corrupted by threads.
          ///
       {


           ScopedMutexLock lock(itemsGuard_);

           output.clear();
           
           CXXTLS_FOREACH(T const &cur, items_) output.push_back(cur);
       }


};

} // cxxtls

#endif
