#include <iostream>
#include <cxxtls/stringexpression.h>
#include <cxxtls/stringexpressionparser.h>
#include <stdlib.h>
#include <cxxtls/lastErrno.h>


using namespace std;
using namespace cxxtls;

int main()
{
    string tmp = lastErrno();  // call but ignore

    bool errorFound = false;

    StringExpr::ExecutionVariables globals;
    StringExpr::ExecutionContext context(&globals);

    globals.includeProgramEnvironment();


    if(!globals.contains("PATH"))
    {
        cout << "Error, could not find the PATH setting" << endl;
        errorFound = true;
    }

    if(!globals.contains("SE_TEST_VAR"))
    {
        cout << "Error, could not find the SET_TEST_VAR setting" << endl;
        errorFound = true;
    }
    else
    {
       string value = globals.get("SE_TEST_VAR");

       if(value != "expectedValue")
       {
            cout << "Error, SET_TEST_VAR value (" << value << ") was not (expectedValue) " << endl;
        errorFound = true;
       }
    }



    context.assign("hank", "HANK");
    context.assign("HOME", "/home/usr/user");
    context.assign("stuff", "STUFF");
    context.assign("sos", "SOS");
    context.assign("lower", "lower and more");

    context.addArgument("programName");   // argument 0 is always the program name and does not count in the $# count!
    context.setScriptName("programName");
    context.addArgument("fred");
    context.addArgument("bill");
    context.addArgument("frank");
    context.addArgument("sue");


    typedef pair<string,string> ExprTestEntry;

    static ExprTestEntry expression[]=
    {

        ExprTestEntry("stuff",  "stuff"),
        ExprTestEntry("'stuff'",  "stuff"),
        ExprTestEntry("\"stuff\"",  "stuff"),

        ExprTestEntry("$stuff",      "STUFF"),
        ExprTestEntry("\"$stuff\"",  "STUFF"),
        ExprTestEntry("'$stuff'",    "$stuff"),

        ExprTestEntry("${stuff}",      "STUFF"),
        ExprTestEntry("\"${stuff}\"",  "STUFF"),
        ExprTestEntry("'${stuff}'",    "${stuff}"),

        ExprTestEntry("${HOME}",      "/home/usr/user"),
        ExprTestEntry("${HOME%/*}",   "/home/usr"),
        ExprTestEntry("${HOME%%/*}",  ""),
        ExprTestEntry("${HOME#/*}",   "home/usr/user"),   // this is what bash does!
        ExprTestEntry("${HOME##/*}",  ""),

        ExprTestEntry("${HOME/\\//-}", "-home/usr/user"),  // replace first / with -
        ExprTestEntry("${HOME//\\//-}", "-home-usr-user"),  // replace all / with -

        ExprTestEntry("${sos/#S/x}", "xOS"),  // replace s only at the beginning

        ExprTestEntry("${HOME/%r/R}", "/home/usr/useR"),  // replace R only at the end

        ExprTestEntry("seEval.*.txt", "seEval.1.txt seEval.2.txt seEval.3.txt"),

#ifndef _MSC_VER
        ExprTestEntry("/u*/l*/lib", "/usr/local/lib"),   // iffy test!  some hosts have
                                                         // other /lib subdirs to confuse
                                                         // the issue
#else
        ExprTestEntry("/W*/SysW*/drivers", "c:/Windows/SysWOW64/drivers"),   
#endif

        ExprTestEntry("${lower^}", "Lower and more"),
        ExprTestEntry("${lower^^}", "LOWER AND MORE"),

        ExprTestEntry("${lower,}", "lower and more"),     // no change
        ExprTestEntry("${lower,,}", "lower and more"),    // no change

        ExprTestEntry("${hank,}", "hANK"),
        ExprTestEntry("${hank,,}", "hank"),

        ExprTestEntry("${#sos}", "3"),
        
        ExprTestEntry("${#@}", "4"),
        ExprTestEntry("${#*}", "4"),

        ExprTestEntry("$*", "fred bill frank sue"),
        ExprTestEntry("$@", "fred bill frank sue"),

        ExprTestEntry("${*}", "fred bill frank sue"),
        ExprTestEntry("${@}", "fred bill frank sue"),

        ExprTestEntry("${*#f}", "red bill rank sue"),
        ExprTestEntry("${@#f}", "red bill rank sue"),

        ExprTestEntry("${*##f*}", "bill sue"),
        ExprTestEntry("${@##f*}", "bill sue"),


        ExprTestEntry("${1}", "fred"),
        ExprTestEntry("${2}", "bill"),
        ExprTestEntry("${3}", "frank"),
        ExprTestEntry("${4}", "sue"),


        ExprTestEntry("$1", "fred"),
        ExprTestEntry("$2", "bill"),
        ExprTestEntry("$3", "frank"),
        ExprTestEntry("$4", "sue"),

    };

    size_t expressions = sizeof(expression) / sizeof(expression[0]);

    for(size_t i = 0; i < expressions; ++i)
    {

      cout << "Evaluating[" << i << "]:  <" << expression[i].first << ">" << endl;

       std::string::iterator first(expression[i].first.begin());
       std::string::iterator last(expression[i].first.end());

        std::auto_ptr<  StringExpression > e(  StringExpr::parse(first, last) );

        if(!e.get())
          {
            cout << __FILE__ << ':' << __LINE__ << ":  Error expression failed to parse!" << endl;
            errorFound = true;
            continue;
          }

        StringArray output;
        e.get()->execute(&context, output);

        string rv = output.combine();

        bool passed = (rv == expression[i].second);

        if(passed)
        {
           cout << "    Passed!" << endl;
        }
        else
        {
           cout << "    FAILED:  Expected: " << expression[i].second << endl;
           cout << "                Found: " << rv << endl;
           cout << __FILE__ << ':' << __LINE__ << ":  failure point" << endl;
           //exit(1);
           errorFound = true;
        }
       
    }

    // the following aren't string evals, but they fit here in the test
    // suite well enough:
    {
       
        cout << "Checking for redirection strings." << endl;

        struct ExprTestEntry
        {
           string expression_;
           string value_;
           bool   isRedir_;
           bool   isSpecial_;

           ExprTestEntry(string const &e, string const &v, bool r, bool s)
           :  expression_(e)
           ,  value_(v)
           ,  isRedir_(r)
           ,  isSpecial_(s)
           {

           }
        };


        static ExprTestEntry expression[]=
        {
                                        //executed     isRedir   isSpecial
            ExprTestEntry("",             "",          false,    false),
            ExprTestEntry("something",    "something", false,    false),
            ExprTestEntry("'something'",  "something", false,    false),
            ExprTestEntry("'>'",          ">",         false,    false),
            ExprTestEntry(">file",        ">file",     true,     true),
            ExprTestEntry(">'file'",      ">file",     true,     true),
            ExprTestEntry("> \"file\"",   ">file",     true,     true),
            ExprTestEntry("> \"$hank\"",   ">HANK",     true,     true),
            ExprTestEntry("<file",        "<file",     true,     true),
            ExprTestEntry("<  file",      "<file",     true,     true),
            ExprTestEntry("2",            "2",         false,    false),
            ExprTestEntry("2>file",       "2>file",    true,     true),
            ExprTestEntry("2> file",      "2>file",    true,     true),
            ExprTestEntry("2<file",       "2<file",    true,     true),
            ExprTestEntry("2< file",      "2<file",    true,     true),
            ExprTestEntry("23< file",     "23<file",   true,     true),
            ExprTestEntry("2>&1",         "2>&1",      true,     true),
            ExprTestEntry("|",            "|",         false,    true),
            ExprTestEntry("&&",           "&&",        false,    true),

        };
        size_t expressions = sizeof(expression) / sizeof(expression[0]);

        for(size_t i = 0; i < expressions; ++i)
        {

            string e = expression[i].expression_;
            string v = expression[i].value_;
            bool   r = expression[i].isRedir_;
            bool   s = expression[i].isSpecial_;
            
            std::string::const_iterator first = e.begin(),
                                        last  = e.end();

            std::auto_ptr<  StringExpression > expr(  StringExpr::parse(first, last) );

            if(!expr.get())
            {
                cerr << __FILE__ << ':' << __LINE__ << ":  expression " << i << " failed to parse" << endl;
                exit(1);
            }

            if(expr->isRedir() != r)
            {
                cerr << __FILE__ << ':' << __LINE__ << ":  expression " << i 
                     << " expr->isRedir()=" << expr->isRedir() << " was wrong, expected " <<  r << endl;
                exit(1);
               
            }

            StringArray output;
            expr->execute(&context, output);
            string rv = output.combine();

            if(rv != v)
            {
                cerr << __FILE__ << ':' << __LINE__ << ":  expression " << i 
                     << " expr->execute()='" << rv << "' was wrong, expected '" <<  v << "'" << endl;
                exit(1);
               
            }

            if(expr->isSpecial() != s)
            {
                cerr << __FILE__ << ':' << __LINE__ << ":  expression " << i 
                     << " expr->isSpecial()=" << expr->isSpecial() << " was wrong, expected " <<  s << endl;
                exit(1);
               
            }

        }


        cout << "     Checking for redirection strings:  PASSED" << endl;

    }

//////////////////
    // the following aren't string evals, but they fit here in the test
    // suite well enough:
    {
       
        cout << "Checking for non-redirection string parsing." << endl;

        struct ExprTestEntry
        {
           string expression_;
           string value_;
           bool   isRedir_;
           bool   isSpecial_;

           ExprTestEntry(string const &e, string const &v, bool r, bool s)
           :  expression_(e)
           ,  value_(v)
           ,  isRedir_(r)
           ,  isSpecial_(s)
           {

           }
        };


        static ExprTestEntry expression[]=
        {
                                        //executed     isRedir   isSpecial
            ExprTestEntry("",             "",          false,    false),
            ExprTestEntry("something",    "something", false,    false),
            ExprTestEntry("'something'",  "something", false,    false),
            ExprTestEntry("'>'",          ">",         false,    false),
            ExprTestEntry(">file",        ">",         false,    false),
            ExprTestEntry(">'file'",      ">",         false,    false),
            ExprTestEntry("> \"file\"",   ">",         false,    false),
            ExprTestEntry("> \"$hank\"",  ">",         false,    false),
            ExprTestEntry(">\"$hank\"",   ">",         false,    false),
            ExprTestEntry(">$hank",       ">",         false,    false),
            ExprTestEntry("<file",        "<",         false,    false),
            ExprTestEntry("<  file",      "<",         false,    false),
            ExprTestEntry("<<",           "<<",        false,    false),
            ExprTestEntry(">>",           ">>",        false,    false),
            ExprTestEntry("<=",           "<=",        false,    false),
            ExprTestEntry(">=",           ">=",        false,    false),
            ExprTestEntry("|",            "|",         false,    true),
            ExprTestEntry("&&",           "&&",        false,    true),

        };
        size_t expressions = sizeof(expression) / sizeof(expression[0]);

        for(size_t i = 0; i < expressions; ++i)
        {

            string e = expression[i].expression_;
            string v = expression[i].value_;
            bool   r = expression[i].isRedir_;
            bool   s = expression[i].isSpecial_;
            
            std::string::const_iterator first = e.begin(),
                                        last  = e.end();

            std::auto_ptr<  StringExpression > expr(  StringExpr::parseStrExpr(first, last, false) );  // parse for non-redirection expressions.
                                                                                                       // that is, treat <, >, and &, like normal
                                                                                                       // operators.

            if(!expr.get())
            {
                cerr << __FILE__ << ':' << __LINE__ << ":  expression " << i << ", e=" << e << ", failed to parse" << endl;
                // exit(1);
            }

            if(expr->isRedir() != r)
            {
                cerr << __FILE__ << ':' << __LINE__ << ":  expression " << i 
                     << " expr->isRedir()=" << expr->isRedir() << " was wrong, expected " <<  r << endl;
                // exit(1);
               
            }

            StringArray output;
            expr->execute(&context, output);
            string rv = output.combine();

            if(rv != v)
            {
                cerr << __FILE__ << ':' << __LINE__ << ":  expression " << i 
                     << " expr->execute()='" << rv << "' was wrong, expected '" <<  v << "'" << endl;
                exit(1);
               
            }

            if(expr->isSpecial() != s)
            {
                cerr << __FILE__ << ':' << __LINE__ << ":  expression " << i 
                     << " expr->isSpecial()=" << expr->isSpecial() << " was wrong, expected " <<  s << endl;
                exit(1);
               
            }

        }


        cout << "     Checking for non-redirection strings:  PASSED" << endl;

    }
/////////////////


    {
       cout <<  "Confirming ExecutionContext::assign:" << endl;

       StringExpr::ExecutionVariables base, unified;
       StringExpr::ExecutionContext context(&base);

       context.superseed("fred","FRED");  // supserseed it to be the same as it always was
       
       if(context.find("fred") != "FRED")
       {
          cerr << __FILE__ << ':' << __LINE__ << ":  variable 'fred' was not found in modified ExecutionContext" << endl;

          exit(1);
          
       }

       if(base.get("fred") == "FRED")
       {
          cerr << __FILE__ << ':' << __LINE__ << ":  variable 'fred' was mistakenly found in supposedly non-modified base ExecutionContext" << endl;

          exit(1);
          
       }

       context.unifyVariables(unified);

       if(unified.get("fred") != "FRED")
       {
          cerr << __FILE__ << ':' << __LINE__ << ":  ExecutionContext::unifyVariables() is broken" << endl;

          exit(1);
          
       }

       context.assign("fred","SUSAN");

       if(context.find("fred") != "SUSAN")
       {
          cerr << __FILE__ << ':' << __LINE__ << ":  variable 'fred' was not replaced in modified ExecutionContext" << endl;

          exit(1);
          
       }

       if(base.get("fred") != "SUSAN")
       {
          cerr << __FILE__ << ':' << __LINE__ << ":  ExecutionContext::assign() is broken" << endl;

          exit(1);
          
       }

       cout << "    Passed!" << endl;



    }

    cout << endl;  // force a blank line to improve fc.exe on windows

    if(errorFound)
    {
       cerr << "Errors found in the test, see above\n";
       exit(1);
    }



    return 0;
}
