#ifndef __linux__
int please_ignore_this_file_on_windows_gdbviewer=0;
#else

///@file
///
///  This file implements the gdb viewer functionality.
///

#include <cxxtls/gdbviewer.h>
#include <cxxtls/viewermanager.h>
#include <cxxtls/lexical_cast.h>
#include <cxxtls/file.h>
#include <cxxtls/fmtd.h>

#include <cxxtls/deltaviewer.h>
#include <cxxtls/dumbterminal.h>
#include <portable_io.h>
#include <cxxtls/strtool.h>

#include <pty.h>

namespace cxxtls
{
//----------------------------------------------------------------------------------

    static inline std::string linePrefix(int depth)
        /// return a string of blanks with the specified length
    {
       std::string rv;
    
       while(depth--)
          rv += ' ';
    
       return rv;
    }

//----------------------------------------------------------------------------------

    static void splitGdbVariable(std::string const &input, std::list<std::string> &output, size_t maxWidth);
        ///  Parse the display format for variables in the gdb interpreter mode command, "-stack-list-arguments 1",
        ///  into a array of strings which represent indentation depth of structures.  The resultant string
        ///  may contain strings, arrays, integers, floats, enumeration names, structures, etc.  Structures will
        ///  be split into multiple lines.  Atomic data types will be displayed all on one line.  
        ///
        ///  This function is used to prepare data for inclusion in a window.
        ///
        ///  @param[in]  input  A "value" string returned by stackFrameParms() or stackFrameVariables()
        ///
        ///  @param[out] output A vector of strings containing the same data as input but arranged for
        ///                     attractive formatting.
        ///
        ///  @param maxWidth -- the maximum width to target (if possible) when generating formatted
        ///                     output.  Typically this is the width of the window into the split variable
        ///                     will be displayed in.

//----------------------------------------------------------------------------------

    static std::string breakpoint("breakpoint");

//----------------------------------------------------------------------------------

    static std::string bpMessage(std::string const &file, std::string const &line, bool deleteIt)
        /// create a string encoded in VarObj format that means to delete a breakpoint at a specified
        /// file and line
        ///
        /// The generated string looks like this:
        ///@code
        ///
        ///    {file="filename",line="lineNumber",action="delete"}
        ///
        ///@endcode
        ///
        /// Without the action="delete" part, this would be a create breakpoint notification
    {
        std::string msg;

        msg += "{file=\"";
        msg += file;
        msg += "\",line=\"";
        msg += line;
        if(deleteIt)
            msg += "\",action=\"delete\"}";
        else
            msg += "\"}";

        return msg;
    }
//----------------------------------------------------------------------------------

    static std::string stopped("stopped");
//----------------------------------------------------------------------------------

    static std::string stopPointMessage(std::string const &file, std::string const &line, bool deleteIt)
        /// create a string encoded in VarObj format that means that the debugger has
        /// stopped a a particular file location.
        ///
        /// The generated string looks like this:
        ///@code
        ///
        ///    {file="filename",line="lineNumber"}
        ///
        ///@endcode
        ///
        /// The stop point message is sent to the editor to inform it to display a marker where the 
        /// the program has stopped.
        ///
    {
        std::string msg;

        msg += "{file=\"";
        msg += file;
        msg += "\",line=\"";
        msg += line;
        if(deleteIt)
        {
            msg += "\",action=\"delete\"}";
        }
        else
            msg += "\"}";

        return msg;
    }


//-----------------------------------------------------------------------------------

static std::string logFromExeName(std::string const &exeName)
{
    FileName tmp(exeName);

    return tmp.dirname() + "gdblog.log";
}



GdbViewer::
GdbViewer(ViewerManager *vm, std::string const &fullname)
:  Viewer(vm)
,  breakpoints_(0)
,  variables_(0)
,  stack_(0)
,  terminal_(0)
,  active_(false)
,  runnable_(false)
,  argsSent_(false)
,  firstStart_(true)         // more aptly named "first split".  See showCurrentFile()
,  gdb_(false, logFromExeName(fullname))
,  savedState_(gdb_.state())
,  stackViewer_(0)
,  gdbStackViewer_(0)
,  logViewer_(0)
,  gdbLogViewer_(0)
,  breakpointViewer_(0)
,  gdbBreakpointViewer_(0)
,  gdbVariableViewer_(0)
,  variableViewer_(0)
,  showLine_(0)  // show file name defaults to empty
,  stopFrames_(0)
,  attachMode_(0) 
{

   std::string tmp(fullname);

   if(tmp.empty())
   {
       FileName::sorted_names_t filesInThisDirectory;

       FileName::find_matching("*", &filesInThisDirectory, 0x05); // give me all files except directories
                                                                  // and include the paths on the names

       typedef std::multimap<FileTime, FileName>  DateSortedStrings;

       DateSortedStrings executables;

       CXXTLS_FOREACH(FileName const &cur, filesInThisDirectory)
       {
           if(cur.is_executable())
           {
              FileTime tmp;

              if(!cur.file_time(&tmp))
              {
                  executables.insert( DateSortedStrings::value_type(tmp, cur) );
              }
           }
       }


       CursorWindow::Selection prompt("Use gdb on an executable");

       DateSortedStrings::iterator l = executables.end(), f = executables.begin();

       while(l != f)
       {
          --l;

          prompt += l->second;
       }

      if(!prompt.popup(vm->window()))
      {
          tmp = prompt.selection_;
      }
   }


   FileName tmpfname = tmp;

   tmpfname.convert_to_absolute_path();

   fullname_ = tmpfname;

   if(0 != access(fullname_.c_str(), X_OK) )
   {
       CursorWindow::Message m("Error");

       m += "";
       m += "Specified file is not executable";
       m += "";
       m += "Press Enter to continue";

       m.popup(manager_->window());

   }
   else
   {
       runnable_ = true;
   }

   handleStackViewerDeath();       // must have stack viewer -- even if not visible
   handleLogViewerDeath();         // must have log viewer -- even if not visible
   handleBreakpointViewerDeath();  // must have a breakpoint viewer -- even if not visible
   handleVariableViewerDeath();    // must have a variable viewer -- even if not visible


}

//-----------------------------------------------------------------------------------
GdbViewer::
~GdbViewer()
{
   // do not delete or manager_->discard() the stackViewer_ here!  Nor any other daughter viewer

   removeStopPointMarker();

   CXXTLS_FOREACH(ActiveBreakpoints::value_type &cur, activeBreakpoints_)
   {
       std::string const &file = cur.second.first;
       std::string const &line = cur.second.second;

       std::string msg = bpMessage(file,line, true);  // create a message to delete the bp annotation

       manager()->broadcast(breakpoint, msg);


   }

}


//-----------------------------------------------------------------------------------
static void hide_cursor(CursorWindow::viewport *vp)
{
   // move the cursor in the specified viewport to an out of the way location.

   // WARNING:  you should always ensure that the vp is valid and displayed before
   //           calling this function.  Some command handlers delete the vp and
   //           you should call manager()->find_viewport(viewer*) before calling
   //           this function.  find_viewport will give you null or a pointer to
   //           the correct viewport pointer.

   CursorWindow::row_col size = vp->size();

   vp->set_curpos(size.row_-1, size.col_-1);
}

//-----------------------------------------------------------------------------------

void 
GdbViewer::
operator() ( CursorWindow::viewport *     vp,
             int                          cmd
           )
{
   int att;

   if(cmd == 0)
       active_ = true;
   else
       active_ = false;


   if( active_ && runnable_)
   {
      att = ViewerManager::active_title_att;
   }
   else
   {
      att = ViewerManager::inactive_title_att;
   }

   vp->set_text_attribute(att);
   vp->set_curpos(0,0);

   vp->write(gdb_.stateName());
   vp->write(":");

   vp->write(fullname_);
   vp->fill_to_eol();
   
   vp->set_text_attribute(ViewerManager::normal_att);

   vp->fill_to_eos();


   if(gdb_.state() == GdbController::notStarted)
   {
       vp->set_curpos(2, 0);

       vp->write("Debugged Target is not yet started:");

       vp->set_curpos(3, 0);

       vp->write("  Press F1 for more help.");

       vp->set_curpos(4, 0);

       vp->write("  You can either use 'A' to attach to an");

       vp->set_curpos(5, 0);

       vp->write("      already running process, or");

       vp->set_curpos(6, 0);

       vp->write("  You can use F4 to set command line parameters");

       vp->set_curpos(7, 0);

       vp->write("  Use F5 to run the process");




   }
   else
   {
       
       vp->set_curpos(2, 0);
       
       if(runnable_)
       {
           if(attachMode_)
               vp->write("F1=help  F5=run F8=stop/kill F9|F10=StepOver  F11=StepIn F12=StepOut");
           else
               vp->write("F1=help  F4=SetArgs  F5=run  F7=FeedSTDIN F8=stop/kill F9|F10=StepOver  F11=StepIn F12=StepOut");
       }
       else
           vp->write("Whoa, hold on there big guy.  No can run.   No, no, no....");
       
       int row=4;
       
       CXXTLS_FOREACH(std::string const &cur, messages_)
       {
           vp->set_curpos(row++, 0);
           vp->write(cur);
       }
   }

   {
       // current vp can get deleted by one or more of the above commands

       vp = manager()->find_viewport(this);

       if(vp)
           hide_cursor(vp);
   }

}
//-----------------------------------------------------------------------------------

void
GdbViewer::
repaintTitle(CursorWindow::viewport *vp)
{
    (*this)(vp, active_);
}

//-----------------------------------------------------------------------------------

static std::string pendingBpDelete;

//-----------------------------------------------------------------------------------
void
GdbViewer::
removeBreakpoint(std::string const &id)
{
   std::map< std::string, std::pair< std::string, std::string > >::iterator where = activeBreakpoints_.find(id);
   
   if(where != activeBreakpoints_.end())
   {
      // remove the breakpoint from the local copy

      std::string error = gdb_.removeBreakpoint(id);

      if(error.empty())
      {
          std::string msg = bpMessage(where->second.first, where->second.second, true); // create a message to delete the annotation of the bp
          
          manager()->broadcast(breakpoint,msg); // let everyone know that this breakpoint is going away
          
          activeBreakpoints_.erase(where);
      }
      else
      {
         CursorWindow::Message m("Error removing breakpoint");

         m += "";
         m += "Breakpoint ID:";
         m += "";
         m += std::string("   ") + id;
         m += "";
         m += "Error Message";
         m += "";
         m += std::string("   ") + error;
         m += "";

         m.popup(manager()->window());
      }


   }

   pendingBpDelete = id; // not currently dealing with the possibility that
                         // a delete will not have completed by the time a
                         // second delete command is issued.

}

//-----------------------------------------------------------------------------------
void
GdbViewer::
enableBreakpoint(std::string const &id, bool enabled)
{
   std::map< std::string, std::pair< std::string, std::string > >::iterator where = activeBreakpoints_.find(id);
   
   if(where != activeBreakpoints_.end())
   {
       gdb_.enableBreakpoint(id, enabled);   // sleeps 1 second

       bpItem_ = id;

       displayBreakpoints(bpItem_);  // make sure bpItem is visible
   }

}

//-----------------------------------------------------------------------------------
bool
GdbViewer::
getBreakpointCondition(std::string const &id, std::string &bpCond)
{
    return gdb_.getBreakpointCondition(id,bpCond);
}

//-----------------------------------------------------------------------------------
void
GdbViewer::
setBreakpointCondition(std::string const &id, std::string const &bpCond)
{
    gdb_.setBreakpointCondition(id,bpCond);

    bpItem_ = id;

    displayBreakpoints(bpItem_);
}



//-----------------------------------------------------------------------------------

Viewer::BackgroundProcessStatus 
GdbViewer::
getBackgroundData(CursorWindow::input_event &e)
{
    // background value not important, just the existence of it

    e.value_ = 0;
    e.type_  = CursorWindow::input_event::BackgroundData;


    GdbController::ChildState currentState = gdb_.state();

    if(currentState != savedState_)
    {
        savedState_ = currentState;

        return   BACKGROUND_DATA_AVAILABLE;
    }

    if(!pendingBpDelete.empty())
    {
        gdb_.removeBreakpoint(pendingBpDelete);
        pendingBpDelete = "";

        return  BACKGROUND_DATA_AVAILABLE;

    }

    GdbController::TargetOutput &childOutputQueue = gdb_.targetOutput();

    if(!childOutputQueue.empty())
    {
       
       return BACKGROUND_DATA_AVAILABLE;
    }



    return NO_BACKGROUND_DATA_AVAILABLE;

}

//-----------------------------------------------------------------------------------
void
GdbViewer::
discardDaughterViewers()
{
    manager_->unify();

    if(stackViewer_)
    {
        manager_->discard(stackViewer_);
        stackViewer_ = 0;
    }

    if(logViewer_)
    {
        manager_->discard(logViewer_);
        logViewer_ = 0;
    }

    if(breakpointViewer_)
    {
        manager_->discard(breakpointViewer_);
        breakpointViewer_ = 0;
    }

    if(variableViewer_)
    {
        manager_->discard(variableViewer_);
        variableViewer_ = 0;
    }

       
}
//-----------------------------------------------------------------------------------

void
GdbViewer::
feedStdin(CursorWindow::viewport *vp)
{
    if(attachMode_)
       return; // no can feed stdin data like this in attachment mode

    CursorWindow::Dialog  d("Enter Target STDIN data");

    d += new CursorWindow::Dialog::String("text", "Text for Target STDIN", "", 40);
    d += new CursorWindow::Dialog::String("yesno", "Add a newline or end of file (y/n/e)?", "y", 2);

    if(!d.popup(manager()->window()))
    {
       std::string data = d.element_value("text");

       bool includeCR = true;
       bool includeEOF= false;

       std::string addcr = d.element_value("yesno");

       if(   !addcr.empty() 
          && ( addcr[0] == 'n' || addcr[0] == 'N')
         )
       {
          includeCR = false;
       }
       else
       if(   !addcr.empty() 
          && ( addcr[0] == 'e' || addcr[0] == 'E')
         )
       {
          includeEOF= true;
       }

       if(includeCR)
          data += '\n';

       if(!data.empty())
           gdb_.writeTargetStdin(data);

       if(includeEOF)
       {
           data = "";
           data += char(0x04);  // ^D
           data += '\n';

           gdb_.writeTargetStdin(data);

       }




    }


}
//-----------------------------------------------------------------------------------

void
GdbViewer::
feedStdinHelper()
{

    if(!attachMode_)
    {
        manager()->activate(this);
        
        CursorWindow::viewport *vp = manager()->find_viewport(this);
        
        if(vp)
        {
           feedStdin(vp);
        }
    }

}

void
GdbViewer::
promptAttachPID(CursorWindow::viewport *vp, ViewerManager *manager)
{
    CursorWindow::Dialog d("Manually Specify a process id?  Press Esc To cancel");
    
    d += new CursorWindow::Dialog::String("pid", "Process ID?", "", 20);
    
    if(!d.popup(manager->window()))
    {
       int pid=0;
    
       sscanf(d.element_value("pid").c_str(), "%d", &pid);
    
       if(pid > 0)
       {
         bool fail = run(vp, CursorWindow::key_f5, pid); // gdb viewer run, not gdb controller run

            if(fail)
            {
                CursorWindow::Message m("Error:  attachment failure");

                m += "";
                m += "    Unable to attach to running process";
                m += "";
                m += "Press Enter to continue";
                m += "";

                m.popup(manager->window());
            }
       }
    
    }
    
}

//-----------------------------------------------------------------------------------

static bool okToExit(bool dirty, CursorWindow *w)
{
    if(!dirty)
       return true;

    CursorWindow::Selection  m("Debbuging is still in progress!");

    m += "  Exit Anyway?";
    m += "";
    m += "Press Enter to exit, or Esc to continue debugging";
    m += "";

    return !m.popup(w);
}

//-----------------------------------------------------------------------------------

bool 
GdbViewer::
handle_event( CursorWindow::input_event const * e,
              CursorWindow::viewport          * vp
            ) 
{
    if(e->value_  == 'q' || e->value_ == 'Q' || e->value_ == 0x0b)  // 0x0b is ^K
    {
       if(okToExit(is_dirty(), manager()->window()))
       {
           discardDaughterViewers();
           return true;
       }
       else 
           vp->beep();

       return false;
    }
    
    if(e->type_ == CursorWindow::input_event::ForceExitKey)
    {

       if(okToExit(is_dirty(), manager()->window()))
       {
           vp->set_curpos(0,0);
           vp->fill_to_eos('@');
           vp->refresh();
           
           discardDaughterViewers();
           
           return true;
       }
       else
          vp->beep();

       return false;
    }

    bool rv=false;

    switch(e->value_)
    {
        case '%':
        {
            std::string count = lexical_cast<std::string>(gdb_.readerThreadCount());

            CursorWindow::Message  m("Debugger internal info:");

            m += "";
            m += "Count of Reader Threads:";
            m += std::string("    ") + count;
            m += "";
            m += "Debugger State:";
            m += std::string("    ") + gdb_.stateName();

            m += "";
            m.popup(manager()->window());


        }
        break;


        case CursorWindow::key_f7:
           {
              feedStdin(vp);
           }
           break;

        case CursorWindow::key_f3:  // view previous stop locations
           {
               CursorWindow::Selection d("Previous program stop points");

               history_t::iterator first = history_.begin(), last = history_.end();

               while(last != first)
               {
                   --last;

                   history_t::value_type &cur = *last;

                   std::string value = cur.first;
                               value += ':';
                               value += cur.second;

                   d += value;
               }

               bool aborted = d.popup(manager()->window());

               if(!aborted)
               {
                  size_t finalColon = d.selection_.find_last_of(':');

                  if(finalColon < d.selection_.size())
                  {
                      std::string filename   = d.selection_.substr(0,finalColon);
                      std::string lineNumber = d.selection_.substr(finalColon+1);
                      
                      showCurrentFile(vp, filename, cxxtls::lexical_cast<int>(lineNumber)); // and all other interesting windows
                  }
               }
              
           }
           break;


        case CursorWindow::key_f4:  // set run arguments
           if(!attachMode_)
           {
               CursorWindow::Dialog   d("Set Program Run Arguments");

               if(firstStart_ && runArgs_.empty())
               {
                  // haven't run yet

                  std::string sf = gdb_.saveFile();

                  VarObj::ParseResult rv = VarObj::readAndParseFile(sf);

                  if(rv.rv_)
                  {
                     VarObj *args = rv.rv_->findChild("args");

                     if(args)
                     {
                        runArgs_ = args->value();
                     }
                  }

               }

               d += new CursorWindow::Dialog::String("args", "Arguments", runArgs_, 40);

               if( !d.popup(manager()->window()) )
               {
                   std::string args = d.element_value("args");

                   if(args.empty())
                   {
                      vp->beep();
                   }
                   else
                       runArgs_ = args;
               }
           }
           break;

        case CursorWindow::key_f8:   
               performHalt();
           break;


        case CursorWindow::key_f11:  // step or step into the function under the cursor.
        case CursorWindow::key_f10:  // step without stepping into functions
        case CursorWindow::key_f9:   // step without stepping into functions
        case CursorWindow::key_f5:   // run without stopping on the next line
        case CursorWindow::key_f12:  // step out

           rv= run(vp, e->value_);    // running is compilicated, so let this one function
                                      // handle it all

           break;

        case 'A':

        if(   !fullname_.empty() 
           && (gdb_.state() == GdbController::notStarted)
          )
        {
            char buffer[2200];

            static std::string psCommandBegin("ps -eo pid,start_time,args | grep ");
            static std::string psCommandEnd  (" | expand | grep -v '^ *[0-9]\\+ *[0-9:]\\+ *\\(grep\\|gdb\\|sh -c\\)'");

            std::string progName;                       

            size_t lastSlash = fullname_.find_last_of('/');

            if(lastSlash <  fullname_.size())
            {
               progName = fullname_.substr(lastSlash+1);  
            }
            else
            {
               progName = fullname_;
            }

            std::string psCommand(psCommandBegin + progName + psCommandEnd);


            FILE *psOutput = popen(psCommand.c_str(), "r");

            if(!psOutput)
            {
               // cannot read from ps command output

               CursorWindow::Message m("Error");

               m += "";
               m += "Unable to perform search for proccesses to attach to";
               m += "";

               m.popup(manager()->window());

               pclose(psOutput);

               promptAttachPID(vp, manager()); // ask the user if he wants to manually specify a process id to
                                               // attach to

               break;

            }
            else
            {
                // we were able to open the ps output stream

                CursorWindow::Selection d("Select a running process to debug or Esc to enter a PID manually");

                while(fgets(buffer, sizeof(buffer), psOutput))
                {
                    buffer[sizeof(buffer)-1] = 0; // guarantee null termination

                    size_t l = strlen(buffer);

                    if(l)
                    {
                        if(buffer[l-1] == '\n')
                        {
                           buffer[l-1] = 0;
                           --l;
                        }

                        if(l && buffer[l-1] == '\r')
                        {
                           buffer[l-1] = 0;
                           --l;
                        }



                    }

                    d += buffer;
                    
                }
                pclose(psOutput);


                if(!d.popup(manager()->window()))
                {
                   // user did not abort

                   int pid = -1;

                   sscanf(d.selection_.c_str(), "%d", &pid);

                   if(pid > 0)
                   {
                     bool fail = run(vp, CursorWindow::key_f5, pid);  // gdbviewer run, not gdb controller run

                       if(fail)
                       {
                           CursorWindow::Message m("Error:  Attach failure");

                           m +="";
                           m +="  Unable to Attach to the running process";
                           m +="";
                           m +="Press Enter to continue";

                           m.popup(manager()->window());
                       }
                   }
                   

                }
                else
                {
                    // the user did not or could not select a process to attach to
                    // so let him manually enter one if so desired

                    promptAttachPID(vp, manager());
                }

            }
        }
        else
        if(gdb_.state() != GdbController::notStarted) 
        {
            CursorWindow::Message m("Error:  Cannot Attach to Process");

            m += "";
            m += "  The debugger is already running.  Exit and start over";
            m += "  to attach to another process.";
            m += "";
            m += "Press Enter to continue";

            m.popup(manager()->window());
           
        }
        break;

        case 'C':
            changeDirectory();
        break;


        case '!':
            {
                CursorWindow::Selection s("Save the GDB Debugger state to a file?");

                s += "";

                s += "Press Enter to continue, or Esc to abandon the save operation";

                if(!s.popup(manager()->window()))
                {
                    bool err = gdb_.saveState();
    
                    if(err)
                    {
                       CursorWindow::Message m("Error Saving State");
                       m += "";
                       m += "The name of the saved state output file is:";
                       m += "";
                       m += std::string("    '") + gdb_.saveFile()  + std::string("'");
    
                       m.popup(manager()->window());
                       
                    }
                }
            }
        break;


        case 'F':
        {
            CursorWindow::Dialog   d("Search for functions in this executable");

            d += new CursorWindow::Dialog::String("name", "Fragment of function name", "", 40);

            if(d.popup(manager()->window()))
               break;

            std::vector<std::string> functions;

            if(gdb_.doneErrorScoop(std::string("info function ") + d.element_value("name"), functions))
            {
                CursorWindow::Message m(std::string("Error Searching for Functions matching: ") + d.element_value("name"));

                std::string msg("unknown error");

                if(!functions.empty())
                    msg= functions[0];

                m += "";
                m += std::string("    ") + msg;
                m += "";
                m += "Press Enter to continue";
                m += "";

                functions.clear();

                m.popup(manager()->window());

                break;
            }
          
            {
                CursorWindow::Message s(std::string("Functions in this executable matching: ") + d.element_value("name"));

                s += "";
                s += "Press Enter or Esc to exit!";
                s += "----------------------------";
                s += "";
                s += "";

                static std::string spaces("    ");

                size_t maxWidth = manager()->window()->size().col_ * 3;

                maxWidth /= 4;
                maxWidth -= (4 +  2 * spaces.size());

                std::string filter = d.element_value("name");

                CXXTLS_FOREACH(std::string const &cur, functions)
                {

                  if(cur.find(filter) >= cur.size())
                     continue; // only include functions that at least contain the string itself


                  if(0 == strncmp("File", cur.c_str(), 4))
                  {
                     // IGNORE "File"
                  }
                  else
                  if(!cur.empty() && cur[cur.size()-1] == ':')
                  {
                     if(cur.size() > maxWidth + 2 * spaces.size())
                     {
                          std::vector<std::string> folded;
                  
                          StrTool::fold(cur, maxWidth, folded);
                  
                          int count=0;

                          CXXTLS_FOREACH(std::string const &foldee, folded)
                          {
                             if(count++ == 0)
                             {
                                 s += foldee;
                             }
                             else
                             {
                                 std::string outputData = "@";
                                 
                                 outputData += foldee;
                                 
                                 s += outputData;
                             }
                          }
                     }
                     else
                         s += cur;
                  }
                  else
                  {
                       if(cur.size() < maxWidth )
                          s += spaces + cur;
                       else
                       {
                          std::vector<std::string> folded;
                  
                          StrTool::fold(cur, maxWidth, folded);
                  
                          int count=0;

                          CXXTLS_FOREACH(std::string const &foldee, folded)
                          {
                             if(count++ == 0)
                             {
                                 std::string outputData = spaces;
                                 
                                 outputData += foldee;
                                 
                                 s += outputData;

                             }
                             else
                             {
                                 std::string outputData = spaces + spaces;
                                 
                                 outputData += '@';
                                 
                                 outputData += foldee;
                                 
                                 s += outputData;
                             }
                          }
                       }
                  }
                }

                s.popup(manager()->window());

            }

        }
        break;

        case 'V':
        {
            CursorWindow::Dialog   d("Search for variables in this executable");

            d += new CursorWindow::Dialog::String("name", "Fragment of variable name", "", 40);

            if(d.popup(manager()->window()))
               break;


            std::vector<std::string> exeVariables;

            if(gdb_.doneErrorScoop("info variables", exeVariables))
            {
                CursorWindow::Message m("Error Searching for Variables");

                std::string msg("unknown error");

                if(!exeVariables.empty())
                    msg= exeVariables[0];

                m += "";
                m += std::string("    ") + msg;
                m += "";
                m += "Press Enter to continue";
                m += "";

                exeVariables.clear();

                m.popup(manager()->window());

                break;
            }
          
            {
                CursorWindow::Message s("Global Variables in this executable");

                s += "";
                s += "Press Enter or Esc to exit!";
                s += "----------------------------";
                s += "";
                s += "";

                static std::string spaces("    ");

                size_t maxWidth = manager()->window()->size().col_ * 3;

                maxWidth /= 4;
                maxWidth -= (4 +  2 * spaces.size());

                std::string filter = d.element_value("name");

                CXXTLS_FOREACH(std::string const &cur, exeVariables)
                {
                   if(cur.find(filter) >= cur.size())
                      continue;


                  if(0 == strncmp("File", cur.c_str(), 4))
                  {
                     // IGNORE "File"
                  }
                  else
                  if(!cur.empty() && cur[cur.size()-1] == ':')
                  {
                     if(cur.size() > maxWidth + 2 * spaces.size())
                     {
                          std::vector<std::string> folded;
                  
                          StrTool::fold(cur, maxWidth, folded);
                  
                          int count=0;

                          CXXTLS_FOREACH(std::string const &foldee, folded)
                          {
                             if(count++ == 0)
                             {
                                 s += foldee;
                             }
                             else
                             {
                                 std::string outputData = "@";
                                 
                                 outputData += foldee;
                                 
                                 s += outputData;
                             }
                          }
                     }
                     else
                         s += cur;
                  }
                  else
                  {
                       if(cur.size() < maxWidth )
                          s += spaces + cur;
                       else
                       {
                          std::vector<std::string> folded;
                  
                          StrTool::fold(cur, maxWidth, folded);
                  
                          int count=0;

                          CXXTLS_FOREACH(std::string const &foldee, folded)
                          {
                             if(count++ == 0)
                             {
                                 std::string outputData = spaces;
                                 
                                 outputData += foldee;
                                 
                                 s += outputData;

                             }
                             else
                             {
                                 std::string outputData = spaces + spaces;
                                 
                                 outputData += '@';
                                 
                                 outputData += foldee;
                                 
                                 s += outputData;
                             }
                          }
                       }
                  }
                }

                s.popup(manager()->window());

            }

        }
        break;


        case 'P':
        {
            CursorWindow::Dialog d("Get the definition of data type");

            d += new CursorWindow::Dialog::String("name", "Full type name", "", 40);

            if(!d.popup(manager()->window()))
            {
                std::string name = d.element_value("name");

                if(!name.empty())
                {
                    std::vector<std::string>  rows;

                    gdb_.doneErrorScoop(std::string("ptype ") + name, rows);


                    CursorWindow::Message m(name + " described");

                    CXXTLS_FOREACH(std::string const &cur, rows)
                       m += cur;

                    m.popup(manager()->window());


                }
            }
        }
        break;


        case 'T':
        {
            if(types_.empty() && gdb_.doneErrorScoop("info types", types_))
            {
                CursorWindow::Message m("Error Searching for Types");

                std::string msg("unknown error");

                if(!types_.empty())
                    msg= types_[0];

                m += "";
                m += std::string("    ") + msg;
                m += "";
                m += "Press Enter to continue";
                m += "";

                types_.clear();

                m.popup(manager()->window());

                break;
            }
          
            {
                CursorWindow::Message s("Types in this executable");

                s += "";
                s += "Press Enter or Esc to exit!";
                s += "----------------------------";
                s += "";
                s += "";

                static std::string spaces("    ");

                size_t maxWidth = manager()->window()->size().col_ * 3;

                maxWidth /= 4;
                maxWidth -= (4 +  2 * spaces.size());

                CXXTLS_FOREACH(std::string const &cur, types_)
                {
                  if(0 == strncmp("File", cur.c_str(), 4))
                  {
                     // IGNORE "File"
                  }
                  else
                  if(!cur.empty() && cur[cur.size()-1] == ':')
                  {
                     if(cur.size() > maxWidth + 2 * spaces.size())
                     {
                          std::vector<std::string> folded;
                  
                          StrTool::fold(cur, maxWidth, folded);
                  
                          int count=0;

                          CXXTLS_FOREACH(std::string const &foldee, folded)
                          {
                             if(count++ == 0)
                             {
                                 s += foldee;
                             }
                             else
                             {
                                 std::string outputData = "@";
                                 
                                 outputData += foldee;
                                 
                                 s += outputData;
                             }
                          }
                     }
                     else
                         s += cur;
                  }
                  else
                  {
                       if(cur.size() < maxWidth )
                          s += spaces + cur;
                       else
                       {
                          std::vector<std::string> folded;
                  
                          StrTool::fold(cur, maxWidth, folded);
                  
                          int count=0;

                          CXXTLS_FOREACH(std::string const &foldee, folded)
                          {
                             if(count++ == 0)
                             {
                                 std::string outputData = spaces;
                                 
                                 outputData += foldee;
                                 
                                 s += outputData;

                             }
                             else
                             {
                                 std::string outputData = spaces + spaces;
                                 
                                 outputData += '@';
                                 
                                 outputData += foldee;
                                 
                                 s += outputData;
                             }
                          }
                       }
                  }
                }

                s.popup(manager()->window());

            }

        }
        break;

        case 'v':
        {
              showCurrentFile(vp, showFilename_, showLine_); // and all other interesting windows

              Viewer* v = manager()->find_viewer(showFilename_);

              if(v)
              {
                  manager()->activate(v);
              }
              return false;
        }
          break;

        case 'b':
          if(breakpointViewer_)
          {
             manager_->activate(breakpointViewer_);

          }
          break;

        case 'o':
          if(!attachMode_ && logViewer_)
          {
             manager_->activate(logViewer_);

          }
          break;

        case 's':
          if(stackViewer_)
          {
             manager_->activate(stackViewer_);

          }
          break;

        case 'd':
          if(variableViewer_)
          {
             manager_->activate(variableViewer_);

          }
          break;

       case  'r':
          {
             manager()->activate(this);
             CursorWindow::viewport *vp = manager()->find_viewport(this);

             if(vp)
             {
                showCurrentFile(vp, stopFilename_, lexical_cast<int>(stopLine_));
             }
          }
          break;


        case 'a':  
        {
            // add a breakpoint

            CursorWindow::Dialog d("Add a breakpoint");

            d += new CursorWindow::Dialog::String("func", "Function Name", "", 40);

            bool rc = d.popup(manager_->window());

            if(!rc)
            {
                std::string funcName = d.element_value("func");

                if(!funcName.empty())
                {
                   std::string error = gdb_.addBreakpoint(funcName);

                   if(error.empty())
                   {
                       bpItem_ = gdb_.getMostRecentBreakpoint();

                       displayBreakpoints(bpItem_);

                       repaintTitle(vp);
                   }
                   else
                   {
                       CursorWindow::Message m("Error setting breakpoint");

                       m += "";
                       m += "Attempting to set breakpoin on:";
                       m += "";
                       m += std::string("   ") + funcName;
                       m += "";
                       m += "Resulted in the following error:";
                       m += "";
                       m += std::string("   ") + error;

                       m.popup(manager()->window());
                   }
                   
                }
            }

        }
        break;

        case 'e':
        case 'E':
        {
            if(gdb_.state() == GdbController::stopped)
            {
                // we have to at least hit the first breakpoint before we can do this
                // or calls to send commands to the debugger will SEGV

                gdb_.exceptionBreakpoints(e->value_ == 'E', true);
                displayBreakpoints(bpItem_);
            }
            else
            {
                manager()->window()->beep();
            }

        }
        break;

        case 'h':
            if(gdb_.state() == GdbController::stopped)
            {
                bool error = gdb_.requestSignalHandlerList(); // waits 1 second
                
                if(error)
                {
                   CursorWindow::Message m("Error:  can not retrieve signal handlers");
                   m += "";
                   m += "Press Enter to Continue";
                   m += "";

                   m.popup(manager()->window());

                }
                else
                {
                   GdbController::SignalHandlers handlers;

                   gdb_.getSignalHandlers(handlers);

                   CursorWindow::Selection m("Signal Handlers            (Enter to edit, Esc to quit)");

                   char buffer[1000];

                   #define SIGHANDLERFORMAT "%-12s %-5s %-5s %-5s %s"

                   snprintf(buffer, sizeof(buffer), SIGHANDLERFORMAT,
                            "SIGNAL NAME",
                            "STOP",
                            "PRINT",
                            "PASS",
                            "DESCRIPTION"
                           );

                   m += buffer;

                   CXXTLS_FOREACH(GdbController::SignalHandlers::value_type const& cur, handlers)
                   {

                       snprintf(buffer, sizeof(buffer), SIGHANDLERFORMAT,
                                cur.first.c_str(),
                                (cur.second.stop_  ? "Y" : "N"),
                                (cur.second.print_ ? "Y" : "N"),
                                (cur.second.pass_  ? "Y" : "N"),
                                (cur.second.description_.c_str())
                               );
                                
                       m += buffer;

                   }

                   if(!m.popup(manager()->window()))
                   {
                      // user pressed enter on a handler line

                      std::string selection = m.selection_;

                      char name[256];
                      char stop[256];
                      char print[256];
                      char pass[256];

                      name[0] = stop[0] = pass[0] = print[0] = 0;

                      sscanf(selection.c_str(), "%s %s %s %s", name, stop, print, pass);

                      std::string title  = "Modify Handling for Signal '";
                                  title += name;
                                  title += "'";

                      static std::string SIGNAL("SIGNAL");

                      if(SIGNAL == name)
                      {
                         break;  // row 1 is a header...
                      }


                      CursorWindow::Dialog d(title);

                      d += new CursorWindow::Dialog::String("stop", "Stop?", stop, 3);
                      d += new CursorWindow::Dialog::String("print", "Print?", print, 3);
                      d += new CursorWindow::Dialog::String("pass", "Pass?", pass, 3);

                      if(!d.popup(manager()->window()))
                      {

                         if(gdb_.handleSignal(name, d.element_value("stop"), d.element_value("print"), d.element_value("pass")))
                         {
                            CursorWindow::Message m("Error or timeout setting signal handler");
                            m += "";
                            m += name;
                            m += "";
                            m += "Press Enter to continue";

                            m.popup(manager()->window());
                         }
                      }
                   }
                }
            }
            break;

        case 0:
            // this is a state change event

            rv = handleStateChange(vp);
            break;

        default:
            break;
    }

   if(rv)
   {
       discardDaughterViewers();
   }

   {
       vp = manager()->find_viewport(this);

       if(vp)
           hide_cursor(vp);
   }

   return rv;

}
//-----------------------------------------------------------------------------------

void
GdbViewer::
displayTargetOutput()
{
    GdbController::TargetOutput &childOutputQueue = gdb_.targetOutput();

    std::deque<char> childOutput;

    childOutputQueue.trySwapAll(childOutput);  // swap the childOutput container (which is empty) with
                                               // the current output contents.

    // split the output into lines and append the last 5
    // lines into the screen output buffer


    if(!childOutput.empty())
    {
        std::vector<std::string> lines;

        lines.push_back("");

        CXXTLS_FOREACH(char c, childOutput)
        {
            if(c == '\r')
            {
                // ignore return chars
            }
            else
            if(c == '\n')
            {
                lines.push_back("");
            }
            else
            if(c == '\t')
            {
                std::string &line = lines.back();

                int inc = 8 - (line.size() % 8);

                while(inc--)
                   line += ' ';

            }
            else
            {
                lines.back() += c;
            }
        }

        if(gdbLogViewer_)
        {
            gdbLogViewer_->appendContent(lines, logViewer_);
        }

    }
}

//-----------------------------------------------------------------------------------
void 
GdbViewer::
displayBreakpoints(std::string const &item)
{
    // display the breakpoint list and make sure item is visible

    GdbController::IdList breakpointNames;
    
    size_t itemColonLoc;

    bool itemIsFileLineCombo = (itemColonLoc = item.find_last_of(':') ) < item.size();

    std::string stopFileName;
    std::string stopLineNo;

    if(itemIsFileLineCombo)
    {
       stopLineNo   = item.substr(itemColonLoc+1);
       stopFileName = item.substr(0, itemColonLoc);
    }

    gdb_.breakpoints(breakpointNames);
    
    std::vector<std::string> breakpoints;

    size_t itemIndex=0;  // the index to the "item" to be guaranteed to be displayed
    
    CXXTLS_FOREACH(std::string const &cur, breakpointNames)
    {
        std::string encodedForm = gdb_.breakpoint(cur);
    
        VarObj::ParseResult bpInfo(VarObj::parse(encodedForm));
    
        if(bpInfo.rv_)
        {
            VarObj *fileInfo    = bpInfo.rv_->findNestedChild("file");
            VarObj *lineInfo    = bpInfo.rv_->findNestedChild("line");
            VarObj *enabledInfo = bpInfo.rv_->findNestedChild("enabled");
            VarObj *fullInfo    = bpInfo.rv_->findNestedChild("fullname");
            VarObj *funcInfo    = bpInfo.rv_->findNestedChild("func");
            VarObj *originInfo  = bpInfo.rv_->findNestedChild("original-location"); 
            VarObj *whatInfo    = bpInfo.rv_->findNestedChild("what");
            VarObj *typeInfo    = bpInfo.rv_->findNestedChild("type");
            VarObj *condInfo    = bpInfo.rv_->findNestedChild("cond");
            VarObj *addrInfo    = bpInfo.rv_->findNestedChild("addr");

            if(addrInfo == 0 || addrInfo->value() == "")
               whatInfo = typeInfo; // may still be 0

            if(fileInfo == 0)
              fileInfo = whatInfo;

            if(fullInfo == 0)
               fullInfo = fileInfo;

            if(funcInfo == 0)
              funcInfo = originInfo;

            // make sure this breakpoint exists in the activeBreakpoints_ list

            ActiveBreakpoints::iterator activePtr = activeBreakpoints_.find(cur);

            if(activePtr == activeBreakpoints_.end())
            {

                std::string filename    (  fullInfo ? fullInfo->value() : "" );
                std::string linenumber  (  lineInfo ? lineInfo->value() : "" );

                activeBreakpoints_.insert(ActiveBreakpoints::value_type(cur, std::pair<std::string,std::string>(filename, linenumber)) );
            }
    
            std::string tmp;
    
            std::string question("?");
    
            if(fileInfo && lineInfo)
            {
               if( activeBreakpoints_.end() == activeBreakpoints_.find(cur) )
               {
                  // need to add this breakpoint to the deletion list

                  activeBreakpoints_.insert( ActiveBreakpoints::value_type(cur, std::pair<std::string,std::string>(fullInfo->value(), lineInfo->value()) ) );

               }
            }

    
            tmp += fmtd<-3>  (cur);                                           tmp += ' ';
            tmp += fmtd<-2>  (enabledInfo ? enabledInfo->value() : question); tmp += ' ';

            std::string nameTmp;

            nameTmp += fmtd<-13> (fileInfo    ? fileInfo->value()    : question); tmp += ' ';

            if(nameTmp.empty())
            {
                nameTmp = "             ";
            }
            else
            if(nameTmp[0] == '@')
            {
                nameTmp = fileInfo    ? fileInfo->value()    : question;

                if(nameTmp.size() > 13)
                   nameTmp = nameTmp.substr(0,13);
            }

            tmp += nameTmp;
            tmp += ' ';
            tmp += (funcInfo ? funcInfo->value() : question);

            if(!itemIsFileLineCombo && cur == item)
            {
               // this breakpoint number is the one we want to guarantee gets displayed
               //
               // if itemIsFileLineCombo true, then the itemIndex is set above
               //
               //  

               itemIndex = breakpoints.size();

            }
            else
              {
                std::string filename    (  fullInfo ? fullInfo->value() : "" );
                std::string linenumber  (  lineInfo ? lineInfo->value() : "" );

                if(itemIsFileLineCombo && fullInfo && lineInfo)
                {
                  if(   fullInfo->value()     == stopFileName 
                        && lineInfo->value()  == stopLineNo
                     )
                   {
                      itemIndex = breakpoints.size();  // mark this as the "current" breakpoint
                   }
                }

              }


            breakpoints.push_back(tmp);
            tmp = "  ";

            tmp += (fullInfo    ? fullInfo->value()    : question);

            if(lineInfo)
            {
               tmp += ':';
               tmp += lineInfo->value();
            }
    
            breakpoints.push_back(tmp);

            if(condInfo)
            {
                tmp = "    (cond=";

                tmp += condInfo->value();

                tmp += ")";
                breakpoints.push_back(tmp);
            }

        }
        else
           breakpoints.push_back(cur);
    
        delete bpInfo.rv_;
    }
    
    gdbBreakpointViewer_->resetContent(breakpoints, breakpointViewer_, itemIndex +2);  // two lines per item

    
}
//-----------------------------------------------------------------------------------

bool
GdbViewer::
handleStateChange(CursorWindow::viewport *&vp)
{
    repaintTitle(vp);

    GdbController::ChildState currentState = gdb_.state();

    if(currentState == GdbController::stopped &&
       savedState_ !=  currentState
      )
    {
       if(gdbLogViewer_)
       {
           std::vector<std::string> text;

           text.push_back("-------------------------------------------------------------------");
           text.push_back("Ending execution");
           text.push_back("");

           gdbLogViewer_->appendContent(text, logViewer_);
       }
    }

    displayTargetOutput();

    switch(currentState)
    {
        case GdbController::stopped:
        {

            std::string stopInfoText = gdb_.stopInfo();

            std::auto_ptr<VarObj>  stopInfo;

            VarObj::ParseResult  result(VarObj::parse(stopInfoText));

            if( result.rv_ )
            {
               stopInfo.reset(result.rv_);
            }
          
            if(1 
               && (!stopInfo.get() || 0 == stopInfo.get()->findNestedChild("frame/file"))
              )
            {
                CursorWindow::Message m("Unable to get stop point from gdb");

                m += "";
                m += "Perhaps the program or module is not compiled for debug.";
                m += "";

                if(stopInfo.get())
                {
                    VarObj *p = stopInfo.get()->findNestedChild("frame/func");  

                    if(p)
                    {
                       std::string tmp("  Stopped in ");

                       tmp += p->value();

                       m += tmp;
                    }
                    else
                        m += "  <Unable to determine stop function name>";

                    m += "";
                }


                m += "Press Enter to continue";

                m.popup(manager_->window());

                repaintTitle(vp);

                if(stopInfo.get() == 0)
                   break;
               
            }


#if 0
            messages_.push_back("Stop Point: ");
            messages_.push_back(std::string("  ") + stopInfo->value());
#endif
          
            VarObj const *func = stopInfo->findNestedChild("frame/func");
          
            if(!func)
            {
                CursorWindow::Message m("Unable to get current function from gdb");

                m += "";
                m += "Press Enter to continue";

                m.popup(manager_->window());

                repaintTitle(vp);
                break;
                
            }

            messages_.push_back("Current Stop Function");
            messages_.push_back(std::string("  ") + func->value());

            VarObj const *fullname     = stopInfo->findNestedChild("frame/fullname");
            VarObj const *shortName    = stopInfo->findNestedChild("frame/file");
            VarObj const *stopLine     = stopInfo->findNestedChild("frame/line");
            VarObj const *reason       = stopInfo->findNestedChild("reason");
            VarObj const *signalName   = stopInfo->findNestedChild("signal-name");
            VarObj const *signalMeaning= stopInfo->findNestedChild("signal-meaning");

            if(fullname && stopLine)
            {
               history_.push_back( std::pair<std::string,std::string>(fullname->value(), stopLine->value() ) );
            }


            std::string sigName;
            std::string sigMeaning;
            std::string reasonName;

            if(reason)
               reasonName = reason->value();

            if(!reasonName.empty())
               messages_.push_back( std::string("Stop Reason:  ") + reasonName);

            if(signalName)    sigName     = signalName->value();
            if(signalMeaning) sigMeaning  = signalMeaning->value();

            if(!sigName.empty())
            {
               messages_.push_back(std::string("Signalled!  Type = ") + sigName);

               if(!sigMeaning.empty())               
                  messages_.push_back(std::string("Signal Meaning   = ") + sigMeaning);

            }

            if(!fullname && reason && !signalMeaning)
            {
                // don't come in here if there is a signal meaning available

                // when we attach to a running process, it is entirely possible that the child
                // is stopped deep inside some operating system call -- such as when it is 
                // waiting in an I/O operation.  Under these circumstances, we should not attempt
                // to step the debugger out of routine which has no debug information because
                // it will cause the debugger to hang.  
                //
                // So don't come in here if there is no reason code...


                // fullname = stopInfo->findNestedChild("frame/file"); // most likely, file not found

                // if we are at a point where there is no source code, let us
                // stop out to the caller function.  However, if we are in the outer stack
                // frame of the program we can't do this.
                // 

                std::string stackInfo = gdb_.stack();
                
                VarObj::ParseResult stackRes = VarObj::parse(stackInfo);
                
                std::auto_ptr<VarObj> stack;
                
                if(stackRes.rv_)
                {
                    stack.reset(stackRes.rv_);
                }

                bool can_step_out=false;
                
                if(gdb_.hasSymbols() && stack.get())
                {
                    ArrayVarObj const*  frames    = dynamic_cast<ArrayVarObj const*>(stack.get());
                    
                    if(frames)
                    {
                        size_t maxFrame = frames->size();
                    
                        if(maxFrame > 0)
                        {
                            // we are not sitting in the top stack frame
                            can_step_out = true;
                        }
                    }
                }
                else
                if(!gdb_.hasSymbols())
                {
                   showCurrentFile(vp, ".", 1);
                }

                if(can_step_out)
                {
                    
                    savedState_ = GdbController::running;
                    
                    gdb_.stepOut();  // can't show source, don't stop on the breakpoint, just keep
                                     // going till we get useful source

                    messages_.clear();

                    messages_.push_back("");
                    messages_.push_back("Automatically continuing");
                    messages_.push_back("due to lack of source availability");

                    repaintTitle(vp);
                    
                }
#if 0
                messages_.push_back("Stop Point: ");
                messages_.push_back(std::string("  ") + stopInfo->value());
#endif

                return false; 

            }

            if(fullname)
            {
                messages_.push_back("Full Filename:");
                messages_.push_back(std::string("  ") + fullname->value());

            }

            VarObj const *line = stopInfo->findNestedChild("frame/line");

            if(fullname && line)
            {
                
                std::string stopMsg = "Stopped [";
                            stopMsg += cxxtls::lexical_cast<std::string>(history_.size());
                            stopMsg += "] ";
                            stopMsg += shortName ? shortName->value() : std::string("");
                            stopMsg += ':';
                            stopMsg += stopLine ? stopLine->value() : std::string("1");

                messages_.push_back(stopMsg);
            }

            if(fullname && line)
            {
                std::string        name = fullname->value();
                int                lno  = lexical_cast<int>(line->value());

                showCurrentFile(vp, name, lno);

                vp = manager_->find_viewport(this);  // find the viewport holding this viewer

                if(!vp)
                {
                    CursorWindow::Message  m("Doh:  after split screen, cannot find original viewer");
                    m += "";
                    m += "Press enter to exit";
                    

                    m.popup(manager_->window());

                    return true;
                }

            }
            else
            if(reason == 0)
            {
               // we have stopped for reasons not associated with the debugger

               std::string name = ".";
               int lno  = 1;

               if(fullname)
                  name = fullname->value();

               if(line)
                  lno = lexical_cast<int>(line->value());


                showCurrentFile(vp, name, lno);

                vp = manager_->find_viewport(this);  // find the viewport holding this viewer

                if(!vp)
                {
                    CursorWindow::Message  m("Doh:  after split screen, cannot find original viewer");
                    m += "";
                    m += "Press enter to exit";
                    

                    m.popup(manager_->window());

                    return true;
                }


            }
            else
                repaintTitle(vp);  // repaint screen


            std::string stackInfo = gdb_.stack();

            VarObj::ParseResult stackRes = VarObj::parse(stackInfo);

            std::auto_ptr<VarObj> stack;

            if(stackRes.rv_)
            {
                stack.reset(stackRes.rv_);
            }

            std::vector< std::string > stackViewerContents;

            std::string stackTopLocation;

            if(stack.get())
            {
                ArrayVarObj const*  frames    = dynamic_cast<ArrayVarObj const*>(stack.get());

                if(frames)
                {
                    size_t maxFrame = frames->size();

                    stopFrames_ = maxFrame;  // save the count of stack frames for use in determining
                                             // how to implement F12 (step out).


                    for(size_t i = 0; i < maxFrame; ++i)
                    {
                        VarObj const *fullname = (*frames)[i]->findNestedChild("frame/fullname");
                        VarObj const *line     = (*frames)[i]->findNestedChild("frame/line");
                        VarObj const *func     = (*frames)[i]->findNestedChild("frame/func");
                        VarObj const *level    = (*frames)[i]->findNestedChild("frame/level");
                        VarObj const *args     = (*frames)[i]->findNestedChild("frame/args");

                        std::string tmp;

                        if(level)
                        {
                           tmp += level->value();
                           tmp += " ";
                        }
                        else
                           tmp += "? ";

                        if(func)
                        {
                            tmp += func->value();
                            tmp += " ";
                        }
                        else
                           tmp += "??????? ";

                        if(args)
                        {
                           tmp += args->value();
                           tmp += " ";
                        }
                        else
                           tmp += "(...) ";

                        stackViewerContents.push_back(tmp);
                        tmp = "  ";


                        if(!fullname)
                        {
                           fullname = (*frames)[i]->findNestedChild("frame/file");
                        }

                        if(fullname)
                        {
                            tmp += fullname->value();
                           
                        }
                        else
                           tmp += "???????.cpp";

                        tmp += ':'; // colon separates filename from line number


                        if(line)
                        {
                            tmp += line->value();
                        }
                        else
                           tmp += "???";

                        if(i == 0)
                        {
                           // set a variable containing the file/line of the top of the stack
                           // for use in deciding which breakpoint we wish to display.

                           if(fullname && line)
                               stackTopLocation = fullname->value() + std::string(":") + line->value();

                        }

                        stackViewerContents.push_back(tmp);
                    }

                }

               
            }

            if(gdbStackViewer_)
            {
                gdbStackViewer_->resetContent(stackViewerContents, stackViewer_);
            }


            if(gdbBreakpointViewer_)
            {
                if(!stackTopLocation.empty())
                    displayBreakpoints(stackTopLocation);
                else
                    displayBreakpoints(bpItem_);
            }

            if(gdbVariableViewer_)
            {
                displayVariables();
            }

            repaintTitle(vp);
                
        }
        break;


        case GdbController::terminated:
        {
            removeStopPointMarker();

            history_.push_back(std::pair<std::string, std::string>("PROGRAM TERMINATED","0"));

            std::vector<std::string> emptyStack;

            gdbStackViewer_->resetContent(emptyStack,stackViewer_);
            repaintTitle(vp);

        }
        break;

        default:
            break;
    }

    return false;
}


//-----------------------------------------------------------------------------------

bool
GdbViewer::
run(CursorWindow::viewport *vp, int keycode, int pid )
{
    if(keycode == CursorWindow::key_f9)
       keycode =  CursorWindow::key_f10;

    if(gdb_.state() == GdbController::terminated)
    {
        if(   keycode == CursorWindow::key_f9
           || keycode == CursorWindow::key_f10
           || keycode == CursorWindow::key_f11
           || keycode == CursorWindow::key_f12
          )
        {
            keycode = keycode == CursorWindow::key_f5;
        }
    }

    int savedStopFrames = stopFrames_;

    stopFrames_ = 0;

    if(!runnable_ || isRunning())
    {
       vp->beep();
       return false;
    }

    messages_.clear();

    switch(gdb_.state())
    {
        case  GdbController::runnable:
        case  GdbController::notStarted:
        case  GdbController::terminated:
            if(gdbLogViewer_ && pid == 0)
            {
                std::vector<std::string> text;

                text.push_back("");
                text.push_back("Starting -------------------------------------------------------------------------");
                text.push_back("");

                gdbLogViewer_->appendContent(text, logViewer_);
            }
            else
            {
                std::vector<std::string> text;

                text.push_back("");
                text.push_back(std::string("Attached to process ") + lexical_cast<std::string>(pid));

                gdbLogViewer_->appendContent(text, logViewer_);
            }


        default:
            break;
    }
  
    switch(keycode)
    {
        case CursorWindow::key_f11:
        case CursorWindow::key_f9:
        case CursorWindow::key_f10:
        case CursorWindow::key_f12:

            if(!isStopped())
            {
               // must be waiting to run

               CursorWindow::Dialog  m("Start / Restart the run?");

               m += new CursorWindow::Dialog::String("run",
                                                     "Press Enter to run the executable or Esc to abort",
                                                     " ",
                                                     2
                                                    );

               if(m.popup(manager_->window()))
               {
                  vp->beep();
                  return false;
               }

            }

            if(isStopped())
            {
               savedState_ = GdbController::runnable; // to trigger a state change

               if(keycode == CursorWindow::key_f12)
               {
                   if(savedStopFrames > 1)
                       gdb_.stepOut();  // returns immediately
                   else
                       gdb_.resume();   // ditto
               }
               else
               if(keycode == CursorWindow::key_f11)
                   gdb_.stepInto();  // returns immediately
               else
                   gdb_.stepOver();  // returns immediately


               return false;
            }

            // drop through

        default:

            if(!argsSent_)
            {
                gdb_.setExecutable(fullname_);
                gdb_.addBreakpoint("main");
                argsSent_ = true;

                bpItem_ = gdb_.getMostRecentBreakpoint();

            }

            if(pid == 0)
            {
                gdb_.setArgs(runArgs_);  // apparently in interpreter mode, each time you run
                                         // you have to reset the run arguments.
                                         //
                                         // I did not realize this when I picked the name "argsSent_"
            }

            savedState_ = GdbController::runnable;

            if(gdb_.state() == GdbController::stopped)
            {
              gdb_.resume();
            }
            else
            {
              bool success = gdb_.run(pid); // returns immediately

              if(pid && !success)
                return true;  // 

              if(pid)
              {
                   // if we get here, successfully attached to a running process

                   attachMode_ = true;
              }



            }
            
            break;

    }


    return false;
}


//-----------------------------------------------------------------------------------

std::string const &
GdbViewer::
application_name() const 
{
    return fullname_;
}

//-----------------------------------------------------------------------------------

std::string 
GdbViewer::
description() const
{
   return "Gdb Debugger";
}

//-----------------------------------------------------------------------------------

void 
GdbViewer::
help()
{
  std::list<std::string> help_text;

  help_text.push_back("");
  help_text.push_back("PROGRAM DEBUGGER HELP");
  help_text.push_back("  The viewer you are looking at is a mode controller that takes over the layout");
  help_text.push_back("  of the entire XTERM screen.  It allows you to debug programs which have been");
  help_text.push_back("  compiled for debug.");
  help_text.push_back("");
  help_text.push_back("  To compile a program for debug, use the linux options:");
  help_text.push_back("");
  help_text.push_back("     -g3 -fno-inline -O0");
  help_text.push_back("");
  help_text.push_back("  To debug a program:");
  help_text.push_back("");
  help_text.push_back("    1.  Navigate to the directory where the executable lies");
  help_text.push_back("    2.  Invoke the debugger using either of the following methods:");
  help_text.push_back("          Using the ^X ^N Menu");
  help_text.push_back("              *  Press ^X ^N to bring up the alternate viewer list");
  help_text.push_back("              *  Select 'gdb_debugger'");
  help_text.push_back("        Or:");
  help_text.push_back("          Editing the .gdbMode file for the executable");
  help_text.push_back("              *  the .gdbMode file has the same name as the exectuable");
  help_text.push_back("                 but with '.gdbMode' appended.");
  help_text.push_back("              *  the act of editing the .gdbMode file will be to bring");
  help_text.push_back("                 up the debugger window on the corresponding executable.");
  help_text.push_back("              *  When using the .gdbMode file, exiting the debugger will");
  help_text.push_back("                 save your breakpoints that were in place on exit.");
  help_text.push_back("    4.  After the debugger window appears, press either 'A' to attach");
  help_text.push_back("        to an already running process or press F5 to start a new process");
  help_text.push_back("");
  help_text.push_back("  .gdbMode files");
  help_text.push_back("");
  help_text.push_back("    If you wish to use the .gdbMode method of starting the debugger, described above,");
  help_text.push_back("    all you need to do is this:");
  help_text.push_back("      *  first, bring up the debugger on the executable using the ^X^N method");
  help_text.push_back("      *  start the executable running.");
  help_text.push_back("      *  when it stops, use the '!' command to save the debugger state.");
  help_text.push_back("    From that point on, the .gdbMode file will exist for your executable and the");
  help_text.push_back("    debugger will automatically read/write it.");  
  help_text.push_back("");
  help_text.push_back("    To stop using the mode, delete the .gdbMode file");
  help_text.push_back("");
  help_text.push_back("  The debugger divides the screen into the following viewers:");
  help_text.push_back("");
  help_text.push_back("    GDB Master               -- you are viewing the help page for it now");
  help_text.push_back("    Source Viewer            -- A text edit viewer showing the source for the program.");
  help_text.push_back("                                Note that you can toggle breakpoints in the source");
  help_text.push_back("                                viewer using the '^X space' key sequence.");
  help_text.push_back("    Stack Viewer             -- lets you see the program call stack");
  help_text.push_back("                                and print data in stack frames.  See it's help page.");
  help_text.push_back("    Breakpoint Viewer        -- lets you see the breakpoints and issue related");
  help_text.push_back("                                commands.  See its help page");
  help_text.push_back("");
  help_text.push_back("    Display Variables Viewer -- Displayed expressions which are automatically updated");
  help_text.push_back("                                whenever the debugger stops in stack frames of interest.");
  help_text.push_back("");
  help_text.push_back("    Log Viewer         -- Lets you scroll through output from a debugged program.");
  help_text.push_back("");
  help_text.push_back("Command keys for the GDB Master Viewer:");
  help_text.push_back("  F1        -- Get help");
  help_text.push_back("  F2        -- switch between viewers in cyclic fashion.");
  help_text.push_back("  F3        -- Review program stop history.  Select file locations for viewing.");
  help_text.push_back("  F4        -- Set the program run arguments.  Only effective when program terminated.");
  help_text.push_back("  F5        -- Run or continue running the program until termination or breakpoint.");
  help_text.push_back("  F7        -- Enter text into the running target program's stdin stream");
  help_text.push_back("  F8        -- Halt the running process or kill a stopped process completely.");
  help_text.push_back("  F9        -- Run one line of code without stepping into function calls.");
  help_text.push_back("  F10       -- Run one line of code without stepping into function calls.");
  help_text.push_back("  F11       -- Run one line of code but do step into functions.");
  help_text.push_back("  F12       -- Run until the current function completes.  Stop in its caller.");
  help_text.push_back("  A         -- Attach to a running process");
  help_text.push_back("  C         -- Change target process's current directory -- use before start");
  help_text.push_back("  a         -- Add a breakpoint on a function name.  Note that after adding,");
  help_text.push_back("               the gdbBreakpointViewer can be used to add condition expressions.");
  help_text.push_back("  b         -- Switch directly to the Breakpoint Viewer.");
  help_text.push_back("  d         -- Switch directly to the Display Variables Viewer.");
  help_text.push_back("  e         -- Catch exception throws (only)");
  help_text.push_back("  E         -- Catch exception throws and catches");
  help_text.push_back("  F         -- list all the functions in the program (takes 10 seconds!)");
  help_text.push_back("  h         -- Display/modify Signal Handling");
  help_text.push_back("  ^K        -- kill the debugged target program and exit the debugger.");
  help_text.push_back("  o         -- Switch directly to the GdbOutputViewer.");
  help_text.push_back("  P         -- Display the definition of a named data type");
  help_text.push_back("  q         -- kill the debugged target program and exit the debugger.");
  help_text.push_back("  Q         -- kill the debugged target program and exit the debugger.");
  help_text.push_back("  r         -- Restore the window layout to default size / position.");
  help_text.push_back("  s         -- Switch directly to the Stack Viewer.");
  help_text.push_back("  T         -- Get a list of the types in this executable");
  help_text.push_back("  v         -- Switch directly to the Source Viewer.");
  help_text.push_back("  V         -- Get a list of the global variables in this executable");


  manager()->help_helper(help_text);
}

//-----------------------------------------------------------------------------------
bool 
GdbViewer::
isRunning()
{
    GdbController::ChildState state = gdb_.state();

    return state == GdbController::running;


}

//-----------------------------------------------------------------------------------
bool 
GdbViewer::
isStopped()
{
    GdbController::ChildState state = gdb_.state();

    return state == GdbController::stopped;
}

//-----------------------------------------------------------------------------------
bool 
GdbViewer::
isTerminated()
{
    GdbController::ChildState state = gdb_.state();

    switch(state)
    {
        case GdbController::terminated:
        case GdbController::unstartable:
        case GdbController::gdbExiting:
            return true;

        default:
            break;
    }

    return false;
}


//-----------------------------------------------------------------------------------
bool 
GdbViewer::
isRunnable()
{
    GdbController::ChildState state = gdb_.state();

    switch(state)
    {
       case GdbController::runnable:
       case GdbController::notStarted:
          return true;

       default:
           break;
    }

    return false; // could be already running, waiting at a breakpoint, stopped by a signal, or
                  // the executable could be bad.

}





//-----------------------------------------------------------------------------------

Viewer* 
GdbViewer::
app(ViewerManager*vm, std::string &fullname)
{
   // create an application window.  
   //
   // If fullname refers to a .gdbMode file, remove the .gdbMode part
   // and use the rest of the fullname as the name of an executable.
   //
   // Also, read the .gdbMode file (ie the full fullname) as a VarObj to find the
   // exeFile child tag.  If you find one, replace the original fullname with the 
   // exefile part.


   std::string originalName;  // only set if the file name extension of fullnaem is .gdbMode

   size_t trailingDot = fullname.find_last_of('.');

   std::string runArgs;   // read from .gdbMode file, if any

   if(trailingDot < fullname.size())
   {
      std::string ext = fullname.substr(trailingDot);

      if(ext == ".gdbMode")
      {
         originalName = fullname;
         fullname = fullname.substr(0, trailingDot);

         VarObj::ParseResult config = VarObj::readAndParseFile(originalName);

         if(config.rv_)
         {
            VarObj* exeFile = config.rv_->findChild("exeFile");

            if(exeFile)
            {
               fullname = exeFile->value();

            }

            VarObj *dir = config.rv_->findChild("dir");

            FileName dirname;

            if(dir)
            {
               dirname = dir->value();

            }

            VarObj *args = config.rv_->findChild("args");

            if(args)
            {
               runArgs = args->value();
            }


            bool badDir = !dirname.is_dir();

            dirname = fullname;

            bool badExe = !dirname.is_executable();

            if(badExe || badDir )
            {
                CursorWindow::Message m("Error:  .gdbMode file is invalid");
                m += "";
                m += "The debugger will not run properly with this .gdbMode file.";
                m += "Either the exeFile or dir setting is invalid.";

                m.popup(vm->window());

            }


         }


      }
   }


   GdbViewer *rv =new GdbViewer(vm, fullname);  

   if(!originalName.empty())
       rv->setSaveFile(originalName);

   if(!runArgs.empty())
      rv->setRunArgs(runArgs);

   fullname = rv->application_name();  // the constructor asks the user for a name
                                       // which is now available as "appliation_name()"
                                       // and we are modifying the input fullname,
                                       // which is null in some circumstances with the
                                       // correct data so that the viewer manager
                                       // will correctly associate the name the user
                                       // entered with the viewer object in the viewer table.



   return rv;
}

//-----------------------------------------------------------------------------------
void
GdbViewer::
showCurrentFile(CursorWindow::viewport *&vp, std::string const &fullname, int lineNumber)
{
  // Warning:  This function changes the value of the viewport -- it's vp parameter is
  // a reference!

  // first, make sure all breakpoints are visible

   CXXTLS_FOREACH(ActiveBreakpoints::value_type &cur, activeBreakpoints_)
   {
       std::string const &file = cur.second.first;
       std::string const &line = cur.second.second;

       std::string msg = bpMessage(file,line, false);  // create a message to show the bp annotation

       manager()->broadcast(breakpoint, msg);


   }


  FileName name = manager_->find_file_in_standard_locations(fullname, ".");

  if(name.empty())
  {
    manager_->window()->beep();

    return ;

  }

  bool previouslyShowingAFile = !showFilename_.empty();

  showFilename_ = fullname;
  showLine_     = lineNumber;


  // prepare to set the sizes of daughter windows
  int currentGdbViewerHeight   = vp->size().row_;   // used to preserve user selection of size of windows
  int currentGdbViewerWidth    = vp->size().col_;   // used to preserve user selection of size of windows
  int currentStackHeight       = 0;
  int currentStackWidth        = 0;    currentStackWidth = currentStackWidth;  //shut up the darn compiler
  int currentBreakpointsHeight = 0;
  int currentBreakpointsWidth  = 0;    currentBreakpointsWidth = currentBreakpointsWidth; // shut up compiler!
  int currentVariablesHeight   = 0;
  int currentVariablesWidth    = 0;

  { // try to preserve current window sizes, if possible


      if(firstStart_)
      {
         // first time through, though the current height is full screen -- so we can't call that
         // a user selection
      
         if(currentGdbViewerHeight > 11)
            currentGdbViewerHeight = 11;
      
         if(currentGdbViewerWidth > 40)
            currentGdbViewerWidth = 40;

         firstStart_ = false;
      }
      else
      if(previouslyShowingAFile)
      {
          // after the first time through, don't let the fact that the user did a ^X 1 to
          // unify the screen totally mess up the 

          CursorWindow::row_col xtermSize = manager()->window()->size();

          if(currentGdbViewerWidth > xtermSize.col_ / 2)
          {
              currentGdbViewerWidth = xtermSize.col_ / 2;
          }
      }
      
      CursorWindow::viewport *tmp = manager()->find_viewport(stackViewer_);
      
      if(tmp)
      {
         currentStackHeight = tmp->size().row_;
         currentStackWidth  = tmp->size().col_;
      }
      
      tmp = manager()->find_viewport(breakpointViewer_);
      
      if(tmp)
      {
         currentBreakpointsHeight = tmp->size().row_;
         currentBreakpointsWidth  = tmp->size().col_;
      }

      tmp = manager()->find_viewport(variableViewer_);
      
      if(tmp)
      {
         currentVariablesHeight = tmp->size().row_;
         currentVariablesWidth  = tmp->size().col_;
      }

     currentVariablesHeight = currentVariablesHeight; // suppress unused variable warning
     currentVariablesWidth  = currentVariablesWidth;  // suppress unused variable warning


     int minLeftColumnDaughterHeight = 3 * 4;  // 3 daughters times 4 rows per lc daughter

     if(currentGdbViewerHeight > manager()->window()->size().row_ - minLeftColumnDaughterHeight)
     {
        currentGdbViewerHeight = manager()->window()->size().row_ - minLeftColumnDaughterHeight;
     }

  }


  Viewer* oldie = manager_->find_viewer(name);

  int oldHeight = 0;

  if(oldie)
  {
    // while the viewer holding the source code for the current lineis still visible,
    // set the desired line number for the repaint of that viewer's window just below.
    // this must be done before the call to unify() because to make it work as desirable,
    // the window has to actually be displayed so that desired line appears somewhere in
    // the middle of the window that will get painted rather than being always forced to
    // be the top displayed line of the page.  While that works, it is annoying.
    //
    // If the window is not already visible, well, top is all you get.

    oldie->set_row_col_hint(lineNumber-1,0); 
                    
    viewport * vp = manager_->find_viewport(oldie);

    if(vp)
    {
        oldHeight = vp->size().row_;

    }
                                             
  }
  manager_->activate(this);
  manager_->unify();

  if(oldie)
  {
    // found a viewer showing the source code line of interest

    if(!manager_->isVisible(oldie))
    {
        manager_->hsplit(oldie, name, currentGdbViewerWidth);

        if(!attachMode_)
        {
            manager_->activate(oldie);

            if(!logViewer_)
                handleLogViewerDeath();

            manager_->vsplit(logViewer_, gdbLogViewer_->executableName(), oldHeight);
        }

    }
  }
  else
  {

    // no viewer had the source code of interest -- so this must be the first time that
    // we have split.


    if(!name.exists())
    {

      messages_.push_back("Sorry, can't find that file.  Press F12 to step out of this function");
      messages_.push_back("  Filename is:");
      messages_.push_back(name);

       return ;
    }

    Viewer *newbie = manager_->open(name);

    if(!newbie)
    {
      manager_->window()->beep();
      
      manager_->vsplit(logViewer_, gdbLogViewer_->executableName());
    }
    else
    {
      newbie->set_row_col_hint(lineNumber-1, 0);  // 0 based line numbers!
      
      manager_->hsplit(newbie, name, currentGdbViewerWidth);

      if(!attachMode_)
      {

          manager_->activate(newbie);

          if(!logViewer_)
             handleLogViewerDeath();

          manager_->vsplit(logViewer_, gdbLogViewer_->executableName());
      }


    }

  }

  manager_->activate(this);

  if(!gdbStackViewer_)
  {
     handleStackViewerDeath();
  }

  if(gdbStackViewer_)
  {
      manager_->vsplit(stackViewer_, gdbStackViewer_->executableName(), currentGdbViewerHeight);

      manager_->activate(stackViewer_);
  }


  if(!gdbBreakpointViewer_)
  {
     handleBreakpointViewerDeath();
  }

  if(breakpointViewer_)
  {
      manager_->vsplit(breakpointViewer_, gdbBreakpointViewer_->executableName(), currentStackHeight);
      manager_->activate(breakpointViewer_);
  }

  if(!gdbVariableViewer_)
  {
     handleVariableViewerDeath();
  }

  if(variableViewer_)
  {
      manager_->vsplit(variableViewer_, gdbBreakpointViewer_->executableName(), currentBreakpointsHeight);
      manager_->activate(this);
  }

  vp = manager_->find_viewport(this);

  {
      // move the stop point marker to the new place

      removeStopPointMarker();

      stopFilename_ = fullname;
      stopLine_     = fmtd<>(lineNumber);

      displayStopPointMarker();
      
  }
  manager_->window()->refresh();

}


//----------------------------------------------------------------------------------
GdbBreakpointViewer::
~GdbBreakpointViewer()
{
    gdbViewer_->clearBreakpointViewerPointer();
}


//----------------------------------------------------------------------------------
std::auto_ptr< std::list< std::string > > 
GdbBreakpointViewer::
help_text()
{
    std::auto_ptr< std::list< std::string> > rv(new std::list< std::string > );

    rv.get()->push_back("");
    rv.get()->push_back("");
    rv.get()->push_back("GDB Breakpoint Viewer Help");
    rv.get()->push_back("");
    rv.get()->push_back("  The breakpoint viewer allows you see a list of the breakpoints");
    rv.get()->push_back("  and execute related commands.");
    rv.get()->push_back("");
    rv.get()->push_back("Command keys:");
    rv.get()->push_back("");
    rv.get()->push_back("  F1       --  Get Help");
    rv.get()->push_back("  F2       --  Switch between viewers in a cyclic fashion");
    rv.get()->push_back("  F5       --  Switch to the main gdb window and run target");
    rv.get()->push_back("  F7       --  Switch to the main gdb window and prompt for text to be");
    rv.get()->push_back("               sent to the target stdin stream");
    rv.get()->push_back("  F8       --  Switch to the main gdb window and halt/kill the program");
    rv.get()->push_back("  F9       --  Switch to the main gdb window and step the program");
    rv.get()->push_back("               skipping over function calls.");
    rv.get()->push_back("  F10      --  Switch to the main gdb window and step the program");
    rv.get()->push_back("               skipping over function calls.");
    rv.get()->push_back("  F11      --  Switch to the main gdb window and step the program");
    rv.get()->push_back("               stepping into function calls.");
    rv.get()->push_back("  F12      --  Switch to the main gdb window and step the program");
    rv.get()->push_back("               until it returns from the current function.");
    rv.get()->push_back("  ^5       --  Widen this window");
    rv.get()->push_back("  ^6       --  Increase this window's height");
    rv.get()->push_back("  Enter    --  When on the filename lines, cause the source viewer");
    rv.get()->push_back("               to display the line where the breakpoint is set.");
    rv.get()->push_back("  C        --  switch to the main gdb viewer and let the user");
    rv.get()->push_back("               change the target's current directory.");
    rv.get()->push_back("  c        --  change the breakpoint's condition expression.");
    rv.get()->push_back("  d        --  disable the breakpoint under cursor");
    rv.get()->push_back("  e        --  enable the breakpoint under cursor");
    rv.get()->push_back("  g        --  Switch directly back to the Gdb main viewer.");
    rv.get()->push_back("  o        --  Switch directly to the GdbTargetOutputViewer");
    rv.get()->push_back("  s        --  Switch directly to the GdbStackViewer");
    rv.get()->push_back("  v        --  Switch directly to the source code viewer.");
    rv.get()->push_back("  x        --  delete the breakpoint under cursor");
    rv.get()->push_back("");
    rv.get()->push_back("");

    return rv;
}
//----------------------------------------------------------------------------------
static void signalBreakpointViewerDeath(ViewerManager*vm, std::string const &gdbviewerInstanceName)
{
    GdbViewer *v = dynamic_cast<GdbViewer*>(vm->find_viewer(gdbviewerInstanceName));

    if(v)
    {
       // found a viewer by the proper name and it is a gdb viewer

       v->handleBreakpointViewerDeath();
    }

    
}
//----------------------------------------------------------------------------------
int
GdbBreakpointViewer::
handle_event(CursorWindow::input_event const *e, 
             size_t                           currentRow,
             ViewerManager*                   vm
            )
{

    if(e->value_  == 'q' || e->value_ == 'Q' || e->value_ == 0x0b)   // 0x0b is ^K
    {
       signalBreakpointViewerDeath(vm, executableName());
       return 2;   // we are done
    }
    
    if(e->type_ == CursorWindow::input_event::ForceExitKey)
    {
       signalBreakpointViewerDeath(vm, executableName());
       return 2; // we are done
    }

    switch(e->value_)
    {
        case CursorWindow::key_f7:
           if(gdbViewer_)
           {
               gdbViewer_->feedStdinHelper();  // the user enter data to be sent to the target's stdin stream
           }
           break;

        case CursorWindow::key_f8:
           if(gdbViewer_)
           {
               gdbViewer_->haltHelper();
           }
           break;


        case CursorWindow::key_f5:
        case CursorWindow::key_f9:
        case CursorWindow::key_f10:
        case CursorWindow::key_f11:
        case CursorWindow::key_f12:
           if(gdbViewer_)
           {
               gdbViewer_->runHelper(e->value_);
           }
           break;

        case  '\n':  // drop through
        case  '\r':
           if(currentRow < strings_.size())
           {
               std::string const &curRow =  strings_[currentRow];

               if(curRow.empty() || curRow[0] != ' ')
                  vm->window()->beep();
               else
               {
                  // on the filename/linenumber rows, cause the viewer to show the desired file/line

                  size_t filenameStart = curRow.find_first_not_of(' ');
                  size_t filenameEnd   = curRow.find_last_of(':');

                  if(filenameEnd < curRow.size())
                  {
                       std::string filename = curRow.substr(filenameStart, (filenameEnd - filenameStart));
                       
                       int line=1;
                       
                       sscanf(curRow.c_str() + filenameEnd+1, "%d", &line);
                       
                       gdbViewer_->showFileLocation(filename, line);
                       return false;
                  }
                  else
                  {
                      vm->window()->beep();
                  }
               }
           
           }                               
           else
              vm->window()->beep();
           break;

        case 'C':
           if(gdbViewer_)
           {
               gdbViewer_->cdHelper();  // let the user change the debugger's current directory
           }
           break;

       case 'x':
       {
          // delete breakpoints

          std::string curLine="-1";


          if(currentRow < strings_.size())
          {
             curLine = strings_[currentRow];
          }

          size_t space = curLine.find_first_of(' ');

          if(space != curLine.npos)
          {
              curLine = curLine.substr(0, space);

              // remove bp from list of active breakpoints


              // remove it from the deb ugger

              gdbViewer_->removeBreakpoint(curLine);       // let the gdb viewer do the work
              gdbViewer_->manager()->activate(gdbViewer_); // make it active
              // this viewer's onscreen content will be updated by the gdbviewer
              // shortly thereafter
          }


       }
       break;

       case 'd':
       case 'e':
       {
          // disable/enable breakpoints

          std::string curLine="-1";


          if(currentRow < strings_.size())
          {
             curLine = strings_[currentRow];
          }

          size_t space = curLine.find_first_of(' ');

          if(space != curLine.npos)
          {
              curLine = curLine.substr(0, space);

              // remove bp from list of active breakpoints


              // remove it from the deb ugger

              gdbViewer_->enableBreakpoint(curLine, e->value_ == 'e' );   // let the gdb viewer do the work
              gdbViewer_->manager()->activate(gdbViewer_); // make it active
              // this viewer's onscreen content will be updated by the gdbviewer
              // shortly thereafter
          }


       }
       break;

       case 'c':
       {
          // modify breakpoint condition

          std::string curLine="-1";


          if(currentRow < strings_.size())
          {
             curLine = strings_[currentRow];
          }

          size_t space = curLine.find_first_of(' ');

          if(space != curLine.npos)
          {
              curLine = curLine.substr(0, space);

              std::string condition;

              if(!gdbViewer_->getBreakpointCondition(curLine, condition))
              {
                  gdbViewer_->manager()->window()->beep(); // no such bp -- something is wrong, but what?
              }
              else
              {
                 // let the user modify the breakpoint condition

                 CursorWindow::Dialog d("Define/Modify Breakpoint Condition Expression");

                 d += new CursorWindow::Dialog::String("condition", std::string("For Breakpoint ") + curLine, condition, 40);

                 if(!d.popup(gdbViewer_->manager()->window()))
                 {
                     condition = d.element_value("condition");

                     gdbViewer_->setBreakpointCondition(curLine, condition);
                     gdbViewer_->manager()->activate(gdbViewer_); // make it active
                     // this viewer's onscreen content will be updated by the gdbviewer
                     // shortly thereafter
                 }
              }


          }


       }
       break;

       case 'g':
              gdbViewer_->manager()->activate(gdbViewer_); // make it active
           break;

       case 'o':
       case 's':
       case 'v':
                gdbViewer_->focusOnOtherViewer(e->value_);  // make corresponding viewer active
            break;

       default:
             gdbViewer_->manager()->window()->beep();      // invalid command
           break;
    }

    return 0;  // we are not done and the screen doesn't need updating

}

//----------------------------------------------------------------------------------
GdbStackViewer::
GdbStackViewer(std::string const &executableName, GdbViewer* viewer)
:  executableName_(executableName)
,  gdbViewer_(viewer)
{
    title_ = "GdbStack";
}


//----------------------------------------------------------------------------------
GdbStackViewer::
~GdbStackViewer()
{
    gdbViewer_->clearStackViewerPointer();
}

//----------------------------------------------------------------------------------
std::string const&
GdbStackViewer::
application_name() const
{
   return executableName_;
}

//----------------------------------------------------------------------------------
std::string
GdbStackViewer::
description() const
{
   return std::string("Stack for ") + executableName_;
}

//----------------------------------------------------------------------------------
bool
GdbStackViewer::
is_marked(size_t index)
{
    return false;
}


//----------------------------------------------------------------------------------
std::auto_ptr< std::list< std::string > > 
GdbStackViewer::
help_text()
{
    std::auto_ptr< std::list< std::string> > rv(new std::list< std::string > );

    rv.get()->push_back("");
    rv.get()->push_back("");
    rv.get()->push_back("GDB Stack Viewer Help");
    rv.get()->push_back("");
    rv.get()->push_back("  The stack viewer is a 'helper' application that is only");
    rv.get()->push_back("  useful in the context of running GdbViewer.  It exists");
    rv.get()->push_back("  primary to assist in the display of stack frames and in");
    rv.get()->push_back("  handling commands associated therewith.  For example:");
    rv.get()->push_back("");
    rv.get()->push_back("    *  listing the function call parameters for a frame.");
    rv.get()->push_back("    *  listing the values of local variables in the function's frame.");
    rv.get()->push_back("    *  printing user supplied expressions within the frame's context.");
    rv.get()->push_back("    *  jumping to the source code associated with the frame.");
    rv.get()->push_back("    *  switching back to the main GdbViewer window.");
    rv.get()->push_back("    *  adding breakpoints which stop the program when variables");
    rv.get()->push_back("       in the stack frame have changed.");
    rv.get()->push_back("");
    rv.get()->push_back("Command keys:");
    rv.get()->push_back("");
    rv.get()->push_back("   F1    -- get help");
    rv.get()->push_back("   F2    -- switch viewers in a cycle fashion");
    rv.get()->push_back("   F5    -- Switch to the main gdb window and run target");
    rv.get()->push_back("   F7    -- Switch to the main gdb window and prompt for text to be");
    rv.get()->push_back("            sent to the target stdin stream");
    rv.get()->push_back("   F8    -- Switch to the main gdb window and halt/kill the target program");
    rv.get()->push_back("   F9    -- Switch to the main gdb window and step the program");
    rv.get()->push_back("            skipping over function calls.");
    rv.get()->push_back("   F10   -- Switch to the main gdb window and step the program");
    rv.get()->push_back("            skipping over function calls.");
    rv.get()->push_back("   F11   -- Switch to the main gdb window and step the program");
    rv.get()->push_back("            stepping into function calls.");
    rv.get()->push_back("   F12   -- Switch to the main gdb window and step the program");
    rv.get()->push_back("            until it returns from the current function.");
    rv.get()->push_back("  ^5     --  Widen this window");
    rv.get()->push_back("  ^6     --  Increase this window's height");
    rv.get()->push_back("   Enter -- If on a stack frame line, show the function parameters.");
    rv.get()->push_back("            Otherwise, instruct the source code viewer to display");
    rv.get()->push_back("            The source for this stack frame's current program count.");
    rv.get()->push_back("   a     -- Add a expression to the GdbDisplayVariable window which");
    rv.get()->push_back("            will be displayed whenever the debugger stops in");
    rv.get()->push_back("            in the selected stack frame.");
    rv.get()->push_back("   b     -- Switch directly to the GdbBreakpointViewer");
    rv.get()->push_back("   C     -- Switch directly to the gdb main viewer and let the user");
    rv.get()->push_back("            change the target's current directory.");
    rv.get()->push_back("   g     -- Switch directly back to the main GdbViewer.");
    rv.get()->push_back("   l     -- Print the local variable values for this frame.  (see below)");
    rv.get()->push_back("   o     -- Switch directly to the GdbTargetOutputViewer");
    rv.get()->push_back("   p     -- Prompt the user for an expression and print it's value");
    rv.get()->push_back("            within this stack frame. (see below)");
    rv.get()->push_back("   s     -- Switch directly to the GdbStackViewer");
    rv.get()->push_back("   v     -- Switch directly to the GdbDisplayVariableViewer");
    rv.get()->push_back("   w     -- Add a breakpoint that stops the program whenever a");
    rv.get()->push_back("            variable in this stack frame is changed");
    rv.get()->push_back("   ?     -- Get the data type of a variable name");
    rv.get()->push_back("");
    rv.get()->push_back("Printing Array and Pointer Values:");
    rv.get()->push_back("");
    rv.get()->push_back("  When using the command keys to print array and pointer values");
    rv.get()->push_back("  You often get only a hex number displayed.  To see array locations");
    rv.get()->push_back("  you can use the 'p' or 'a' command as in these examples:");
    rv.get()->push_back("");
    rv.get()->push_back("   p *pointer");
    rv.get()->push_back("   a array[4]");
    rv.get()->push_back("   p array[9]");
    rv.get()->push_back("");
    rv.get()->push_back("Printing Character Arrays:");
    rv.get()->push_back("");
    rv.get()->push_back("  When printing a character array, the debugger tries to print");
    rv.get()->push_back("  the data as an ARRAY, not a string.  This means you get text");
    rv.get()->push_back("  past the null terminator displayed -- this is typically not");
    rv.get()->push_back("  what you really want.");
    rv.get()->push_back("");
    rv.get()->push_back("  To avoid this, use the cast operator to cast the array to a");
    rv.get()->push_back("  a (char*) when printing or displaying.  For example:");
    rv.get()->push_back("");
    rv.get()->push_back("    p (char*)arrayName");
    rv.get()->push_back("    a (char*)arrayName");
    rv.get()->push_back("");
    rv.get()->push_back("Printing Reference Variables:");
    rv.get()->push_back("");
    rv.get()->push_back("  When printing a reference to a pointer, the debugger tries to print");
    rv.get()->push_back("  the pointer as a HEX number.  This is typically not");
    rv.get()->push_back("  what you really want.");
    rv.get()->push_back("");
    rv.get()->push_back("  To avoid this, use the cast operator to cast the reference variable to a");
    rv.get()->push_back("  pointer to the type of interest.  For example:");
    rv.get()->push_back("");
    rv.get()->push_back("    p (char*)refToPtr");
    rv.get()->push_back("    a (char*)refToPtr");
    rv.get()->push_back("");



    return rv;
}



//----------------------------------------------------------------------------------
static void signalStackViewerDeath(ViewerManager*vm, std::string const &gdbviewerInstanceName)
{
    GdbViewer *v = dynamic_cast<GdbViewer*>(vm->find_viewer(gdbviewerInstanceName));

    if(v)
    {
       // found a viewer by the proper name and it is a gdb viewer

       v->handleStackViewerDeath();
    }

    
}

//----------------------------------------------------------------------------------
int
GdbStackViewer::
handle_event(CursorWindow::input_event const *e, 
             size_t                           currentRow,
             ViewerManager*                   vm
            )
{

    if(e->value_  == 'q' || e->value_ == 'Q' || e->value_ == 0x0b)   // 0x0b is ^K
    {
       signalStackViewerDeath(vm, executableName_);
       return 2;   // we are done.  Note, this is a temporary death only because
                   // main viewer will restart this window
    }
    
    if(e->type_ == CursorWindow::input_event::ForceExitKey)
    {
       signalStackViewerDeath(vm, executableName_);
       return 2; // see above.
    }

    switch(e->value_)
    {
        case CursorWindow::key_f7:
           if(gdbViewer_)
           {
               gdbViewer_->feedStdinHelper();  // the user enter data to be sent to the target's stdin stream
           }
           break;
        case CursorWindow::key_f8:
           if(gdbViewer_)
           {
               gdbViewer_->haltHelper();
           }
           break;
        case CursorWindow::key_f5:
        case CursorWindow::key_f10:
        case CursorWindow::key_f9:
        case CursorWindow::key_f11:
        case CursorWindow::key_f12:
           if(gdbViewer_)
           {
               gdbViewer_->runHelper(e->value_);
           }
           break;

        case '?':
           if(gdbViewer_ && currentRow < strings_.size())
           {
               std::string const &curRow =  strings_[currentRow];

               if(curRow.empty() || ! isdigit(curRow[0]))
               {
                  vm->window()->beep();
                  break;
               }

               std::string frameDepth;

               size_t firstNonDigit = curRow.find_first_not_of("01234567890");

               if(firstNonDigit == 0 || firstNonDigit >= curRow.size())
               {
                  vm->window()->beep();
                  break;
               }

               frameDepth = curRow.substr(0, firstNonDigit);

               CursorWindow::Dialog d("Get Data Type from gdb");

               d += new CursorWindow::Dialog::String("expression", "Expression", "", 40);

               if(!d.popup(vm->window()))
               {
                  std::string expression = d.element_value("expression");
                  std::string type       = gdbViewer_->whatisHelper(frameDepth, expression);

                  if(type.empty())
                  {
                      CursorWindow::Message m("Unable to obtain data type");

                      m += "";
                      m += std::string("  for: ") + expression;
                      m += "";

                      m.popup(vm->window());
                  }
                  else
                  {
                      CursorWindow::Message m("Data type                              ");

                      m += "";
                      m += std::string("  for: ") + expression;
                      m += "";
                      m += "is:";
                      m += "";
                      m += std::string("    ") + type;

                      m.popup(vm->window());
                     
                  }


               }


           }
           break;


        case 'C':
           if(gdbViewer_)
           {
               gdbViewer_->cdHelper(); 
           }
           break;


        case 'l': // list the local variables in this stack frame
           if(currentRow < strings_.size())
           {
               std::string const &curRow =  strings_[currentRow];

               if(curRow.empty() || ! isdigit(curRow[0]))
               {
                  vm->window()->beep();
               }
               else
               {
                  std::string frameDepth;

                  size_t firstNonDigit = curRow.find_first_not_of("01234567890");

                  frameDepth = curRow.substr(0, firstNonDigit);

                  std::string frameVars = gdbViewer_->stackFrameVariables(frameDepth);

                  if(frameVars.empty())
                      vm->window()->beep();
                  else
                  {
                      std::string msg("Variables in stack frame ");

                      msg += frameDepth;

                      gdbViewer_->displayGdbVariableList(vm, msg, frameVars);

                  }
                   
               }

           }
           else
              vm->window()->beep();

           break;

        case 'p':  // print a specific expression, one time
           if(currentRow < strings_.size())
           {
               std::string const &curRow =  strings_[currentRow];

               if(curRow.empty() || ! isdigit(curRow[0]))
               {
                  vm->window()->beep();
               }
               else
               {
                  std::string frameDepth;

                  size_t firstNonDigit = curRow.find_first_not_of("01234567890");

                  frameDepth = curRow.substr(0, firstNonDigit);

                  std::string expression;

                  std::string title = "Enter expression for stack frame ";

                  title += frameDepth;


                  CursorWindow::Dialog d(title);

                  d += new CursorWindow::Dialog::String("e", "Expression", expression, 40);

                  if(d.popup(vm->window()) )
                  {
                     vm->window()->beep();
                     break;
                  }

                  expression = d.element_value("e");

                  std::string printedValue = gdbViewer_->evaluateTmpExpression(frameDepth, expression);

                  {
                      title = "Value of ";
                      title += expression;

                      gdbViewer_->displayGdbValue(vm, title, printedValue);

                  }
                   
               }

           }
           else
              vm->window()->beep();

           break;

        case 'w':  // create a watch-expression style of breakpoint
           if(currentRow < strings_.size())
           {
               std::string const &curRow =  strings_[currentRow];

               if(curRow.empty() || ! isdigit(curRow[0]))
               {
                  vm->window()->beep();
               }
               else
               {
                  std::string frameDepth;

                  size_t firstNonDigit = curRow.find_first_not_of("01234567890");

                  frameDepth = curRow.substr(0, firstNonDigit);

                  std::string expression;

                  std::string title = "Enter Watch Expression for this stack frame ";

                  title += frameDepth;

                  CursorWindow::Dialog d(title);

                  d += new CursorWindow::Dialog::String("e", "Expression", expression, 40);
                  d += new CursorWindow::Dialog::String("m", "Access mode (R/W/Any)", "W", 4);

                  if(d.popup(vm->window()) )
                  {
                     vm->window()->beep();
                     break;
                  }

                  expression = d.element_value("e");

                  std::string mode = d.element_value("m");

                  char const *modePtr = mode.c_str();

                  if(modePtr[0] == 'A' || modePtr[0] == 'a')
                     modePtr="-a";  // any access
                  else
                  if(modePtr[0] == 'R' || modePtr[0] == 'r')
                     modePtr="-r";  // read access
                  else
                     modePtr="";  // write mode

                  gdbViewer_->watchLocation(frameDepth, expression, modePtr);

                  gdbViewer_->manager()->activate(gdbViewer_);  

               }

           }
           else
              vm->window()->beep();

           break;

        case 'a': // add a variable to the GdbDisplayVariables window
           if(currentRow < strings_.size())
           {
               std::string const &curRow =  strings_[currentRow];

               if(curRow.empty() || ! isdigit(curRow[0]))
               {
                  vm->window()->beep();
               }
               else
               {
                  std::string frameDepth;

                  size_t firstNonDigit = curRow.find_first_not_of("01234567890");

                  frameDepth = curRow.substr(0, firstNonDigit);

                  std::string expression;

                  std::string title = "Enter Display Expression for this stack frame ";

                  title += frameDepth;

                  CursorWindow::Dialog d(title);

                  d += new CursorWindow::Dialog::String("e", "Expression", expression, 40);

                  if(d.popup(vm->window()) )
                  {
                     vm->window()->beep();
                     break;
                  }

                  expression = d.element_value("e");

                  gdbViewer_->addDisplayExpression(frameDepth, expression);

                  gdbViewer_->manager()->activate(gdbViewer_);  

               }

           }
           else
              vm->window()->beep();

           break;

        case  '\n':  // drop through
        case  '\r':  // Either display the function parameters or the source code associated
                     // with program counter in this stack frame.
           if(currentRow < strings_.size())
           {
               std::string const &curRow =  strings_[currentRow];

               if(curRow.empty())
                  vm->window()->beep();
               else
               if(curRow[0] != ' ')
               {
                  // handle rows defining the function call level and function name.

                  std::string depthBuffer;

                  size_t firstNonDigit = curRow.find_first_not_of("01234567890");

                  depthBuffer = curRow.substr(0, firstNonDigit);

                  std::string frameArgs = gdbViewer_->stackFrameParms(depthBuffer);

                  if(frameArgs.empty())
                      vm->window()->beep();
                  else
                  {
                      std::string msg("Function call parameters for stack frame ");

                      msg += depthBuffer;

                      gdbViewer_->displayGdbVariableList(vm, msg, frameArgs);



                  }
               }
               else
               {
                  // on the filename/linenumber rows, cause the viewer to show the desired file/line

                  size_t filenameStart = curRow.find_first_not_of(' ');
                  size_t filenameEnd   = curRow.find_last_of(':');

                  std::string filename = curRow.substr(filenameStart, (filenameEnd - filenameStart));

                  int line=1;

                  sscanf(curRow.c_str() + filenameEnd+1, "%d", &line);

                  gdbViewer_->showFileLocation(filename, line);
                  return false;

               }
           
           }                               
           else
              vm->window()->beep();
           break;

        default:
             gdbViewer_->manager()->window()->beep();      // invalid command
           break;

       case 'o':
       case 'd':
       case 'b':
       case 'v':
                gdbViewer_->focusOnOtherViewer(e->value_);  // make corresponding viewer active
            break;

        case 'g': // return to the main gdb viewer
            gdbViewer_->manager()->activate(gdbViewer_);
        break;
    }

    return 0;  // we are not done and the screen doesn't need updating

}

//----------------------------------------------------------------------------------
void
GdbStackViewer::
resetContent(std::vector<std::string> &content, Viewer *v)
{
    ListViewer *lv = dynamic_cast<ListViewer*>(v);

    if(lv)
    {
       lv->resetContent();
    }
    

    strings_.swap(content);

    CursorWindow::viewport *vp = v->manager()->find_viewport(v);

    if(vp)
    {
       (*v)(vp, 1);  // repaint the viewer in the associated viewport.  display it as inactive
    }

}

//-----------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------
GdbLogViewer::
GdbLogViewer(std::string const &executableName, GdbViewer *vm)
:  executableName_(executableName)
,  gdbViewer_(vm)
{
    title_ = "GdbTargetOutputViewer";
}

//----------------------------------------------------------------------------------
GdbLogViewer::
~GdbLogViewer()
{
    gdbViewer_->clearLogViewerPointer();
}

//----------------------------------------------------------------------------------
std::string const&
GdbLogViewer::
application_name() const
{
   return executableName_;
}

//----------------------------------------------------------------------------------
std::string
GdbLogViewer::
description() const
{
   return std::string("stdout for ") + executableName_;
}

//----------------------------------------------------------------------------------
bool
GdbLogViewer::
is_marked(size_t index)
{
    return false;
}

//----------------------------------------------------------------------------------
std::auto_ptr< std::list< std::string > > 
GdbLogViewer::
help_text()
{
    std::auto_ptr< std::list< std::string> > rv(new std::list< std::string > );

    rv.get()->push_back("");
    rv.get()->push_back("");
    rv.get()->push_back("GDB TARGET PROCESS LOG VIEWER");
    rv.get()->push_back("");
    rv.get()->push_back("  The log viewer exists so that you can watch output from");
    rv.get()->push_back("  the debugged child executable.  You can really only scroll");
    rv.get()->push_back("  up and down within this viewer and perform simple searches.");
    rv.get()->push_back("");
    rv.get()->push_back("Command Keys:");
    rv.get()->push_back("");
    rv.get()->push_back("  F1       --  Get help");
    rv.get()->push_back("  F2       --  Switch other viewers in a cyclic pattern");
    rv.get()->push_back("  F7       --  Switch to the main gdb window and prompt for text to be");
    rv.get()->push_back("               sent to the target stdin stream");
    rv.get()->push_back("  F8       --  Switch to the main gdb window and prompt for a new");
    rv.get()->push_back("               current directory for the target process.");
    rv.get()->push_back("  ^5       --  Widen this window");
    rv.get()->push_back("  ^6       --  Increase this window's height");
    rv.get()->push_back("  ^S       --  Search for a string.");
    rv.get()->push_back("  ^N       --  Repeat a previous search.");
    rv.get()->push_back("  g        --  Jump directly to the gdb main viewer");
    rv.get()->push_back("");

    return rv;
}

//----------------------------------------------------------------------------------
static void signalLogViewerDeath(ViewerManager*vm, std::string const &gdbviewerInstanceName)
{
    GdbViewer *v = dynamic_cast<GdbViewer*>(vm->find_viewer(gdbviewerInstanceName));

    if(v)
    {
       // found a viewer by the proper name and it is a gdb viewer

       v->handleLogViewerDeath();
    }

    
}

//----------------------------------------------------------------------------------
int
GdbLogViewer::
handle_event(CursorWindow::input_event const *e, 
             size_t                           currentRow,
             ViewerManager*                   vm
            )
{

    if(e->value_  == 'q' || e->value_ == 'Q' || e->value_ == 0x0b)  // 0x0b is ^K
    {
       signalLogViewerDeath(vm, executableName_);
       return 2;   // we are done
    }
    
    if(e->type_ == CursorWindow::input_event::ForceExitKey)
    {
       signalLogViewerDeath(vm, executableName_);
       return 2; // we are done
    }

    switch(e->value_)
    {
        case CursorWindow::key_f7:
           if(gdbViewer_)
           {
               gdbViewer_->feedStdinHelper();  // the user enter data to be sent to the target's stdin stream
           }
           break;
        case CursorWindow::key_f8:
           if(gdbViewer_)
           {
               gdbViewer_->haltHelper();  // halt or kill the target
           }
           break;

        case 'C':
           if(gdbViewer_)
           {
               gdbViewer_->cdHelper();  // let the user change the target's current directory
           }
           break;

        case 'g':
            {
                Viewer* gdb = vm->find_viewer(executableName_);
                
                if(gdb)
                    vm->activate(gdb);
                
            }
            break;

        default:
            vm->window()->beep();
            break;
            
    }
    return 0;  // we are not done and the screen doesn't need updating

}

//----------------------------------------------------------------------------------
void
GdbLogViewer::
appendContent(std::vector<std::string> &content, Viewer *v)
{
    CXXTLS_FOREACH(std::string const &cur, content)
        strings_.push_back(cur);

    CursorWindow::viewport *vp = v->manager()->find_viewport(v);

    if(!strings_.empty())
        v->set_row_col_hint(strings_.size()-1, 0, true);  

    if(vp)
    {
       (*v)(vp, 1);  // repaint the viewer in the associated viewport.  display it as inactive
    }

}

//----------------------------------------------------------------------------------
void
GdbLogViewer::
set_row_col_hint(ListViewer*v, size_t line, size_t column, bool repaint, size_t &display_top, size_t &current_row)
{

    CursorWindow::viewport *vp = v->manager()->find_viewport(v);

    if(line >= strings_.size())
        return;   // don't position to invalid location

    if(!vp || !repaint)
    {
        //
        //   viewer not visible, just move the cursor to the desired location.
        //
        display_top  = line;
        current_row  = line;
    }
    else
    {
       //
       //  viewer is visible, make a goodly amount of the strings visible
       //

       // CursorWindow::row_col size = vp->size();

       int rowsAvailable = v->rowsAvailable(vp);

       if(strings_.size() < size_t(rowsAvailable))
       { 
          display_top  = 0;
          current_row  = line;
       }
       else
       {
          //
          //  make the requested line show up in the bottom of the visible
          //  region of the screen
          //

          display_top   = line - rowsAvailable+1;
          current_row   = line;
       }


    }
    
}

//----------------------------------------------------------------------------------
void
GdbBreakpointViewer::
set_row_col_hint(ListViewer*v, size_t line, size_t column, bool repaint, size_t &display_top, size_t &current_row)
{

    CursorWindow::viewport *vp = v->manager()->find_viewport(v);

    if(line >= strings_.size())
        return;   // don't position to invalid location

    if(!vp || !repaint)
    {
        //
        //   viewer not visible, just move the cursor to the desired location.
        //
        display_top  = line;
        current_row  = line;
    }
    else
    {
       //
       //  viewer is visible, make a goodly amount of the strings visible
       //

       // CursorWindow::row_col size = vp->size();

       int rowsAvailable = v->rowsAvailable(vp);

       if(strings_.size() < size_t(rowsAvailable))
       { 
          display_top  = 0;
          current_row  = line;
       }
       else
       {
          //
          //  make the requested line show up in the bottom of the visible
          //  region of the screen
          //

          display_top   = line - rowsAvailable+1;
          current_row   = line;
       }


    }
    
}


//----------------------------------------------------------------------------------
void
GdbBreakpointViewer::
resetContent(std::vector<std::string> &content, Viewer *v, size_t itemIndex)
{
    ListViewer *lv = dynamic_cast<ListViewer*>(v);

    if(lv)
    {
       lv->resetContent();

    }

    strings_.swap(content);


    CursorWindow::viewport *vp = v->manager()->find_viewport(v);

    if(vp)
    {
        CursorWindow::row_col size = vp->size();   size = size; // shut up compiler!


        if(lv)
        {
           if(itemIndex > 2)
              itemIndex -= 2;

           lv->set_row_col_hint(itemIndex, 0, false);
        
        }
        

       (*v)(vp, 1);  // repaint the viewer in the associated viewport.  display it as inactive
    }
    else
    {
        if(lv)
        {
        
           lv->set_row_col_hint(itemIndex, 0, false);
        
        }
        
    }

}

//----------------------------------------------------------------------------------
void
GdbViewer::
handleBreakpointViewerDeath()  // also, create the darn thing in the first place
{
   // create new breakpointvier viewer

   breakpointViewer_ = new ListViewer(gdbBreakpointViewer_ = new GdbBreakpointViewer(fullname_, this), manager_);

   manager_->add(breakpointViewer_, fullname_);  // not made visible, just added.  Visibility comes on stop/terminate
    
}

//----------------------------------------------------------------------------------
void
GdbViewer::
handleStackViewerDeath()
{
   // create new stack viewer

   stackViewer_ = new ListViewer(gdbStackViewer_ = new GdbStackViewer(fullname_, this), manager_);

   manager_->add(stackViewer_, fullname_);  // not made visible, just added.  Visibility comes on stop/terminate
    
}

//----------------------------------------------------------------------------------
void
GdbViewer::
handleLogViewerDeath()
{
    // create new log viewer

   logViewer_   = new ListViewer(gdbLogViewer_ = new GdbLogViewer(fullname_, this), manager_);

   manager_->add(logViewer_, fullname_); // not made visible, just added.  Visibility comes on stop/terminate

}

//----------------------------------------------------------------------------------
void
GdbViewer::
listen(std::string const &type, std::string const &msg)
{
    // this function is invoked when the user presses ^X SPACE in the editor to set or remove a breakpoint

    static std::string fileLocation("fileLocation");

    if(type == fileLocation)
    {
         VarObj::ParseResult parseResults = VarObj::parse(msg);  // the message is assumed to be VarObj encoded.
                                                                 // the following fields are expected:
                                                                 //
                                                                 //  *  file
                                                                 //  *  line
                                                                 //  *  action  (add/delete)

         if(parseResults.rv_)
         {
             VarObj *filePtr = parseResults.rv_->findNestedChild("file");
             VarObj *linePtr = parseResults.rv_->findNestedChild("line");

             if(filePtr && linePtr)
             {
                 std::string file = filePtr->value();
                 std::string lineValue = linePtr->value();

                 int line=-1;

                 sscanf(lineValue.c_str(), "%d", &line);

                 if(line > 0)
                 {
                     VarObj *actPtr = parseResults.rv_->findNestedChild("action");

                     static std::string deleteIt("delete");

                     if(actPtr && deleteIt == actPtr->value())
                     {
                        // find the file/line pair in the activeBreakpoints_ map and by 
                        // linear search and then send the delete command


                        CXXTLS_FOREACH(ActiveBreakpoints::value_type const &cur, activeBreakpoints_)
                        {
                           if(cur.second.first == file && cur.second.second == lineValue)
                           {
                               gdb_.removeBreakpoint(cur.first);
                               break;
                           }
                        }

                     }
                     else
                     {
                         gdb_.addBreakpoint(file, line);

                         bpItem_ = gdb_.getMostRecentBreakpoint();
                     }

                     displayBreakpoints(bpItem_);

                     CursorWindow::viewport *vp = manager()->find_viewport(this);

                     if(vp)
                         repaintTitle(vp);


                     manager()->broadcast(breakpoint, msg);

                 }
                 else
                 {
                     manager_->window()->beep();
                 }
             }

             delete parseResults.rv_;
         }
    }
}

//----------------------------------------------------------------------------------

void 
GdbViewer::
removeStopPointMarker()
{

      if(stopLine_.empty())
         return;  // already done

      // erase the old

      std::string msg = stopPointMessage(stopFilename_, stopLine_, true); // remove a stop point marker in editors showing this file/line
      
      manager()->broadcast(stopped, msg);

      stopLine_ = "";

}
//----------------------------------------------------------------------------------

void 
GdbViewer::
displayStopPointMarker()
{
      // create the new

      std::string msg = stopPointMessage(stopFilename_, stopLine_, false); // create a stop point marker in editors showing this file/line
      
      manager()->broadcast(stopped, msg);
}

//----------------------------------------------------------------------------------

void
GdbViewer::
showFileLocation(std::string const &file, int line)
{
  CursorWindow::viewport *vp = manager()->find_viewport(this);

    if(vp)
        showCurrentFile(vp, file, line);
}

//----------------------------------------------------------------------------------
// Gdb Data Grammer
//
//  value                   := compositeValue | simpleNumber | characterValue | 
//                             stringValue | idValue | wrappedText
//
//  simpleNumber            := ['-'] {integerOrFloatCharacter}
//
//  integerOrFloatCharacter := '0'-'9' | '+' | 'e' | 'a'-'f' | 'A'-'F' | 'x' | 'o' | '.'
//
//  characterValue          := simpleNumber  "'" anyCharacter "'" 
//
//  stringValue             := simpleNumber  '"' anyCharacter '"'
//
//  anyCharacter            := '\\' | '\'' | \" | '\n' | '\r' | '\t' | '\e' |
//                             <all other single characters>
//
//  idValue                 := {idCharacter}
//
//  idCharacter             := '0'-'9' | 'a'-'z' | '_' | 'A'-'Z' | ':'
//
//  compositeValue          := '{' compositeMemberList '}'
//
//  compositeMemberList     := compositeMember [ ',' compositeMemberList ]
//
//  compositeMember         := (idValue | wrapptedText) [ '=' value ] | value
//
/// wrappedText             := '<' ... '>'  (note that the <> pairs must be balanced)
//

struct GdbDataValue
{
    std::string                 number_;
    std::string                 charString_;
    std::list< GdbDataValue *>  members_;

    std::auto_ptr<GdbDataValue> value_;  // only for "ID=VALUE" style members

    GdbDataValue()
    {
    }

    ~GdbDataValue()
    {
        CXXTLS_FOREACH(GdbDataValue *cur, members_)
           delete cur;
    }

    bool isUseless() const
        // determine if this particular value is useless.  A useless value 
        // basically is one with no number or identifier or member list.  Uselessness
        // also includes the possibility that the debugger has printed the name of
        // a structure which has no members.  This doesn't sound useless, but it is --
        // it conveys no useful information.  No decisions about program correctness
        // can be made by a user debugging a program without data.  Standard template library containers often
        // have this kind of object in them -- and this greatly bulks up the debugger
        // output and makes it confusing.
    {
        static std::string noData("<No data fields>");
        static std::string optimizedOut("<optimized out>");

        if(number_.empty())
        {
            // if number_ is empty, then probably we are dealing with a composite
            
            if(members_.empty())
               return true;

            bool allMembersUseless = true;

            CXXTLS_FOREACH(GdbDataValue *cur, members_)
            {
               if(!cur->isUseless())
               {
                   allMembersUseless = false;
                   break;
               }
            }

            return allMembersUseless;

        }
        else
        if( 0 == value_.get())
        {
            // we have a number_ (which could be an identifier as well)
            // but we have no value for the identifier.


            if(    noData       == number_
               ||  optimizedOut == number_
              )
                return true;
        }
        else
        {
           // have an identfier and a value

           if(value_->isUseless())
              return true;

           if(    number_ == noData
              ||  number_ == optimizedOut
             )
               return true;

        }

        return false;  // normally, things are useful
    }

private:
    // no can copy big guy.
    GdbDataValue &operator=(GdbDataValue const &);
    GdbDataValue(GdbDataValue const&);

};

//----------------------------------------------------------------------------------


static GdbDataValue* parseNumString(char const *&fc, char const *lc);
    ///<  parse a number, character, or a string value.  See grammer for details.
    ///
    ///  @return A pointer to the parsed value, or null.  @b You @b Must @b Delete
    ///          @b the @b return @b value.
    ///
    ///  @param[in,out] fc   The first character in a range to be parsed.  This
    ///                      parameter is a reference and it moves the referenced
    ///                      variable past the parsed data towards lc.
    ///  @param[out[    lc   The last character in the range to be parsed.  As usual
    ///                      lc is outside the range.

static GdbDataValue* parseId(char const *&fc, char const *lc);
    ///<  parse an idvalue.  See grammer above
    ///
    ///  @return A pointer to the parsed value, or null.  @b You @b Must @b Delete
    ///          @b the @b return @b value.
    ///
    ///  @param[in,out] fc   The first character in a range to be parsed.  This
    ///                      parameter is a reference and it moves the referenced
    ///                      variable past the parsed data towards lc.
    ///  @param[out[    lc   The last character in the range to be parsed.  As usual
    ///                      lc is outside the range.

static GdbDataValue* parseComposite(char const *&fc, char const *lc);
    ///<  parse a compositeValue, including its list.  See grammer above
    ///
    ///  @return A pointer to the parsed value, or null.  @b You @b Must @b Delete
    ///          @b the @b return @b value.
    ///
    ///  @param[in,out] fc   The first character in a range to be parsed.  This
    ///                      parameter is a reference and it moves the referenced
    ///                      variable past the parsed data towards lc.
    ///  @param[out[    lc   The last character in the range to be parsed.  As usual
    ///                      lc is outside the range.

static GdbDataValue* parseCompositeMember(char const *&fc, char const *lc);
    ///<  parse a compositeValue, including its list.  See grammer above
    ///
    ///  @return A pointer to the parsed value, or null.  @b You @b Must @b Delete
    ///          @b the @b return @b value.
    ///
    ///  @param[in,out] fc   The first character in a range to be parsed.  This
    ///                      parameter is a reference and it moves the referenced
    ///                      variable past the parsed data towards lc.
    ///  @param[out[    lc   The last character in the range to be parsed.  As usual
    ///                      lc is outside the range.

static GdbDataValue* parseWrappedText(char const *&fc, char const *lc);
    ///<  parse a "<...>" See grammer above
    ///
    ///  @return A pointer to the parsed value, or null.  @b You @b Must @b Delete
    ///          @b the @b return @b value.
    ///
    ///  @param[in,out] fc   The first character in a range to be parsed.  This
    ///                      parameter is a reference and it moves the referenced
    ///                      variable past the parsed data towards lc.
    ///  @param[out[    lc   The last character in the range to be parsed.  As usual
    ///                      lc is outside the range.


static GdbDataValue* parseGdbValue(char const *&fc, char const *lc)
    ///  parse a string looking for a gdb value.  
    ///
    ///  @return A pointer to the parsed value, or null.  @b You @b Must @b Delete
    ///          @b the @b return @b value.
    ///
    ///  @param[in,out] fc   The first character in a range to be parsed.  This
    ///                      parameter is a reference and it moves the referenced
    ///                      variable past the parsed data towards lc.
    ///  @param[out[    lc   The last character in the range to be parsed.  As usual
    ///                      lc is outside the range.
{
   //  value                   := compositeValue | simpleNumber | characterValue | 
   //                             stringValue | idValue
   //

   std::auto_ptr<GdbDataValue> rv;

   StrTool::skipBlanks(fc, lc);

   if(fc != lc)
   {
       switch(*fc)
       {

           case '{':  
               // handle compositeValue
               return parseComposite(fc,lc);

           // handle numbers, strings, and characters

           case '-': case '0': case '1': case '2': case '3': case '4': case '5': case '6': case '7': case '8': case '9': case '.':
               return parseNumString(fc,lc);

           case '<':
               return parseWrappedText(fc,lc);

           case '(':

               // skip "(typename &) @0x1234:" then try parsing again
               while(fc != lc && *fc != ')')
                  ++fc;  

               while(fc != lc && *fc != ':')
                  ++fc;  

               if(fc != lc)
                  ++fc;

               return parseGdbValue(fc,lc);         

           case '@':
              {
                 char const *curlySearch = fc;

                 while(    (curlySearch != lc)
                       &&  (*curlySearch != '{')
                      )
                     ++curlySearch;

                 if(curlySearch == lc)
                 {
                    // did NOT find @value: {...}

                    ++fc;
                    return parseGdbValue(fc,lc);

                 }

                 fc = curlySearch;

                 return parseGdbValue(fc,lc);

              }
              break;

           case '"':
                  // parse strings that look like this:
                  //
                  //   "stuff" [, <error message>]

                  {
                      rv.reset( new GdbDataValue );

                      // "stuff" [, <error message>

                      ++fc; // skip leading "

                      while(fc != lc)
                      {
                          char c = *fc++;

                          if(c == '"')
                          {
                             break;  // fc is already incremented
                          }

                          if(c == '\\')
                          {

                              if(fc != lc)
                              {
                                  c = *fc++;

                                  switch(c)
                                  {
                                      case 'r': rv->number_ += '\r'; break;
                                      case 'n': rv->number_ += '\n'; break;
                                      case 't': rv->number_ += '\t'; break;

                                      case 'a': rv->number_ += '\a'; break;
                                      case 'e': rv->number_ += '\e'; break;

                                      default:  rv->number_ += c;
                                                break;
                                  }

                              }
                              else
                                 rv->number_ += '\\';
                          }
                          else
                              rv->number_ += c;

                      }

                      StrTool::skipBlanks(fc,lc);

                      if(    (fc != lc)
                         &&  *fc == ',')
                      {
                          ++fc;

                          rv->charString_.append(fc,lc);
                      }


                  }

              break;


           // handle idValue and error conditions

           default:
              if(   (*fc == '_')
                 || (*fc == ':')
                 || isalpha(*fc)
                )
                  return parseId(fc,lc);

              // errors:

              rv.reset( new GdbDataValue );

              rv->number_     = "Error, invalid token:";
              rv->charString_.append(fc,lc);

              fc = lc; // end parsing
               
              break;
       }
   }

   return rv.release();
}

//----------------------------------------------------------------------------------
static GdbDataValue* parseNumString(char const *&fc, char const *lc)
{
   // parse a number or a hex number followed by a string:
   //
   //     1234
   //     0xbadbeef "stuff"
   //     0xbadbeef L"stuff"


    std::auto_ptr<GdbDataValue> rv( new GdbDataValue );

    std::string &result = rv->number_;  

    while(fc != lc) 
    {
       char c = *fc;

       bool is_a_digit = false;
       
       if(c >= '0' && c <= '9')
          is_a_digit = true;
       else
       if(c == 'o' || c == 'x' || c == '+' || c == '-' || c == 'e' || c == '.')
          is_a_digit = true;
       else
       if(c >= 'a' && c <= 'f')
          is_a_digit = true;
       else
       if(c >= 'A' && c <= 'F')
          is_a_digit = true;

       if(!is_a_digit)
       {
          break;
       }

       result += *fc++; 
    }

    if(rv->number_.empty())
    {
        rv->number_ = "Error:  expected numeric:";
        rv->charString_.append(fc,lc);
        fc = lc;
    }

    StrTool::skipBlanks(fc,lc);

    if(fc != lc)
    {
       if(*fc == 'L')
       {
         // allow for unicode strings

          ++fc;

          if(*fc != '"')
            --fc;
          else
          {
             rv->charString_ += 'L';
          }
       }


       if(*fc == '"')
       {
          // we got us your standard gdb formatted string representation, right here!

          rv->charString_ += '"';

          ++fc;

          while(   (fc != lc)
                && (*fc != '"')
               )
          {
             char c = *fc;

             rv->charString_ +=  c;

             if( c == '\\')
             {
               ++fc;

               if(fc != lc)
                  rv->charString_ += *fc;
             }

             if(fc != lc)
                 ++fc;
          }

          if(fc != lc)
             ++fc; // skip trailing "

          rv->charString_ += '"';

       }
       else
       if(*fc == '\'')
       {
          rv->charString_ += '\'';

          ++fc;  //skip leading '

          if(*fc == '\\')
          {
            ++fc;   // skip slash
            rv->charString_ += '\'';

            if(fc != lc)
                rv->charString_ += *fc++;  // print and skip /
          }
          else
            rv->charString_ += *fc++; // save the normal characters

          if(fc != lc)
             ++fc; // skip the trailing '
          
          rv->charString_ += '\'';


          // we got us your standard gdb formatted string representation, right here!

       }
    }


    return rv.release();
}

//----------------------------------------------------------------------------------
static GdbDataValue* parseId(char const *&fc, char const *lc)
{
    std::auto_ptr<GdbDataValue> rv( new GdbDataValue );

    std::string &result = rv->number_;  // yeah, I know.  It's not anumber

    // right now, I'm assuming that there will be no intervening spaces
    // in the generated typename.

    while(   (fc != lc) 
          && (   isalnum(*fc) 
              || *fc == '_'
              || *fc == ':'
             )
         )
    {
       result += *fc++; 
    }

    StrTool::skipBlanks(fc,lc);

    if(fc != lc)
    {
       if(isalpha(*fc) || *fc == '_')
       {
          // we have something like 'static npos'

          return parseId(fc,lc);  // discard static and keep npos

       }
       else
       if( *fc == '<')
       {
          // something like 'static <base<class>>'

          return parseWrappedText(fc,lc);  // discard static and keep the rest
                                         
       }


    }

    return rv.release();
}

//----------------------------------------------------------------------------------
static GdbDataValue* parseComposite(char const *&fc, char const *lc)
{
    //  compositeValue          := '{' compositeMemberList '}'
    //
    //  compositeMemberList     := compositeMember [ ',' compositeMemberList ]
    //

    std::auto_ptr<GdbDataValue> rv( new GdbDataValue );

    if(   (fc != lc)
       && (*fc == '{')
      )
    {
        ++fc;
        StrTool::skipBlanks(fc,lc);

        while(   (fc != lc)
              && (*fc != '}')
             )
        {
            char const *oldFC = fc;

            rv->members_.push_back(parseCompositeMember(fc,lc) );

            if(fc == oldFC)
            {
               break;
            }

            StrTool::skipBlanks(fc,lc);

            if(fc != lc && *fc == ',')
               ++fc;

        }
    }

    if(fc != lc && *fc == '}')
      ++fc;

    if(rv->members_.empty())
    {
       rv->number_ = "Error:  expected {...}, found:";
       rv->charString_.append(fc,lc);
    }


    return rv.release();
}

//----------------------------------------------------------------------------------
static GdbDataValue* parseCompositeMember(char const *&fc, char const *lc)
{
    std::auto_ptr<GdbDataValue> rv( parseGdbValue(fc,lc) );

    StrTool::skipBlanks(fc,lc);

    if(   (fc != lc)
       && (*fc == '=')
      )
    {
       ++fc;
       rv->value_.reset(parseGdbValue(fc,lc));
    }

    if(    rv->number_.empty() 
       &&  rv->members_.empty() 
       && (rv->value_.get() == 0)
      )
    {
       rv->number_ = "Error: missing value:";
       rv->charString_.append(fc,lc);
       fc = lc;

    }

    return rv.release();
}

//----------------------------------------------------------------------------------
static GdbDataValue* parseWrappedText(char const *&fc, char const *lc)
{
    // wrappedText             := '<' ... '>'  (note that the <> pairs must be balanced)
    //

    std::auto_ptr<GdbDataValue> rv( new GdbDataValue );

    StrTool::skipBlanks(fc,lc);

    int depth=0;

    while(fc != lc)
    {
       if(*fc == '<')
       {
          rv->number_ += *fc++;
          ++depth;
       }
       else
       if(*fc == '>')
       {
          rv->number_ += *fc++;
          --depth;

          if(depth == 0)
             break;
       }
       else
          rv->number_ += *fc++;
    }

    if(rv->number_.empty())
    {
       rv->number_ = "Error:  mising <...>:";

       rv->charString_.append(fc,lc);
       fc = lc;
    }

    return rv.release();
}

//----------------------------------------------------------------------------------

static void rowify(GdbDataValue const &input, std::list<std::string> &output, std::string const &prefix, size_t maxWidth)
   ///  Convert the GdbDataValue object into rows of text in a list.
   ///  @param input  The variable to convert to a string and stores as a multi-row table output
   ///  @param output The list of output strings
   ///  @param maxWidth  A loose hint about how wide the strings should be.  May be violated.
{
   if(maxWidth > 40)
      maxWidth -= 10;



   if(input.members_.empty())
   {
       // we are dealing with a non-compound object

       if(input.isUseless())
          return;


       if(input.value_.get() && input.value_->isUseless())
       {
          return; // we have "name = value" where the value is useless  -- just ignore the whole thing
       }

       output.push_back(prefix);

       std::string & line = output.back();

       size_t oldLineSize = line.size();

       line += input.number_;

       if(!input.number_.empty() && !input.charString_.empty())
       {
           // we are dealing with a string or a character
           //
           //   0x12345 'x'
           ///  0x2468A "some string"

           if(   (line.size() < maxWidth/2) 
              && (line.size() + input.charString_.size() > maxWidth - 4)
             )
           {
               // line too long to fit on the screen in the dialog -- split it up

               std::string subPrefix;

               for(size_t i = 0; i < line.size()-1; ++i)   subPrefix += ' ';

               std::string tmp = input.charString_;

               size_t maxAvail = (maxWidth-4) - (line.size() + 1);

               if(tmp.size() >= maxAvail)
               {
                  std::string firstPart = tmp.substr(0,maxAvail);

                  tmp = tmp.substr(maxAvail);

                  line += ' ';
                  line += firstPart;

               }


               while(tmp.size() >= maxAvail)
               {
                   output.push_back(subPrefix);

                   output.back() += ' ';
                   output.back() += tmp.substr(0, maxAvail);

                   tmp = tmp.substr(maxAvail);
                  
               }

               if(tmp.size())
               {
                   output.push_back(subPrefix);

                   output.back() += ' ';

                   output.back() += tmp;
               }



           }
           else
           {

               line += ' ';
               line += input.charString_;

           }
       }
       else
       if(input.value_.get())
       {
          // we have
          //
          //  { id = something }

          line += '=';

          if(!input.value_->members_.empty())
          {
             // we are dealing with 
             //
             //   id = { nested child }

             std::string childPrefix;

             size_t newLineSize = line.size();

             if(newLineSize - oldLineSize  > 15)
             {
                newLineSize = oldLineSize + 15;
             }

             for(size_t i = 0; i < newLineSize; ++i)
                childPrefix += ' ';

             rowify(*input.value_, output, childPrefix, maxWidth);

          }
          else
          {
             line += input.value_->number_;

             if(!input.value_->number_.empty() && !input.value_->charString_.empty())
             {
               line += ' ';
               line += input.value_->charString_;
             }


          }

       }
   }
   else
   {
      // dealing with a composite object

      if(!input.isUseless())
      {

          output.push_back(prefix + "{");
    
          std::string memberPrefix = prefix + " ";
    
          CXXTLS_FOREACH(GdbDataValue* cur, input.members_)
          {
             if(    cur->value_.get()
                &&  !cur->number_.empty()
                &&  cur->number_[0] == '<'  
               )
             {
                 // the current item looks like this:
                 //  <baseClassDataMember> = something
                 // skip the base class part and just print the value
                 // but take into account the possibility that the
                 // value has a member list.  If so, print the members,
                 // not the whole gdb because printing the whole thing
                 // will cause a nesting

                 if(cur->value_->members_.empty())
                 {
                     rowify(*cur->value_.get(), output, memberPrefix, maxWidth);
                 }
                 else
                 {
                     CXXTLS_FOREACH(GdbDataValue const *sub, cur->value_->members_)
                     {
                         rowify(*sub, output, memberPrefix, maxWidth);
                     }
                 }



             }
             else
             {
                 rowify(*cur, output, memberPrefix, maxWidth);
             }

          }
    
          output.push_back(prefix + "}");

      }

   }


}

//----------------------------------------------------------------------------------

static void splitGdbVariable(std::string const &input, std::list<std::string> &output, size_t maxWidth)
{
    std::auto_ptr<GdbDataValue> result;

    char const *fc = input.data();
    char const *lc = fc + input.size();

    result.reset(parseGdbValue(fc, lc));

    if(result.get())
        rowify(*result, output, "", maxWidth);
    else
      {
        output.clear();
      }
}

//----------------------------------------------------------------------------------
void
GdbViewer::
displayGdbValue(ViewerManager* vm, std::string const &title, std::string const &valueValue)
{
    CursorWindow::Message m(title);

    std::list<std::string> valueRows;
    
    splitGdbVariable(valueValue, valueRows, vm->window()->size().col_);
    
    static std::string equals("=");
    static std::string space(" ");
    
    if(!valueRows.empty())
    {
        int count=0;
    
        std::string secondaryLinePrefix=linePrefix(1);
    
        CXXTLS_FOREACH(std::string const &cur, valueRows)
        {
           std::string totality;
           
           if(count++ == 0)
           {
              totality = cur;
           }
           else
           {
              totality = secondaryLinePrefix + space + cur;
           }
    
           m += totality;
        }
    }
    else
    {
         m += "Error:  no value found.";
         m += "";
         m += "Possible syntax error in expression?";
    }

    m.popup(vm->window());


}

//----------------------------------------------------------------------------------
void 
GdbViewer::
displayGdbVariableList(ViewerManager* vm, std::string const &title, std::string const &data)
{
    CursorWindow::Message m(title);
    
    VarObj::ParseResult parseResult = VarObj::parse(data);
    
    ArrayVarObj *parmArray = dynamic_cast<ArrayVarObj*>(parseResult.rv_);
    
    if(!parmArray)
    {
        m += std::string("error parsing frame variables:");
        m += std::string("   ") + parseResult.error_;
    }
    else
    {
        size_t count = parmArray->size();
    
        for(size_t i = 0; i < count; ++i)
        {
            std::string line("  ");
    
            VarObj *parm = (*parmArray)[i];
    
            if(parm)
            {
               VarObj *name = parm->findNestedChild("name");
               VarObj *value= parm->findNestedChild("value");
    
               if(name && value)
               {
                  std::string nameValue = name->value();
                  std::string valueValue= value->value();
    
                  std::list<std::string> valueRows;
    
                  splitGdbVariable(valueValue, valueRows, vm->window()->size().col_);
    
                  static std::string equals("=");
                  static std::string space(" ");
    
                  if(valueRows.size() == 1)
                  {
                      std::string totality = nameValue + equals + valueValue;
                      m += totality;
                  }
                  else
                  {
                      int count=0;
    
                      std::string secondaryLinePrefix=linePrefix(1 + nameValue.size());
    
                      CXXTLS_FOREACH(std::string const &cur, valueRows)
                      {
                         std::string totality;
                         
                         if(count++ == 0)
                         {
                            totality = nameValue + equals + cur;
                         }
                         else
                         {
                            totality = secondaryLinePrefix + space + cur;
                         }
    
                         m += totality;
                      }
                  }
    
               }
               else
               {
                  m += "  <missing>";
               }
    
            }
            else
            {
               m += "  <missing>";
            }
        }
    
    }
    
    
    m.popup(vm->window());
    
    delete parseResult.rv_;

}

//----------------------------------------------------------------------------------

std::string 
GdbViewer::
evaluateTmpExpression(std::string const &frameDepth, std::string const &expression)
{
    return gdb_.evaluateTmpExpression(frameDepth,expression);
}

//----------------------------------------------------------------------------------

void 
GdbViewer::
watchLocation(std::string const &frame, std::string const &expression, char const *access)
{
   gdb_.watchLocation(frame, expression, access);

   displayBreakpoints(bpItem_);

}

//----------------------------------------------------------------------------------

void 
GdbViewer::
addDisplayExpression(std::string const &frame, std::string const &expression)
{
   if(gdb_.createDisplayExpression(frame, expression))
   {
       CursorWindow::Message m("Error:");

       m += "";
       m += std::string("Error creating display expression:  ") + expression;
       m += "";
       m += "Press Enter to continue";

       m.popup(manager()->window());

   }
   else
       displayVariables();
}


//----------------------------------------------------------------------------------
void
GdbViewer::
handleVariableViewerDeath()
{
   // create new stack viewer

   variableViewer_ = new ListViewer(gdbVariableViewer_ = new GdbDisplayVariableViewer(fullname_, this), manager_);

   manager_->add(variableViewer_, fullname_);  // not made visible, just added.  Visibility comes on stop/terminate
    
}

//----------------------------------------------------------------------------------
GdbDisplayVariableViewer::
GdbDisplayVariableViewer(std::string const &executableName, GdbViewer* viewer)
:  executableName_(executableName)
,  gdbViewer_(viewer)
{
    title_ = "GdbDisplayVariables";
}


//----------------------------------------------------------------------------------
GdbDisplayVariableViewer::
~GdbDisplayVariableViewer()
{
    gdbViewer_->clearVariableViewerPointer();
}

//----------------------------------------------------------------------------------
std::string const&
GdbDisplayVariableViewer::
application_name() const
{
   return executableName_;
}

//----------------------------------------------------------------------------------
std::string
GdbDisplayVariableViewer::
description() const
{
   return std::string("Display Variables for ") + executableName_;
}

//----------------------------------------------------------------------------------
bool
GdbDisplayVariableViewer::
is_marked(size_t index)
{
    return false;
}


//----------------------------------------------------------------------------------
std::auto_ptr< std::list< std::string > > 
GdbDisplayVariableViewer::
help_text()
{
    std::auto_ptr< std::list< std::string> > rv(new std::list< std::string > );

    rv.get()->push_back("");
    rv.get()->push_back("");
    rv.get()->push_back("GDB Display Variable Viewer Help");
    rv.get()->push_back("");
    rv.get()->push_back("  The display variable viewer is a 'helper' application that is only");
    rv.get()->push_back("  useful in the context of the running GdbViewer.  It exists");
    rv.get()->push_back("  primary to assist in the display of expressions that the user");
    rv.get()->push_back("  needs printed repeatedly while single stepping, or when stopping");
    rv.get()->push_back("  at a breakpoint");
    rv.get()->push_back("");
    rv.get()->push_back("  Display variables are added in the stack viewer because the command");
    rv.get()->push_back("  to add a display variable implies a particular stack frame at which");
    rv.get()->push_back("  the variable should be display (and can only be displayed)");
    rv.get()->push_back("");
    rv.get()->push_back("Command keys:");
    rv.get()->push_back("");
    rv.get()->push_back("   F1    -- get help");
    rv.get()->push_back("   F2    -- switch viewers in a cycle fashion");
    rv.get()->push_back("   F5    -- Switch to the main gdb window and run target");
    rv.get()->push_back("   F8    -- Switch to the main gdb window and halt/kill the program");
    rv.get()->push_back("   F7    -- Switch to the main gdb window and prompt for text to be");
    rv.get()->push_back("            sent to the target stdin stream");
    rv.get()->push_back("   F9    -- Switch to the main gdb window and step the program");
    rv.get()->push_back("            skipping over function calls.");
    rv.get()->push_back("   F10   -- Switch to the main gdb window and step the program");
    rv.get()->push_back("            skipping over function calls.");
    rv.get()->push_back("   F11   -- Switch to the main gdb window and step the program");
    rv.get()->push_back("            stepping into function calls.");
    rv.get()->push_back("   F12   -- Switch to the main gdb window and step the program");
    rv.get()->push_back("            until it returns from the current function.");
    rv.get()->push_back("  ^5     -- Widen this window");
    rv.get()->push_back("  ^6     -- Increase this window's height");
    rv.get()->push_back("   b     -- Switch directly back to the GdbBreakpointViewer.");
    rv.get()->push_back("   C     -- Switch to the main gdb window and prompt for a new");
    rv.get()->push_back("            current directory for the target program");
    rv.get()->push_back("   d     -- Does nothing.  See 'u' and 'x' to delete expressions");
    rv.get()->push_back("   g     -- Switch directly back to the main GdbViewer.");
    rv.get()->push_back("   o     -- Switch directly back to the GdbTargetOutputViewer.");
    rv.get()->push_back("   s     -- Switch directly back to the GdbStackViewer.");
    rv.get()->push_back("   u     -- undisplay an expression.  Delete it.");
    rv.get()->push_back("   v     -- Switch directly back to the source code viewer.");
    rv.get()->push_back("   x     -- undisplay an expression.  Delete it.");
    rv.get()->push_back("");




    return rv;
}



//----------------------------------------------------------------------------------
static void signalDisplayVariableViewerDeath(ViewerManager*vm, std::string const &gdbviewerInstanceName)
{
    GdbViewer *v = dynamic_cast<GdbViewer*>(vm->find_viewer(gdbviewerInstanceName));

    if(v)
    {
       // found a viewer by the proper name and it is a gdb viewer

       v->handleVariableViewerDeath();
    }

    
}

//----------------------------------------------------------------------------------
int
GdbDisplayVariableViewer::
handle_event(CursorWindow::input_event const *e, 
             size_t                           currentRow,
             ViewerManager*                   vm
            )
{

    if(e->value_  == 'q' || e->value_ == 'Q' || e->value_ == 0x0b)   // 0x0b is ^K
    {
       signalDisplayVariableViewerDeath(vm, executableName_);
       return 2;   // we are done
    }
    
    if(e->type_ == CursorWindow::input_event::ForceExitKey)
    {
       signalDisplayVariableViewerDeath(vm, executableName_);
       return 2; // we are done
    }

    switch(e->value_)
    {
        case CursorWindow::key_f7:
           if(gdbViewer_)
           {
               gdbViewer_->feedStdinHelper();  // let the user enter data to be sent to the target's stdin stream
           }
           break;
        case CursorWindow::key_f8:
           if(gdbViewer_)
           {
               gdbViewer_->haltHelper();  // halt/kill the target proces
           }
           break;
        case CursorWindow::key_f5:
        case CursorWindow::key_f9:
        case CursorWindow::key_f10:
        case CursorWindow::key_f11:
        case CursorWindow::key_f12:
           if(gdbViewer_)
           {
               gdbViewer_->runHelper(e->value_);
           }
           break;

        default:
             gdbViewer_->manager()->window()->beep();      // invalid command
           break;

        case 'C':
           if(gdbViewer_)
           {
               gdbViewer_->cdHelper();  // halt/kill the target proces
           }
           break;

       case 'o':
       case 's':
       case 'b':
       case 'v':
                gdbViewer_->focusOnOtherViewer(e->value_);  // make corresponding viewer active
            break;

        case 'g':
            gdbViewer_->manager()->activate(gdbViewer_);
        break;
   
        case 'u':
        case 'x':
        {
           if(currentRow < strings_.size())
           {
               std::string const &curRow =  strings_[currentRow];

               if(curRow.empty() || ! isdigit(curRow[0]))
               {
                  vm->window()->beep();
               }
               else
               {
                  size_t colon = curRow.find_first_not_of("0123456789");

                  if(colon >= curRow.size())
                       vm->window()->beep();
                  else
                  {
                     std::vector<std::string> remainingStrings;

                     size_t i=0;

                     bool skipping = false;

                     CXXTLS_FOREACH(std::string const &cur, strings_)
                     {
                        if(i == currentRow)
                            skipping=true;
                        else
                        if(skipping)
                        {
                            if(!cur.empty() && isdigit(cur[0]))
                            {
                               skipping = false;
                            }
                        }

                        if(!skipping)
                           remainingStrings.push_back(cur);

                        ++i;
                     }

                     gdbViewer_->deleteDisplayExpression(curRow.substr(0, colon), remainingStrings);


                     return 1;  // repaint this window
                  }

               }
           }
        }
        break;
    
    }

    return 0;  // we are not done and the screen doesn't need updating

}

//----------------------------------------------------------------------------------
void
GdbDisplayVariableViewer::
resetContent(std::vector<std::string> &content, Viewer *v)
{
    ListViewer *lv = dynamic_cast<ListViewer*>(v);

    if(lv)
    {
       lv->resetContent();
    }
    

    strings_.swap(content);

    CursorWindow::viewport *vp = v->manager()->find_viewport(v);

    if(vp)
    {
       (*v)(vp, 1);  // repaint the viewer in the associated viewport.  display it as inactive
    }

}

//----------------------------------------------------------------------------------
void
GdbViewer::
displayVariables()
{
    
    size_t windowWidth = 40;  // choose a reasonable default window width

    CursorWindow::viewport* myvp = manager()->find_viewport(this);

    if(myvp)
    {
       windowWidth = myvp->size().col_;
    }


    std::vector<std::string> displayVariables;
    std::vector<std::string> formattedVariables;

    gdb_.fetchDisplayExpressionResults(displayVariables);


    // copy the displayVariables to the formattedVariables vector
    // if the size of the individual display variables fits in the window.
    // if not, the split the display variables up into rows and 
    // put the rows in the formattedVariables vector individually --
    // with leading blanks to aid in the formatting.

    CXXTLS_FOREACH(std::string const &cur, displayVariables)
    {
        size_t curSize = cur.size();

        if(curSize > windowWidth)
        {
            size_t equalSign = cur.find_first_of('=');

            if(equalSign < curSize)
            {
               std::string actualValue = cur.substr(equalSign+1);
               std::string prefix      = cur.substr(0, equalSign);

               std::string blanks;

               CXXTLS_FOREACH(char c, prefix)
               {
                  blanks += ' ';

                  c = c; // suppress unused variable warning
               }

               char const *fc = actualValue.data();
               char const *lc = fc + actualValue.size();

               std::auto_ptr<GdbDataValue> parsedDataStructure( parseGdbValue(fc,lc) );
               
               if(parsedDataStructure.get())
               {
                   std::list<std::string> reformattedForm;

                   rowify(*parsedDataStructure, reformattedForm, "", windowWidth-blanks.size());

                   bool firstLine = true;

                   CXXTLS_FOREACH(std::string const &subExpression, reformattedForm)
                   {
                      if(firstLine)
                         formattedVariables.push_back( prefix + subExpression );
                      else
                         formattedVariables.push_back( blanks + subExpression);

                      firstLine=0;
                   }
                  
               }
               else
                  formattedVariables.push_back(cur);
            }
            else
            {
                formattedVariables.push_back(cur);
            }

        }
        else
        {
           formattedVariables.push_back(cur); // just pass short lines through
        }
    }
    

    
    gdbVariableViewer_->resetContent(formattedVariables, variableViewer_);

}

//----------------------------------------------------------------------------------
void
GdbViewer::
runHelper(int keycode)
{
    switch(keycode)
    {
        case CursorWindow::key_f5:
        case CursorWindow::key_f10:
        case CursorWindow::key_f9:
        case CursorWindow::key_f11:
        case CursorWindow::key_f12:
        {
           manager()->activate(this);

           CursorWindow::viewport *vp = manager()->find_viewport(this);

           if(vp)
           {
              run(vp, keycode);
           }
        }
        break;


        default:
           manager()->window()->beep();
    }
}
//----------------------------------------------------------------------------------
void 
GdbViewer::
focusOnOtherViewer(int keycode)
{
    Viewer* vp=0;

    switch(keycode)
    {
        default:
            break;

        case 's':  vp = stackViewer_;              break;
        case 'b':  vp = breakpointViewer_;         break;
        case 'd':  vp = variableViewer_;           break;

        case 'o':  

             if(attachMode_)
                return;  // no can focus lon target output log viewer in attatchment mode

             vp = logViewer_;                
             break;

        case 'v': 

              CursorWindow::viewport *v = manager()->find_viewport(this);

              if(v)
              {
                  showCurrentFile(v, showFilename_, showLine_); // and all other interesting windows

                  vp = manager()->find_viewer(showFilename_);
              }
    }

    if(vp)
    {
        manager()->activate(vp);
    }

}

//----------------------------------------------------------------------------------

void 
GdbViewer::
performHalt()
{
           if(gdb_.state() == GdbController::running)
           {
              if(gdb_.halt())
              {
                  CursorWindow::Message m("Error");
                  m += "";
                  m += "  Unable to interrupt the process";
                  m += "";
                  m += "Press Enter to continue";

                  m.popup( manager()->window() );

              }
           }
           else
           {
              if(gdb_.kill())
              {
                  CursorWindow::Message m("Error");
                  m += "";
                  m += "  Unable to kill the process";
                  m += "";
                  m += "Press Enter to continue";

                  m.popup( manager()->window() );

              }
           }
}

//----------------------------------------------------------------------------------
void
GdbViewer::
changeDirectory()
{
            std::string curdir = gdb_.getcwd();

            if(curdir.empty())
            {
               CursorWindow::Message m("Error:  unable to get the debugger's current directory");
               m += "";

               if(    gdb_.state() != GdbController::stopped
                  &&  gdb_.state() != GdbController::terminated
                 )
               {
                  m += "   Try running to main first";
                  m += "";
               }

               m += "Press Enter to continue";
               m += "";

               m.popup(manager()->window());
            }
            else
            {
               CursorWindow::Dialog  d("Specify new directory for debugged target process.");

               d += new CursorWindow::Dialog::String("path", "Debugging Current Directory", curdir, 40);

               if(!d.popup(manager()->window()))
               {
                   if(gdb_.setcwd(d.element_value("path")))
                   {
                       CursorWindow::Message m("Error:  unable to set the debugger's current directory");
                       m += "";
                       m += "Press Enter to continue";
                       m += "";
                       
                       m.popup(manager()->window());
                      
                   }
                   else
                   if(!firstStart_)
                   {
                      // cur dir set correctly.

                      CursorWindow::Message m("Change will take affect ONLY at the beginning of the next program run.");
                      m += "";
                      m += "    Use F8 to halt the program followed by";
                      m += "    F5 to start a new run with the change in place.";

                      m.popup(manager()->window());


                   }
               }


            }
}

//----------------------------------------------------------------------------------
void
GdbViewer::
cdHelper()
{
    manager()->activate(this);
    changeDirectory();
}
//----------------------------------------------------------------------------------
void
GdbViewer::
haltHelper()
{
    manager()->activate(this);
    performHalt();
}

//----------------------------------------------------------------------------------
bool
GdbViewer::
is_dirty() const
{
   GdbViewer* tmp = const_cast<GdbViewer*>(this);   // state function is safe to call

   switch(tmp->gdb_.state())
   {
       case GdbController::running:
       case GdbController::stopped:
           return true;

       default:
           break;
   }

   return false;
}
//----------------------------------------------------------------------------------
void
GdbViewer::
setSaveFile(std::string const &sf)
{
   gdb_.setSaveFile(sf);
}
//----------------------------------------------------------------------------------

bool
GdbViewer::
deleteDisplayExpression(std::string const &expressionNumber, std::vector<std::string> &remainingStrings)
{
    gdbVariableViewer_->resetContent(remainingStrings, variableViewer_);

    return gdb_.deleteDisplayExpression(expressionNumber);
  
}

//----------------------------------------------------------------------------------
} // namespace cxxtls
#endif
