#ifndef processviewer_included_h
#define processviewer_included_h
//
// Copyright 2013, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.
//
// Warning:  not all files in this directory structure are covered by the
// same copyright.  Some of them are part of the GNU source distribution
// and you must obey the GPL copyright for those files.

///@file
///
/// This file defines the interface to the ProcessViewer class.

#include <cxxtls/viewer.h>
#include <memory>
#include <cxxtls/thread.h>
#include <cxxtls/file.h>
#include <cxxtls/deltaviewer.h>


namespace cxxtls
{

class ControlledProcess;

class ProcessViewer
: public Viewer
  ///  A ProcessViewer is a viewer that allows the user interact with
  ///  a child process.  The process viewer creates a thread that 
  ///  launches a child process and feeds characters into its input
  ///  stream and displays output from its output stream.  stdout and 
  ///  stderr should be merged in the command invocation to the process
  ///  to prevent child process from messing up the screen.
  ///
  ///  @Warning:  I meant this to be a base class for a controlled processes
  ///  but, right now, it is highly specific to the gdb executable and its
  ///  behavior of debugging a child process.
  ///
  ///  The ProcessViewer will only pass basic characters to the child
  ///  process.  These will include normal printable characters, unused
  ///  control character sequences, tabs, and the enter key.  That is,
  ///  characters normally consumed by the CXX tookit for itself will not
  ///  be sent to the child process.  Further, escape sequences (such as
  ///  function keys) will not be sent to the child process.
  ///
  ///  Programs that run as child processes should not use curses, ncurses,
  ///  or ANSI characters sequences that affect the terminal in non-trivial
  ///  ways.  Tab keys and newline characters will be interpreted in the
  ///  program output.  Most other non-printable characters will be ignored.
  ///
  ///  This class is designed, first and foremost, as a mechanism for driving
  ///  gdb inside a window.  Don't over-extend the metaphor that it provides.
  ///  The user is allowed to issue commands to the controlled process (for example
  ///  gdb debugging of some child program).  The controlled program is expected to
  ///  produce a highly unique sequence of characters as a prompt to the user.  
  ///  Gdb, for example, normally produces "(gdb) " as a prompt.  However, it does
  ///  allow you to force the prompt, so we force it to be "(<gdb>) " to make it
  ///  be even more unique.  You can drive the bourne shell from this viewer, but
  ///  you cannot use 'bash -i' because the -i tells bash to run interactively and
  ///  it starts trying to control the terminal using fcnt() and termios() calls
  ///  which fail and causes bash to die.
  ///
  ///  The ProcessViewer looks for the special command prompts designed for it ( "(<gdb>) " )
  ///  and takes action based on the text between most recent two instances.  In gdb's case,
  ///  the behavior is to look for file names and line numbers to show in another window.
  ///  
{
    ViewerManager *vm_;
    std::string    cmdLine_;

    std::auto_ptr<ControlledProcess> process_;

    typedef std::pair<std::string, bool>  LineInfo;  // text of a line a flag indicating that it contains '(<gdb>) '.

    std::vector<LineInfo>  outputLines_;

    CursorWindow::row_col curpos_;  // current cursor position

    enum constants
    {
        TITLE_REGION_HEIGHT = 1,     // where title of window goes (1 line)
        CMD_REGION_HEIGHT   = 1,     // where user input of cmds go (1 line)
    };

    size_t topVisibleLine_;   // index into outputLines
    size_t leftOffset_;       // offset from beginning of rows to leftmost visible char
    int    repaintMode_;      // repaint mode


    void addLineInfo(std::string const &li, bool commandLine=false)
        // add a line to the end of the list of lines produced by the controlled program (and merged with user command text).
    {
       outputLines_.push_back(LineInfo(li, commandLine));
    }

    std::string &lastLine()
        // get a reference to the text of the last line produced by the program (and possibly merged with commands the user typed)
    {
        return outputLines_.back().first; 
    }

    void addOutputCharacter(char c)
       // a single character to the end of the output from the controlled program
    {
       if(outputLines_.empty())
           addLineInfo("");

       lastLine() += c;
    }

    void markOutputLineAsCommand()
        // call this function when mostRecentCharsIsCommandPrompt() returns tree.
    {
        if(!outputLines_.empty())
          outputLines_.back().second = true;
    }


    int  pageHeight(CursorWindow::viewport *vp); 
                                           // return height of window in lines

    int  visibleLines();    // return count of lines actually visible in window
    
    int  usableLines(CursorWindow::viewport *vp)  // size where process output data gets printed
    {
       return pageHeight(vp) - TITLE_REGION_HEIGHT - CMD_REGION_HEIGHT;
    }

    void paintPage(CursorWindow::viewport *vp);         // repaint entire window
    void paintTitleRegion(CursorWindow::viewport *vp);  // top line
    void paintCmdRegion(CursorWindow::viewport *vp);    // bottom line
    void paintOutputRegion(CursorWindow::viewport *vp); // process output region
    void addOutputRegionCharacter(CursorWindow::viewport *vp, char c, bool interpretCharacter=true);  // interpretCharacter=false is used for printing the screen only

    void pageDown(CursorWindow::viewport *);                   // adjust topVisibleLine_ and repaint
    void pageUp(CursorWindow::viewport *);                     // adjust topVisibleLine_ and repaint
    void lineDown(CursorWindow::viewport *);                   // adjust topVisibleLine_ and repaint
    void lineUp(CursorWindow::viewport *);                     // adjust topVisibleLine_ and repaint
    void slideLeft(CursorWindow::viewport *, size_t count);    // adjust leftOffset_ and repaint
    void slideRight(CursorWindow::viewport *, size_t count);   // adjust leftOffset and repaint

    ThreadQueue<char>  stdout_;      // the contents that have been read
                                     // by the consumeSlaveData() thread
                                     // and stored here.

     Mutex threadDataGuard_;
     bool  threadRunning_;
     bool  threadComplete_;

     std::string inputBuffer_;

     char mostRecentChars_[8];  // "(<gdb>) ".  Must match mostRecentCharsIsCommandPrompt() function, below

     bool handlingGdbEmacsAnnotation_;                         // when you executed the gdb command, set annotate 1, you change the output format for single step and continue.
     char prevOutputCharacter_;                                // we use this var to detect the arrival of gdb emacs annotations.

     std::string gdbEmacsAnnotation_;                          // buffer holding gdb emacs annotations while in that mode.


     std::list<DeltaViewer::NameValuePair>  displayVariables_; // the current list of display variables.

     std::string deltaViewerName_;                             // the name of the companion viewer, if any.  empty string means none.
     bool running_;                                            // true if waiting for a gdb command prompt to re-appear


     void pushMRC(char c)
         // slide mostRecentChars_ left by one byte and put c at the end
     {
       for(size_t i = 0; i < sizeof(mostRecentChars_)-1; ++i)
          mostRecentChars_[i] = mostRecentChars_[i+1];

       mostRecentChars_[sizeof(mostRecentChars_)-1] = c;
     }

     bool mostRecentCharsIsCommandPrompt()
     {
         char const *p = &mostRecentChars_[0];

         char const *e = &mostRecentChars_[sizeof(mostRecentChars_)];

         if(*p == '(')  ++p;
         if(*p == '<')  ++p;
         if(*p == 'g')  ++p;
         if(*p == 'd')  ++p;
         if(*p == 'b')  ++p;
         if(*p == '>')  ++p;
         if(*p == ')')  ++p;
         if(*p == ' ')  ++p;

         return p == e;
     }



     typedef std::pair<size_t, size_t>  LineRange;

     LineRange  findMostRecentCommandRange(); 
         ///<   Return the range of lines that encloses the output
         ///    of the most recent command send to the controlled
         ///    process.


     void narrowLineRange(LineRange &r);
         ///<   Reduce the specified line range to the tail end components
         ///    (just before the trailing "(<gdb>)") that are actually gdb commands outputs.
         ///    That is, skip the leading components that are not valid gdb command
         ///    outputs.  
         ///  
         ///    The lines so deleted will include program output as well as commands
         ///    to gdb and also ignored gdb output. 
         ///
         ///    This function does its job by working backwards and looking for
         ///    lines like this:
         ///    @li  Breakpoint, ...
         ///    @li  Breakpoit ....
         ///    @li  <integer> ....
         ///    @li  <integer> ....


      void interpretCommandRegion(CursorWindow::viewport *vp);
          ///<  Interpret the most recent command line range so that
          ///   we can update the position within any files displayed
          ///   to show us the current line number being debugged.

      bool parseAndInterpretLineRange(CursorWindow::viewport *vp, LineRange &range);
          ///<  Consume lines out of the specified line range and
          ///   update the currentFile_ and currentLine_.  Produce any
          ///   needed error messages.
          ///
          ///   @return true if the currentFile_ or currentLine_ have been
          ///    modified by the current command line fragment.

      void showCurrentFile();
          ///<  Open another window with the currentFile_ and currentLine_ visible and
          ///   highlighted.

      void updateDisplayVariablesCompanionWindow();
          ///<  Manage the display variable's companion window whenever display variables
          ///   have changed.

      void addDisplayVariable(std::string const &line);
          ///<  Add the gdb output text associated with a display variable line.  It looks
          ///   like this:
          ///
          ///@code
          ///
          ///    1:   expression = value
          ///
          ///@endcode


public:

  ProcessViewer(ViewerManager *vm, std::string const &cmdLine);
      ///<  Create the process viewer and associate it with the
      ///   specified viewer manager.
      ///
      ///  The cmdLine parameter is the command that will be
      ///  invoked in a child thread to create the child process.
      ///
      ///  The command line should do the following:
      ///
      ///  @li  Define a valid command to be executed
      ///
      ///  @li  Either merge stderr with stdin (using "2>&1") or
      ///       discard stderr (using "2>&1") or send it to some
      ///       disk file (using "2>/path/file").
      ///
      ///  @li  Make sure that curses terminal sequences are not
      ///       sent to the screen -- they will mess up the 
      ///       output.  This can be be done in several ways --
      ///       or it might be not possible.  You could invoke a
      ///       script and in the script, "export TERM=dumb" before
      ///       invoking a program.  On the other hand, many programs
      ///       won't require any effort at all.
      ///
      ///  To invoke a bourne shell command, you would use a command
      ///  line like the following:
      ///
      ///@code
      ///
      ///  std::string cmdline="TERM=dump /bin/bash -i 2>&1"; 
      ///
      ///@endcode

 ~ProcessViewer();

  virtual void operator() ( CursorWindow::viewport *     vp,
                            int                          cmd
                          );
     ///< Handle screen repaint commands.
     ///
     ///  @param vp
     ///  @param cmd

 
  virtual BackgroundProcessStatus getBackgroundData(CursorWindow::input_event &e);
     ///<  Get any pending background process data and store it in the 
     ///   specified event.  Return one of the enumeration values in
     ///   BackgroundProcessStatus to indicate what has happened.
     ///
     ///   In the event that only a single character is returned, 
     ///   e.type_ will be set to CursorWindow::input_event::ProcessOutput.
     ///   
     ///   If there are multiple characters waiting, e.type_ will be set to
     ///   CursorWindow::input_event::BackgroundData
     ///
     ///   @param e  An initialized but not populated cursor window event.
     ///   @return   One of the BackgroundProcessStatus values.
     ///


  virtual bool handle_event( CursorWindow::input_event const * event,
                             CursorWindow::viewport          * viewport
                           );
     ///< Handle input events.
     ///
     /// @param event 
     /// @param viewport

  virtual void help();  ///< Present the user with help information that is
                        ///  specific to this viewer's operation and contains
                        ///  information about system keep bindings, etc.

  static Viewer* app(ViewerManager*vm, std::string &fullname);
    ///<
    ///  Application "hook" function for the ProcessViewer class.
    ///  Use it as a parameter to add_app, like this:
    ///
    ///  @code
    ///    vm.add_app(ProcessViewer::app, "ProcessViewer");
    ///  @endcode

  static std::string app_name;
    ///<
    ///  The name to prompt the user with in the ViewerManager::appChooser
    ///  menu.
    ///

  std::string const& application_name() const { return app_name; }
    ///<
    ///  Return the name of the application to use when describing this
    ///  viewer in the ViewerManager::appChooser or in the
    ///  ViewerManager::ViewerSelector.
    ///

  std::string description() const;
    ///<
    ///  Returns a description of this session.
    ///

  static void*  consumeSlaveData(void *parm);
    ///< The function that pulls data from the slave process and stores it
    ///  in memory for the process viewer to display
    ///

  ThreadQueue<char>& stdout() { return stdout_; }
      ///<  The output queue from the slave process that is being displayed
      ///   by this ProcessViewer

  void startThread();
      ///<  Start the thread running so that it can consume data from the child
      ///   process and store it in the stdout() buffer.  This function is
      ///   called by ViewerManager::registerProcessViewer();

  void setThreadComplete();
     ///<  Set by the thread when it has consumed all the data.

  bool threadRunning();
      ///<  Returns true if the thread is actively running.

  bool threadComplete();
      ///<  Returns true if the thread has finished its work.


  void kill();
      ///<  kill the process in preparation from ProcessViewer destruction

private:
  FileName currentFile_;
  size_t   currentLine_;
  FileName startDir_;


};
} // namespace cxxtls
#endif
