#ifndef CURSOR_WINDOW_H_INCLUDED
#define CURSOR_WINDOW_H_INCLUDED


//
// Copyright 2002, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.

///@file
///
///  This file defines the class, CursorWindow, which provides an interface
///  to curses style windows -- including sub-windows, called CursorWindow::viewport's.
///  The viewports form a mechanism for writing text windowing applications.
///  Before directly using any of the classes in this file, you should consider
///  reading text_windows.
/// 
///  OK, so this is going out of style, but this particular implementation should
///  at least eliminate the problems with function key incompatibilities and
///  problems using rsh to remotely launch curses applications.  It tries anyway.

#include <string>
#include <vector>
#include <deque>
#include <string.h>
#include <list>
#include <map>
#include <cxxtls/streamable.h>
#include <memory>

namespace cxxtls
{

class Viewer; // see viewer.h

class CursorWindow
  ///
  /// A CursorWindow is a rectangular grid of character cells
  /// which can be written into, read from, etc.  It is used for
  /// interacting with human beings inside of xterm and console
  /// windows.  A CursorWindow represents the entire xterm window or screen
  /// window on NT.  If you want to use subwindows, use a
  /// CursorWindow::viewport.  There are some builtin popup
  /// windows, and you can make more of your own using viewports.  See
  /// CursorWindow::Dialog and CursorWindow::Selection.
  ///
  /// Before directly using this class, you should consider
  /// reading @ref text_windows or @ref CXXCURSORWINDOWS
  ///
  /// Note!  when debugging programs using viewers, you must start the program
  /// first then attach to it using the debugger's 'attach processid' feature
  /// (or dbx -debug pdi feature).  To debug the startup code, before the first
  /// keyboard input wait, set the environment variable, WAIT_FOR_DEBUGGER=1
  /// before running the program.  When the debugger attaches to the running
  /// process, you will be in a wait loop.  Use the debugger's print mechanism
  /// or the assign mechanism to change the stop variable to 0.  Then you can
  /// continue debugging as normal.
  ///
  /// A CursorWindow is a thin wrapper around some commonly used
  /// "curses" functions but does attempt to include them all.
  /// On NT, these functions are simulated as well.  The kinds of
  /// textual decorations are limited.
  ///
  /// You can only have one 'opened' cursor window at a time.  If you need
  /// multiple CursorWindow objects, make sure only one is open at a time.
  /// The close() method and the open() method can be used to control this.
  ///
  /// A CursorWindow should be viewed as a simplistic curses window --
  /// it is by no means a sophisticated one.  Here are the
  /// limitations:
  ///
  ///   1 foreground color and 1 background color defined for the
  ///   whole window.  It is possible to have reversed video
  ///   characters, however.
  ///
  ///   individual character cells can be normal, reverse video,
  ///   underlined, or blinking -- but not combinations thereof.
  ///
  ///   the upper left if 0,0.
  ///
  ///   writing into the window moves the cursor.
  ///
  ///   it is possible to read input from the user or to read
  ///   character cell contents (or their attributes).
  ///
  ///   user input can contain normal characters, control characters,
  ///   function keys, arrow keys, paging keys, screen resize flags,
  ///   etc.  It is not possible to detect modifier keys separately
  ///   from the data keys they affect.
  ///
  ///   no alternate character sets are supported and the exact
  ///   character set used is dependent on the platform and other
  ///   factors not under program control.
  ///
  /// As stated earlier, a CursorWindow represents an entire xterm
  /// screen -- or console window on NT.  If you want to access
  /// subwindows, use a CursorWindow::viewport.  See the nested class
  /// below.
  ///
  /// CursorWindow has some important nested types:
  ///
  /// @code
  ///  CursorWindow::viewport  -- A subwindow on top of a CursorWindow.
  ///                                 If you need popup windows or window
  ///                                 partitions, use these.
  ///
  ///  CursorWindow::Dialog    -- An input form which can be popped
  ///                                 up in a subwindow as needed to collect
  ///                                 information from the user.  You can
  ///                                 prompt for a Enter key press (Ok field),
  ///                                 string inputs (String field), etc.
  ///
  ///  CursorWindow::Selection -- An input form that lets the user select
  ///                                 one of a list of items.
  ///
  /// @endcode
  ///
  /// Associated with all cursor windows, there is a static map of keys to
  /// edit functions.  See func -- which can be used as an array of key values which
  /// map to edit functions one might need in dealing with Dialog's.  You can
  /// change the key mappings, add new ones, or use the map for your own purposes
  /// so long as you don't eliminate key mappings needed to fill out dialogs.
  ///
  /// The Esc key has an annoying limitation.  Whenever you press it, you must
  /// wait 1 second (1 whole second) before pressing any other keys or the
  /// Esc key will be lost.  This is due to unix wierdness with key mappings.
  /// On NT, this wait is not needed.
  ///
  /// There is also a std::string of word characters which.  You can use the
  /// std::find_first_of() algorithm with this string to find words in strings.
  ///
  /// On unix, the CursorWindow blocks and unblocks the following signals:
  ///
  ///   \li SIGALRM
  ///   \li SIGWINCH
  ///
  /// Note also that if you use other signals it is possible that their occurrence
  /// will damage the screen layout though the curses calls to minimize this are
  /// used.
  ///
  /// Note that on unix, with a true curses implementation of this functionality,
  /// the TERMINFO used will not be the default terminfo.  Instead, a directory will
  /// be created in the user's home directory that contains both the source for the
  /// terminfo actually used and the compiled terminfo.  In this way, you can edit
  /// the source and manually recompile the terminfo once it is automatically
  /// created for you.  See "$HOME/.tools_home".  This directory is created the
  /// first time you run any program using this curses interface.  If you have
  /// trouble with the function keys, you can edit the xterm.info files and recompile
  /// using "/usr/bin/tic".
  ///
  /// CursorWindow derives from Ostreamable -- which means that can be used
  /// as the target of operator<< for a small subset of the stream output
  /// operations.  See Ostreamable for more details.
  //
: public Ostreamable<CursorWindow>
{
    class   Window;
    friend class SaveRestoreActiveWindow;

public:

    static int normal_att;        ///<  normal text color
    static int active_title_att;  ///<  window title bar color
    static int inactive_title_att;///<  window title bar color
    static int active_mark_att;   ///<  highlighted (selected) and marked lines
    static int inactive_mark_att; ///<  marked but not highlighted (not selected)
    static int bottom_att;        ///<  bottom line of window highlights
    static int highlighted_att;   ///<  highlighted (selected) but not marked
    static int dialog_normal_att; ///<  body of dialogs
    static int dialog_title_att;  ///<  title of dialogs

    class    viewport;

    friend class viewport;

    struct row_col
      //
      /// A pair of integers representing a row and a column.
      //
    {
      int row_;
      int col_;

      row_col(int row, int col) ///  construct a cursor position object
      : row_(row),
        col_(col)
      {
      }

      row_col()                 ///  default constructor
      : row_(0),
        col_(0)
      {
      }

      bool operator==(row_col const &r)  ///  compare to equal
      {
        return row_ == r.row_ && col_ == r.col_;
      }

    };

    char const *error() const;
      
      ///<  Get a string representation of the most recent error
      ///  if any.  If no error has occurred, then null is returned.
      ///  All methods first clear the error condition and set it
      ///  only if there is an error.

    void set_error(char const *);
      
      ///< Set a new error string.  The window does not take ownership
      /// of the error string.

    void clear_error() { set_error(0); }
      
      ///< Clear out the current error flag.  Doesn't delete the string.

    CursorWindow();
      
      ///< Construct a cursor window but do not open it.

   ~CursorWindow();
      
      ///<  Close the cursor window if it is open and then
      ///   destruct it.

    bool open();

      ///< Attempt to open the cursor window -- return true on
      ///  failure.  If possible, save the current terminal
      ///  screen contents.  Note that this only opens the
      ///  curses terminal window, it does not affect any data
      ///  structures.  See resize_viewports() below.

    void close();
      ///< Close the cursor window if it is open.  Repaint the
      ///  terminal screen with the saved contents if possible.
      ///  Note that this does not damage any data structures -- it
      ///  just closes the window and sets the flag.  See resize_viewports()
      ///  below.

    void resize_viewports(row_col const &old_size, row_col const &new_size);
      ///< After you have manually closed and then reopened the terminal
      /// window usinging open() and close(), above, you should call
      /// resize_viewports() to re-layout the viepworts given any screen
      /// sized changes that occurred while the window was closed.  This means
      /// that before closing, you should save the old screen size and then
      /// inquire it again after the open.

    void repaint_all();
      ///<  repaint everything.

    struct input_event
      /// An input_event is _any_ kind of input from the user,
      /// including keys, screen resizes, etc.  If NoKey is the
      /// type, then the event should be ignored.  The input_event
      /// objects can also store background data being sent to
      /// a cursor window from a different thread.  Only one thread
      /// is allowed to interact with the curses package.  So, the
      /// ViewerManager::run() function, which reads input events
      /// from the user and sends them to the active window must also
      /// be able to handle the reading of background data being
      /// sent to the viewer.  It sends said background data using
      /// the input_event data structure.  Viewer objects, which define
      /// the handle_event() method that inteprets the event objects
      /// must also provide a mechanism for detecting background data
      /// so as to be sent to the handle_event function by the run()
      /// function.  The threads.h file defines the ThreadQueue class
      /// which is wrapper around a mutex locked std::deque&lt;char&gt;.  
      /// This is the natural way to send data to the viewer --
      /// you have the implementation of the Viewer populate a
      /// ThreadQueue with data and the run() function can call
      /// ThreadQueue&lt;char&gt;::trySwapAll() to get the waiting data.
      /// It could also call tryPop(), but that would only get you
      /// one character at time.  The ProcessOutput type_constants
      /// value means 1 key read from a background process,  and
      /// BackgroundData means a deque thereof.
      ///
      /// Keys are categorized in to one of the type_constants
      /// described below.  Further categorization of keys can
      /// be done using the CursorWindow::func -- which is a
      /// map of key values into logical rather than physical
      /// function keys.  For example, there is a physical
      /// F1 function key, which is mapped to CursorWindow::key_f1.
      /// However, what does F1 mean?  By default, it has no
      /// mapping to any edit function that would be used by a
      /// CursorWindow::dialog object.  You could map key_f1 to
      /// something in your program using a switch statement --
      /// or you could add a mapping between f1 and some standard
      /// edit function -- anything in the edit_func_names list
      /// or anything you add to that list.
    {
      enum type_constants ///  Symbolic names for event types
      {
        NoKey=0,          ///<  No key or other event actually read
        DataKey,          ///<  Normal ascii key read
        FunctionKey,      ///<  Function key read
        ResizeKey,        ///<  Window resized -- use size() to get info about the new
                          ///<  window size.  Note that you will never get this event
                          ///<  if you use a resize_handler functor.  See CursorWindow::resize_handler.
        ForceExitKey,     ///<  exit the viewer and do not ask any stupid questions
        ProcessOutput,    ///<  value_ contains a byte read from some non-keyboard process
        MouseEvent,       ///<  Not supported, but mouse events if they were
        BackgroundData,   ///<  some form of background data event.  See backgroundData_ below
      };

      int type_;   ///< Data key, function key, resize

      int value_;  ///< Key value, mouse state, or background data type.  Key values are unique 
                   ///  across function and data keys.  For mouse events, value is the 
                   ///  cursesinterface.h 'mouse_info::state_'  and reflects a bit mask of the 
                   ///  left, middle, right buttons are down flags plus '@' -- so as to give a 
                   ///  valid ascii char that is readable.
                   ///
                   ///  In BackgroundData events, the value_ is the sub-type of data of interest.
                   ///  the actual data can be found in backgroundData_ and can be of any
                   ///  size.
                   

      int row_;    ///< valid only on mouse vents
      int col_;    ///< valid only on mouse events

      std::deque<char>  backgroundData_;  ///<  Data 

      enum buttons
      {
        left  = 1,
        right = 2,
        middle= 4 // powers of two...
      };

      input_event(int type, int value, int row=0, int col=0)   ///  construct an event
      : type_(type),
        value_(value),
        row_(row),
        col_(col)
      {
      }

      input_event()                      ///  default constructor
      : type_(NoKey),
        value_(0),
        row_(0),
        col_(0)
      {
      }

      void clear()
      {
          type_  = NoKey;
          value_ = 0;
          row_   = 0;
          col_   = 0;
          backgroundData_.clear();
      }

    };

    bool kbhit();
        ///<  Return true if there are any input events to read using
        ///   the read_input() method

    input_event read_input();
        ///< Attempt to read an input event.  If wait is true, then
        ///  wait indefinitely for such an event.  If wait is false,
        ///  wait about 1/10 of a second before return a NoKey event
        ///  if there are no keys available.  Note that this call
        ///  also flushes any pending output to the screen so that
        ///  you can see it before pressing any keys.  See CursorWindow::
        ///  refresh() or viewport::refresh().   If the screen is
        ///  resized, you may or may not receive an input_event --
        ///  based on whether a resize handler has been registered
        ///  with the window.
        /// 
        ///  Note that on unix, this function uses SIGALRM, and it
        ///  does not save/restore the alarm timer value or signal handler.


    row_col curpos() const;
      ///< Get the current cursor position

    void set_curpos(int row, int col)
      ///< Move the cursor to the specified row and column.
    {
      set_curpos( row_col(row, col) );
    }

    void set_curpos(row_col rc);
      ///< Move the cursor to the specified row and column

    row_col size() const;

      ///< Get the width in columns and the height in rows
      /// of the actual window.  Note that you cannot force
      /// this information but must live with the size
      /// you get.

    void beep();
      ///< Produce an audible alarm.

    void write(char const *, size_t length);

      ///< Write a specified number of characters to the current cursor
      /// position and move the cursor.

    void write(char const *s)
      ///
      /// Write a string in the current text attributes at
      /// the current cursor position and move the cursor.
    {
       write(s, strlen(s));
    }

    void write(std::string const &s, size_t length=1000000)
      //
      /// Ditto
    {
      size_t actual_length = s.size();

      write(s.data(), actual_length < length ? actual_length : length );
    }

    void write(long c, int length=1);

      ///< Write specified number of characters to the window in the
      ///  default text attribute at the current cursor position.  The
      ///  default repeat count is 1.


    void fill_to_eol(char c=' ');
      ///< Fill the rest of the current line, starting at the current
      ///  cursor position, with a specified character -- defaulting to space.

    void box(int width, 
             int height, 
             bool filled,
             char const *title_string, 
             unsigned length
            );
      ///< Draw a box starting at the current cursor position that is
      ///  width columns wide and height columns tall (running down from
      ///  the current cursor position.  
      ///
      /// If the filled flag is true, the box will be filled with blanks
      /// before the border is drawn.
      ///
      /// If the string is not null, and the length is not 0, the string
      /// will be drawn inside the top border of the box to indicate that
      /// it is the 'title' of the box

    void box(int width, int height, bool filled, char const *title=0)
      ///
      /// Draw a box starting at the current cursor position that is
      /// width columns wide and height columns tall (running down from
      /// the current cursor position.  
      ///
      /// If the filled flag is true, the box will be filled with blanks
      /// before the border is drawn.
      ///
      /// Draw the title string, if specified, inside the top border of the
      /// box to serve as a title.
    {
      box(width, height, filled, title, title ? strlen(title): 0);
    }

    void box(int width, int height, bool filled, std::string title)
      //
      /// Draw a box starting at the current cursor position that is
      /// width columns wide and height columns tall (running down from
      /// the current cursor position.  
      ///
      /// If the filled flag is true, the box will be filled with blanks
      /// before the border is drawn.
      ///
      /// Draw the title string, if specified, inside the top border of the
      /// box to serve as a title.
    {
      box(width, height, filled, title.data(), title.size());
    }


    void fill_to_eos(char c=' ');
      ///< Fill the rest of the current screen, starting at the current
      ///  cursor position, with a specified character -- defaulting to space.

    enum text_attributes
      //
      /// Names for text attribute values.  Note that they
      /// are mutually exclusive.
    {
      normal,       ///<  normal text with decoration
      reversed,     ///<  reversed video
      underlined,   ///<  normal text + underlining
      blinking,     ///<  normal text + blinking
      bold,         ///<  normal text + bold
      reverse_ul,   ///<  reverse video + underlined
      reverse_bold, ///<  reverse video + bold
      rev_ul_bold,  ///<  reverse video + bold +underlined
      bold_ul       ///<  bold + underlined
    };

    int text_attribute() const;
      ///<  Get the current text attribute

    void set_text_attribute(int);
      ///<  Set the current text attribute


    enum key_names ///  the names of function keys
    {
       first_function_key=0x100,
       //< first the unshifted function keys
       key_f1,  key_f2,  key_f3, key_f4, key_f5, key_f6, key_f7, key_f8, key_f9, key_f10,
       key_f11, key_f12,
       key_up, key_down, key_left, key_right, key_home, key_end, key_prior, key_next,
       key_ic,  //< Insert key
       key_dc,  //< Delete key
       key_bs,  //< Backspace key
       key_btab,
       ///<  next the shifted function keys -- which might be alt or ctrl function
       /// keys depending on the implementation -- but only one form is supported
       /// here -- and further shifted versions may (like the function keys themselves)
       /// may not be available
       key_sf1,  key_sf2,  key_sf3, key_sf4, key_sf5, key_sf6, key_sf7, key_sf8, key_sf9, key_sf10,
       key_sf11, key_sf12,
       key_sup, key_sdown, key_sleft, key_sright, key_shome, key_send, key_sprior, key_snext,
       key_sic, key_sdc,

       last_function_key  ///  don't use this for anything

    };

    static std::string key_name(int keyvalue);

    enum edit_function_names
      /// Symbolic names for the normal text edit functions you might need
      /// to perform in a dialog window.  Associated with a CursorWindow
      /// there is a mapping between key values (integers) and edit function
      /// enumeraion values.  You can add your own functions to the window's
      /// edit function map -- but dialog handler won't use them for anything.
      /// To add your own functions, use numeric identifiers bigger than
      /// 'user_function_0'.  There can be only intepretation of a given
      /// curses key value, but more than one key may map to the same
      /// edit function.
      ///
      /// If you add new items to the edit_function_names list, make sure
      /// that you add mappings and names for those functions and in cursorwindow.c.
      /// See the CursorWindow::CursorWindow() for an example.
      ///
      /// Feel free to stick application specific functions in the func[] map
      /// but don't create names for them here -- simply keep your own enum
      /// of the func names and make the first one have the value 'user_function_0'
      /// plus one.
      ///
      /// Note also that if you wish to change the key bindings used by the
      /// dialog input mechanism, you need only modify CursorWindow::func to
      /// suit your liking.
    {
      func_unknown, ///< (0) unknown function -- not a real function
      func_up, 
      func_down, 
      func_left, 
      func_right,
      func_ic,      ///<  (5) toggle insert character mode (rarely used, not implemented)
      func_dc,      ///<  delete the character under the cursor
      func_home,    ///<  move to left of line
      func_end,     ///<  move to right of line
      func_prior,   ///<  prev page -- page up
      func_next,    ///<  (10) next page -- page down
      func_dc_prev, ///<  delete char before current one (backspace key?)
      func_tab,     ///<  tab right
      func_btab,    ///<  back tab
      func_top,     ///<  (14) top of this file or page being edited
      func_bottom,  ///<  bottom of this file or page being edited
      func_esc,     ///<  esc key pressed
      func_enter,   ///<  enter key pressed
      func_data,    ///<  normal data keys
      func_clreol,  ///<  clear to end of line
      func_nextwd,  ///<  move to next word
      func_prevwd,  ///<  move to prev word
      func_upword,  ///<  uppercase the current word and move to next
      func_dnword,  ///<  lowercase the current word and move to the next
      func_undo,    ///<  undo most recent change
      func_find,    ///<  find first
      func_findnxt, ///<  find next
      func_findprv, ///<         find prev
      func_repl,    ///<  replace first
      func_replnext,///<         replace n
      func_help,    ///<  get application specific help (default is F1 key)
      func_mark,    ///<  mark the current line
      func_dl,      ///<  delete line
      func_db,      ///<  delete block
      func_cb,      ///<  copy block
      func_wb,      ///<  write block to a file
      func_paste,   ///<  paste text into edit session
      func_goto,    ///<  prompt for a line number to goto
      func_matching,///<  find matching character -- used on {...}, [...], etc
      func_dw,      ///<  delete word
      func_join,    ///<  join this line with the next line
      func_insline, ///<  insert blank line
      func_to_edit, ///<  convert a view only session to an edit session
      func_reread,  ///<  verify that the disk copy of your edit session hasn't changed
      func_switch,  ///<  switch to alternate source (given your current edit file, find another)
      func_insfile, ///<  insert a file above the current line

      user_function_0=0x100  ///  don't add any builtins below here
                             ///  user defined key mappings should be bigger
                             ///  than user_function_0!
    };

    static std::string        func_name(int funcvalue);
      ///< func_name is a function that returns a string value for each editor
      ///  function in the above list.  If you add your own, they will show up
      ///  in this list ass 'user function' or some such thing.

    static std::map<int,int>  func;

      ///< func is the all important key to edit function map.  See
      ///  CursorWindow::edit_function_names commentary above.  Whenever you
      ///  press a key, the dialog popup() method uses this table to figure out
      ///  if you are pressing a key which is bound to an edit function or just
      ///  a plain old data key.  Feel free to add new bindings if you want
      ///  but be careful about removing existing bindings -- like the arrow
      ///  keys!

    static std::string word;

      ///< This string is nothing more than a list of characters which are valid
      ///  word characters.  Feel free to use this but be careful about elimnating
      ///  letters, numbers, etc.  See std::find_first_of() for a way to use this
      ///  easily.


    struct resize_handler
      /// A resize_handler gets invoked if the user resizes the window.
      /// But only if you have specified the resize handler to use by
      /// calling the set_resize_handler() method.  If you don't make
      /// such a call, the window size change event will be reported
      /// by the read_input() function.  If you do make the setting,
      /// you will not receive any ResizeKey events.  If you are
      /// using viewports, the resize_handler's job is mainly to
      /// enforce your viewport resize policy.  That is, you use this
      /// function to scale the viewports after a resize.  If you are
      /// not using viewports, this function's job will be to repaint
      /// the window correctly.  If you are using viewports, the
      /// resizing event will automatically invoke their repaint methods
      /// after first invoking your resize_handler.  See the file
      /// lib/cursorwindow.c and look for function resize_viewports()
      /// to see what a resize_handler has to do.  Note that you can't
      /// use that code directly.  If you are using a resize_handler of
      /// your own construction, you should maintain your own list of
      /// viewports over which can iterate.
    {
      virtual void operator() (CursorWindow*)=0;
      virtual ~resize_handler();
    };

    void set_resize_handler(resize_handler *);
      ///< Instruct the CursorWindow to inform you of resize events
      ///  by calling the operator method of a specified object rather
      ///  than by returning a ResizeKey event when you call read_input.


    class viewport
      //
      /// Before directly using this class, you should consider
      /// reading @ref text_windows.
      ///
      /// A viewport is a simplistic representation of a subwindow of
      /// a CursorWindow.  You can draw in it, position the cursor in it, etc.
      /// To use a view port, you must provide a repaint object which
      /// will repaint the window on command.  Derive from viewport::
      /// repaint_handler to implement yours.  Each CursorWindow keeps
      /// a list of viewports.  Only one may be active at any given time.
      /// You can manually activate a viewport, using the 'activate()'
      /// method of the viewport -- it will in turn invoke your repaint
      /// object's function call operator method.  Attempts to write into
      /// a deactivated viewport will be ignored.
      ///
      /// WARNING!  Don't allow your repaint_handler's to go out of scope
      /// before your viewports that referring to them do!  You have been
      /// warned!
      ///
      /// You cannot get input from a viewport, per se.  Instead, you can
      /// activate a viewport and position the one and only text cursor
      /// available in the viewport -- thus making it look like you are
      /// getting input there.  In truth, all input comes from the cursor
      /// window as a whole -- so you are responsible for giving the user
      /// the hints you choose as to "where" the input seems to be going.
      ///
      /// To use a viewport, you ask the CursorWindow to make you one.
      /// See the method, CursorWindow::new_viewport().  This function
      /// returns your viewport as an alt_tools::auto_ptr<viewport>.
      ///
      /// When drawing into a viewport, the upper left of the viewport is
      /// 0,0.  Attempts to draw outside the viewport will be silently
      /// ignored.  Each viewport has its own text attributes.
      ///
      /// Note that there is no read_input() method of a viewport.
      /// All input goes to the currently active viewport and needs to
      /// be handled as a global feature.  You can call restore_curpos()
      /// to make sure that the cursor is in the current cursor position
      /// for this viewport.
      ///
      /// When the user resizes his xterm window, the CursorWindow's
      /// resize_handler (see above) is responsible for scaling the
      /// viewports to account for the change.  It can draw if it likes,
      /// but the viewport repaint_handlers will be invoked as well.
      ///
      /// Note that a viewport, like a CursorWindow, is an Ostreamable
      /// object.  That is, you can perform 'iostream' operations to
      /// it.  But don't get too fancy, simple formatting is the best
      /// you can do.  See Ostreamable for more details.
      ///
      /// Note also that there are some builtin input mechanisms that
      /// use viewports.  They are nested classes inside the viewport
      /// class.  See viewport::Dialog, viewport::Selection,
      /// viewport::select_file.
      ///
      /// When creating a viewport, you must specify the cursorwindow,
      /// the repaint handler, and you are optionally allowed to specify
      /// the position hint.  The positions are as follows:
      ///
      /// @code
      ///   viewport::tiled -- means that the window is part of the
      ///                      "background" of the window and should
      ///                      always be positioned in such a way as to
      ///                      ensure that the windows touch one another
      ///                      with no gaps.
      ///
      ///   viewport::floating -- means that this is a pop-up style window
      ///                         that can be anywhere and minor postion
      ///                         round off errors should be ignored.  This
      ///                         is typically only used for dialog windows.
      ///
      /// @endcode
      ///
      /// The default value of the position is 'tiled'.
      ///
      /// @code
      /// Advice:  creating a viewport doesn't cause it to be painted -- or
      ///          to become the active viewport.  Manually invoke the
      ///          viewport's activate() method to force it to do both.
      ///          Deactivation is fully automatic.  If you want a window to
      ///          be active, call its activate method.  The currently active
      ///          window will be deactivated automatically -- ie its
      ///          repaint_handler will be invoked with cmd=deactivate.
      ///
      /// @endcode
      ///
      /// The default mechanisms for handling viewport creation and destruction
      /// are:
      ///
      /// @code
      ///    CursorWindow::new_viewport
      ///    CursorWinddow::viewport::vsplit
      ///    CursorWinddow::viewport::hsplit
      ///    ~viewport
      /// @endcode
      ///
      /// However these mechanisms should be viewed as tools for
      /// implementing an application specific behavior -- not defining
      /// a paradigm for proper use.  The functions in the above list which
      /// return an auto_ptr<viewport> are returning a special kind of viewport
      /// which "owns" the physical screenspace.  Deleting these viewport pointers
      /// also deletes the screen space (and invokes a resizing algorithm to
      /// cover the hole left by its deletion).
      ///
      /// When you copy a viewport, however, you do not copy the screenspace,
      /// nor do you copy the ownership.  That is, only the viewports returned
      /// by the above functions automatically clean up the screen space when
      /// the viewport is destructed.  There is no way you can copy this feature.
      /// The destructor for copied viewports, constructed like this:
      ///
      /// @code
      ///    extern alt_tools::auto_ptr<CursorWindow::viewport> realvp;
      ///    viewport temp(*realvp.get());
      ///
      /// @endcode
      ///
      /// do not clean up the screen space.  However, the destruction of
      /// 'realvp' will free up the screen space.
      ///
      /// In general, your application should make its own list of 'real' viewports
      /// and control the creation and destruction thereof to implement your own
      /// screen appearance.  You can then treat the other viewports created via
      /// various queries as merely temporary objects to be used and then discarded
      /// without concern.
      ///
      /// CursorWindow derives from Ostreamable -- which means that can be used
      /// as the target of operator<< for a small subset of the stream output
      /// operations.  See Ostreamable for more details.
      //
      : public Ostreamable<viewport>
    {
    public:

      struct repaint_handler
       /// @warning You probably want to be using a Viewer than a repaint handler.
       ///  repaint_handler is a a low level detail.  The strategy of using the
       ///  the ViewerManager to control viewers is preferred to actually implementing
       ///  a repaint handler yourself.  Note that a Viewer is a repaint_handler, so
       ///  the further reduction of effort that it provides with worth using.
       ///
       /// Derive from repaint_handler to implement your repaint
       /// object which will be invoked whenever there is a resize
       /// event or if some other viewport is created or deleted
       /// or activated.  Your repaint handler is expected to
       /// ask for the viewports size and completely repaint this
       /// region.  The cmd passed to your repaint handler is one
       /// of the cmd_constant enum values.
       ///
       ///  Maintenance note:  a repaint_handler should not store a pointer to its viewport
       ///  because the viewport associated with a viewer will change over time.
       ///  A viewer has the responsibility of completely repainting the viewport it
       ///  given when asked to do so.  This means, you can't store the size of the
       ///  viewport either -- but must inquire it from the viewport passed during
       ///  resizes and handle_event() calls.

      {
        enum cmd_constant
          ///  values that will be passed to the function call
          ///  operator method.
        {
          activate,     ///< means you have been activated
          deactivate,   ///< some other viewport has taken focus
          resize        ///< only the size has changed
        };

        virtual void operator() ( viewport * vp, int cmd_constant_value ) = 0;
        virtual ~repaint_handler();

        bool isTemporary() const { return isTemporary_; } ///<  Delete the file managed by this
                                                          ///   repaint_handler -- actually managed
                                                          ///   by it's derived class.  The
                                                          ///   repaint_handler is meant to manage
                                                          ///   screen regions but since we put
                                                          ///   editors in the these screen regions,
                                                          ///   this flag refers to the file managed
                                                          ///   by the editor in this screen region.

        repaint_handler()
        :  isTemporary_(false)
        {
        }

        void setTemporaryFlag(bool f) { isTemporary_ = f; }
            ///<  Set the flag that indicates to derived classes that this screen region contains
            ///   a "temporary" file.  During destruction of the derived class, temporary files
            ///   should be deleted by the derived class.
        
        private:

           bool isTemporary_;   ///< Means that the file managed by the editor, which is a derived class
                                ///  of the repaint handler, is a temporary file.


      };

      enum position
        ///  A specification of how constrained the window position and
        ///  size is.
      {
         tiled,   ///<  Always align this window with others to establish a
                  ///   complete covering of the CursorWindow area
         floating ///<  let this window float around on the screen where it may
      };

      viewport(CursorWindow*, repaint_handler*, position p=tiled);
                                                 ///<  constructor
     ~viewport();                                ///<  destructor


      viewport(viewport const &);                ///<  Not for general use!
                                                 ///   only meant to be used by
                                                 ///   get_viewports() but it has a
                                                 ///   container that requires it!  Darn!
                                                
     private:

      viewport &operator= (viewport const &r);   ///<  not implemented!
                                        
      public:

      void set_repaint_handler(repaint_handler *r);

      repaint_handler* get_repaint_handler();               ///< get the "viewer" associated with this viewport.
      repaint_handler const* get_repaint_handler() const;   ///< get the "viewer" associated with this viewport.


      std::auto_ptr<viewport> vsplit(repaint_handler* viewer, int keep_hint=0);

        ///< Split the viewport vertically and return handle to the
        ///  bottom half.  The repaint_handler you specify is used to
        ///  handle repaint requests there.  Note that if the window
        ///  is too small to handle the request, you will receive a
        ///  new viewport completely overlaying this viewport.  Note that
        ///  this is meant to be used with tiled windows which are contiguous
        ///  but you can move a viewport using the 'move()' method.
        ///
        ///  
        ///  @param viewer     The the viewer to split.
        ///  @param keep_hint  A suggestion of how many rows to leave in
        ///                    in the donating viewer.
        
      std::auto_ptr<viewport> hsplit(repaint_handler*, int keep_hint=0);

        ///< Split the viewport horizontally and return handle to the
        ///  right half.  The repaint_handler you specify is used to
        ///  handle repaint requests there.   Note that if the window
        ///  is too small to handle the request, you will receive a
        ///  new viewport completely overlaying this viewport.  Note that
        ///  this is meant to be used with tiled viewports -- but you can
        ///  manually move a viewport using the 'move()' method.
        ///  
        ///  @param viewer     The the viewer to split.
        ///  @param keep_hint  A suggestion of how many cols to leave in
        ///                    in the donating viewer.

      row_col curpos() const;

        ///< Get the current cursor position within the viewport

      void set_curpos(int row, int col)
        //
        ///  Move the cursor to the specified row and column within the
        ///  viewport.
      {
        set_curpos( row_col(row, col) );
      }

      void set_curpos(row_col rc);
        ///< Move the cursor to the specified row and column within the
        ///  viewport.
        
      void restore_curpos();
        ///< Make sure the cursor is in the correct position for this
        ///  viewport.

      row_col size() const;

        ///< Get the width in columns and the height in rows
        ///  of the viewport.  Note that you cannot force
        ///  this information but must live with the size
        ///  you get.  Well, ok, you can call 'move' which will
        ///  force the size but generally this will mess up screen
        ///  layout -- only do it as part of the resize_handler logic.
        
      row_col origin() const;

        ///< Get the absolute position of the upper left corner of the
        ///  viewport.

      void beep();
        ///< Produce an audible alarm.
        
      template<class Iterator>
      void write(Iterator first, Iterator last) /// Write text to the window
      {
        while(first != last) write(*first++);
      }
        
      void write(char const *, size_t length);

        ///<  Write a specified number of characters to the current cursor
        ///   position and move the cursor -- all within the viewport.

      void write(char const *s)
        ///  Write a string in the current text attributes at
        ///  the current cursor position and move the cursor --
        ///  all within the viewport.
      {
         write(s, strlen(s));
      }

      void write(std::string const &s, size_t length=1000000)
        //
        ///  Ditto
      {
        size_t actual_length = s.size();
        
        write(s.data(), actual_length < length ? actual_length : length );
      }

      void write(long c, int length=1);
        ///< write specified number of characters to the window in the
        ///  default text attribute at the current cursor position.  The
        ///  default repeat count is 1.


      void fill_to_eol(char c=' ');
        ///< fill the rest of the current line, starting at the current
        ///  cursor position, with a specified character -- defaulting to space.

      void box(int width, 
               int height, 
               bool filled,
               char const *title_string, 
               unsigned length
              );
        ///< Draw a box starting at the current cursor position that is
        ///  width columns wide and height columns tall (running down from
        ///  the current cursor position.  
        /// 
        ///  If the filled flag is true, the box will be filled with blanks
        ///  before the border is drawn.
        /// 
        ///  If the string is not null, and the length is not 0, the string
        ///  will be drawn inside the top border of the box to indicate that
        ///  it is the 'title' of the box
      
      void box(int width, int height, bool filled, char const *title=0)

        ///  Draw a box starting at the current cursor position that is
        ///  width columns wide and height columns tall (running down from
        ///  the current cursor position.  
        /// 
        ///  If the filled flag is true, the box will be filled with blanks
        ///  before the border is drawn.
        /// 
        ///  Draw the title string, if specified, inside the top border of the
        ///  box to serve as a title.
      {
        box(width, height, filled, title, title ? strlen(title): 0);
      }
      
      void box(int width, int height, bool filled, std::string title)
        //
        ///  Draw a box starting at the current cursor position that is
        ///  width columns wide and height columns tall (running down from
        ///  the current cursor position.  
        /// 
        ///  If the filled flag is true, the box will be filled with blanks
        ///  before the border is drawn.
        /// 
        ///  Draw the title string, if specified, inside the top border of the
        ///  box to serve as a title.
      {
        box(width, height, filled, title.data(), title.size());
      }

      void fill_to_eos(char c=' ');
        //
        ///< Fill the rest of the current VIEWPORT, starting at the current
        ///  cursor position, with a specified character -- defaulting to space.

      int text_attribute() const;
        //
        ///<  Get the current text attribute of this viewport

      void set_text_attribute(int);
        ///< Set the current text attribute of this view port.


      void move(row_col where, row_col size);
        ///< Move the viewport around within the CursorWindow.  This
        ///  call sets the position and size of the viewport.  This
        ///  will trigger a repaint of the whole cursor window when
        ///  next input occurs -- with the cmd of 'resize'.
        
      void activate();
        ///<  repaint the window and give it a command of 'activate'.

      void refresh();
        //
        ///<  Flush pending output to the physical window.  You call this
        ///   when you have written stuff that doesn't get displayed because
        ///   you are not calling read_event();


      struct port;

      port* port_;  ///< this variable should be private but my crufty old
                    ///  MSVC 6.0 compiler requires it to be public

      bool  owned_; ///< indicates that when the viewport is destroyed, deleted
                    ///  the port.  This is almost but not always true.  The
                    ///  get_viewports() function sets this to false.


      friend class  CursorWindow;


    public:

    
      bool  constructed_;  ///<  debug measure.  

      viewport(port*);

        ///< make a second viewport that uses the same port (safe but rarely needed )

      CursorWindow* window();
        ///< Get the window on which this viewport is drawn.

      void ostreamable_helper(char c)
      {
        write(&c, 1);
      }

    }; // end of class viewport

    std::auto_ptr<viewport> new_viewport(viewport::repaint_handler*,
                                viewport::position p = viewport::tiled);
  
      ///< Create a new viewport.
      /// The default position descriptor is 'tiled'.  See CursorWindow::viewport
      /// for details.
      ///
      /// Note that you can also create a viewport using viewport::vsplit()
      /// or viewport::hsplit().  These are meant for use with tiled windows
      /// which will be contiguous with one another -- although you can
      /// manually adjust the positions.
      ///
      /// Creating a viewport does not activate it and does not draw it.
      /// Use the 'activate()' method of the viewport to do both.

    void needs_resized();
    
      ///< Called to force a complete repaint of the window.


    void refresh();
    
      ///< Flush pending output to the physical window.  You call this
      /// when you have written stuff that doesn't get displayed because
      /// you are not calling read_event();

    class Dialog
      //
      /// A Dialog is a list of fields which which will be displayed
      /// in a viewport and the user can edit them.  Users of a Dialog
      /// initalize the list of fields, each of which must be a
      /// Dialog::Element.  They then let the user edit the fields.
      /// After editing is complete, the caller iterates over the fields
      /// to see what changed.
      ///
      /// To allow user interaction, call 'popup()'.  After the user has
      /// finished editing the last field (including, simply ok'ing its
      /// contents), the dialog automatically pops down and you can
      /// iterate over the fields.
      ///
      /// If the user presses Esc, the dialog terminates early.  The
      /// popup() function handles the user edits and returns true if the
      /// user aborts using the Esc key.
      ///
      /// See the comments in the class Elements for a general description
      /// of a dialog's use.
    : public viewport::repaint_handler
    {
    public:
      class Impl;
      friend class Impl;

    private:

      Impl *impl_;

    public:

      Dialog(std::string title);
       
        ///< All dialogs must have a title string
        
        
     ~Dialog();
       
       ///< Ah, the _destructor_.  Hmmm.

      void operator() ( viewport * vp, int cmd );
        
        ///< Handles repaints when window sizes change.

      bool popup(CursorWindow*);

        ///< Popup the dialog and let the user edit the fields.  If
        ///  this function returns true, the user aborted.
        
      void set_completion_directory(std::string const &s);

        ///<  Set the 'current directory' which should be observed
        ///   when tab completion key strokes are interpreted by
        ///   elements.  This value is by default empty and
        
      std::string const &completion_directory() const;

        ///<  Get a reference to the 'current directory' which should be
        ///   used by string elements when the user presses the tab key to
        ///   activate 'tab completion' of file names.

      void set_first_input_field(std::string const &s);

        ///<  Set the name of the field the cursor will go on when the
        ///   popup function is called.

      class Element
        ///  An input element in a Dialog.  This is an abstract
        ///  base class from which you can derive -- although there are
        ///  some builtins (Ok,string, etc).  The popup() function will handle
        ///  some keys and some will be sent to the element's handle_key method.
        ///  The draw() method may be invoked multiples times within
        ///  the popup() function call.
        /// 
        ///  @code
        ///
        ///  To add Elements to a dialog, do this:
        /// 
        ///    Dialog d("title");
        /// 
        ///      d += new ElementType(initializer);
        ///      d += new ElementType(initializer);
        ///      d += new ElementType(initializer);
        /// 
        ///  To allow user's to edit, then do this:
        /// 
        ///      if(! d.popup() )
        ///      {
        ///        iterate over the elements in d and extract the
        ///        names and updated values for your use.  use begin(),
        ///        and end() to get iterators to the elements
        ///      }
        ///
        ///  @endcode
        ///  The popup() function draws the labels and the dialog
        ///  title.  The element is responsible for drawing its
        ///  input area -- but not violate the input_width_ which may
        ///  not be the same as what is specified during the element's
        ///  construction.
        /// 
        ///  the popup() function handles keyboard input and distributes
        ///  some keys to the element's handle_key function:  data keys
        ///  and left and right arrows but no other keys -- except key_dc.
      {
      protected:

        Dialog*     dialog_; //< when you append a dialog element to a dialog
                             //  this field gets set by the Dialog::operator+= method.
                        
        friend class Dialog;
        
      public:

        std::string name_;
        std::string label_;
        std::string value_;
        int         input_width_;

        Element(std::string const& name, std::string const &label, int width=0)
        : name_(name),
          label_(label),
          input_width_(width)
        {
        }
        

        virtual ~Element() {}

        virtual void draw_input_area(viewport *v, bool highlighted) = 0;

          ///<  Draw the input area of this field in the specified viewport.
          ///   The cursor will already be positioned for you.  Do not
          ///   draw more than 'input_width_' characters.  If the
          ///   highlighted flag is passed in, the field is active, otherwise
          ///   it is not.

        virtual void handle_key(viewport *v, int key) = 0;
          
          ///< Interpret keys passed to this field by popup() as the
          ///  user wants to make changes.  Only some keys get to the
          ///  field -- data keys, and left and right arrow keys for
          ///  example, but not up and down or paging keys.


      };

      typedef std::list<Element*>  element_list_t;
      typedef element_list_t::iterator iterator;

    public:

      Dialog& operator += (Element *e);

      iterator begin();
      iterator end();

      class Ok
        //
        ///  A dialog element that does nothing -- you typically
        ///  use this as the last element of dialog so that the
        ///  user has one last chance after entering data in the
        ///  last real field to edit his work.  It has not input
        ///  semantics -- except wasting time.
        //
      : public Element
      {
      public:

        Ok(std::string const &name, std::string const &label)
        : Element(name, label, 4)
        {
          ///<  Construct an OK field.
        }

        void draw_input_area(viewport *v, bool highlighted);
          ///<  paints [ ok ]

        void handle_key(viewport*v, int key);
          ///< ignores all keys
      };

      class String
        //
        ///  A String is a dialog element that prompts the user for a
        ///  character string -- providing a default value.  Normal keyboard
        ///  editing keys are active.  The 'tab' key has special interpretation.
        ///  It completes the string under the cursor.  That is, if you press the
        ///  tab key, the string under the cursor is treated as a partial file
        ///  name fragment and you are presented with the valid completions for
        ///  your selection.  If there is only one valid selection, it is chosen
        ///  for you automatically.
        /// 
        ///  In order for tab completion to work correctly, you will probably have
        ///  to use Dialog::set_completion_directory() based on the 'current' directory
        ///  you wish to present the user with.  This can be obtained from the
        ///  ViewerManager using ViewerManager::active_directory().
        //
      : public Element
      {
        int col_;     ///<  current input column (where cursor  goes) (relative to
                      ///   the string, not the window.

        int offset_;  ///<   from beginning of string to first displayed char -- ie the
                      ///    character displayed in the input_ col.

        int width_;   ///<  ACTUAL input width given window size limitations

        int input_;   ///<  actual_input_col_ obtained by draw_input_area when
                      ///   ever it is invoked.  This is the left most column
                      ///   where input of this string occurs in its dialog
                      ///   window.
      protected:

        bool password_;  ///<  this is a password string -- dont' echo its contents

      public:

        String(std::string const &name,
               std::string const &label,
               std::string const &default_value,
               int                input_width)
        : Element(name, label, input_width),
          col_(0),
          offset_(0),
          width_(input_width),
          input_(0),
          password_(false)
        {
          ///<  Construct a string field.
        
          value_ = default_value;
        }

        void draw_input_area(viewport *v, bool highlighted);
          ///<  paints [ ok ]

        void handle_key(viewport*v, int key);
          ///<  ignores all keys
        
        bool compute_actual_width(viewport *v);
          ///<
          ///  Compute actual width of input area given window size limitations.
          ///  The function return value indicates that the offset_ member
          ///  changed -- this implies a need to repaint the text input region.
        
        void handle_paging(viewport *v);
          ///<
          ///  Use horizontal paging if necessary to insure that the character
          ///  in the string under col_ is actually visible.

        void setColumn(int col);
            ///<  Set the current input column.

      };

      class Password
       //
       /// Same as String, except that the user never gets to see what he has typed
       /// in.
      : public String
      {
      public:

        Password(std::string const &name,
                 std::string const &label,
                 std::string const &default_value,
                 int                input_width)
        : String(name, label, default_value, input_width)
        {
          password_ = true;
        }
      };

      std::string element_value(std::string const &name);
        ///<
        ///   Search all elements until you find the one with the specified
        ///   name and return its value.  If the string is not found, you get
        ///   an empty string back.  If you specify an invalid name, the
        ///   string will be empty as well.  So specify wisely...
    };

    class Selection
      //
      /// A Selection object is a simpler version of a Dialog.
      /// It presents the user with a list of strings of which he/she is
      /// allowed to select exactly one -- or press the escape key to
      /// abort.  To use a Selection object, do the following:
      ///
      ///  Create your object
      ///  then append strings to the selection using operator+= ()
      ///  then call popup()
      ///  if popup returns false, the use the selection_ member
      ///  as the value chosen.
      ///
      /// To append a string, use operator+=.  For example:
      ///
      /// @code
      ///    Selection s("Some title");
      ///
      ///      s += "line 1";
      ///      s += "line 2";
      ///
      ///    if(!s.popup(window))
      ///    {
      ///       // use the s.selection_ variable
      ///    }
      /// @endcode
      ///
    : public viewport::repaint_handler
    {
    public:

      class   Impl;
      typedef std::list<std::string> rep_type;  ///<  actual implementation

    private:

      Impl* impl_;                              ///<  private data not to be
                                                ///   messedwith

    public:

      rep_type rep_;                            ///< container of strings
                                                ///  that will be presented
                                                ///  to the user.  They are
                                                ///  your strings, dork with them
                                                ///  to your hearts content.

      std::string selection_;                   ///< the user's actual selection
                                                ///  made during the popup
                                                ///  call.  Will be null if the
                                                ///  user aborts.


      Selection(std::string title);
        ///<  All Selections must have a title string
        
        
     ~Selection();
       ///< Ah, the _destructor_.  Hmmm.

      void operator() ( viewport * vp, int cmd );

        ///<
        ///  Handles repaints when window sizes change.
        /// 

      bool popup(CursorWindow*);

        ///<
        ///  Popup the Selection and let the user select an item or
        ///  abort.  If this function returns true, the user aborted.
        //
        
      void operator+= (std::string const &s)
      {

        ///<   Append strings to rep_
        //
        rep_.push_back(s);
      }
        
    };


    class Message
      //
      /// A Message object is a simpler version of a Dialog.  It does not
      /// allow the user to make any selections -- it just presents the
      /// user with a scrolloing list of lines of test.  When the user
      /// presses enter, the message goes away.
      ///
      /// To use a Message object, do the following:
      ///
      /// @code
      ///
      ///  Create your object
      ///  then append strings to the Message using operator+= ()
      ///  then call popup()
      ///
      /// @endcode
      ///
      /// To append a string, use operator+=.  For example:
      ///
      /// @code
      ///    Message s("Error:");
      ///
      ///      s += "line 1";
      ///      s += "line 2";
      ///
      ///    s.popup(window));
      /// @endcode
      ///
      /// There is no return value.
      ///
    : public viewport::repaint_handler
    {
    public:

      class   Impl;

      typedef std::list<std::string> rep_type;  ///<  actual implementation

    private:

      Impl* impl_;                              ///<  private data not to be
                                                ///  messedwith

    public:

      rep_type rep_;                            ///<  container of strings
                                                ///  that will be presented
                                                ///  to the user.  They are
                                                ///  your strings, dork with them
                                                ///  to your hearts content.

      Message(std::string title);
        ///<  All Messages must have a title string
        
     ~Message();
       ///< Ah, the _destructor_.  Hmmm.

      void operator() ( viewport * vp, int cmd );
  
        ///<
        ///  Handles repaints when window sizes change.
        /// 

      void popup(CursorWindow*);
  
        ///<
        ///   Present the message to the user and wait until he presses
        ///   enter.  There is no return value as the user is only allowed
        ///   to acknowledge the message.
        //
        
      void operator+= (std::string const &s)
      {
  
        ///
        ///   Append strings to rep_
  
        rep_.push_back(s);
      }
        
    };


    class FileSelector
    : public viewport::repaint_handler
    {

    public:

       FileSelector(std::string const &title,
                    std::string const &filenamePattern,
                    std::string const &startFileName
                   );
       ///<
       ///    Create a file sector dialog which is not yet
       ///    popped up.  Use the popup() method to display
       ///    it and let the user interact.
       ///
       ///    @param title            The title of the dialog
       ///    @param filenamePattern  The "dir/*" string to prime
       ///                            the pump for the user.
       ///    @param startFileName    Optional string meaning
       ///                            that when the user first gets
       ///                            input, try to make sure that the
       ///                            cursor is sitting on the
       ///                            startFileName
      
      void operator() ( viewport * vp, int cmd );
        ///<
        ///  Handles repaints when window sizes change.
        /// 

      std::string popup(CursorWindow*window, Viewer *parent);
        ///<
        ///   Present the user with dialog box that
        ///   lets him select a file or directory from a list
        ///   of files.  The user can navigate up and down in
        ///   the list, move up a directory, down a directory,
        ///   or jump to another directory.  
        ///
        ///   @return The function returns empty string to indicate that
        ///   the user aborted.
        ///
        ///   @param window  The parent window
        ///   @param parent  The parent viewer
        ///
        ///

    private:

        std::string title_;
        std::string searchPattern_;
        std::string searchDir_;             // populated by readFileInfo
        std::string startFile_;

        std::vector<std::string> rowText_;  // names and perhaps other info
        size_t maxNameWidth_;               // longest string in rowText_;

        int row_;               // screen relative position of dialog (top row)
        int col_;               // left coloumn
        int dialogHeight_;      // in rows
        int dialogWidth_;       // in cols


        int displayAreaRow_;    // top of display area
        int displayAreaCol_;    // left of display area
        int displayAreaHeight_;
        int displayAreaWidth_;  

        int inputIndex_;     // into rowText_
        int inputTopIndex_;  // index into rowText_ of top item displayed in the display area

        bool noRepaint_;     // true only in minute instants when popup is still running
                             // but we want to ignore calls to operator()()


        void adaptToWindow(viewport *); // recalculate display coordinates and active
                                        // window regions based on actual window sizes

        void paintDialogBackground(viewport *); // paint the dialog region in the appropriate colors

        void readFileInfo();

        std::string handleInput(CursorWindow *w, Viewer *parent);
                                              
        void paintRow(CursorWindow *w, int displayAreaRow);  // repaint current specified row

        void eraseMe(CursorWindow *w);   // remove this dialog from the screen


    };

    // bogus do not use!
    static std::string select_file(std::string const &title,
                                   std::string const &dir_file_pattern,
                                   CursorWindow *w
                                  );

    void ostreamable_helper(char c)
      //
      /// Helper function used by base class, Ostreamable, to actually
      /// perform io when you use syntax of the form:
      ///
      ///   window << "stuff";
      //
    {
      write(&c, 1);
    }




    int get_viewports(std::list<viewport> *viewports);

      ///<
      ///  Get a list of the currently extant viewports.  Returns the
      ///  count of viewports as well.  The viewports returned are not the
      ///  owner of the physical screen are valid ONLY until the next creation
      ///  or deletion of a viewport.

private:

                        
    Window * window_; ///< the real implementation of window

    void destroy_viewport(viewport*);


}; // CursorWindow 
} // namespace cxxtls

#endif
