#ifndef __deltaviewer_header_included_p
#define __deltaviewer_header_included_p
//
// Copyright 2013, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.
//
// Warning:  not all files in this directory structure may be covered by the
// same copyright.  Some of them may be part of the GNU source distribution
// and you must obey the GPL copyright for any such files.
#include <cxxtls/foreach.h>
#include <cxxtls/viewermanager.h>
///@file
///   This file defines the DeltaViewer class which is used for showing
///   and highlighting changes to short lists of name/value pairs.

namespace cxxtls
{

class DeltaViewer
    ///  DeltaViewer is a kind of viewer targetting the displaying short
    ///  lists of name-value pairs in a tabular format.  This kind of
    ///  viewer emphasizes the dynamic nature of changes to the values
    ///  by displaying recently changed values.  That is, each time there
    ///  is an update to the data in the viewer, all lines of text are
    ///  repainted.  Changed lines are displayed with underscores, but
    ///  lines which have not changed are displayed plainly.  
: public Viewer
{
public:

    struct NameValuePair
    {
        bool changed_;
        std::string name_;
        std::string value_;

        NameValuePair(std::string const &name, std::string const &value)
        :  changed_(true)
        ,  name_(name)
        ,  value_(value)
        {
        }

        void update(std::string const &value)
        {
            if(value_ != value)
                changed_ = true;
            else
                changed_ = false;

            value_ = value;
        }

    };

private:

    std::string                 title_;
    std::list<NameValuePair>    data_;
    ViewerManager*              vm_;
    int                         displayMode_;
    int                         cursorRow_;           // offset from the top row of the data_ table where the
                                                      // cursor lies.  range: 0 -> data_.size()-1
    int                         topDisplayedRow_;     // range:  0 -> data_.size()-1

    NameValuePair * find(std::string const &name)   // return 0 if not found
    {
        CXXTLS_FOREACH(NameValuePair & i, data_)
        {
            if(name == i.name_)
               return &i;
        }

        return 0;
    }

    void update(std::string const &name, std::string const &value)
    {
        NameValuePair * where = find(name);

        if(!where)
        {
            data_.push_back(NameValuePair(name, value));
        }
        else
        {
            where->update(value);
        }
    }




    void paintTitle(CursorWindow::viewport*);                 //  paint the top line of the window in title highlights
    void paintPage (CursorWindow::viewport*);                 //  paint the highlighted title line and the bulk of the window in normal highlights
    void paintCurrentExpression(CursorWindow::viewport * vp); //  paint the wider form of the name/value pair for the currently highlighted value

public:

    DeltaViewer(ViewerManager* vm, std::string const &title)
    :  Viewer(vm)
    ,  title_(title)
    ,  vm_(vm)
    ,  displayMode_(0)
    ,  cursorRow_(0) 
    ,  topDisplayedRow_(0) // range:  0 -> data_.size()
    {
    }

    void replace(std::list<NameValuePair> const &replacementData);  //  replace, delete, or update as needed.
    
  virtual void operator() ( CursorWindow::viewport *     vp,
                            int                          cmd
                          );
     ///< Handle screen repaint commands.  Store cmd in displayMode_ member
     ///
     ///  @param vp
     ///  @param cmd

  virtual bool handle_event( CursorWindow::input_event const * event,
                             CursorWindow::viewport          * viewport
                           );
     ///< Handle input events.
     ///
     /// @param event 
     /// @param viewport

  virtual void help();  ///< Present the user with help information that is
                        ///  specific to this viewer's operation and contains
                        ///  information about system keep bindings, etc.

  static Viewer* app(ViewerManager*vm, std::string &fullname);
    ///<
    ///  Application "hook" function for the ProcessViewer class.
    ///  Use it as a parameter to add_app, like this:
    ///
    ///  @code
    ///    vm.add_app(DeltaViewer::app, "DeltaViewer");
    ///  @endcode

  static std::string app_name;
    ///<
    ///  The name to prompt the user with in the ViewerManager::appChooser
    ///  menu.
    ///

  std::string const& application_name() const { return app_name; }
    ///<
    ///  Return the name of the application to use when describing this
    ///  viewer in the ViewerManager::appChooser or in the
    ///  ViewerManager::ViewerSelector.
    ///

  std::string description() const;
    ///<
    ///  Returns a description of this session.
    ///



}; // class DeltaViewer
}; // namespace cxxtls
#endif
