#ifndef __dumbterminal_h_included_p
#define __dumbterminal_h_included_p
#include <cxxtls/viewer.h>
#include <cxxtls/pseudoterminal.h>
#include <cxxtls/thread.h>
#include <deque>
#include <memory>

namespace cxxtls
{

class DumbTerminalViewer
: public Viewer
    /// A dumb terminal emulator that runs in a CursorWindow::view_port.  That is, a "viewer".
    ///
    /// The DumbTerminalViewer class interacts with with a dumb terminal, but it does not create any
    /// process that attaches to it.  You can create your own process that talks to the viewer, outside
    /// the scope of this class, or in a derived class.
    ///
    /// The DumbTerminalViewer creates the pseudo terminal, puts it in no-echo mode, then lets the user
    /// enter text which will appear in the input stream of the child process reading from the terminal --
    /// if any.  You must first create the DumbTerminalViewer, then once it is initialized and the name
    /// of the terminal is available, you can create a child process that interacts with it.
    ///
    /// Here's how:
    ///
    ///@code
    ///
    ///    DumbTerminalViewer *dt = new DumbTerminalViewer(currentVM);
    ///
    ///    if(!dt->ptyGood())
    ///       exit(1);
    ///
    ///    std::string ptyName = dt->ptyName(); //  "/dev/pty06" or something like this
    ///
    ///    std::string command( std::string("/bin/bash &lt;") + ptyName + " &gt;" + ptyName + " 2&lt;&1 &" );
    ///
    ///    // Create a detatched bash process reading and writing to the terminal
    ///
    ///    system(command.c_str());
    ///
    ///    ...
    ///
    ///@endcode
{

public:

  DumbTerminalViewer(ViewerManager* vm, int masterFD);
     ///  Construct a DumbTerminalViewer given the viewer manager for the application
     ///  and the master file descriptor for the pseudo terminal being used.

  virtual ~DumbTerminalViewer();

  std::string ptyName() const { return terminalName_; }
      ///<  Get the name of the slave terminal.  Use this name as stdin/stdout for processes.

  bool ptyGood() const { return ptyGood_; }
     ///<  Return true if the pseudo terminal is up and running correctly

  virtual void operator() ( CursorWindow::viewport *     vp,
                            int                          cmd
                          );
     ///< Handle screen repaint commands
     ///
     ///  @note The vp parameter gets deleted regularly, don't save
     ///        it!

  virtual BackgroundProcessStatus getBackgroundData(CursorWindow::input_event &e);
     ///<  Get any pending background process data and store it in the 
     ///   specified event.  Return one of the enumeration values in
     ///   BackgroundProcessStatus to indicate what has happened.
     ///
     ///   In the event that only a single character is returned, 
     ///   e.type_ will be set to CursorWindow::input_event::ProcessOutput.
     ///   
     ///   If there are multiple characters waiting, e.type_ will be set to
     ///   CursorWindow::input_event::BackgroundData
     ///
     ///   @param e  An initialized but not populated cursor window event.
     ///   @return   One of the BackgroundProcessStatus values.
     ///
     ///   The default implementation of this function always returns
     ///   NO_BACKGROUND_PROCESS.  If the viewer does implement one, it must
     ///   return some other value.  Fair warning:  only the main thread of the
     ///   application can make CursorWindow calls!  No other thread should.


  virtual bool handle_event( CursorWindow::input_event const * e,
                             CursorWindow::viewport          * vp
                           );
     ///<
     /// Handle input events -- and return true if the event is meant to
     /// terminate the viewer -- and make it go away.  When handling
     /// key events -- either FunctionKey or DataKey, you will want to
     /// consult the CursorWindow::func mapping to figure out if the
     /// key corresponds to 'system wide' edit function.  For example,
     ///
     /// @code
     ///    int edit_func = vp->window()->func[e->value_];
     ///    switch(edit_func)
     ///    {
     ///       case CursorWindow::func_up:    // the up arrow key
     /// @endcode
     ///
     ///
     /// Good example:  see handle_event in the deltaviewer.cxx file.
     ///
     /// Choose this approach instead of looking at the key values
     /// directly to allow for consistence in system wide input remapping
     /// that occurs when the application modifies the CursorWindow::func
     /// map.  Note that the CursorWindow::func_data is returned for all
     /// keys not mapped into one of the other named func_* keys.
     ///
     /// Usage note:  If your handle event creates a new viewport that
     /// covers this viewport, make sure that your current viewport doesn't
     /// draw anything until the next repaint request.  Otherwise, your
     /// screen will get very jumbled up!

  virtual std::string const &application_name() const;
     ///<
     ///  Return a const references to a string which defines the name
     ///  of the application being implemented in the viewer.  This function
     ///  must always return the same address and the string at that address
     ///  must never change;

  virtual std::string description() const;
     ///<
     ///  Get a description of this viewer.  Typically, this just returns
     ///  the name of the directory or the name of the file the viewer is
     ///  being applied to.  However, with FTP viewers, POP3 viewers, etc
     ///  this may not refer to any real file.

  virtual void help();  ///< Present the user with help information that is
                        ///  specific to this viewer's operation and contains
                        ///  information about system keep bindings, etc.
                        ///  Note that there is a default implementation of
                        ///  this function, but it should be overridden.
                        
  virtual bool is_dirty() const;
     ///<  Returns true if there are unsaved edit changes in this viewer.
     ///  the default return value is false.  Override this function if
     ///  your viewer has the ability to have unsaved edit changes -- most
     ///  don't.

  static Viewer* app(ViewerManager*vm, std::string &fullname);
     ///<  create a new dumb terminal app

  void outputLine(std::string const &);  ///<  Append a line of text into the output.
  void appendOutput(char);               ///<  Append a single character into the output -- adds it to the end of the most recent line. 
                                         ///   Creates a line if none exists

  void setTitle(std::string const &s) { title_ = s; }     ///<  Set the window title
  std::string const &title() const    { return title_; }  ///<  Get the window title

  void terminate();  ///<  set the shuttingDown_ flag so that handle_event will know to return true to tell the vm we are done

  std::list<int> getProcesses(); ///<  Get a list of the process ids which are reading/writing to this terminal

private:

    std::string                   inputLine_;         // current input line
    std::deque<std::string>       output_;            // output read from the terminal, used for scrolling
    Mutex                         outputMutex_;       // don't write on output without owning a lock on outputMutex_
    std::string                   title_;             // program being monitored by this viewer

    size_t                        pageBottom_;        // bottom of page (first output_ row currently visible)
    int                           titleRows_;         // height of title in rows
    int                           inputRows_;         // height of input region in rows (long input lines change this)
    int                           inputRow_;          // offset within input region
    int                           inputOffset_;       // byte offset within input region to cursor
    int                           availInputCols_;    // available input columns
    int                           leftInputCol_;      // offset from window left to input region (0 or 1)
    int                           inputCol_;          // column within the input region where cursor lies
    int                           windowRows_;        // height of window in rows
    int                           windowCols_;        // width of window in columns
    int                           windowOffset_;      // offset to left column of window from left edge of xterm we are in
    int                           scrollRows_;        // height of scrollable region in rows (where output_ is displayed)
    size_t                        leftOffset_;        // offset from beginning to text visible in leftmost window column
    size_t                        cursorRow_;         // 0 -> oo.  if in output_.size(), displays data.  else nothing
    int                           cmdType_;           // cmd parameter to operator()()
    int                           inputAtt_;          // text attribute of the input region
    int                           bottomAtt_;         // text attribute of last line of output_ displayed
    int                           outputAtt_;         // text attribute of most of the output_ region
    int                           titleAtt_;          // text attribute of top line of window
    int                           cursorAtt_;         // text attribute of cursor line in the output_
    int                           markedAtt_;         // text attribute of output_ regions which have been marked
    bool                          ctrlX_;             // most recent char was a control x.  Hack to handle a few commands



    ThreadId                      thread_;            // reads from pty_ so as to display in the window
    bool                          shuttingDown_;      // destruction is in progress
    Mutex                         shutDownMutex_;     // destruction is in progress
    Mutex                         dataArrivedMutex_;  // guards dataArrived_
    bool                          dataArrived_;       // indicates output_ has changed since the last screen paint
    int                           masterFD_;          // pseudo terminal master file descriptor;
    std::string                   terminalName_;      // pseudo terminal name
    bool                          ptyGood_;           // psuedo terminal running properly
    std::deque<std::string>       input_;             // text that the user has input
    size_t                        inputIndex_;        // next index to fetch input from when you press ^U or ^L
    size_t                        markTop_;           // oldest line in the output buffer that is marked
    size_t                        markBottom_;        // youngest marked line
    int                           markedCursorAtt_;   // att of marked rows when cursor is on them
    int                           botMarkCursorAtt_;  // bottom, marked, cursor
    int                           botCursorAtt_;      // bottom, not marked, cursor
    bool                          suppressRepaint_;   // used in tab-completion logic, see func_tab handling in handle_event()

    void getWindowSize(CursorWindow::viewport *vp, int cmd); // get windowRows_, windowCols_, windowOffset_, cmd
    void paintRow(CursorWindow::viewport *vp, int row);      // display output_[row + pageBottom_] in the proper attributes
protected:
    void paintWindow(CursorWindow::viewport *vp);            // paint the whole window
private:
    void paintOutput(CursorWindow::viewport *vp);            // paint visible parts of output_
    void paintCommandLine(CursorWindow::viewport *vp);       // paint input region, including handling line wrap
    void paintTitle(CursorWindow::viewport *vp);             // paint the top line of the window

    int inputArea() const                    // given the string in inputLine_, compute inputRows_
    {
       std::string tmp;
       return inputArea(tmp);
    }
    
    int inputArea(std::string &text) const;  // get the needed row count and the text to display

    static void* reader(void*);   // the thread that reads from the terminal and displays data in the window

    bool isMarked(size_t rowOffset)
    {
        if(    rowOffset < markBottom_
           ||  rowOffset > markTop_
          )
        {
           return false;
        }

        return true;
    }

protected:
    void addOutput(std::string const &line) 
        // does not lock mutex!  Not thread safe, 
    {
        output_.push_front(line);  // not thread safe!
    }

    std::string const &inputString() const { return inputLine_; }

    size_t inputOffset() const { return inputOffset_; }

    void setInputString(CursorWindow::viewport *vp, std::string const &s, size_t inputOffset)
    {
       inputLine_ = s;
       inputOffset_ = inputOffset;

       getWindowSize(vp, cmdType_);

       inputRow_ = inputOffset_ / availInputCols_;
       inputCol_ = inputOffset_ % availInputCols_ + leftInputCol_;

       paintWindow(vp);
    }

}; // class DumbTerminalViewer

} // namespace cxxtls
#endif // __dumbterminal_included_p
