#ifndef TABLE_EDITOR_HEADER_INCLUDED_p
#define TABLE_EDITOR_HEADER_INCLUDED_p
//
// Copyright 2010, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.

#include <cxxtls/tableviewer.h>
#include <cxxtls/file.h>

namespace cxxtls
{

class TableEditor
: public TableViewer
    ///
    ///  An editor for table files whose fields are separated by some
    ///  user defined character.
    ///
    ///  The user is allowed to change the contents of the file 
    ///  and save those changes or write them to a different file.
    ///
    ///  See @ref CsvViewer for a derived class that does useful work.
    ///
{

    size_t fieldEditColumn_;


public:

    TableEditor(ViewerManager *vm,
	      int delimChar=','
	     );
	///<  Construct a TableEditor and open its file and read
	///   it into memory.  The lines in the file are 
	///   assumed to be delimited by the specified delimiter character.
	///   This constructor assumes that the first line of the file
	///   contains the names of the columns rather than the
	///   data for a row of text.  The column headers must be delimited
	///   as well.
	///   
	///   @param vm       The viewermanager controlling this viewer.
	///   
	///   @param delimChar defines the character that delimits the
	///                    values -- the default is of course comma
	///                    but other characters can be used -- tab,
	///                    "|" or others as needed.
	///   
    
    virtual ~TableEditor();


protected:

    int    delimiter_;
    bool   keyKludge_;  // previous input key was ^X

    virtual bool handle_event( CursorWindow::input_event const * e,
                             CursorWindow::viewport          * vp
                           );
        ///< Handle keystrokes and other events by superceding the
        ///  the behavior of the TableViewer.


   virtual void setCursorInfo(CursorWindow::viewport *vp,
                              size_t worldRow, 
                              size_t worldCol, 
                              bool displayed
                             );
       ///<  Superseed of base class version only for the purpose of detecting
       ///   field changes so that edit sessions on fields can be terminated.
       ///   @see TableViewer::setCursorInfo

   void terminateInput();
       ///<  If the current cell has been edited, finish the editing process.

   CellInfo *editCell(); 
       ///<  Get a pointer, possibly null, to the current cell being edited.
       ///
       ///   @return a pointer cell information for the current cursor position.
       ///
       ///   If the return value is null, then table is empty. 

   virtual void paintField(CursorWindow::viewport *vp,
                           size_t          vpRow, 
                           size_t          vpCol, 
                           CellInfo const *cell, 
                           size_t          cellRow, 
			   size_t          vpColumns,
			   bool            editCursor
			  );
       ///<  Called by paintCursor to draw the text of the field on the screen.
       ///   This function handles clipping and truncaton of text that is too
       ///   long in coordination with fieldEditColumn_.
       ///
       ///   @param vp         the viewport into which to draw the field
       ///   @param vpRow      the row within the viewport to paint the field
       ///   @param vpCol      the column within the viewport to start the painting
       ///   @param cell       the table cell to paint
       ///   @param vpColumns  the number columns the field should take up
       ///   @param editCursor indicates that the cursor should be left in the
       ///                     field edit position (or not).
       ///
       ///   If @b editCursor is false, the cursor will be left at the end of the
       ///   the field.
       ///
       ///
       ///   @see TableViewer::paintField for an explanation of the parameters and
       ///        the environment under which this function is called.
       ///
       ///   @note This function leaves the cursor in first column of the
       ///         the field, not at the end!  PaintCell calls this function
       ///         and this fact needs to be understood when writing paint
       ///   	   handlers.

    static bool isTableEditKey(CursorWindow::input_event const * e);
       ///<  Returns true if the specified event is an "edit" key.
       /// 
       ///   @param e a pointer to an input event that should treated as if it
       ///            were printable character or delete character, that sort of thing.

    bool fieldEditable(size_t tableRow, size_t tableCol) const
        ///
        ///  Determine if a specified row and column is and editable
        ///  column.
        ///
        ///  @return true if the specified row and column is editable.
        ///
        ///  @param tableRow -- the row within the table, has nothing
        ///                     to do with screen coordinates.
        ///  @param tableCol -- the column within the table, has nothing
        ///                     to do with screen coordinates.

    {
       if(    tableRow >= rows_.size()
          ||  tableCol >= cols_.size()
         )
            return false; // invalid table location

       return rows_[tableRow].editable_ && cols_[tableCol].editable_;
    }

    bool fieldEditable(CursorInfo const &cur) const
       ///
       ///  Determine if the specified table cursor cell is
       ///  editable or not.
       ///
       ///  @return true if the table row and column, specified in @b cur
       ///               refers to an editable location.
       ///
       ///  @param cur The table cursor position (not screen, table)
       ///
    {
      return fieldEditable(cur.row_, cur.col_);
    }


    void snatchEdit(CursorWindow::viewport *vp, FileName const &directory);
        ///<
        ///  Given the current cursor position within the current cell, snatch
        ///  a filename / line number pair out of the text and open an edit
        ///  session on that file/line.
        ///
        ///  @param[in]  vp         The viewport to draw in.
        ///  @param[out] directory  The directory that the file name,
        ///                         extracted from the cell text, is meant to
        ///                         be found relative to. WARNING:   The directory
        ///                         name must have a trailing separator (/ or \).
        ///

};

} // namespace cxxtls
#endif
