#ifndef TREE_VIEWER_HEADER_INCLUDED_p
#define TREE_VIEWER_HEADER_INCLUDED_p
//
// Copyright 2011, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.

#include <cxxtls/viewer.h>
#include <cxxtls/simple_regex.h>
#include <cxxtls/file.h>

#include <list>
#include <vector>
#include <utility>  // to get std::pair

//============================================================================================
//============================================================================================
//============================================================================================

namespace cxxtls
{

    //----------------------------------------------------------------------------------------

    struct TreeViewerNodeInfo
        /// Derive from this class to implement data associated with tree
        /// viewer nodes.  New up objects of your derived type and give them
        /// to tree viewer nodes using TreeViewerNode::adoptInfo(TreeViewerNodeInfo *);
        /// When the node gets deleted, the node info will get deleted.
    {
       virtual ~TreeViewerNodeInfo()=0;

       virtual bool operator() (ViewerManager*vm, CursorWindow::viewport *vp, int command, FileName const &) = 0;
           ///<  Interpret command keys sent to this node.  Silently ignore unknown
           ///   command keys.
           ///
           ///   @param vm       The viewer manager (needed to create new windows)
           ///   @param vp       The window drawing viewport.
           ///   @param command  The value of a key event sent to this node.
           ///   @param rootDir  The root directory associated with the tree nodes.
           ///                   If not empty, it must end in either \ or /
           ///
           ///   @return true only to indicate the window should close

    };

    //----------------------------------------------------------------------------------------

    class TreeViewerNode
        /// A string object that is displayable in a tree structure.  A node
        /// can have children, see methods children(), open(), and close().
        ///
        /// There is no flag in the data that indicates what kind of tree
        /// viewer node you have, but you can use dynamic cast to see if a given
        /// node is a container or not.
        ///
        /// Upon destruction, the children are also deleted.
        ///
        /// Note that no child in the children list will be null (unless there is a
        /// major bug).
        ///
        /// TreeViewerNodes are meant for raw data storage and do not have features
        /// needed to animate the nodes on the screen.
        ///
        /// The nodes can have associated information that can be used to perform
        /// command handling.  See adoptInfo() below and TreeViewerNodeInfo above.

    {
    public:
        typedef  std::list<TreeViewerNode> ChildrenList;                  ///< Holds children
        typedef  std::vector<std::pair<TreeViewerNode*, int> > FlatForm;  ///< Holds child*'s and depth

    private:

        ChildrenList children_;  // can have TreeViewerContains in them

        bool opened_;            // Indicates that the user has requested to see the 
                                 // child members displayed.  True by default.

        std::string name_;       // The text of this tree viewer node.

        TreeViewerNodeInfo *info_; // user defined extension

        bool highlighted_;          // display this node underlined

    public:

        TreeViewerNode(std::string const &name);
        TreeViewerNode();

        virtual ~TreeViewerNode(); 

        void setName(std::string const &n); ///<  Set the node name

        void adoptInfo(TreeViewerNodeInfo *p); ///< Take ownership of node info.  Delete
                                               ///  any previous node info.

        TreeViewerNodeInfo *info(); ///<  Return any user supplied information associated
                                    ///   with this node.  

        TreeViewerNode& addChild(std::string const &name);
            ///< Take ownership of a new tree viewer node.  Append it to the list of
            /// children.

        ChildrenList const &children() const; ///<  List of child nodes
        ChildrenList       &children();       ///<  List of child nodes

        bool hasChildren() const;        ///<  Return true if this node has children.
        bool opened() const;             ///<  Return true if the children should be displayed
        std::string const &name() const; ///<  Get this node's name
        virtual int height() const;      ///<  Get height of self plus children


        void close(bool childrenToo=true);  ///< Mark this and optionally all child nodes as closed.
        void open(bool childrenToo=true);   ///< Mark this and optionally all child nodes as closed.

        void flatten(FlatForm &output, int depth=0);
           ///< Appends this node to a flattend version of the TreeViewerNode structure.  
           ///  A flat version does not contain the nested tree structure, per se, but does
           ///  have an indication of the depth of the nodes in the tree to aid in the
           ///  display thereof.

        void setHighlighted();  ///<  Force this node to be displayed in highlighted text

        bool highlighted() const { return highlighted_; }  ///< @return true if highlighted
    };

    //----------------------------------------------------------------------------------------

    class TreeViewer
    : public Viewer
        ///  A TreeViewer allows the display of an acyclic directed graph of TreeViewerNodes.
        ///  
        ///  The TreeViewer class is does not provide a useful definition of TreeViewerNodes
        ///  and you must derive from this class 
    {
        bool           active_;   ///< the current window is active
        TreeViewerNode nodes_;    ///< All nodes in the tree are found as children of this node.

        TreeViewerNode::FlatForm flattenedNodes_;  ///< Linear list of all visible nodes

        FileName rootDir_; ///< Root directory of the tree.  This filename is used by the
                           ///  handle_events() method to associate tree nodes with actual
                           ///  files.  If non-empty, then this string must end in \ or /

    public:
        std::string application_name_;
    private:

        int topNode_;        // top item in the flattenedNodes_ list to be displayed.
        int displayNodes_;   // amount of flattenedNodes_ to be displayed in the viewport
        int cursorNode_;     // node under the cursor

        int treeStartRow_;  // Top row in viewport for display of tree info.
                            // The titles etc occur above this row.
        int displayRows_;   // rows where displaying of tree data can occur.

        bool keyKludge_;    // ^X has been pressed

        bool is_regex_;            // search type flag
        std::string search_text_;  // most recent search string
        SimpleRegex regexp_;       // most recent regex search

    public:

        TreeViewerNode &topNode();   ///< Get pointer to the top node

        void flatten() { nodes_.flatten(flattenedNodes_); }

        virtual ~TreeViewer();

        void setApplicationName(std::string const &name);

        static Viewer* app(ViewerManager*vm, std::string &fullname);
          ///<
          ///  Application "hook" function for the KeyViewer class.
          ///  Use it as a parameter to add_app, like this:
          ///
          ///    vm.add_app(TreeViewer::app, "TreeViewer");


        TreeViewer(ViewerManager* vm); 

        virtual void operator() ( CursorWindow::viewport *     vp,
                                  int                          cmd
                                );
           ///< Handle screen repaint commands
           ///
           ///  @note The vp parameter gets deleted regularly, don't save
           ///        it!
        
        virtual bool handle_event( CursorWindow::input_event const * e,
                                   CursorWindow::viewport          * vp
                                 );
           ///<
           /// Handle input events -- and return true if the event is meant to
           /// terminate the viewer -- and make it go away.  When handling
           /// key events -- either FunctionKey or DataKey, you will want to
           /// consult the CursorWindow::func mapping to figure out if the
           /// key corresponds to 'system wide' edit function.  For example,
           ///
           /// @code
           ///    int edit_func = vp->window()->func[e->value_];
           ///    switch(edit_func)
           ///    {
           ///       case CursorWindow::func_up:    // the up arrow key
           /// @endcode
           ///
           /// Choose this approach instead of looking at the key values
           /// directly to allow for consistence in system wide input remapping
           /// that occurs when the application modifies the CursorWindow::func
           /// map.  Note that the CursorWindow::func_data is returned for all
           /// keys not mapped into one of the other named func_* keys.
           ///
           /// Usage note:  If your handle event creates a new viewport that
           /// covers this viewport, make sure that your current viewport doesn't
           /// draw anything until the next repaint request.  Otherwise, your
           /// screen will get very jumbled up!
        
        virtual std::string const &application_name() const;
           ///<
           ///  Return a const references to a string which defines the name
           ///  of the application being implemented in the viewer.  This function
           ///  must always return the same address and the string at that address
           ///  must never change;
        
        virtual std::string description() const;
           ///<
           ///  Get a description of this viewer.  Typically, this just returns
           ///  the name of the directory or the name of the file the viewer is
           ///  being applied to.  However, with FTP viewers, POP3 viewers, etc
           ///  this may not refer to any real file.

        virtual void help();
          ///<  The interface to the help system.


        void setRootDir(FileName const &dirname); ///< Set the root directory for file opens


        virtual void helpHeader(std::list<std::string> &helpText);  ///<  Append help to helpText 

    private:
        void acknowledgeViewport(CursorWindow::viewport * vp); // Set display height, etc
        void paintViewport(CursorWindow::viewport * vp);       // fill the viewport text
        void paintTitle(CursorWindow::viewport * vp);          // display window title
        void moveUp(CursorWindow::viewport * vp);              // cursor up
        void moveDown(CursorWindow::viewport * vp);            // cursor down
        void pageTop(CursorWindow::viewport * vp);             // move to top of list
        void pageBottom(CursorWindow::viewport * vp);          // move to bottom of list
        void pageDown(CursorWindow::viewport * vp);            // page down
        void pageUp(CursorWindow::viewport * vp);              // page up
        void findFirst(CursorWindow::viewport * vp);           // find string
        void findNext(CursorWindow::viewport * vp);            // find next instance of prev string
        bool findNextHelper(CursorWindow::viewport * vp, bool); // used by findfirst & next
        void make_line_visible(CursorWindow::viewport * vp);
        void displayCurrentLine(CursorWindow::viewport * vp, bool highlighted);
        void paintNodeText(CursorWindow::viewport * vp, int nodeNumber, bool isCursor=false);  
        void positionCursor(CursorWindow::viewport *vp); // for next user input

        void close(CursorWindow::viewport * vp);     // close the current item and redisplay the page
        void open(CursorWindow::viewport * vp);      // close the current item and redisplay the page
        void toggle(CursorWindow::viewport * vp);    // toggle open/close on current item
        void toggleAll(CursorWindow::viewport * vp); // toggle open/close on all items

        void findMyParent(CursorWindow::viewport * vp);     // search up for parent of curre item
        void findNextParent(CursorWindow::viewport * vp);   // search downfor next parent node, higher than current item

        
    };


//----------------------------------------------------------------------------------------

inline 
TreeViewerNode::TreeViewerNode(std::string const &name)
: opened_(true)
, name_(name)
, info_(0)
, highlighted_(false)
{
}

//----------------------------------------------------------------------------------------

inline 
TreeViewerNode::TreeViewerNode()
: opened_(true)
, info_(0)
, highlighted_(false)
{
}

//----------------------------------------------------------------------------------------

inline 
void
TreeViewerNode::
setName(std::string const &name)
{
   name_ = name;
}

//----------------------------------------------------------------------------------------

inline 
std::string const &
TreeViewerNode::
name() const
{
  return name_;
}

//----------------------------------------------------------------------------------------

inline
TreeViewerNode&
TreeViewerNode::
addChild(std::string const &name)
{
   children_.push_back(name);
   return children_.back();
}
//----------------------------------------------------------------------------------------

inline
TreeViewerNode::ChildrenList const &
TreeViewerNode::
children() const 
{ 
    return children_; 
}
//----------------------------------------------------------------------------------------
inline
TreeViewerNode::ChildrenList &
TreeViewerNode::
children()
{ 
    return children_; 
}

//----------------------------------------------------------------------------------------

inline
bool 
TreeViewerNode::
hasChildren() const 
{ 
    return !children_.empty(); 
}  

//----------------------------------------------------------------------------------------

inline
bool 
TreeViewerNode::
opened() const 
{ 
    return opened_; 
}  

//----------------------------------------------------------------------------------------

inline 
TreeViewer::
TreeViewer(ViewerManager* vm)
: Viewer(vm)
, active_(false)
, application_name_("TreeViewer")
, topNode_(0)
, displayNodes_(0)
, cursorNode_(0)
, treeStartRow_(0)
, displayRows_(0)
, keyKludge_(false)
, is_regex_(false)
{
}
//----------------------------------------------------------------------------------------


inline 
TreeViewerNode &
TreeViewer::
topNode() 
{ 
   return nodes_; 

} 

//----------------------------------------------------------------------------------------


inline 
void
TreeViewerNode::
adoptInfo(TreeViewerNodeInfo *p) 
{ 
    delete info_;

    info_ = p;
} 

//----------------------------------------------------------------------------------------


inline 
TreeViewerNode::
~TreeViewerNode()
{ 
    delete info_;

} 

//----------------------------------------------------------------------------------------


inline 
TreeViewerNodeInfo *
TreeViewerNode::
info()
{ 
    return info_;
} 

//----------------------------------------------------------------------------------------

inline
void 
TreeViewer::
setApplicationName(std::string const &name)
{
   application_name_ = name;
}

//----------------------------------------------------------------------------------------

inline
void 
TreeViewer::
setRootDir(FileName const &name)
{
   rootDir_ = name;
}

//----------------------------------------------------------------------------------------
inline
void
TreeViewerNode::
setHighlighted()
{
   highlighted_ = true;
}



//----------------------------------------------------------------------------------------

} // namespace cxxtls
#endif

