#ifndef viewermanager_h_included
#define viewermanager_h_included

//
// Copyright 2002, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.

///@file
///
///  This file defines ViewerManager.  That class is a container of
///  Viewer and CursorWindow::viewport's.   The viewer manager
///  maintains a mapping between Viewer's and the currently visible viewports.
///  It also contains a 'viewer' of its own that lets you select from
///  currently active viewers.  A Viewer, of course, can modify the
///  items being viewed.
/// 
///  Before directly using any of the classes in this file, you should consider
///  reading @ref text_windows or @ref CXXCURSORWINDOWS
///
///  In addition to being a container of viewports, the ViewerManager class
///  also provides methods that help viewports communicate with one another and
///  share common code.  For example: 
///
///  @li  The copy/paste buffer is in the viewer manager
///
///  @li  The etags and cpptags databases are stored in the viewer manager so that
///       viewers don't have to read multiple copies thereof.
///
///  @li  The "find_file_in_standard_locations()" function lets viewers 
///       convert file name fragments into their full pathname by considering
///       the etags or cpptags databases that might exist.
//

#include <cxxtls/viewer.h>
#include <utility>
#include <cxxtls/sequence_map.h>
#include <cxxtls/file.h>
#include <cxxtls/mru.h>

namespace cxxtls
{
class ViewerSelector;
class appChooser;
class EtagsDB;
namespace CppTagDB { class Info; }

class ProgramOptions;  // for use by the default_etags_path() method

class ViewerManager
/// 
///  This class is a container for a collection of CursorWindow::viewport objects
///  plus a collection of Viewers which can draw in the viewports.  A
///  "viewport" is a sub-window.  There is a single CursorWindow which 
///  wraps all aspects of the concept of a terminal that are interesting
///  to viewers.  Within the single cursor window, there are sub-windows
///  called "viewports".
/// 
///  See also, helper class ViewportAdjust, a template
///  that helps re-organize viewports on the screen.
///
///  Before directly using any of the classes in this file, you should consider
///  reading doc/text_windows.
/// 
///  Note!  when debugging programs using viewers, you must start the program
///  first then attach to it using the debugger's 'attach processid' feature
///  (or dbx -debug pdi feature).  To debug the startup code, before the first
///  keyboard input wait, set the environment variable, WAIT_FOR_DEBUGGER=1
///  before running the program.         When the debugger attaches to the running
///  process, you will be in a wait loop.  Use the debugger's print mechanism
///  or the assign mechanism to change the stop variable to 0.  Then you can
///  continue debugging as normal.
///
///  The ViewerManager associates viewers
///  with viewports as needed to make them visible.  The ViewerManager controls
///  the visibility and placement of the Viewer's within the outer Xterm or
///  windows console window.  Viewers fit into tiled windowing regions -- with
///  popup style windows (CursorWindow::Dialog) appearing in the center.
/// 
///  The ViewerManager maintains control of the active viewport and filters all
///  user input -- deciding which events belong to itself and which belong to the
///  currently active viewport.  When you open a ViewerManager, you are also
///  opening a CursorWindow but it is hidden from view because the
///  ViewerManager asserts control of resizes, input, viewport management, etc.
///  You can get a pointer to this window if you need it.
/// 
///  Every Viewer controlled by a ViewerManager has a name.  The names are
///  std::strings -- they don't have to be unique.  If the viewers are associated
///  with files, the name should be the fully qualified file name.  A special kind of
///  viewer is provided by the ViewerManager.  It allows the user to select a
///  desired viewer from the list of all active viewers -- it then associates
///  that viewer with a viewport.  To use a ViewerManager, you create it,
///  then use add(), vsplit(), or hsplit(), to insert viewers.  To
///  insert a viewer, create the viewer, then call one of the aforementioned
///  functions.  The viewer manager will delete the viewer as it sees fit.
///  Also, you must call the open() method to open the window before creating
///  any viewers or viewports.  To 'run' the viewers, execute the member
///  function 'run()'.  When run completes, it means that all viewers are
///  in a state that is safe for the ViewerManager to be destroyed -- that
///  doesn't mean the window is closed -- you have to do that yourself -- or
///  let the destructor do it.
/// 
///  To aid in the creation of viewers whose type can be determined by the
///  file name extension or the stat information for the file, the ViewerManager
///  provides a method, open(std::string filename), which determines what kind
///  of viewer you need to use on this filename.  There are no default behaviours.
///  To use this version of the open method, see the instructions below for
///  the open() method, set_directory_app(), and add_viewer_app().
/// 
///  Note that a given viewer pointer can only be inserted into the viewport
///  manager once -- you can't stick the same thing in several times.  However,
///  the same viewer may appear in several viewports.  As mentioned earlier, you
///  can have multiple Viewers with the same name -- just not the same address.
///  (NO YOU CAN'T!  I never implemented THIS!)
/// 
///  The ViewerManager provides a mechanism for displaying help associated with
///  viewers.  The viewer class must provide a hook function that generates a
///  list of help text strings.  The ViewerManager knows about the help function
///  key bindings and does all the work of calling the help hook function and
///  displaying the text.
/// 
///  The ViewerManager also maintains the 'paste buffer'.  That is, when user's
///  cut or copy a block of selected lines using editor functions, rather than
///  using the mouse and xterm mechanism, the text goes into the paste_buffer_
///  of the viewer manager.
/// 
///  Please @sa text_windows for more details of use.
///
///  In addition to being a container of viewports, the ViewerManager class
///  also provides methods that help viewports communicate with one another and
///  share common code.  For example: 
///
///  @li  The copy/paste buffer is in the viewer manager
///
///  @li  The etags and cpptags databases are stored in the viewer manager so that
///       viewers don't have to read multiple copies thereof.
///
///  @li  The "find_file_in_standard_locations()" function lets viewers 
///       convert file name fragments into their full pathname by considering
///       the etags or cpptags databases that might exist.
{
public:

  static int normal_att;        ///< normal text color
  static int active_title_att;  ///< window title bar color
  static int inactive_title_att;///< window title bar color
  static int active_mark_att;   ///< highlighted (selected) and marked lines
  static int inactive_mark_att; ///< marked but not highlighted (not selected)
  static int bottom_att;        ///< bottom line of window highlights
  static int highlighted_att;   ///< highlighted (selected) but not marked

  static std::list<std::string> key_bindings_doc;  ///< public so it can be
                                                   ///  used by help generators.


  friend class ViewerSelector; ///< lets you select and activate a viewer
  friend class appChooser;     ///< lets you create a new viewer with an app in it


  typedef CursorWindow::viewport            viewport;        ///< a VERY TEMPORARY handle to a screen
                                                             ///  region (at least as used by the
                                                             ///  viewer manager.

  typedef CursorWindow::input_event         input_event;     ///< key or mouse button info
  typedef std::auto_ptr<viewport>           vp_handle;       ///< typename alias
  typedef CursorWindow::row_col             row_col;         ///< typename alias
  typedef std::pair<Viewer*,std::string>    vpentry;         ///< Entry in the viewers_ table.
  typedef viewport::repaint_handler         repaint_handler; ///< Base class ov Viewers

  typedef Viewer* (*application)(ViewerManager*vm, std::string &fullname);
    ///<
    /// An application is a pointer to a function that returns a Viewer*.
    /// It is also called the 'application hook function'.
    ///
    /// It must be declared like this:
    ///
    ///   Viewer* your_application(ViewerManager* w, std::string &fullname);
    ///
    /// It must not be a class member -- unless it is static.
    ///
    /// The application may prompt the user with dialogs or it may
    /// simply return a Viewer*.  It may also return 0 to indicate
    /// that the user aborted any input prompts required by the
    /// application in order to create the viewer.
    ///
    /// The parameter fullname could be a file name, program invocation parms,
    /// etc.  If it is an empty string, the application should prompt the
    /// user for a proper name.  If it is not empty, the application may
    /// choose, if so desired, to produce error prompts.  If fullname is not
    /// empty, the application should NOT prompt.
    ///
    /// If the ViewerManager supplies a non-empty fullname, it will be the
    /// full pathname of the target file.

  typedef std::pair<application,std::string>  appentry;
     ///<
     /// An entry in the applications table


  CursorWindow* window() { return &cw_; } 

  struct FileLocation
     ///< The name of a file and a line number within it.
  {
     std::string filename_;
     int         lineNumber_;

     FileLocation(std::string const &file, int line)
     : filename_(file)
     , lineNumber_(line)
     {
     }

     FileLocation()
     : lineNumber_(0)
     {
     }


  };

  typedef std::list<FileLocation> FileLocations;

  bool isVisible(Viewer const *) const;  ///<  Return true if the specified viewer appears in an active viewport.

  void broadcast(std::string const &type, std::string const &msg);
      ///<  Broadcast a message to all viewers who are interested.
      ///  
      ///   The Viewer class has a virtual method, listen() which receives the message.  By default, all messages
      ///   are ignored.  There must be cooperation between broadcasters and listeners with respect to the typename and the format of message value strings.  
      ///   In general, viewers should simply ignore message types that they are not expecting.
      ///   The VarObj class can be useful for structuring complex message bodies.  It is JSON-like and fairly easy to use to hold complex
      ///   messages.  There is no requirement to use it however.
      ///
      ///   The principle use of the broadcast() function is expected to be debugging.  When the debugger is active,
      ///   text editor's can broadcast the current file and line information as a "fileLocation" message type.  The
      ///   debugger will listen for this message and can then set a breakpoint and broadcast a "breakpointSet" message
      ///   of it's own in response.  Then any viewer showing the file file and line can annotate said line in response.
      ///
      ///   @param type  A string which represents the type of message being sent.
      ///   @param msg   The message being sent (possibly formatted as VarObj).



private:

  std::string            tool_dir_;  ///< ~/.tools_home

  CursorWindow           cw_;        ///< the CursorWindow assocated with this viewer

  std::vector<viewport*> viewports_; ///< The list of active viewports.  Note that
                                     ///  while a viewport points to a viewer,
                                     ///  there is no one to one mapping between
                                     ///  them.

  std::vector<vpentry>   viewers_;   ///< Active viewers.  See the commentary
                                     ///  with respect to the relationship between
                                     ///  active viewports and active viewers under
                                     ///  ViewportManager::viewports_.
                                     ///  Note that viewers are "owned" by the
                                     ///  viewer manager.  When one is removed
                                     ///  from this list, it is deleted as well.

  MRU<std::string>       order_;     ///<  The order in which viewers have been activated.

  size_t                 active_viewport_;
                                     ///<
                                     ///  index in viewports_ to active viewport
                                     ///  You can get the viewer from the the
                                     ///  viewport using get_repaint_handler().

  void setup_key_bindings();         ///< Initialize master_keys_ map.

  void dispatch_event(input_event const &e);
                                     ///  Give event to active viewport
                                
  void handle_master_command(int cmd); ///<  Interpret commands in the ViewerManager::master_commands list

  std::auto_ptr<Viewer>       vs_;     ///< special Viewer selector viewer (recursive ain't it)

  std::vector<appentry>       apps_; ///< list of available applications (ie viewer types)

  size_t override_active_viewport_;  ///< only used occasionally to override cur
                                     ///  active_viewport_ member.  mostly a kludge

  typedef std::pair<application, std::string> opener_map_entry; ///< entries in the table used by open()

  std::list<opener_map_entry> openers_;  ///< list of applications which can be used
                                         ///  to open files with various extensions.
                                        
  std::list<opener_map_entry> editors_;  ///< list of applications which can be used
                                         ///  to edit files with various extensions.
                                        
                                        
  application                 dir_opener_;  ///< application handling directories.

  std::string                 exit_dir_; ///< If not empty, its the dir at the
                                         ///  time of an force exit key

  void recordExitDir() const;  ///<  Called just before exiting the application to
                               ///   write the active_directory() to ~/.tools_home/exitDir.txt.

  FileLocations fileLocations_; ///<  Places in various files saved for later revisiting.

public:

  void findMatchingViewers(FileName namePattern, std::list<std::pair<Viewer*, FileName> > &viewers);
      ///<  Search the list of viewers and return those that match the specified pattern.
      ///
      ///  @param[in]  namePattern  The full pathname of the file of interest or a *.cpp style
      ///                           pathname pattern.
      ///
      ///  @param[out]  viewers     Pointers to the viewers that match the pattern.
      ///
      ///  @warning:  do not store the viewers returned by this function since they can be
      ///             deleted at any time after the caller returns.


  ViewerManager(); ///<  Hmm!  A constructor, what'll they thank of next.
 ~ViewerManager(); ///<  Aww!  A destructor.  Bad ole' puddy tats use destructors


  FileName find_file_in_standard_locations(FileName const &fragment, FileName const &directory);
      ///<
      ///   Search for a specified file (assuming it is not an absolute path) in the set
      ///   of standard directories.  This set include:
      ///
      ///   A full directory pathname passed in (including trailing /).
      ///
      ///   The SOURCEPATH directory list
      ///
      ///   All directories mentioned in file TAGPP.tagpp (if any).
      ///
      ///   Directories mentioned in TAGS (the etags database -- if any).
      ///
      ///   Note that the "current" directory is explicitly NOT in the search location set
      ///   unless you pass it in as "directory".
      ///
      ///   @param[in] fragment    The file name extracted from a log file, or whatever
      ///                          that maybe relative to a different directory than
      ///                          the current directory.
      /// 
      ///   @param[in] directory   The name of the directory that should be treated as
      ///                          the "current directory" for opening the file -- it may
      ///                          not be the real current directory.

  bool popFileLocation(FileLocation &output)
  {
     if(fileLocations_.empty())
        return false;

     output = fileLocations_.front();
     fileLocations_.pop_front();

     return true;
  }

  template<class Container>
  void setFileLocations(Container const &c)
  {
     typename Container::const_iterator first = c.begin(), last = c.end();

     fileLocations_.clear();

     while(first != last)
     {
        fileLocations_.push_back(*first++);
        
     }
  }

  void add_app(application a, std::string const &s)
    /// Add an application to this viewermanager.
  {
    apps_.push_back(appentry(a,s));
  }

  void vsplit(Viewer* viewer, std::string fullname, int keep_hint=0);
     ///<
     /// Split the current viewport vertically and give space to a new viewer.
     /// If the viewer* is already in the viewers_ list, the fullname will be
     /// ignored in favor of the extant viewer's name.
     ///
     /// @warning:  this function call deletes viewports!  If the user
     ///            is looking at a viewer and splits it, then the original
     ///            viewport instantly becomes invalid.
     ///            Call ViewerManager::find_viewport(this) immediately
     ///            after a call to vsplit to get the proper value.
     ///  
     ///  @param viewer     The the viewer to split.
     ///  @param fullname   The full pathname of the file or object to
     ///                    show in the bottom portion of the split.
     ///  @param keep_hint  A suggestion of how many rows to leave in
     ///                    in the donating viewer.

  void hsplit(Viewer*, std::string fullname, int keep_hint=0);
     ///<
     /// Split the current viewport horizontally and give space to a new viewer.
     /// If the viewer* is already in the viewers_ list, the fullname will be
     /// ignored in favor of the extant viewer's name.
     ///
     ///
     /// @warning:  this function call deletes viewports!  If the user
     ///            is looking at a viewer and splits it, then the original
     ///            viewport instantly becomes invalid.
     ///            Call ViewerManager::find_viewport(this) immediately
     ///            after a call to vsplit to get the proper value.
     ///
     ///  @param keep_hint  A suggestion of how many colss to leave in
     ///                    in the donating viewer.


  void add(Viewer*, std::string fullname);
    ///<
    /// Replace the current viewport's viewer with this new one.  Or create
    /// the first one if it isn't.  Note that the current viewer is unaffected
    /// by this call -- only the binding to the viewport is lost.  Note that
    /// if the viewer* is alread in the viewers_ list, the fullname will be
    /// ignored in favor of the extant name.

  bool activate(Viewer* v);
    ///<
    /// Make the specified viewer visible.  If it already is, then activate
    /// that viewport.  If not, replace the current viewport with a viewport
    /// containing the named viewer.  If you can't find the specified viewer,
    /// return an true.  Otherwise, return false.

  void pushbackViewport(viewport *vp);
     ///<  Push the specified viewport onto viewports_ if it is not already there

  void discard(Viewer* v);
    ///<
    /// Discard the specified viewer by hiding it then deleting it.  If was previously
    /// visible, find a different viewer and make it full screen.
    ///
    /// @param v  A pointer to the viewer to be discarded.

  bool is_visible(Viewer* v);
    ///<
    ///  Return true if the specified viewer is displayed in a viewport that
    ///  is currently visible.  This envolves a search of all vieweports.

  viewport *unify(); ///< Eliminate all viewports but a single full screen viewport
                     ///  holding the currently active viewer.  This collapses viewports
                     ///  but does not close the viewers.
                     ///
                     ///  @return The @b new viewport * of the current viewer.
                     ///
                     ///  The unify function will hide all viewers except the current
                     ///  one which means that viewport objects all get recreated.
                     ///  This function returns the viewport pointer that is recreated
                     ///  to hold the current viewer.
                     ///
                     ///  @note  Other viewer's are not just hidden.  The viewports associated
                     ///         with them are destroyed, so you should never hang onto a viewport
                     ///         because it could be destroyed on  you by the user unifying some
                     ///         other window.

  bool isUnified() const;  ///<  returns true if exactly one viewers are visible

  void growTaller(); ///<  if possible make this window 1 row taller.  This is only
                     ///   possible if the screen is split vertically and if there are windows above or
                     ///   below the current window which can give up screen space to this window.
       
  void growWider();  ///<  if possible make this window 1 col wider.  This is only
                     ///   possible if the screen is split horizontally and if there are windows left or
                     ///   rigth the current window which can give up screen space to this window.

  void vs();    ///< Replace the current viewer with the viewport selector
                ///  viewer.  It will let the user choose another viewer to
                ///  store in the current viewport.

  void open(); ///< Open the cursor window for drawing.  You must do this
               ///  before you create any viewers or viewports.  This does
               ///  not occur automatically.
        
  void close(); ///< Close the cursor window.  This happens automatically in
                ///  the destructor.
                
  void exit();   ///< Close the window and terminate the program.
                
  void run();   ///< Prompt the user for input until all windows close.

  enum master_commands   ///  commands interpreted by the ViewerManager
  {                      ///  not sent to the active viewer.

     vertical_split,     ///<  split active window in 1/2 vertically
     horizontal_split,   ///<  split active window in 1/2 horizontally
     unsplit,            ///<  hide all but the active window and maximize it
     next,               ///<  move to the next active viewport
     selector,           ///<  popup the viewer selector menu
     die,                ///<  attempt to close all windows and exit
     die_and_jump,       ///<  same as die but save current dir in ~/.tools_home for
                         ///   bash edit() alias to use in cd'ing to the current directory
                         ///   after the editor terminates
     new_application,    ///<  pop up the application selector
     new_viewer,         ///<  open a new _viewer_ over current viewport
     new_editor,         ///<  open a new _editor_ over current viewport
     display_help,       ///<  pop up a help viewer describing the active viewer's application
     interactive_shell,  ///<  goto an interactive command shell.  (user must use exit to get rid of it)
     grow_taller,        ///<  increase height of this window
     grow_wider,         ///<  increase width of this window
     last_command
  };

  typedef Sequence_Map<int,int> master_key_map;
    ///<
    /// type of map of key sequences intercepted by the run command and not
    /// sent to the active window

  master_key_map  master_keys_;
                                     ///< Key sequences absorbed by the viewer
                                     ///  manager that are not passed through.
                                     ///  Sequences of keys get mapped into
                                     ///  one of the master_commands below.

  Viewer* active_viewer()
  {
    size_t viewport_no = active_viewport_;

    if(override_active_viewport_ < viewports_.size())
      viewport_no = override_active_viewport_;

    if(viewports_.size())
      return static_cast<Viewer*>( viewports_[viewport_no]->get_repaint_handler() );

    return 0;
  }

  viewport* active_viewport()
    ///  Get the currently active viewport in case someone wants to screw
    ///  around with the cursor position and then restore it without calling
    ///  the activate method to repaint the whole window.
  {
    size_t viewport_no = active_viewport_;

    if(override_active_viewport_ < viewports_.size())
      viewport_no = override_active_viewport_;

    return viewports_[viewport_no];
  }

  viewport *find_viewport(Viewer *v);
    ///<
    ///   Search the list of active viewports for the one holding the specified viewer.
    ///   @return null if not found, otherwise the pointer to the viewport.
    ///
    ///   @note  Viewport pointers should not be held for any length of time.  They are
    ///          temporary objects that come and go frequently.
    ///
    ///   @b Usage
    ///
    ///   After you split the screen, the original viewport becomes invalid, so if you need to
    ///   draw in the viewport after splitting it, you should call ViewerManager::find_viewport()
    ///   to get the proper pointer.


  Viewer* find_viewer(std::string const& viewername);
    ///<
    /// Search the vector of viewers in creation order to find the viewer whose name is provided.
    /// If you can't find it, return 0, otherwise, return its pointer.  You can
    /// then use activate() to make it the active viewer.

  std::string viewerFullname(CursorWindow::viewport *); 
    ///<
    /// Return the fullname associated with the specified viewport.  This function is mainly used for debugging
    /// as you rarely have a viewport pointer with which to search in any other situation.
    /// @return empty string or the "fullname" stored in the vpentry that is kept in the viewers_ table.

  Viewer* most_recent_viewer();
    ///<
    /// Return a pointer to the most recently activated viewer.  Most of the time this means
    /// the currently active viewer.  However, when this function is called after a viewer
    /// closes, then the call gives you "previous" most recently activated viewer so that you
    /// reactivate it and make the user happy.
    ///
    /// @return null or a pointer to a viewer.  Null is only returned when there are no viewers.

  std::string find_viewer_name(Viewer *vp);
    ///<  Given a viewer pointer, find the viewer name 

  size_t find_viewport_index(std::string const &viewername);  ///< @return the index into the
                                                              ///  viewports_ table for the
                                                              ///  specified viewer.  
                                                              ///  
                                                              ///  Search the viewers_ table
                                                              ///  then map that to the the
                                                              ///  viewports_ table.


  void help();
    ///<
    ///  Present the user with help information specific to the currently
    ///  active viewer.

  void help_helper( std::list<std::string>  &  help_text);
    ///<
    ///  This function can be called by viewer's help() methods to actually
    ///  popup the help viewer.  All you have to do is make a
    ///  std::list<std::string> containing the text to be displayed then
    ///  call ViewerManager::help_helper(list).

  Viewer* open(std::string const &fullname);
    ///<
    ///  Return a viewer opening the specified filename -- which should be the
    ///  fullname of the target file.  This does not add the viewer to the displayed
    ///  viewers nor does it it put it in the list of managed viewers.  You must
    ///  call add(), vsplit(), or hsplit(), on your own.  The viewer returned will
    ///  be one defined by the type of the named file (directory or not), and by
    ///  the viewer associated with the file name extension in the list.  The default
    ///  directory viewer application is specified using ViewerManager::set_directory_app().
    ///  To specify which applications open which file name extensions, make calls to
    ///  add_viewer_app().
    ///
    ///  Note, if the basename part of the specified name contains '*' or '?', it will
    ///  be assumed that this is a directory search -- ie a directory open, not some
    ///  other kind.  So, '/dir/*.c' is an attempt to open directory /dir, but restrict
    ///  the display to the '*.c' files.
    ///
    ///  Note that this creates a new viewer.  If you want to
    ///  activate an existing viewer -- which is often the case -- you will want
    ///  to use 'find_viewer(fullname)' to find it.  Then use activate(viewer)
    ///  to make the the active_viewer().

  Viewer* edit(std::string const &fullname);
    ///<
    ///  Same as open(name) except that it edits instead of just viewing.  Note that a
    ///  viewer may also edit, but and editor explicitly edits as opposed to safely
    ///  viewing the file.
    ///
    ///  Note that this creates a new viewer.  If you want to
    ///  activate an existing viewer -- which is often the case -- you will want
    ///  to use 'find_viewer(fullname)' to find it.  Then use activate(viewer)
    ///  to make the the active_viewer().

  void set_directory_app(application);
    ///<
    ///  Specify which application handles directory browsing.  If you don't call this
    ///  routine, the open() function won't be able to open directories for you.

  void add_viewer_app(application, std::string filename_extension);
    ///<
    ///  Specify which applications handle which file name extensions.  This list is
    ///  searched in reverse order (LIFO).  Thus you can superseed the previous
    ///  specifications.  If the open method cannot find a match in the list created
    ///  by calls to set_viewer_app, it will use the very first one specified (ie the
    ///  first element in the list.  If no element exists, you will get a null value from
    ///  open().

  void add_editor_app(application, std::string filename_extension);
    ///<
    ///  Same as add_viewer_app except that it adds an editor application.

  Viewer::paste_buffer_type  paste_buffer_;
    ///<
    /// When viewers copy lines into the "paste buffer", this is where they go.

  void popup_new_viewer(bool edit=false);
    ///<
    /// Prompt the user for the name of a file or directory, starting in the
    /// directory where the currently active viewer is situated.  If the user
    /// doesn't abort, use the 'open(name)' method to find a viewer for the
    /// file the user has selected.  If the parameter, edit is set to true then
    /// edit(name) will be used instead of open(name) to define the viewer
    /// activated.

  void splitscreen_new_editor(bool horizontal=false);
    ///<
    /// Prompt the user for the name of a file or directory, starting in the
    /// directory where the currently active viewer is situated.  If the user
    /// doesn't abort, use the 'edit(name)' method to find a viewer for the
    /// file the user has selected.  Split the current window vertically, or
    /// horizontally if the horizontal flag is specified and show the editor
    /// in the newly created split.

  std::string active_directory() const;
    ///<
    /// Get the directory part, including the trailing /, of the name of the
    /// current viewer (active_viewer()).
    ///
    /// Note that this data is only as good as was specified at the time the
    /// viewer was 'add'ed.

  void handle_die_request();
    ///<
    ///  Check to see if there are any unsaved edit changes.  If so,
    ///  print an error and return.  If not, exit().  This function is
    ///  called when the user presses the 'die' key sequence.

  std::string executable_version_;
    ///<
    ///  The executable_version_ is a string set by the application that creates
    ///  the viewer manager as a hint to be used to identify the version of the
    ///  executable.  The executable version is made visible to the user when the
    ///  help command is executed -- assuming that help_helper() is invoked.

  std::string executable_name_;
    ///<
    ///  The name of the main program that is currently in use.  This variable
    ///  must be set by the program that initially creates the viewer, or
    ///  the value will be null.

  static std::string default_etags_path(std::string &pathlist, char const *file="TAGS");
    ///<
    ///  Get the full pathname of the TAGS file -- if any. Also return the
    ///  colon separated list of directories searched by the function
    ///  The pathlist will contain $ETAGSPATH, $HOME, and $CDPATH.

  EtagsDB* default_etags_db();
    ///<
    ///  Return a pointer to a new'ed up etags database read from the default location.
    ///  @warning You must delete this when you are done with it.
    ///

  CppTagDB::Info* default_cpptags_db();
    ///<
    ///  Return a pointer to a new'ed up CppTagDb database read from the default location.
    ///  @warning You must delete this when you are done with it.
    ///


  int   find_files(std::string operand, std::list<std::string> *result);
    ///<
    ///
    /// This function finds files in a list of standard locations independently
    /// of any viewers that might be showing any of those files.  It strictly
    /// access the file system and etags database if any.
    ///
    /// @param operand  A file name fragment (base name or non-fully qualified
    ///                 relative path reference).
    ///
    /// @param result   A list of matching files.  See below.
    ///
    /// @return The count of matching files.
    /// 
    /// It searches for a specified file (operand) in the "standard" locations and return
    /// the full[er] name of the file.  That is, given a file name specified
    /// as a simple file basename or relative path, search for any file that 
    /// matches that and return a list thereof.  Return a count of said
    /// matches.  The "standard" locations are defined as follows:
    ///
    ///  If operand is a fully qualified pathname, return only it.
    ///
    ///  If the operand is an empty string, return all the files in the 
    ///  etags database file.
    ///
    ///  If the operand is path fragment, search for all matches in the 
    ///  the following locations:
    ///
    ///  @li  current directory
    ///  @li  $SOURCEPATH
    ///  @li  $CDPATH
    ///  @li  TAGS list of files (ie etags generated file list)
    ///
    ///  When generating the etags database, it is desireable to specify the
    ///  fully qualified pathnames on the etags command line otherwise the
    ///  returned filenames are likely to be unusuable.
    ///
    ///  Note that the output list of strings is not cleared by this function
    ///  call -- it appends new filenames to the resultant list and returns the
    ///  count of the strings appended.
    ///

};
} // namespace cxxtls
#endif
