//
// Copyright 2002, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.
//
// Warning:  not all files in this directory structure are covered by the
// same copyright.  Some of them are part of the GNU source distribution
// and you must obey the GPL copyright for those files.

///@file
/// Implements members of the TextViewer class.

#include <cxxtls/textviewer.h>
#include <cxxtls/viewermanager.h>
#include <fstream>
#include <cxxtls/strtool.h>
#include <portable_strstream.h>
#include <vector>
#include <portable_io.h>
#include <cxxtls/cursesinterface.h>

namespace cxxtls
{

struct TextViewer::impl
  //
  ///  This class holds the actual guts of a viewer.
  //
{
   FileName                 filename_;   ///< saved name of the file
   std::vector<std::string> lines_;      ///< actual text of the file
   bool                     active_;     ///< does this viewer have focus
   int                      left_;       ///< column bias to skip the bordering '|'s
   int                      first_row_;  ///< where file text is displayed
   size_t                   first_mark_; ///< first marked line
   size_t                   last_mark_;  ///< last marked line ( 1 past last marked line, really)
   size_t                   page_top_;   ///< top of displayed page
   size_t                   cur_;        ///< current cursor position
   size_t                   col_;        ///< column cursor is in, 0 origin
   size_t                   page_left_;  ///< column of left side of visible page
   std::string              search_;     ///< string to use in find string operations

   void set_text_attribute(TextViewer*,
                           viewport*,
                           size_t line_no,
                           bool is_bottom,
                           bool highlight=true
                           );
   void display_line(viewport*, size_t line);  // doesn't set text attributes

   bool find_string(size_t line_no);  ///< search for string starting at line_no + 1
                                      ///< return true if the string is found.

};

TextViewer::
~TextViewer()
{
}

TextViewer::
TextViewer(ViewerManager*vm, FileName filename)
:  Viewer(vm),
   impl_(new impl)
{
   // construct a text viewer and read the file corresponding thereto.

   impl_->filename_   = filename;
   impl_->active_     = false;
   impl_->left_       = 0;
   impl_->first_row_  = 1;
   impl_->first_mark_ = 0;
   impl_->last_mark_  = 0;
   impl_->page_top_   = 0;
   impl_->cur_        = 0;
   impl_->col_        = 0;
   impl_->page_left_  = 0;

   using namespace std;

   std::list<std::string> lines;

   fstream file;

   file.open(filename.c_str(), ios::in);

   if(file.fail())
   {
     return;
   }

   istreambuf_iterator<char> nc(file);
   istreambuf_iterator<char> lc;

   string line;

   size_t line_count=0;

   while(nc != lc)
   {
     char c = *nc++;

     if(c == '\n')
     {
       lines.push_back(StrTool::expand_tabs(line,0,'~'));
       line.resize(0);
       ++line_count;
     }
     else
       line += c;

   }

   if(line.size())
   {
     lines.push_back(StrTool::expand_tabs(line,0,'~'));
     ++line_count;
   }

   file.close();

   impl_->lines_.resize(line_count);

   std::copy(lines.begin(), lines.end(), impl_->lines_.begin());

}

void
TextViewer::impl::
display_line(viewport *vp, size_t line)
{

  if(line >= lines_.size())
  {
    vp->fill_to_eol();
    return;
  }
  

  std::string const & str = lines_[line];

  size_t width = str.size();

  if(width >= page_left_ && line <= lines_.size() )
  {
    std::string::const_iterator start = str.begin() + page_left_,
                                end   = str.end();
                        
    vp->write(start, end);

  }

  vp->fill_to_eol();

}


void
TextViewer::
operator() ( CursorWindow::viewport *     vp,
                            int           cmd
                          )
 /// Handle screen repaint commands
{
  // WARNING:  do not save vp -- it gets deleted regularly!

  if(cmd == CursorWindow::viewport::repaint_handler::activate)
    impl_->active_ = true;
  else
  if(cmd == CursorWindow::viewport::repaint_handler::deactivate)
    impl_->active_ = false;


  // determine if we need to write a column of '|'s to separate us from
  // the window to the left -- and subsequently not write in that column.

  row_col size    = vp->size();
  row_col origin  = vp->origin();

  if(origin.col_ != 0)
    impl_->left_ = 1;
  else
    impl_->left_ = 0;

  int i;

  vp->set_text_attribute(manager_->inactive_mark_att);

  for(i=0; impl_->left_ && i < size.col_; ++i)
  {
    vp->set_curpos(i,0);

    using namespace CursesInterface;

    vp->write(line_chars[VL_MIDDLE]);
  }

  // now write the title

  vp->set_curpos(0,impl_->left_);
  vp->set_text_attribute( impl_->active_ ? manager_->active_title_att
                                         : manager_->inactive_title_att
                        );
                        
  *vp << "Viewing " << impl_->filename_.shorten(size.col_ - 8);

  vp->fill_to_eol();

  // now write the displayable portions of the file text -- handling
  // marked lines, left margin and special highlighting of the bottom
  // line.

  int row=impl_->first_row_;

  int displayable_rows = size.row_ - row;

  size_t cur_line = impl_->page_top_;

  for(i=0; i < displayable_rows; ++i)
  {
    // for each displayable row in the window

    if(cur_line >= impl_->lines_.size())
      break;

    // we haven't dropped off the bottom of the file data

    impl_->set_text_attribute(this, vp, cur_line, row == size.row_-1);

    vp->set_curpos(row, impl_->left_);

    impl_->display_line(vp, cur_line);

    ++row;
    ++cur_line;
  }


  bool first_time=true;

  while(row < size.row_)
  {
    vp->set_curpos(row,impl_->left_);

    if(first_time)
    {
      first_time=false;

      vp->set_text_attribute(ViewerManager::active_title_att);
      vp->write("*eof*");
      impl_->set_text_attribute(this, vp, cur_line, false, false); // not current and not bottom
    }
    else
    {
      if(row == size.row_-1)
      {
        vp->set_text_attribute(manager_->bottom_att);
      }
      else
        impl_->set_text_attribute(this, vp, cur_line, false, false); // not current and not bottom

    }

    vp->fill_to_eol();

    ++row;
  }

  {
    // make sure cursor is in a valid position

    size_t width = size.col_ - impl_->left_;

    size_t cur_col = impl_->col_ - impl_->page_left_;

    if(cur_col >= width)
    impl_->col_ = impl_->page_left_ + width-1;

  }


  vp->set_curpos(impl_->first_row_ + impl_->cur_ - impl_->page_top_,
                 impl_->left_ + (impl_->col_ -  impl_->page_left_)
                );

}


bool
TextViewer::
handle_event( CursorWindow::input_event const * e,
                             CursorWindow::viewport          * vp
                           )
  /// Handle input events.  Return false means we didn't terminate
{

  if(e->type_ == input_event::ForceExitKey)
    return true; // terminate window operations


  if(e->type_ != input_event::FunctionKey &&
     e->type_ != input_event::DataKey
    )
  return false;  // huh?  Ignore unexpected events

  if(e->value_ == ('K' - '@'))
    return true;

  int edit_func = (*manager_->window()).func[e->value_]; // map user key to function

  // unhighlight the currently selected line to handle the key rehighlighting
  // that may be needed

  row_col size    = vp->size();
  row_col origin  = vp->origin();

  if(origin.col_ != 0)
    impl_->left_ = 1;

  int row=impl_->first_row_;   // row within displayable region of viewport

  size_t cur_line = impl_->cur_;

  row += cur_line - impl_->page_top_;

  impl_->set_text_attribute(this, vp, cur_line, row == size.row_-1, false);
       // don't highlight but do all else

  vp->set_curpos(row, impl_->left_);

  if(impl_->lines_.size() > cur_line)
    impl_->display_line(vp, cur_line);

  // handle the function -- assuming that the current line is not currently drawn highlighted

  switch(edit_func)
  {
    case CursorWindow::func_down:  // cursor down 1 line

      if(impl_->lines_.size() == 0)
        break;

      if(cur_line < impl_->lines_.size()-1 &&
         row < size.row_ - impl_->first_row_
        )
      {
        // not at bottom of displayable area, just move down a line
        
        ++impl_->cur_; // change cur line to force highlight color change
        
        impl_->set_text_attribute(this, vp, cur_line, row == size.row_-1);
             // re-highlight the current line

        vp->set_curpos(row, impl_->left_);

        impl_->display_line(vp, cur_line);
        
        ++row;
        ++cur_line;  // change active line
      }
      else
      if(cur_line != impl_->lines_.size()-1)
      {
        // we are on last line of viewport but not last line of the
        // file -- scroll page down 1 line
        
        
        ++cur_line;
        ++impl_->cur_;
        ++impl_->page_top_;
        
        (*this)(vp,0);  // repaint window
        
      }

      break;

    case CursorWindow::func_up:  // cursor down 1 line

      if(impl_->lines_.size() == 0)
        break;

      if(row != impl_->first_row_)
      {
        // not at bottom of displayable area, just move down a line

        --impl_->cur_; // change cur line to force highlight color change
        
        impl_->set_text_attribute(this, vp, cur_line, row == size.row_-1);
             // re-highlight the current line

        vp->set_curpos(row, impl_->left_);

        impl_->display_line(vp, cur_line);

        --row;
        --cur_line;
      }
      else
      if(cur_line != 0)
      {
        // scroll up one line
        --cur_line;
        --impl_->cur_;
        --impl_->page_top_;
        
        (*this)(vp,0);  // repaint window
        
      }

      break;


    case CursorWindow::func_next:  // cursor down 1 page

      if(impl_->lines_.size() == 0)
        break;

      {
        int page_size = size.row_ - impl_->first_row_;
        
        impl_->cur_ += page_size;
        
        if(impl_->cur_ >= impl_->lines_.size())
        {
          impl_->cur_ = impl_->lines_.size() -1;
        }
        
        if( (int)(impl_->cur_) < 0 )
          impl_->cur_ = 0;
        
        impl_->page_top_ += page_size;
        
        if(impl_->page_top_ >= impl_->cur_)
          impl_->page_top_ = impl_->cur_;
        
        row = impl_->first_row_ + (impl_->cur_ - impl_->page_top_);
        cur_line = impl_->cur_;
        
        (*this)(vp,0);
        
      }
      break;


    case CursorWindow::func_prior:  // cursor up 1 page

      if(impl_->lines_.size() == 0)
        break;

      if(impl_->page_top_ == 0)
      {
        impl_->cur_ = 0;
        
        impl_->set_text_attribute(this, vp, cur_line, row == size.row_-1);
             // re-highlight the current line

        vp->set_curpos(row, impl_->left_);

        impl_->display_line(vp, cur_line);
        
        row = impl_->first_row_;
        cur_line = 0;
      }
      else
      {
        int page_size = size.row_ - impl_->first_row_;
        
        impl_->cur_ -= page_size;



        
        if(impl_->cur_ >= impl_->lines_.size())  // cur is unsigned ;-<
        {
          impl_->cur_ = 0;
        }
        
        if( (int)(impl_->cur_) < 0 )
          impl_->cur_ = 0;
        
        impl_->page_top_ -= page_size;
        
        if(impl_->page_top_ >= impl_->cur_)  // page top is unsigned
          impl_->page_top_ = impl_->cur_;
        
        row = impl_->first_row_ + (impl_->cur_ - impl_->page_top_);
        cur_line = impl_->cur_;
        
        (*this)(vp,0);
        
      }
      break;


    case CursorWindow::func_top:  // goto top of file

      if(impl_->lines_.size() == 0)
        break;

      row = impl_->first_row_;
      cur_line = 0;
      impl_->cur_ = 0;
      impl_->page_top_ = 0;

      (*this)(vp,0);

      break;


    case CursorWindow::func_bottom:  // goto top of file

      if(impl_->lines_.size() == 0)
        break;

      {
        int page_size = size.row_ - impl_->first_row_;

        impl_->cur_ = impl_->lines_.size()-1;
        
        impl_->page_top_ = impl_->cur_ - page_size + 1;
        
        if(impl_->page_top_ > impl_->cur_ )
          impl_->page_top_ = 0;
        
        cur_line = impl_->cur_;
        
        row = impl_->first_row_ + (impl_->cur_ - impl_->page_top_);

        (*this)(vp,0);
      }

      break;

    case CursorWindow::func_right:
      {
        size_t width = size.col_ - impl_->left_;
        
        ++impl_->col_;
        
        if(impl_->col_ - impl_->page_left_ >= width )
        {
          impl_->page_left_ += 10;
          impl_->col_       = impl_->page_left_ + width - 1;
        
          (*this)(vp,0);
        }
        
      }
      break;


    case CursorWindow::func_left:
      {
        if(impl_->col_)
        {
          --impl_->col_;
        
          if(impl_->col_ < impl_->page_left_)
          {
            impl_->page_left_ -= 10;
        
            if( (int)(impl_->page_left_) < 0 )
              impl_->page_left_ = 0;
        
            (*this)(vp,0);
        
          }
        
        }
      }
      break;

    case CursorWindow::func_tab:
      {
        impl_->page_left_ += 10;
        impl_->col_       += 10;
        (*this)(vp,0);
      }
      break;

    case CursorWindow::func_btab:
    scroll_left:
      {
        impl_->page_left_ -= 10;
        impl_->col_       -= 10;
        
        if( (int)(impl_->page_left_) < 0 )
          impl_->page_left_ = 0;
        
        if( (int)(impl_->col_) < 0)
          impl_->col_ = 0;
        
        (*this)(vp,0);
      }
      break;

    case CursorWindow::func_home:
      {
        bool repaint = false;
        
        if(impl_->page_left_)
        {
          impl_->page_left_ = 0;
          repaint = true;
        }
        
        impl_->col_       = 0;
        
        if(repaint)
          (*this)(vp,0);
      }
      break;


    case CursorWindow::func_end: // move cursor to end of current line

      if(impl_->lines_.size())
      {
        size_t width = size.col_ - impl_->left_;

        std::string const &cur = impl_->lines_[cur_line];
        
        impl_->col_ = cur.size();
        
        if(impl_->col_ < impl_->page_left_)
        {
          impl_->page_left_ = impl_->col_ - 10;

          if( (int)(impl_->page_left_) < 0)
          {
            impl_->page_left_ = 0;
          }
        
          (*this)(vp,0);
        }
        else
        if(impl_->col_ >= impl_->page_left_ + width)
        {
          impl_->page_left_ = impl_->col_ - 10;
        
          if( (int)(impl_->page_left_) < 0)
          {
            impl_->page_left_ = 0;
          }
        
          (*this)(vp,0);
        }
        
      }
      break;

    case CursorWindow::func_find:  //find first
      {
        CursorWindow::Dialog d("Find String");
        
        d += new CursorWindow::Dialog::String("name",
                                              "String to search for",
                                              "",
                                              15
                                             );
                                
        if(d.popup(manager_->window()))
          return false;
        
        impl_->search_ = d.element_value("name");
        
        if(impl_->search_.size() == 0)
           return false;

        if( impl_->find_string(cur_line-1) )
        {
          ensure_cursor_visible(vp);
        
          row = impl_->first_row_ + impl_->cur_ - impl_->page_top_;

          cur_line = impl_->cur_;
        }
        
      }
      break;

    case CursorWindow::func_findnxt:  //find first

        if( impl_->find_string(cur_line) )
        {
          ensure_cursor_visible(vp);
        
          row = impl_->first_row_ + impl_->cur_ - impl_->page_top_;

          cur_line = impl_->cur_;
        }
        break;
        
    case CursorWindow::func_mark:
      goto mark_line;

    case CursorWindow::func_data:  // normal keys

      if(impl_->lines_.size() == 0 && e->value_ != 'q' )
        break;

      switch(e->value_)
      {

        case 'q':  return true;
        
        case 'p':
        
          if(impl_->first_mark_ != impl_->last_mark_)
          {


            CursorWindow::FileSelector d("Put Marked lines to a file", impl_->filename_.dirname() + "*", "");

            FileName fullname = d.popup(manager_->window(), this);

        
            if(fullname == "")
              return 0;
        
        
            FileName tmp(fullname);
            tmp.convert_to_absolute_path();
        
            fullname = tmp;
        
            {
              using namespace std;
        
              fstream file;
        
              file.open(fullname.c_str(), ios::out);
        
              if(file.fail())
              {
                CursorWindow::Message m("Error");
                
                m += "Sorry, could not open the target file for write";
                m += "";
                m += "Press enter to continue";
        
                m.popup(manager_->window());
                return 0;
        
                break;
              }
        
              ostreambuf_iterator<char> nc(file);
        
              size_t i;
        
              for(i= impl_->first_mark_; i < impl_->last_mark_; ++i)
              {
                std::string &s = impl_->lines_[i];
                
                std::copy(s.begin(), s.end(), nc);
                
                nc = copy_end_of_line(nc);
                
              }
        
              file.close();
        
            }
        
                
          }
          break;
        
        case 'm':
        mark_line:
          {
            if(impl_->first_mark_ == impl_->last_mark_)
            {
              // no lines currently marked
        
              impl_->first_mark_ = impl_->cur_;
              impl_->last_mark_  = impl_->cur_ + 1;
            }
            else
            {
              // there is already a marked region
        
              if(impl_->last_mark_ == impl_->first_mark_ + 1)
              {
                // the marked region is exactly one line
                
                if(impl_->cur_ == impl_->first_mark_)
                {
                  // and the cursor is sitting on that one line,
                  // then unmark
                
                  impl_->last_mark_ = impl_->first_mark_ = 0;
                
                }
                else
                {
                  // the cursor is not sitting on the currently marked line
                
                  if(impl_->cur_ < impl_->first_mark_)
                    impl_->first_mark_ = impl_->cur_;
                  else
                    impl_->last_mark_ = impl_->cur_ + 1;
                }
                
              }
              else
                impl_->first_mark_ = impl_->last_mark_ = 0;
        
            }
        
            (*this)(vp,0); // repaint to highlight the marks
        
          }
          break;
      }
      break;

    default:  // ignore random commands

      if(e->value_ == ('\\' - '@'))
        goto scroll_left;

      break;
  }

  // re-highlight the current row

  impl_->set_text_attribute(this, vp, cur_line, row == size.row_-1); // don't highlight but do all else

  vp->set_curpos(row, impl_->left_);

  impl_->display_line(vp, cur_line);

  vp->set_curpos(row,
                 impl_->left_ + (impl_->col_ -  impl_->page_left_)
                );


 return 0;

}

void
TextViewer::
help()
{
  std::auto_ptr< std::list< std::string > >
    viewer_help( new std::list< std::string > );

  viewer_help->push_back("Help for the text viewer");
  viewer_help->push_back("");
  viewer_help->push_back("  Use the normal scrolling and cursor movement keys.");
  viewer_help->push_back("");
  viewer_help->push_back("  The 'm' key toggles inclusion of the current line in");
  viewer_help->push_back("  marked region.");
  viewer_help->push_back("");
  viewer_help->push_back("  The 'p' key lets you saved the marked lines to a file.");
  viewer_help->push_back("");
  viewer_help->push_back("  The '^S' key lets you begin a search for a string");
  viewer_help->push_back("");
  viewer_help->push_back("  The '^N' key lets repeat a search for same string used before");
  viewer_help->push_back("");
  viewer_help->push_back("  The 'Tab' key lets scroll to the right by 10 characters");
  viewer_help->push_back("");
  viewer_help->push_back("  The 'Ctrl-\\' key lets scroll to the left by 10 characters");
  viewer_help->push_back("");
  viewer_help->push_back("  The 'shift Tab' key also lets scroll to the left by 10 characters");



  manager_->help_helper(*viewer_help);
}

Viewer*
TextViewer::
app(ViewerManager*vm, std::string &fullname)
{
  if(fullname.size() == 0)
  {
    // do not use the FileSelector here because it can't repaint the screen

    CursorWindow::Dialog d("File name");

    d += new CursorWindow::Dialog::String("name",
                                          "path",
                                          "*",
                                          40
                                         );
                                
    if(d.popup(vm->window()))
      return 0;

    fullname = d.element_value("name");

    if(fullname == "")
      return 0;


    FileName tmp(fullname);
    tmp.convert_to_absolute_path();

    fullname = tmp;

  }
  return new TextViewer(vm,fullname);

}

std::string TextViewer::app_name = "VIEW TEXT";

std::string TextViewer::description() const
{
  return impl_->filename_;
}

void
TextViewer::impl::
set_text_attribute(TextViewer* viewer,
                   viewport*   vp,
                   size_t      line,
                   bool        is_bottom,
                   bool        highlight
                  )
{
  if(line >= first_mark_ && line < last_mark_)
  {
    if(highlight)
    {
      if(active_)
      {
        vp->set_text_attribute(ViewerManager::highlighted_att);
      }
      else
        vp->set_text_attribute(ViewerManager::inactive_mark_att);
    }
    else
        vp->set_text_attribute(ViewerManager::inactive_mark_att);

  }
  else
  if(line == cur_)
    vp->set_text_attribute(active_ ? (is_bottom ? ViewerManager::bottom_att
                                                : ViewerManager::normal_att)
                                   : ViewerManager::inactive_mark_att);
  else
  if(is_bottom)
    vp->set_text_attribute(ViewerManager::bottom_att);
  else
    vp->set_text_attribute(ViewerManager::normal_att);
}

struct caseless_compare
{
  bool operator() (char a, char b)
  {
    if(a == b)
      return true;
    else
    {
      // the characters are not equal

      if( (a < 'a' && b < 'a')
         ||
          (a >= 'a' && b >= 'a')
        )
      {
        // the characters are either both upper case or both lower case
        // but they are different

        return false;
      }
      else
      {
        //
        //  the case of the characters is different, so try comparing them again
        //
        if(islower(a))
          a = toupper(a);
      
        if(islower(b))
          b = toupper(b);
      }
    }

    return a == b;

  }
};

bool
TextViewer::impl::
find_string(size_t line_before)
{
  size_t line = line_before + 1;

  while(line < lines_.size())
  {
    std::string::iterator start = std::search(lines_[line].begin(),
                                              lines_[line].end(),
                                              search_.begin(),
                                              search_.end(),
                                              caseless_compare()
                                             );
    if(start != lines_[line].end())
    {
      col_ = start - lines_[line].begin();
      break;
    }

    ++line;
  }

  if(line == lines_.size())
    return false;

  // at this point, we have found a line containing the specified
  // search string.

  cur_ = line;

  return true;

}


void
TextViewer::
ensure_cursor_visible(viewport* vp)
{
  row_col size = vp->size();

  size_t usable_columns = size.col_ - impl_->left_;
  size_t usable_rows    = size.row_ - impl_->first_row_;

  size_t column_offset  = impl_->col_ - impl_->page_left_;
  size_t row_offset     = impl_->cur_  - impl_->page_top_;

  if(   (row_offset >= usable_rows)
     ||
        (column_offset >= usable_columns)
     ||
        (impl_->col_ < impl_->page_left_)
    )
  {
    // cursor is not on current page

    if(row_offset >= usable_rows )
    {
      impl_->page_top_ = impl_->cur_ - size.row_/2;

      if( (int)(impl_->page_top_) < 0)
        impl_->page_top_ = 0;
    }

    if(  impl_->col_ < impl_->page_left_
       ||
         column_offset >= usable_columns
      )
    {
      impl_->page_left_ = impl_->col_ - 10;

      if( (int)(impl_->page_left_) < 0 )
        impl_->page_left_ = 0;
    }

    (*this)(vp,0);

  }
}

} // namespace cxxtls
